/*
 * JCards
 * Copyright (C) 2002-2009 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package nrs.encryption;

import java.io.FilterInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.bouncycastle.crypto.CryptoException;
import org.bouncycastle.crypto.engines.BlowfishEngine;
import org.bouncycastle.crypto.modes.CBCBlockCipher;
import org.bouncycastle.crypto.paddings.PaddedBufferedBlockCipher;
import org.bouncycastle.crypto.params.KeyParameter;

/**
 * Decrypt an input stream
 *
 * Copyright &#169; 2002-2009 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class Decryptor extends FilterInputStream
{
	private PaddedBufferedBlockCipher cipher;
	private KeyParameter key;
	private int olen;
	private byte[] tmpresult;
	private byte[] decryptBuffer;
	private int decryptBufferSize;
	private boolean doFinalCalled;
	private final static Logger log = Logger.getLogger(Decryptor.class.getName());

	public Decryptor(InputStream is, byte[] password)
	{
		super(is);
		cipher = new PaddedBufferedBlockCipher(new CBCBlockCipher(new BlowfishEngine()));
		key = new KeyParameter(password.clone());
		cipher.init(false, key);
		decryptBuffer = new byte[256];
	}

	@Override public int read() throws IOException
	{
		log.entering("Decryptor", "read()");
		return read(new byte[1]);
	}

	@Override public int read(byte b[]) throws IOException
	{
		log.entering("Decryptor", "read(byte[])");
		return read(b, 0, b.length);
	}

	@Override public int read(byte b[], int off, int len) throws IOException
	{
		int dsize;

		log.entering("Decryptor", "read(byte[], int, int)");

		if (doFinalCalled)
		{
			return -1;
		}

		// return any buffered data first
		if (decryptBufferSize > 0)
		{
			dsize = (decryptBufferSize > len ? len : decryptBufferSize);
			log.log(Level.FINEST, "Returning buffered data: {0} bytes", dsize);
			System.arraycopy(decryptBuffer, 0, b, off, dsize);

			// any more still in the buffer?
			if (dsize < decryptBufferSize)
			{
				System.arraycopy(decryptBuffer, dsize, decryptBuffer, 0, decryptBufferSize - dsize);
			}

			decryptBufferSize -= dsize;
			log.exiting("Decryptor", "read(byte[], int, int)", dsize);

			return dsize;
		}

		// read encrypted bytes
		byte[] encrypted = new byte[len];
		int bread = in.read(encrypted, 0, len);
		try
		{
			// decrypt
			if (bread > 0)
			{
				// decrypting data from stream
				tmpresult = new byte[cipher.getOutputSize(len)];
				olen = cipher.processBytes(encrypted, 0, bread, tmpresult, 0);
				log.log(Level.FINEST, "Deciphered {0} bytes", olen);
			}
			else
			{
				// no more data from stream
				olen = cipher.doFinal(tmpresult, 0);
				if (olen > 0)
				{
					byte tmp[] = new byte[olen];
					System.arraycopy(tmpresult, 0, tmp, 0, olen);
					tmpresult = tmp;
				}
				doFinalCalled = true;
			}

			// copy possible data into buffer
			dsize = (olen > len ? len : olen);
			System.arraycopy(tmpresult, 0, b, off, dsize);

			// buffer any overflow
			if (olen > len)
			{
				System.arraycopy(tmpresult, len, decryptBuffer, 0, olen - len);
				decryptBufferSize = olen - len;
			}

			log.exiting("Decryptor", "read(byte[], int, int)", dsize);

			return dsize;
		}
		catch (CryptoException cex)
		{
			log.log(Level.SEVERE, "Cannot decrypt stream", cex);
			throw new IOException(cex);
		}
	}

	@Override public boolean markSupported()
	{
		return false;
	}

	@Override public int available() throws IOException
	{
		return cipher.getOutputSize(in.available());
	}
}
