/*
 * JCards
 * Copyright (C) 2002-2009 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package nrs.encryption;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.logging.Logger;
import org.bouncycastle.crypto.CryptoException;
import org.bouncycastle.crypto.Digest;
import org.bouncycastle.crypto.digests.*;
import org.bouncycastle.crypto.engines.BlowfishEngine;
import org.bouncycastle.crypto.modes.CBCBlockCipher;
import org.bouncycastle.crypto.paddings.PaddedBufferedBlockCipher;
import org.bouncycastle.crypto.params.KeyParameter;

/**
 * Wrapper to encryption routines
 *
 * @author Copyright &#169; 2002-2009 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class Encryptor extends OutputStream
{
	private PaddedBufferedBlockCipher cipher;
	private KeyParameter key;
	private OutputStream stream = null;
	private int size = 0,  olen = 0;
	private byte[] tmpresult;
	private boolean finished = false;
	private ByteArrayOutputStream result = null;
	private final static Logger log = Logger.getLogger(Encryptor.class.getName());

	public enum DIGEST { MD5, SHA1, SHA256, SHA512, Tiger, Whirlpool }
	public enum CRYPT_ALGORITHM { AES128, AES256, Blowfish }

	/** Initialize the cryptographic engine */
	public Encryptor(byte[] password, boolean encrypt)
	{
		log.entering("Encryptor", "Encryptor(password, encrypt)", encrypt);

		cipher = new PaddedBufferedBlockCipher(
			new CBCBlockCipher(
			new BlowfishEngine()));

		key = new KeyParameter(password.clone());
		cipher.init(encrypt, key);

		log.exiting("Encryptor", "Encryptor()");
	}

	public void init()
	{
		log.entering("Encryptor", "init");

		olen = size = 0;
		finished = false;
		result = new ByteArrayOutputStream();

		log.exiting("Encryptor", "init");
	}

	public void setOutputStream(OutputStream stream)
	{
		this.stream = stream;
	}

	/**
	 * Routine that does the gritty work. Need to call doFinal at the end.
	 * @param data array of bytes to encode
	 * @see #doFinal
	 */
	public void callCipher(byte[] data) throws CryptoException, IOException
	{
		log.entering("Encryptor", "callCipher");

		size = cipher.getOutputSize(data.length);
		tmpresult = new byte[size];
		olen = cipher.processBytes(data, 0, data.length, tmpresult, 0);

		if (stream != null)
		{
			stream.write(tmpresult, 0, olen);
		}
		else
		{
			result.write(tmpresult, 0, olen);
		}

		log.exiting("Encryptor", "callCipher");
	}

	/**
	 * Finishes encoding/decoding the data
	 */
	public void doFinal() throws CryptoException, IOException
	{
		log.entering("Encryptor", "doFinal");

		if (!finished)
		{
			olen = cipher.doFinal(tmpresult, olen);

			if (stream != null)
			{
				stream.write(tmpresult, 0, olen);
				stream.flush();
				stream.close();
			}
			else
			{
				if (olen > 0)
				{
					byte[] tmp = new byte[olen];
					System.arraycopy(tmpresult, 0, tmp, 0, olen);
					tmpresult = tmp;
				}
				result.write(tmpresult, 0, olen);
			}

			finished = true;
		}

		log.exiting("Encryptor", "doFinal");
	}

	/**
	 * Get an MD5 digest for a set of bytes
	 * @param phrase the phrase to calculate a digest on
	 * @return the digest as a byte array
	 * @see #getDigest(byte[], nrs.encryption.Encryptor.DIGEST_SUPPORTED)
	 */
	public static byte[] getDigest(byte[] phrase)
	{
		return getDigest(phrase, DIGEST.MD5);
	}

	public static byte[] getDigest(byte[] phrase, DIGEST digestType)
	{
		log.entering("Encryptor", "getDigest", digestType);

		Digest digest;
		switch (digestType)
		{
			case SHA1:
				digest = new SHA1Digest();
				break;

			case SHA256:
				digest = new SHA256Digest();
				break;

			case SHA512:
				digest = new SHA512Digest();
				break;

			case Tiger:
				digest = new TigerDigest();
				break;

			case Whirlpool:
				digest = new WhirlpoolDigest();
				break;

			default:
			case MD5:
				digest = new MD5Digest();
				break;
		}

		digest.update(phrase, 0, phrase.length);
		byte[] out = new byte[digest.getDigestSize()];
		digest.doFinal(out, 0);

		log.exiting("Encryptor", "getDigest");
		return out;
	}

	@Override
	public void write(byte[] b) throws IOException
	{
		log.entering("Encryptor", "write(byte[])");

		try
		{
			callCipher(b);
		}
		catch (CryptoException cex)
		{
			throw new IOException(cex);
		}
	}

	@Override
	public void write(byte[] b, int off, int len) throws IOException
	{
		log.entering("Encryptor", "write(byte[], int, int)");

		try
		{
			byte[] tmpbytes = new byte[len];
			System.arraycopy(b, off, tmpbytes, 0, len);
			callCipher(tmpbytes);
		}
		catch (CryptoException cex)
		{
			throw new IOException(cex);
		}
	}

	@Override public void write(int b) throws IOException
	{
		log.entering("Encryptor", "write(int)");

		Integer i = Integer.valueOf(b);
		byte[] by = {i.byteValue()};
		try
		{
			callCipher(by);
		}
		catch (CryptoException cex)
		{
			throw new IOException(cex);
		}
	}

	@Override
	public void flush() throws IOException
	{
	// do nothing
	}

	@Override
	public void close() throws IOException
	{
		log.entering("Encryptor", "close");

		try
		{
			doFinal();
		}
		catch (CryptoException cex)
		{
			throw new IOException(cex);
		}
		olen = size = 0;
		result = null;
	}
}
