/*
 * JCards
 * Copyright (C) 2002-2009 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package nrs.jcards;

import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.io.IOException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import nrs.jcards.fields.FieldDef;
import nrs.jcards.fields.FieldType;

/**
 * A card is one <i>page</i> of the database
 *
 * @author Copyright Copyright &#169; 2002-2009 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class Card implements Cloneable, Transferable
{
	private Map<String, Object> values = new HashMap<String, Object>();
	
	/** Create a new Card without any values */
	public Card()
	{}
	
	/**
	 * Create a new Card with the specified values
	 * @param newValues card values
	 */
	public Card(Map<String, Object> newValues)
	{
		values = newValues;
	}
	
	/**
	 * Overwrite any existing values with the specified values
	 * @param newValues new card values
	 */
	public void setValues(Map<String, Object> newValues)
	{
		values = newValues;
	}
	
	/**
	 * Add a value to this Card
	 * @param name field name
	 * @param val value of field
	 */
	public void setValue(String name, Object val)
	{
		if (val == null)
		{
			values.remove(name);
		}
		else
		{
			values.put(name, val);
		}
	}
	
	/**
	 * Get the value for the specified field
	 * @param fieldName name of the field
	 * @return the value as an Object or null if no value set
	 */
	public Object getValue(String fieldName)
	{
		return values.get(fieldName);
	}
	
	/**
	 * Get the combined values for the specified fields. Used by the Cards.quicksort method.
	 * @param firstField name of the first field
	 * @param secondField name of the second field which may be null
	 * @return the value as an Object or null if no value set
	 */
	public Object getValue(String firstField, String secondField)
	{
		String sval = null;
		Object val1 = getValue(firstField);
		if (val1 != null) {
			sval = val1.toString();
		}
		if (secondField != null)
		{
			Object val2 = getValue(secondField);
			if (val2 != null)
			{
				if (sval == null) {
					sval = val2.toString();
				} else {
					sval += val2.toString();
				}
			}
		}
		
		return sval;
	}
	
	/**
	 * Get the value as a String.
	 * Booleans are returned as Y/N.
	 * @param fieldName name of the field
	 * @return the value as a String or null if no value set
	 */
	public String getStringValue(String fieldName)
	{
		String val = null;
		
		Object obj = values.get(fieldName);
		
		if (obj != null)
		{
			DBDef def = DBDef.getInstance();
			FieldDef fd = def.getFieldDef(fieldName);
			val = fd.formatValue(obj);
		}
		
		return val;
	}
	
	/**
	 * Get all values stored by this Card
	 * @return a Map of values
	 */
	public Map<String, Object> getValues()
	{
		return values;
	}
	
	/**
	 * Return the number of values stored by this Card
	 * @return number of values
	 */
	public int getNumberOfValues()
	{
		int numValues = (values == null ? 0 : values.size());
		return numValues;
	}
	
	/**
	 * Remove a value from the card
	 * @param fieldName field name to remove
	 */
	public void removeValue(String fieldName)
	{
		values.remove(fieldName);
	}
	
	/**
	 * Change a field name
	 * @param oldName the existing field name
	 * @param newName the new field name
	 */
	public void changeFieldName(String oldName, String newName)
	{
		Object val = values.remove(oldName);
		if (val != null)
		{
			values.put(newName, val);
		}
	}
	
	/**
	 * Change a field type. If the new and old field types are the same don't
	 * do anything. The new type can only be String or Memo so the value will
	 * be stored as a String.
	 * @param fieldName field name
	 * @param oldType old field type
	 * @param newType new field type
	 */
	public void changeFieldType(String fieldName, FieldType oldType, FieldType newType)
	{
		if (oldType == newType)
		{
			return;
		}
		
		String val = getStringValue(fieldName);
		if (oldType == FieldType.MEMO)
		{
			val = val.replace('\n', ' ');
		}
		values.put(fieldName, val);
	}
	
	/**
	 * Get a String representation of this object suitable
	 * for saving to a file
	 * @return String version of this class
	 */
	@Override
	public String toString()
	{
		StringBuilder sb = new StringBuilder();
		String fieldName;
		
		Iterator it = values.keySet().iterator();
		while (it.hasNext())
		{
			fieldName = it.next().toString();
			sb.append(fieldName);
			sb.append("=");
			sb.append(values.get(fieldName));
			sb.append("\n");
		}
		
		return sb.toString();
	}
	
	/** Get Card data as text */
	@Override public Object getTransferData(DataFlavor flavor) throws UnsupportedFlavorException, IOException
	{
		if (!flavor.equals(DataFlavor.stringFlavor)) {
			throw new UnsupportedFlavorException(flavor);
		}
		
		return toString();
	}
	
	/** Get DataFlavour[] for clipboard transfer */
	@Override public DataFlavor[] getTransferDataFlavors()
	{
		DataFlavor[] dfs = { DataFlavor.stringFlavor };
		return dfs;
	}
	
	/** Check to see if a DataFlavor is supported */
	@Override public boolean isDataFlavorSupported(DataFlavor flavor)
	{
		return ((flavor.equals(DataFlavor.stringFlavor)) ? true : false);
	}
	
	/** Clone this object */
	@Override
	protected Object clone() throws CloneNotSupportedException
	{
		return super.clone();
	}
}
