/*
 * JCards
 * Copyright (C) 2002-2009 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package nrs.jcards;

import java.io.IOException;
import java.io.OutputStream;
import java.io.Writer;
import java.util.*;
import java.util.logging.Level;
import java.util.logging.Logger;
import nrs.jcards.fields.*;
import nrs.jcards.io.Elements;
import nrs.jcards.io.JCAReader;
import org.jdom.Element;
import org.jdom.output.XMLOutputter;

/**
 * Class to reference all the cards in a database.
 * This class is observed by the TableHeader and observes
 * the DBDef.
 *
 * @author Copyright &#169; 2002-2009 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class Cards extends Observable implements Observer
{
	private final static Logger log = Logger.getLogger(Cards.class.getName());
	
	private String info;
	private List<Card> cards = new ArrayList<Card>();
	private List<Card> foundCards;
	private Card deletedCard;
	private Card rowCard;
	private DBDef dbDef = DBDef.getInstance();
	private int oldRowIndex = -1;
	private boolean replacedText;
	
	private final static String CARDS_START = "<"+Elements.CARDS+">\n";
	private final static String CARDS_END = "</"+Elements.CARDS+">\n";
	
	/**
	 * Set some information about the cards. For example, "Music Database"
	 * @param newInfo the comment/note/info
	 */
	public void setInfo(String newInfo)
	{
		if (newInfo != null) {
			newInfo = newInfo.trim();
			if (newInfo.length() == 0) {
				newInfo = null;
			}
		}
		info = newInfo;
		
		setChanged();
		int set = (newInfo == null ? 0 : 1);
		notifyObservers(new CardUpdate(CardUpdate.Event.SET_INFO, set, cards.size(), 0));
	}
	
	/**
	 * Get any info that has been previously set
	 * @return info
	 */
	public String getInfo()
	{
		return info;
	}
	
	/**
	 * Add a card without informing the listeners
	 * @param card the Card to add
	 */
	public void addCardSilently(Card card)
	{
		log.finest("Adding card");
		cards.add(card);
	}
	
	/**
	 * This method should be called to indicate that either a file or CSV
	 * read has finished and that the listeners can be informed.
	 * @see addCSVCard(String, char, char)
	 */
	public void finishedLoad()
	{
		log.entering("Cards", "finishedLoad");
		setChanged();
		notifyObservers(new CardUpdate(CardUpdate.Event.LOAD, -1, cards.size(), 0));
		log.exiting("Cards", "finishedLoad");
	}
	
	/**
	 * Add a card to the list.
	 * @param newCard the Card to add
	 * @return the index of the new card
	 */
	public int addCard(Card newCard)
	{
		log.entering("Cards", "addCard", newCard);
		
		cards.add(newCard);
		
		// add first field as visible
		List<String> visFields = dbDef.getVisibleFields();
		if (visFields.isEmpty())
		{
			FieldDef fd = dbDef.getFieldDef(0);
			visFields.add(fd.getName());
		}
		int index = cards.indexOf(newCard);
		
		foundCards = null;
		
		setChanged();
		notifyObservers(new CardUpdate(CardUpdate.Event.CARD_NEW, index, cards.size(), 0));
		
		log.exiting("Cards", "addCard", index);
		
		return index;
	}
	
	/**
	 * Delete a card
	 * @param index index of the card
	 * @return the deleted card
	 */
	public void deleteCard(int index)
	{
		log.entering("Cards", "deleteCard", index);
		
		List<Card> listCards = (foundCards == null ? cards : foundCards);
		deletedCard = listCards.get(index);
		cards.remove(deletedCard);
		int foundCardsCount = 0;
		if (foundCards != null)
		{
			foundCards.remove(deletedCard);
			if (foundCards.isEmpty()) {
				resetFind();
			} else {
				foundCardsCount = foundCards.size();
			}
		}
		
		log.finer("Notifying observers");
		setChanged();
		notifyObservers(new CardUpdate(CardUpdate.Event.CARD_DELETE, index, cards.size(), foundCardsCount));
		
		log.exiting("Cards", "deletedCard");
	}
	
	/** Restore a previously deleted card */
	public void undeleteCard()
	{
		log.entering("Cards", "undeleteCard");
		
		if (deletedCard != null)
		{
			log.finer("Undeleting card");
			
			cards.add(deletedCard);
			int index = cards.indexOf(deletedCard);
			deletedCard = null;
			int foundCardCount = 0;
			if (foundCards != null) {
				foundCards.add(deletedCard);
				foundCardCount = foundCards.size();
			}
			
			log.finer("Notify observers");
			setChanged();
			notifyObservers(new CardUpdate(CardUpdate.Event.CARD_NEW, index, cards.size(), foundCardCount));
		}
		
		log.exiting("Cards", "undeleteCard");
	}
	
	/**
	 * Get a Card
	 * @param index the index of the card to get
	 * @return the requested card or null if it doesn't exist
	 */
	public Card getCard(int index)
	{
		List<Card> cardList = (foundCards == null ? cards : foundCards);
		assert index > -1 : "Cannot getCard index="+index;
		return cardList.get(index);
	}
	
	/**
	 * Update a card definition
	 * @param index index of the card
	 * @param newCard new new Card definition
	 * @return the index of the updated card
	 */
	public int setCard(int index, Card newCard)
	{
		log.entering("Cards", "setCard", index);
		
		// just update the values referenced by the card in the right place
		List<Card> cardList = (foundCards == null ? cards : foundCards);
		Card c = cardList.get(index);
		c.setValues(newCard.getValues());
		
		int foundCardCount = (foundCards == null ? 0 : foundCards.size());
		
		log.finer("Notify observers");
		setChanged();
		notifyObservers(new CardUpdate(CardUpdate.Event.CARD_UPDATE, index, cards.size(), foundCardCount));
		
		log.exiting("Cards", "setCard");
		
		return cards.indexOf(c);
	}
	
	/**
	 * The underlying database definition has changed.
	 */
	@Override public void update(Observable o, Object arg)
	{
		log.entering("Cards", "update");
		
		assert arg != null : "Cards update arg is null";
		
		DBDefUpdate update = (DBDefUpdate)arg;
		log.log(Level.FINEST, "Event = {0}", update.getEvent());
		switch (update.getEvent())
		{
			case FIELD_DELETE:
				log.finer("Removing field from cards");
				// remove value for field from cards
				for (Card card : cards)
				{
					card.removeValue(update.getOldName());
				}
				break;
				
			case FIELD_CHANGE:
				log.finer("Changing fields in cards");
				FieldDef oldFD = (FieldDef)update.getOldObject();
				FieldDef newFD = (FieldDef)update.getNewObject();
				for (Card card : cards)
				{
					card.changeFieldName(oldFD.getName(), newFD.getName());
					if (oldFD.getType() != newFD.getType())
					{
						card.changeFieldType(newFD.getName(), oldFD.getType(), newFD.getType());
					}
				}
				break;
				
			case CATEGORY_ITEM_RENAMED:
				log.finer("Changing category item");
				String catName = update.getNewName();
				String newCatItem = (String)update.getNewObject();
				FieldDef fd;
				CategoryFieldDef cfd;
				for (String fieldName : dbDef.getFieldNames())
				{
					fd = dbDef.getFieldDef(fieldName);
					if (fd.getType() == FieldType.CATEGORY)
					{
						cfd = (CategoryFieldDef)fd;
						if (cfd.getCategory().equals(catName))
						{
							for (Card card : cards)
							{
								card.setValue(fieldName, newCatItem);
							}
						}
					}
				}
				break;
		}
		
		log.exiting("Cards", "update");
	}
	
	/**
	 * Get the data as XML. Empty fields will be omitted.
	 * @see #getXML(boolean, boolean)
	 */
	public Element getXML()
	{
		return getXML(false, false);
	}
	
	/**
	 * Get the data as XML
	 * @param includeEmptyFields if true empty fields will be included but with no value;
	 * otherwise, the field will be omitted
	 * @param onlyIncludeVisibleFields if true only the displayed fields will be included; otherwise
	 * all fields will be included
	 * @return the data as a XMLElement
	 */
	public Element getXML(boolean includeEmptyFields, boolean onlyIncludeVisibleFields)
	{
		log.entering("Cards", "getXML");
		
		Element cardsEl = new Element(Elements.CARDS);
		
		if (info != null)
		{
			Element infoTag = new Element(Elements.INFO);
			infoTag.setText(info);
			cardsEl.addContent(infoTag);
		}
		
		for (Card c : cards)
		{
			cardsEl.addContent(getCardXML(c, includeEmptyFields, onlyIncludeVisibleFields, false));
		}
		
		log.exiting("Cards", "getXML");
		
		return cardsEl;
	}
	
	/**
	 * Iterate over the Card objects and write them to the specified output stream
	 * @param xmlOut XMLOutputter that may have been configured
	 * @param os the output stream to write to
	 * @param includeEmptyFields
	 * @param onlyIncludeVisibleFields
	 * @param includeComments
	 * @throws IOException
	 * @see #writeCardsToWriter(XMLOutputter, Writer, boolean, boolean)
	 */
	public void writeCardsToOutputStream(XMLOutputter xmlOut, OutputStream os, boolean includeEmptyFields,
		boolean onlyIncludeVisibleFields, boolean includeComments) throws IOException
	{
		log.entering("Cards", "writeCardsToOutputStream");
		
		// An element isn't created for the top-level Cards as you'd have to add all the
		// Card objects to it to get it to output correctly. If there's a lot of Card objects
		// this could take up a lot of memory.
		os.write(CARDS_START.getBytes(JCAReader.UTF8));
		
		if (includeComments && info != null)
		{
			Element infoEl = new Element(Elements.INFO);
			infoEl.setText(info);
			xmlOut.output(infoEl, os);
			os.write("\n".getBytes());
		}
		
		// output each card
		for (Card c : cards)
		{
			xmlOut.output(getCardXML(c, includeEmptyFields, onlyIncludeVisibleFields, true), os);
			os.write("\n".getBytes());
		}
		
		os.write(CARDS_END.getBytes("UTF-8"));
		
		log.exiting("Cards", "writeCardsToOutputStream");
	}

	/**
	 * Get an XML representation of each card
	 * @param card
	 * @param includeEmptyFields
	 * @param onlyIncludeVisibleFields
	 * @param format
	 * @return
	 */
	private Element getCardXML(Card card, boolean includeEmptyFields, boolean onlyIncludeVisibleFields, boolean format)
	{
		String fieldValue;
		Element fieldEl;
		Object objValue;
		FieldDef fieldDef;
		Map<String, Object> cardValues = card.getValues();
		
		Element cardEl = new Element(Elements.CARD);
		
		List<String> fields = (onlyIncludeVisibleFields ? dbDef.getVisibleFields() : dbDef.getFieldNames());
		for (String fieldName : fields)
		{
			objValue = cardValues.get(fieldName);
			
			if (objValue != null || includeEmptyFields)
			{
				fieldEl = new Element(Elements.FIELD);
				fieldEl.setAttribute(Elements.NAME, fieldName);
				
				if (objValue != null)
				{
					fieldDef = dbDef.getFieldDef(fieldName);
					if (format || fieldDef instanceof DateFieldDef) {
						fieldValue = fieldDef.formatValue(objValue);
					} else {
						fieldValue = objValue.toString().trim();
					}
					fieldEl.setText(fieldValue);
				}
				cardEl.addContent(fieldEl);
			}
		}
		
		return cardEl;
	}
	
	/**
	 * Iterate over the Card objects and write them to the specified writer
	 * @param xmlOut XMLOutputter that may have been configured
	 * @param writer the write to write to
	 * @param includeEmptyFields
	 * @param onlyIncludeVisibleFields
	 * @throws java.io.IOException
	 * @see #writeCardsToOutputStream(XMLOutputter, OutputStream, boolean, boolean)
	 */
	public void writeCardsToWriter(XMLOutputter xmlOut, Writer writer, boolean includeEmptyFields, boolean onlyIncludeVisibleFields) throws IOException
	{
		log.entering("Cards", "writeCardsToWriter");
		
		writer.write(CARDS_START);
		
		if (info != null)
		{
			Element infoEl = new Element(Elements.INFO);
			infoEl.setText(info);
			xmlOut.output(infoEl, writer);
		}
		
		for (Card c : cards)
		{
			xmlOut.output(getCardXML(c, includeEmptyFields, onlyIncludeVisibleFields, false), writer);
		}
		writer.write(CARDS_END);
		
		log.exiting("Cards", "writeCardsToWriter");
	}
	
	/**
	 * Get the data in CSV format
	 * @param fieldDelim field delimiter
	 * @param textDelim text delimiter
	 * @param concatNL if true new lines are removed from the text
	 * @return CSV data
	 */
	public String getCSV(String fieldDelim, String textDelim, boolean concatNL)
	{
		log.entering("Cards", "getCSV");
		
		StringBuilder sb = new StringBuilder();
		Card c;
		Iterator cardIt, fieldIt;
		Object val;
		String field, sval;
		Date dval;
		
		cardIt = cards.iterator();
		while (cardIt.hasNext())
		{
			c = (Card)cardIt.next();
			fieldIt = dbDef.getFieldNames().iterator();
			while (fieldIt.hasNext())
			{
				field = (String)fieldIt.next();
				val = c.getValue(field);
				if (val == null)
				{
					sb.append(textDelim);
					sb.append(textDelim);
				}
				else
				{
					// remove newlines
					if (val instanceof String)
					{
						sval = textDelim+(String)val+textDelim;
						if (concatNL)
						{
							sval = sval.replace('\n', ' ');
						}
					}
					else if (val instanceof Date)
					{
						dval = (Date)val;
						DateFieldDef dfd = (DateFieldDef)dbDef.getFieldDef(field);
						sval = dfd.formatDate(dval);
					}
					else
					{
						sval = val.toString();
					}
					
					sb.append(sval);
				}
				sb.append(fieldDelim);
			}
			// remove last comma
			sb.setLength(sb.length()-1);
			sb.append("\n");
		}
		// remove last new line
		sb.setLength(sb.length()-1);
		
		log.exiting("Cards", "getCSV");
		
		return sb.toString();
	}
	
	/**
	 * See if the search criteria match a field
	 * @param fieldText text value of the field
	 * @param frDetails search criteria
	 * @return true if a match is found; false otherwise
	 */
	private boolean matchField(Card card, String fieldName, String fieldText, FindReplaceDetails frDetails)
	{
		boolean matched = false;
		String lowerFieldText = fieldText.toLowerCase();
		String lowerSearchText = frDetails.searchText.toLowerCase();
		int startIndex = -1;
		
		if ((startIndex = lowerFieldText.indexOf(lowerSearchText)) > -1)
		{
			matched = true;
			if (frDetails.matchCase)
			{
				matched = false;
				startIndex = -1;
				if ((startIndex = fieldText.indexOf(frDetails.searchText)) > -1)
				{
					matched = true;
				}
			}
			
			if (matched && frDetails.wholeWords)
			{
				matched = false;
				startIndex = -1;
				
				// see if the user meant to match the entire string or individual words
				if (lowerSearchText.equals(lowerFieldText))
				{
					matched = true;
					startIndex = 0;
				}
				else
				{
					int tokenIndex = 0;
					String[] ltokens = lowerFieldText.split("\\s");
					for (int i=0; i<ltokens.length; i++)
					{
						if (lowerSearchText.equals(ltokens[i]))
						{
							matched = true;
							// set approx position as we don't know how many whitespace
							// characters there are
							startIndex = tokenIndex;
							break;
						}
						tokenIndex += ltokens[i].length()+1;
					}
				}
			}
		}
		
		// handle a replace if necessary
		if (matched && frDetails.replaceOp && startIndex > -1)
		{
			// have to set it again as whole word search is imprecise
			startIndex = lowerFieldText.indexOf(lowerSearchText, startIndex);
			assert startIndex > -1 : "matchField cannot find search text";
			if (startIndex > -1)
			{
				StringBuilder sb = new StringBuilder(fieldText.length());
				sb.append(fieldText.substring(0, startIndex));
				sb.append(frDetails.replaceText);
				sb.append(fieldText.substring(startIndex+frDetails.searchText.length()));
				card.setValue(fieldName, sb.toString());
				replacedText = true;
			}
		}
		
		return matched;
	}
	
	/** Checks to see if a particular card matches the find/replace details */
	private boolean matchCard(Card c, FindReplaceDetails frDetails)
	{
		Object fieldValue = null;
		String fieldText = null;
		boolean matched = false;
		
		if (frDetails.field == null) // match all fields
		{
			Map<String, Object> fhash = c.getValues();
			
			for (String fieldName : fhash.keySet())
			{
				fieldValue = fhash.get(fieldName);
				if (fieldValue != null)
				{
					fieldText = fieldValue.toString();
					matched = matchField(c, fieldName, fieldText, frDetails);
					if (matched)
					{
						break;
					}
				}
			}
		}
		else
		{
			fieldValue = c.getValue(frDetails.field);
			if (fieldValue != null)
			{
				fieldText = fieldValue.toString();
				matched = matchField(c, frDetails.field, fieldText, frDetails);
			}
		}
		
		return matched;
	}
	
	/**
	 * Search for matching cards
	 * @param frDetails find/replace details
	 * @return number of cards found or modified
	 */
	public int findCards(FindReplaceDetails frDetails)
	{
		log.entering("Cards", "findCards", frDetails);
		
		Card c;
		Iterator cit;
		List<Card> findCards = new ArrayList<Card>();
		replacedText = false;
		
		assert frDetails.searchText != null : "Cannot findCards - search is null";
		
		if (frDetails.findInResults && foundCards != null && !foundCards.isEmpty())
		{
			cit = foundCards.iterator();
		}
		else
		{
			cit = cards.iterator();
		}
		
		// iterate through the cards to find matches
		while (cit.hasNext())
		{
			c = (Card)cit.next();
			if (matchCard(c, frDetails))
			{
				findCards.add(c);
			}
		}
		
		// only change the display if some cards were found
		if (!findCards.isEmpty())
		{
			setChanged();
			if (frDetails.replaceText == null)
			{
				foundCards = findCards;
				notifyObservers(new CardUpdate(CardUpdate.Event.FIND, -1, cards.size(), foundCards.size()));
			}
			else if (replacedText)
			{
				notifyObservers(new CardUpdate(CardUpdate.Event.REPLACE, -1, cards.size(), 0));
			}
		}
		else
		{
			foundCards = null;
		}
		
		log.exiting("Cards", "findCards");
		
		return findCards.size();
	}
	
	/**
	 * Reset the search results
	 */
	public void resetFind()
	{
		log.entering("Cards", "resetFind");
		
		if (foundCards != null)
		{
			foundCards = null;
			setChanged();
			notifyObservers(new CardUpdate(CardUpdate.Event.FIND, -1, cards.size(), 0));
		}
		
		log.exiting("Cards", "resetFind");
	}
	
	/**
	 * Return the number of cards
	 * @return number of cards
	 */
	public int size()
	{
		return cards.size();
	}
	
	/** Clear existing data */
	public void reset()
	{
		log.entering("Cards", "reset");
		
		cards.clear();
		foundCards = null;
		deletedCard = null;
		info = null;
		
		log.finer("Notify observers");
		setChanged();
		notifyObservers(new CardUpdate(CardUpdate.Event.LOAD, -1, cards.size(), 0));
		
		log.exiting("Cards", "reset");
	}
	
	/**
	 * Get date statistics for the specified field
	 * @param fieldName field to collect statistics on
	 * @return statistics as an array
	 */
	public String getDateStatistics(String fieldName)
	{
		log.entering("Cards", "getDateStatistics");
		
		Card card;
		Date date, earliest=new Date(), latest=new Date(), average;
		
		date = (Date)getCard(0).getValue(fieldName);
		earliest = date;
		latest = date;
		average = date;
		
		for (int i=0; i<cards.size(); i++)
		{
			card = getCard(i);
			date = (Date)card.getValue(fieldName);
			
			if (date != null)
			{
				if (earliest == null)
				{
					earliest = date;
					latest = date;
				}
				if (date.before(earliest))
				{
					earliest = date;
				}
				else if (date.after(latest))
				{
					latest = date;
				}
			}
		}

		StringBuilder stats = new StringBuilder();
		// check for no data
		if (earliest != null)
		{
			// calculate median date
			long diff = latest.getTime() - earliest.getTime();
			if (diff > 0)
			{
				diff /= 2;
				average = new Date(earliest.getTime() + diff);
			}
			else
			{
				average = earliest;
			}

			DateFieldDef dfd = (DateFieldDef)DBDef.getInstance().getFieldDef(fieldName);
			stats.append(JCards.getFormattedMessage("stats.earliest", dfd.formatDate(earliest)));
			stats.append("\n");
			stats.append(JCards.getFormattedMessage("stats.latest", dfd.formatDate(latest)));
			stats.append("\n");
			stats.append(JCards.getFormattedMessage("stats.median", dfd.formatDate(average)));
		}
		
		log.exiting("Cards", "getDateStatistics", stats);
		
		return stats.toString();
	}
	
	/**
	 * Get numeric statistics for the specified field
	 * @param fieldName field to collect statistics on
	 * @return statistics as a String
	 */
	public String getNumericStatistics(String fieldName)
	{
		log.entering("Cards", "getNumericStatistics", fieldName);
		
		Card card;
		Number num;
		double val, total = 0, average;
		double min=Double.MAX_VALUE, max=Double.MIN_VALUE;
		int dataCount = 0;
		
		num = (Number)getCard(0).getValue(fieldName);
		if (num != null)
		{
			min = max = num.doubleValue();
		}
		
		for (int i=0; i<cards.size(); i++)
		{
			card = getCard(i);
			num = (Number)card.getValue(fieldName);
			
			if (num != null)
			{
				val = num.doubleValue();
				if (val < min)
				{
					min = val;
				}
				else if (val > max)
				{
					max = val;
				}
				
				total += val;
				dataCount++;
			}
		}
		
		StringBuilder stats = new StringBuilder();
		// check for no data
		if (dataCount > 0)
		{		
			average = total / dataCount;


			stats.append(JCards.getFormattedMessage("stats.min", new Double(min)));
			stats.append("\n");
			stats.append(JCards.getFormattedMessage("stats.max", new Double(max)));
			stats.append("\n");
			stats.append(JCards.getFormattedMessage("stats.total", new Double(total)));
			stats.append("\n");
			stats.append(JCards.getFormattedMessage("stats.avg", new Double(average)));
		}
		
		log.exiting("Cards", "getNumericStatistics", stats);
		
		return stats.toString();
	}
	
	/** Return an Iterator over the Card objects */
	public Iterator<Card> iterator()
	{
		return cards.iterator();
	}
	
	/**
	 * Get the value of a card
	 * @param rowIndex row
	 * @param columnIndex column
	 */
	public Object getValueAt(int rowIndex, int columnIndex)
	{
		// cache the Card for the row to speed up subsequent calls
		// when accessing values for the table display
		if (rowIndex != oldRowIndex)
		{
			List cardList = (foundCards == null ? cards : foundCards);
			rowCard = (Card)cardList.get(rowIndex);
		}
		oldRowIndex = rowIndex;
		FieldDef fd = dbDef.getVisibleFieldDef(columnIndex);
		assert fd != null : "Cards.getValueAt cannot get FieldDef for "+rowIndex+","+columnIndex;
		return rowCard.getValue(fd.getName());
	}
	
	/**
	 * Return the number of cards depending on whether a search has occurred
	 * @return number of cards
	 */
	public int getRowCount()
	{
		List cardList = (foundCards == null ? cards : foundCards);
		return cardList.size();
	}
	
	/**
	 * Return the number of visible fields
	 * @return number of visible fields
	 */
	public int getColumnCount()
	{
		int colCount = 0;
		
		List visFields = DBDef.getInstance().getVisibleFields();
		if (visFields != null) {
			colCount = visFields.size();
		}
		
		return colCount;
	}
	
	/**
	 * Is the cards object referencing all cards or just the found ones
	 * @return true if referencing found cards
	 */
	public boolean isFindSubset()
	{
		return foundCards != null;
	}

	/**
	 * Set image file names before saving to JCards DB
	 * @param fieldName field being modified. Should only apply to image fields set to be stored in DB
	 * @return Map between JCards DB ref and system path. System path may be null if image is already in DB.
	 */
	public Map<String, String> setDBImageFileNamesBeforeSave(String fieldName)
	{
		log.entering("Cards", "setImageFileNamesBeforeSave", fieldName);

		Map<String, String> images = new HashMap<String, String>();
		String path, ref;
		long fname;
		for (Card card : cards)
		{
			path = (String)card.getValue(fieldName);
			if (path != null)
			{
				if (path.startsWith(ImageFieldDef.PIC)) {
					images.put(path, null);
				}
				else
				{
					// convert to internal format reference
					JCAReader.SplitFilename sf = JCAReader.splitFilename(path);
					String ext = null;
					if (sf.extension != null) {
						ext = "." + sf.extension;
					}
					fname = new Date().getTime() + (long)(Math.random() * 1000);
					ref = ImageFieldDef.PIC+fname+ext;

					card.setValue(fieldName, ref);
					images.put(ref, path);
				}
			}
		}

		log.exiting("Cards", "setImageFileNamesBeforeSave", images.size());

		return images;
	}

	/**
	 * Remove internal DB references from image fields that are no longer set to store in DB
	 * @param fieldName
	 */
	public void removeImageDBFileRefs(String fieldName)
	{
		log.entering("Cards", "removeImageDBFileRefs", fieldName);

		String path;
		for (Card card : cards)
		{
			path = (String)card.getValue(fieldName);
			if (path != null && path.startsWith(ImageFieldDef.PIC)) {
				card.removeValue(fieldName);
			}
		}

		log.exiting("Cards", "removeImageDBFileRefs");
	}
}
