/*
 * JCards
 * Copyright (C) 2002-2009 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package nrs.jcards;

import java.util.*;
import java.util.logging.Level;
import java.util.logging.Logger;
import nrs.jcards.fields.*;
import nrs.util.PostfixCalculator;

/**
 * Contains the definition of the fields in the database and preferences.
 * @todo shouldn't be able to re-use the same counter
 * @author Copyright &#169; 2002-2009 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public final class DBDef extends Observable
{
	private final static Logger log = Logger.getLogger(DBDef.class.getName());
	
	private static DBDef instance;
	private Map<String, FieldDef> fieldDefs = new LinkedHashMap<String, FieldDef>();
	private List<String> fieldOrder = new ArrayList<String>();
	private List<String> visibleFields = new ArrayList<String>();
	private Map<String, List<String>> categories = new HashMap<String, List<String>>();
	private Map<String, Counter> counters = new HashMap<String, Counter>();
	private boolean displayFieldNames = true;
	private boolean displayColons = true;
	private static PostfixCalculator calculator;
	/** Used to denote a field in a calculation */
	private static final String FIELD_SEPARATOR = "'";
	
	private DBDef()
	{}
	
	/**
	 * Singleton patten for getting only one instance of DBDef.
	 * It isn't synchronised as there is no threading in JCards.
	 */
	public static DBDef getInstance()
	{
		if (instance == null)
		{
			instance = new DBDef();
		}
		
		return instance;
	}
	
	/**
	 * Add a field definition to the database. Field names must be unique.
	 * @param newFieldDef the field definition to add
	 */
	public boolean addFieldDef(FieldDef newFD)
	{
		log.entering("DBDef", "addFieldDef", newFD);

		boolean addedFieldDef = false;
		
		// check to see if the field name already exists
		FieldDef fd = fieldDefs.get(newFD.getName());
		if (fd != null)
		{
			log.fine("Field definition already exists");
			addedFieldDef = false;
		}
		else
		{
			// first field is always mandatory
			if (fieldDefs.isEmpty())
			{
				newFD.setMandatory(true);
				log.fine("First field, so making mandatory");
			}

			fieldDefs.put(newFD.getName(), newFD);
			fieldOrder.add(newFD.getName());

			log.finer("Notifying observers");
			setChanged();
			DBDefUpdate update = new DBDefUpdate(DBDefUpdate.Event.FIELD_ADD);
			update.setNewObject(newFD);
			update.setNewName(newFD.getName());
			update.setNewSize(fieldDefs.size());
			notifyObservers(update);

			// update the visible fields if empty
			if (visibleFields.isEmpty())
			{
				log.fine("Adding to visible fields");
				visibleFields.add(newFD.getName());
				setChanged();
				update = new DBDefUpdate(DBDefUpdate.Event.VISIBLE_FIELDS);
				update.setNewName(newFD.getName());
				notifyObservers(update);
			}
			addedFieldDef = true;
		}

		log.exiting("DBDef", "addFieldDef", addedFieldDef);
		
		return addedFieldDef;
	}
	
	/**
	 * Get the number of fields
	 */
	public int getNumberOfFields()
	{
		return fieldDefs.size();
	}
	
	/**
	 * Remove the specified field def
	 * @param name name of the field to remove
	 */
	public void removeFieldDef(String name)
	{
		log.entering("DBDef", "removeFieldDef", name);
		
		FieldDef fd = fieldDefs.remove(name);
		if (fd != null)
		{
			int index = fieldOrder.indexOf(name);
			if (fieldOrder.remove(name))
			{
				setChanged();
				DBDefUpdate update;
				if (visibleFields.remove(name))
				{
					update = new DBDefUpdate(DBDefUpdate.Event.VISIBLE_FIELDS);
					update.setOldName(name);
					notifyObservers(update);
				}
				update = new DBDefUpdate(DBDefUpdate.Event.FIELD_DELETE);
				update.setOldObject(fd);
				update.setOldName(fd.getName());
				update.setOldIndex(index);
				notifyObservers(update);
			}
		}
		
		log.exiting("DBDef", "removeFieldDef");
	}
	
	/**
	 * Get a FieldDef based on its position
	 * @param fieldNum field position
	 * @return the FieldDef
	 */
	public FieldDef getFieldDef(int fieldNum)
	{
		assert fieldNum < fieldOrder.size() : "fieldNum "+fieldNum+" is greater than fields defined";
		String fieldName = fieldOrder.get(fieldNum);
		return fieldDefs.get(fieldName);
	}
	
	/**
	 * Get a FieldDef based on its position within the visible fields
	 * @param fieldNum visible field index
	 * @return FieldDef
	 */
	public FieldDef getVisibleFieldDef(int fieldNum)
	{
		assert fieldNum > -1 : "getVisibleFieldDef: Incorrect index specified -1";
		String name = visibleFields.get(fieldNum);
		return getFieldDef(name);
	}
	
	/**
	 * Get a FieldDef based on its name
	 * @param fieldName field name
	 * @return the FieldDef or null if it doesn't exist
	 */
	public FieldDef getFieldDef(String fieldName)
	{
		return fieldDefs.get(fieldName);
	}
	
	/**
	 * Set the field definitions
	 * @param fds Vector of FieldDef objects
	 */
	public void setFieldDefs(Map<String, FieldDef> fds)
	{
		log.entering("DBDef", "setFieldDefs", fds);
		// copy into new Map to prevent inadvertent changes to the original list
		assert fds != null : "FieldDefs are null!";
		fieldDefs = new HashMap<String, FieldDef>(fds);
		fieldOrder.addAll(fieldDefs.keySet());
		log.exiting("DBDef", "setFieldDefs");
	}
	
	/**
	 * Move a fields position in the list.
	 * @param index current position of the field
	 * @param incr whether to move the field up or down
	 * @param the new position of the field
	 * @return the new position
	 */
	public int moveField(int index, int incr)
	{
		int pos = 0;
		String name;
		
		log.entering("DBDef", "moveField", index);
		
		// start or end of list
		if (index == 0 && incr == -1)
		{
			pos = 0;
		}
		else if (incr == 1 && ((index+1) == fieldDefs.size()))
		{
			pos = fieldDefs.size();
		}
		else // middle somewhere
		{
			name = fieldOrder.remove(index);

			if ((index+incr) > fieldDefs.size())
			{
				fieldOrder.add(index+incr, name);
				pos = fieldDefs.size()-1;
			}
			else
			{
				fieldOrder.add(index+incr, name);
				// first field is always mandatory
				if ((index+incr) == 0)
				{
					FieldDef fd = getFieldDef(name);
					fd.setMandatory(true);
				}
				pos = index + incr;
			}

			setChanged();
			DBDefUpdate update = new DBDefUpdate(DBDefUpdate.Event.FIELD_MOVE);
			update.setOldIndex(index);
			update.setNewIndex(pos);
			notifyObservers(update);
		}
		
		log.exiting("DBDef", "moveField", pos);
		
		return pos;
	}
	
	/**
	 * Replace a field definition
	 * @param index position of the field
	 * @param newFieldDef the new field definition
	 */
	public void replaceFieldDef(int index, FieldDef newFD)
	{
		log.entering("DBDef", "replaceFieldDef", index);
		
		String oldName = fieldOrder.remove(index);
		FieldDef oldFieldDef = fieldDefs.remove(oldName);
		
		fieldDefs.put(newFD.getName(), newFD);
		fieldOrder.add(index, newFD.getName());
		
		int visIndex = visibleFields.indexOf(oldName);
		if (visIndex > -1)
		{
			visibleFields.remove(visIndex);
			visibleFields.add(visIndex, newFD.getName());
			setChanged();
			DBDefUpdate update = new DBDefUpdate(DBDefUpdate.Event.VISIBLE_FIELDS);
			update.setOldName(oldName);
			update.setNewName(newFD.getName());
			notifyObservers(update);
		}
		
		setChanged();
		DBDefUpdate update = new DBDefUpdate(DBDefUpdate.Event.FIELD_CHANGE);
		update.setOldIndex(index);
		update.setOldObject(oldFieldDef);
		update.setNewIndex(index);
		update.setNewObject(newFD);
		update.setNewName(newFD.getName());
		notifyObservers(update);
		
		log.exiting("DBDef", "replaceFieldDef");
	}
	
	/**
	 * Add/modify a category
	 * @param catName category name
	 * @param catItems List of items
	 */
	public void setCategory(String catName, List<String> catItems)
	{
		log.entering("DBDef", "setCategory");
		// copy into new List to remove possibility of inadvertent changes to the original list
		categories.put(catName, catItems);
		setChanged();
		DBDefUpdate update = new DBDefUpdate(DBDefUpdate.Event.CATEGORY_UPDATED);
		update.setNewName(catName);
		update.setNewSize(catItems.size());
		update.setNewObject(catItems);
		notifyObservers(update);
		log.exiting("DBDef", "setCategory");
	}
	
	/**
	 * Get a Category or null if it doesn't exist
	 * @param catName Category name
	 */
	public List<String> getCategory(String catName)
	{
		return categories.get(catName);
	}
	
	/**
	 * Return a list of all categories
	 * @return the Map of categories
	 */
	public Map<String, List<String>> getCategories()
	{
		return categories;
	}
	
	/** Return true if categories are defined */
	public boolean hasCategories()
	{
		return !categories.isEmpty();
	}
	
	/**
	 * Set the categories
	 * @param cats Map of categories
	 */
	public void setCategories(Map<String, List<String>> cats)
	{
		log.entering("DBDef", "setCategories");
		// copy into new Map to prevent inadvertent changes to the original list
		assert cats != null : "Categories are null";
		categories = new HashMap<String, List<String>>(cats);
		setChanged();
		notifyObservers(new DBDefUpdate(DBDefUpdate.Event.CATEGORY_UPDATED));
		log.exiting("DBDef", "setCategories");
	}
	
	/**
	 * Return a list of counters
	 * @return the Hashtable of counters
	 */
	public Map<String, Counter> getCounters()
	{
		return counters;
	}
	
	/** Return true if counters are defined */
	public boolean hasCounters()
	{
		return !counters.isEmpty();
	}
	
	/**
	 * Add a counter
	 * @param counter new Counter
	 */
	public void addCounter(Counter counter)
	{
		assert counter != null : "Counter cannot be null";
		
		log.entering("DBDef", "addCounter", counter);
		counters.put(counter.getName(), counter);
		setChanged();
		DBDefUpdate update = new DBDefUpdate(DBDefUpdate.Event.COUNTER_ADDED);
		update.setNewName(counter.getName());
		update.setNewObject(counter);
		update.setNewSize(counters.size());
		notifyObservers(update);
		log.exiting("DBDef", "addCounter");
	}
	
	/**
	 * Set the counters
	 * @param counts Map of counters
	 */
	public void setCounters(Map<String, Counter> counts)
	{
		log.entering("DBDef", "setCounters");
		// copy into new Map to prevent inadvertent changes
		assert counts != null : "Counters is null";
		counters = new HashMap<String, Counter>(counts);
		setChanged();
		DBDefUpdate update = new DBDefUpdate(DBDefUpdate.Event.COUNTER_UPDATED);
		update.setNewObject(counts);
		notifyObservers(update);
		log.exiting("DBDef", "setCounters");
	}
	
	/**
	 * Delete a counter if it is not in use
	 * @param counterName name of the counter
	 * @return either return null if the counter isn't in use or the name of the field using this counter
	 */
	public String removeCounter(String counterName)
	{
		log.entering("DBDef", "removeCounter", counterName);
		
		FieldDef fd;
		CounterFieldDef cfd;
		String fieldInUseName = null;

		// check if in use by field
		for (String fieldName : fieldOrder)
		{
			fd = getFieldDef(fieldName);
			assert fd != null : "removeCounter - cannot get FieldDef for "+fieldName;
			if (fd.getType() == FieldType.COUNTER)
			{
				cfd = (CounterFieldDef)fd;
				if (cfd.getCounter().equals(counterName))
				{
					fieldInUseName = fd.getName();
					break;
				}
			}
		}

		if (fieldInUseName == null) // not in use so can be removed
		{
			counters.remove(counterName);
			setChanged();
			DBDefUpdate update = new DBDefUpdate(DBDefUpdate.Event.COUNTER_DELETED);
			update.setOldName(counterName);
			notifyObservers(update);
		}

		log.exiting("DBDef", "removeCounter", fieldInUseName);
		
		return fieldInUseName;
	}
	
	/** Update a counter */
	public void updateCounter(String oldName, Counter counter)
	{
		Object[] params = { oldName, counter };
		log.entering("DBDef", "updateCounter", params);
		
		// check for name change
		if (!oldName.equals(counter.getName()))
		{
			FieldDef fd;
			CounterFieldDef cfd;
			
			for (String fieldName : fieldOrder)
			{
				fd = getFieldDef(fieldName);
				assert fd != null : "updateCounter - cannot get FieldDef for "+fieldName;
				if (fd.getType() == FieldType.COUNTER)
				{
					cfd = (CounterFieldDef)fd;
					if (cfd.getCounter().equals(oldName))
					{
						cfd.setCounter(counter.getName());
					}
				}
			}
		}
		
		counters.remove(oldName);
		counters.put(counter.getName(), counter);
		
		setChanged();
		DBDefUpdate update = new DBDefUpdate(DBDefUpdate.Event.COUNTER_UPDATED);
		update.setNewObject(counter);
		update.setOldName(oldName);
		update.setNewName(counter.getName());
		notifyObservers(update);
		
		log.exiting("DBDef", "updateCounter");
	}
	
	/** Get a Counter */
	public Counter getCounter(String counterName)
	{
		return counters.get(counterName);
	}
	
	/**
	 * Are any counters defined
	 * @return true if the database has counters
	 */
	public boolean areCountersDefined()
	{
		return !counters.isEmpty();
	}
	
	/**
	 * Are any categoried defined
	 * @return true if the database has categories
	 */
	public boolean areCategoriesDefined()
	{
		return !categories.isEmpty();
	}
	
	/**
	 * Are any calculated fields defined
	 * @return true if the database has calculations
	 */
	public boolean areCalculationsDefined()
	{
		log.entering("DBDef", "areCalculationsDefined");
		
		boolean calcsDefined = false;
		FieldDef fd;
		for (String name : fieldOrder)
		{
			fd = getFieldDef(name);
			assert fd != null : "areCalculationsDefined - cannot get FieldDef for "+name;
			if (fd.getType() == FieldType.CALCULATED)
			{
				calcsDefined = true;
				break;
			}
		}
		
		log.exiting("DBDef", "areCalculationsDefined", calcsDefined);
		
		return calcsDefined;
	}
	
	/**
	 * Calculated fields are only allowed if there are numeric fields already defined
	 * @return true if calculated fields are allowed
	 */
	public boolean areCalculationsAllowed()
	{
		log.entering("DBDef", "areCalculationsAllowed");
		
		boolean calcsAllowed = false;
		FieldDef fd;
		for (String name : fieldOrder)
		{
			fd = getFieldDef(name);
			assert fd != null : "areCalculationsAllowed - cannot get FieldDef for "+name;
			switch (fd.getType())
			{
				case NUMBER:
				case COUNTER:
					calcsAllowed = true;
					break;
					
				default:
					break;
			}
		}
		
		log.exiting("DBDef", "areCalculationsAllowed", calcsAllowed);
		
		return calcsAllowed;
	}
	
	/**
	 * Returns data for CSV file
	 * @param fieldDelimiter the delimiter used to separate field names
	 * @return CSV data
	 */
	public String getCSV(String fieldDelimiter)
	{
		log.entering("DBDef", "getCSV", fieldDelimiter);
		
		StringBuilder sb = new StringBuilder();
		for (String name : fieldOrder)
		{
			sb.append(name);
			sb.append(fieldDelimiter);
		}
		
		// remove last comma
		if (sb.length() > 0) {
			sb.setLength(sb.length()-1);
		}
		
		log.exiting("DBDef", "getCSV");
		
		return sb.toString();
	}
	
	/**
	 * Get a list of all the field names
	 * @return a List of field names
	 * @see getFieldNamesArray()
	 */
	public List<String> getFieldNames()
	{
		return fieldOrder;
	}
	
	/**
	 * Get field names as an array
	 * @see #getFieldNames()
	 */
	public String[] getFieldNamesArray()
	{
		String[] fnArray = new String[fieldOrder.size()];
		fieldOrder.toArray(fnArray);
		return fnArray;
	}
	
	/** Delete all field definitions and reset the date format */
	public void reset()
	{
		log.entering("DBDef", "reset");
		
		fieldDefs.clear();
		fieldOrder.clear();
		categories.clear();
		counters.clear();
		visibleFields.clear();
		displayFieldNames = true;
		
		setChanged();
		notifyObservers(new DBDefUpdate(DBDefUpdate.Event.RESET));
		
		log.exiting("DBDef", "reset");
	}
	
	/**
	 * Remove a category
	 * @param catNam category name
	 * @return null if the category was removed, name of field dependency otherwise
	 */
	public String removeCategory(String catName)
	{
		log.entering("'DBDef", "removeCategory", catName);
		
		String fieldName = null;
		if (catName != null)
		{
			// check to see if any fields use this category
			CategoryFieldDef cfd;
			for (FieldDef fd : fieldDefs.values())
			{
				if (fd.getType() == FieldType.CATEGORY)
				{
					cfd = (CategoryFieldDef)fd;
					if (cfd.getCategory().equals(catName))
					{
						fieldName = cfd.getName();
						break;
					}
				}
			}
			
			if (fieldName == null)
			{
				// they don't so remove it
				categories.remove(catName);
				setChanged();
				DBDefUpdate update = new DBDefUpdate(DBDefUpdate.Event.CATEGORY_REMOVED);
				update.setOldName(catName);
				notifyObservers(update);
			}
		}
		
		log.exiting("DBDef", "removeCategory", fieldName);
		
		return fieldName;
	}
	
	/**
	 * Rename a category
	 * @param oldCatName old category name
	 * @param newCatName new category name
	 * @return return true if updated; false otherwise
	 */
	public boolean renameCategory(String oldCatName, String newCatName)
	{
		Object[] params = { oldCatName, newCatName };
		log.entering("DBDef", "renameCategory, ", params);

		assert oldCatName != null : "Existing category name cannot be null";
		assert newCatName != null : "New category name cannot be null";
		
		boolean updated = false;
		if (oldCatName != null && newCatName != null && !newCatName.equals(oldCatName))
		{
			List<String> catItems = categories.remove(oldCatName);
			categories.put(newCatName, catItems);

			// update any FieldDefs
			FieldDef fd;
			CategoryFieldDef cfd;
			for (String name : fieldOrder)
			{
				fd = getFieldDef(name);
				if (fd.getType() == FieldType.CATEGORY)
				{
					cfd = (CategoryFieldDef)fd;
					if (cfd.getCategory().equals(oldCatName))
					{
						cfd.setCategory(newCatName);
					}
				}
			}
			setChanged();
			DBDefUpdate update = new DBDefUpdate(DBDefUpdate.Event.CATEGORY_RENAMED);
			update.setOldName(oldCatName);
			update.setNewName(newCatName);
			notifyObservers(update);
			updated = true;
		}

		log.exiting("DBDef", "renameCategory", updated);
		
		return updated;
	}
	
	/**
	 * Add a new category item
	 * @param catName category name
	 * @param item category item
	 */
	public void addCategoryItem(String catName, String item)
	{
		Object[] params = { catName, item };
		log.entering("DBDef", "addCategoryItem", params);
		List<String> catItems = categories.get(catName);
		assert catItems != null : "Cannot get category "+catName;
		if (catItems != null && item != null && item.trim().length() > 0)
		{
			catItems.add(item);
			setChanged();
			DBDefUpdate update = new DBDefUpdate(DBDefUpdate.Event.CATEGORY_ITEM_ADDED);
			update.setNewName(catName);
			update.setNewObject(item);
			update.setNewIndex(catItems.size()-1);
			update.setNewSize(catItems.size());
			notifyObservers(update);
		}
		log.exiting("DBDef", "addCategoryItem");
	}
	
	/**
	 * Rename a category item
	 * @param catName category name
	 * @param oldItem old item name
	 * @param newItem new item name
	 */
	public void renameCategoryItem(String catName, String oldItem, String newItem)
	{
		Object[] params = { catName, oldItem, newItem };
		log.entering("DBDef", "renameCategoryItem", params);
		List<String> catItems = categories.get(catName);
		if (catItems != null)
		{
			int pos = catItems.indexOf(oldItem);
			catItems.remove(pos);
			catItems.add(pos, newItem);
			
			setChanged();
			DBDefUpdate update = new DBDefUpdate(DBDefUpdate.Event.CATEGORY_ITEM_RENAMED);
			update.setNewName(catName);
			update.setOldObject(oldItem);
			update.setNewObject(newItem);
			update.setNewIndex(pos);
			notifyObservers(update);
		}
		log.exiting("DBDef", "renameCategoryItem");
	}
	
	/**
	 * Remove a category item
	 * @param catName category name
	 * @param item category item to remove
	 */
	public void removeCategoryItem(String catName, String item)
	{
		Object[] params = { catName, item };
		log.entering("DBDef", "removeCategoryItem", params);
		List catItems = categories.get(catName);
		if (catItems != null)
		{
			int index = catItems.indexOf(item);
			catItems.remove(item);
			setChanged();
			DBDefUpdate update = new DBDefUpdate(DBDefUpdate.Event.CATEGORY_ITEM_REMOVED);
			update.setNewName(catName);
			update.setOldObject(item);
			update.setOldIndex(index);
			notifyObservers(update);
		}
		log.exiting("DBDef", "removeCategoryItem");
	}
	
	/**
	 * Move a category item up or down in the list
	 * @param catName category name
	 * @param item category item
	 * @param delta 1 or -1
	 */
	public void moveCategoryItem(String catName, String item, int delta)
	{
		Object[] params = { catName, item, delta };
		log.entering("DBDef", "moveCategoryItem", params);
		List<String> catItems = categories.get(catName);
		if (catItems != null)
		{
			int pos = catItems.indexOf(item);
			int change = pos+delta;
			if (change < 0 || change >= catItems.size())
			{ // check change is allowed
				return;
			}
			catItems.remove(item);
			catItems.add(change, item);
			setChanged();
			DBDefUpdate update = new DBDefUpdate(DBDefUpdate.Event.CATEGORY_ITEM_MOVED);
			update.setOldIndex(pos);
			update.setNewIndex(change);
			notifyObservers(update);
		}
		
		log.exiting("DBDef", "moveCategoryItem");
	}
	
	/**
	 * Test for whether any data definitions have been set up
	 * @return true if some definition has been set up
	 */
	public boolean hasDataBeenDefined()
	{
		boolean dataDefined = (fieldDefs.isEmpty() && categories.isEmpty()) ? false : true;
		
		return dataDefined;
	}
	
	/** Set the visible fields based on the field name */
	public void setVisibleFields(List<String> visFields)
	{
		setVisibleFields(visFields, true);
	}
	
	/**
	 * Set the visible fields based on the field name
	 * @param visFields visible field names
	 * @param notifyObservers if true observers will be notified
	 */
	public void setVisibleFields(List<String> visFields, boolean notifyObservers)
	{
		log.entering("DBDef", "setVisibleFields", visFields);
		
		assert visFields != null : "Visible fields is null";
		
		// copy into new List to prevent inadvertent changes to the original list
		visibleFields = new ArrayList<String>(visFields);
		
		if (notifyObservers)
		{
			setChanged();
			DBDefUpdate update = new DBDefUpdate(DBDefUpdate.Event.VISIBLE_FIELDS);
			update.setNewSize(visibleFields.size());
			notifyObservers(update);
		}
		
		log.exiting("DBDef", "setVisibleFields");
	}
	
	/**
	 * Add a visible field to the table. This will notify any observers.
	 * @param fieldName field to make visible
	 * @see #addVisibleField(String, boolean)
	 */
	public void addVisibleField(String fieldName)
	{
		addVisibleField(fieldName, true);
	}
	
	/**
	 * Add a new visible field, and optionally inform any observers
	 * @param fieldName field name
	 * @param notifyObservers if true notify observers
	 * @see #addVisibleField(String)
	 */
	public void addVisibleField(String fieldName, boolean notifyObservers)
	{
		log.entering("DBDef", "addVisibleField", fieldName);
		
		assert fieldName != null : "addVisibleField: fieldName is null";
		
		if (!visibleFields.contains(fieldName))
		{
			visibleFields.add(fieldName);
			if (notifyObservers)
			{
				setChanged();
				DBDefUpdate update = new DBDefUpdate(DBDefUpdate.Event.VISIBLE_FIELDS);
				update.setNewName(fieldName);
				notifyObservers(update);
			}
		}
		
		log.exiting("DBDef", "addVisibleField");
	}
	
	/**
	 * Get which fields are currently visible
	 * @return an array of visible fields or null
	 */
	public List<String> getVisibleFields()
	{
		return new ArrayList<String>(visibleFields);
	}
	
	/**
	 * Get a list of date and numeric field names
	 * @return the field names which might be empty
	 */
	public List<String> getDateOrNumericFields()
	{
		List<String> fieldNames = new ArrayList<String>(fieldDefs.size());
		
		FieldDef fd;
		for (String name : fieldOrder)
		{
			fd = getFieldDef(name);
			assert fd != null : "getDateOrNumericFields - cannot get FieldDef for "+name;
			switch (fd.getType())
			{
				case DATE:
				case NUMBER:
					fieldNames.add(name);
					break;
					
				default:
					break;
			}
		}
		
		return fieldNames;
	}
	
	/** Set whether field names should be displayed */
	public void setDisplayFieldNames(boolean dfn)
	{
		displayFieldNames = dfn;
	}
	
	/** Should field names be displayed */
	public boolean displayFieldNames()
	{
		return displayFieldNames;
	}
	
	/** Set whether a colon should be added to the end of every field */
	public void setDisplayColons(boolean dc)
	{
		displayColons = dc;
	}
	
	/** Should a colon be added to the end of every field */
	public boolean displayColons()
	{
		return displayColons;
	}
	
	/**
	 * Updates the calculation for field fieldName
	 * @param fieldDef FieldDef for the calculated field
	 * @param values a Map of fields and their values involved in the calculation
	 * @return the result of the calculation
	 * @exception IllegalArgumentException if one of the arguments is incorrect
	 */
	public double updateCalculation(FieldDef fieldDef, Map<String, Object> values) throws IllegalArgumentException
	{
		log.entering("DBDef", "updateCalculation", fieldDef);
		
		Object value;
		String valueS;
		String calc = fieldDef.getCalculation();
		StringBuilder calcBuilder = new StringBuilder(calc.length());
		String token;
		StringBuilder tokenBuilder;
		String tokens[] = calc.split(" ");
		for (int i=0; i<tokens.length; i++)
		{
			token = tokens[i];
			if (token.startsWith(FIELD_SEPARATOR))
			{
				tokenBuilder = new StringBuilder(token);
				while (!token.endsWith(FIELD_SEPARATOR))
				{
					tokenBuilder.append(" ");
					token = tokens[++i];
					tokenBuilder.append(token);
				}
				// remove quotes
				token = tokenBuilder.substring(1, tokenBuilder.length()-1);
				
				// get field value
				value = values.get(token);
				if (value == null || value.toString().trim().length() == 0)
				{
					log.log(Level.FINE, "Cannot get value for {0}", token);
					calcBuilder.append("0");
				}
				else
				{
					valueS = value.toString().trim();
					if (valueS.length() == 0)
					{
						calcBuilder.append("0");
					}
					else
					{
						calcBuilder.append(valueS);
					}
				}
			}
			else
			{
				calcBuilder.append(token);
			}
			calcBuilder.append(" ");
		}
		
		// calculate the value
		if (calculator == null)
		{
			calculator = new PostfixCalculator();
		}
		calculator.parseCalculation(calcBuilder.toString());
		if (Level.FINEST == log.getLevel())
		{
			log.finest("Calculator stack contents");
			log.finest(calculator.toString());
		}
		
		return calculator.getResult();
	}
	
	/**
	 * Update all calculated fields based on values in Card
	 * @param updateField field that changed
	 * @param card Card containing values
	 * @return true if calculations have been updated; false otherwise
	 */
	public boolean updateAllCalculations(String updateField, Card card)
	{
		log.entering("DBDef", "updateAllCalculations", updateField);
		
		boolean calcUpdate = false;
		if (areCalculationsDefined())
		{
			FieldDef fd;
			double result;
			String calc;
			for (String field : fieldOrder)
			{
				fd = getFieldDef(field);
				if (fd.getType() == FieldType.CALCULATED)
				{
					calc = ((CalculatedFieldDef)fd).getCalculation();
					assert calc != null : "DBDef.updateAllCalculations - cannot get calculation for "+fd.getName();
					if (calc.indexOf(updateField) != -1)
					{
						result = updateCalculation(fd, card.getValues());
						card.setValue(fd.getName(), result);
						calcUpdate = true;
					}
				}
			}
		}
		
		return calcUpdate;
	}
	
	/**
	 * Are text fields defined
	 * @return true if a String or Memo field is defined; false otherwise
	 */
	public boolean areTextFieldsDefined()
	{
		log.entering("DBDef", "areTextFieldsDefined");
		
		boolean textFields = false;
		FieldDef fd;
		
		Iterator<FieldDef> it = fieldDefs.values().iterator();
		while (it.hasNext())
		{
			fd = it.next();
			if (fd.getType() == FieldType.STRING || fd.getType() == FieldType.MEMO)
			{
				textFields = true;
				break;
			}
		}
		
		log.exiting("DBDef", "areTextFieldsDefined", textFields);
		return textFields;
	}
	
	/**
	 * Checks to see if a field is part of a calculated field's calculation. Only the first match is returned.
	 * @param fd the field to check
	 * @return if the field is part of a calculation return the field's name otherwise null
	 */
	public String isPartOfCalculation(FieldDef fd)
	{
		assert fd != null : "FieldDef must be specified";
		log.entering("DBDef", "isPartOfCalculation", fd.getName());
		
		String fieldName = null;
		if (areCalculationsDefined())
		{
			FieldDef fdef;
			CalculatedFieldDef cfd;
			Iterator<FieldDef> it = fieldDefs.values().iterator();
			while (it.hasNext())
			{
				fdef = it.next();
				if (fdef.getType() == FieldType.CALCULATED)
				{
					cfd = (CalculatedFieldDef)fdef;
					if (cfd.getCalculation().contains(FIELD_SEPARATOR+fd.getName()+FIELD_SEPARATOR))
					{
						fieldName = cfd.getName();
						break;
					}
				}
			}
		}
		
		log.exiting("DBDef", "isPartOfCalculation", fieldName);
		return fieldName;
	}

	/**
	 * Get a Set of image FieldDefs
	 * @return Set of image FieldDefs
	 */
	public Set<ImageFieldDef> getImages()
	{
		log.entering("DBDef", "getImages");
		
		Set<ImageFieldDef> imageFieldDefs = new HashSet<ImageFieldDef>();
		for (FieldDef fd : fieldDefs.values())
		{
			if (fd.getType() == FieldType.IMAGE)
			{
				imageFieldDefs.add((ImageFieldDef)fd);
			}
		}

		log.exiting("DBDef", "getImages", imageFieldDefs.size());

		return imageFieldDefs;
	}
}
