/*
 * JCards
 * Copyright (C) 2002-2011 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package nrs.jcards;

import java.awt.Dimension;
import java.awt.EventQueue;
import java.awt.Point;
import java.awt.Toolkit;
import java.awt.datatransfer.Clipboard;
import java.io.*;
import java.net.URL;
import java.net.URLConnection;
import java.text.MessageFormat;
import java.util.*;
import java.util.logging.ConsoleHandler;
import java.util.logging.Handler;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.JFileChooser;
import javax.swing.JOptionPane;
import javax.swing.filechooser.FileFilter;
import nrs.jcards.fields.*;
import nrs.jcards.gui.JCAPreviewPanel;
import nrs.jcards.gui.JCardsFrame;
import nrs.jcards.gui.action.JCardsAction;
import nrs.jcards.gui.dialog.PasswordDialog;
import nrs.jcards.io.FileUtils;
import nrs.jcards.io.JCAException;
import nrs.jcards.io.JCAReader;
import nrs.jcards.io.filter.JCardsFileFilter;
import nrs.jcards.io.filter.JCardsV5FileFilter;

/**
 * Main application class for JCards.
 * See http://www.nsydenham.net/java/JCards/JCards.shtml for details.
 *
 * @todo alignment of table columns
 * @todo highlight table rows when meet/exceed/etc certain values
 * @todo fixed category list - new items can't be added via a card
 * @todo formatted text entry field?
 * @todo grouping via tabs
 * @todo updated icon
 *
 * @author Copyright &#169; 2002-2011 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class JCards implements Observer
{
	public final static String VERSION = "6.5.1";
	public final static String VERSION5 = "5.1";
	
	private final static Logger log = Logger.getLogger(JCards.class.getName());
	
	private DBDef def = DBDef.getInstance();
	private Cards cards;
	private static UserSettings settings = new UserSettings();
	private File cardFile = null;
	private boolean version5 = false;
	private transient String password;
	private long creationDate;
	private long modifiedDate;
	private JCardsFrame jcardsFrame;

	// resource bundles
	private static ResourceBundle labelsBundle, menusBundle, buttonsBundle, msgsBundle;
	private static enum RESOURCE_BUNDLE { BUTTONS, LABELS, MENUS, MESSAGES };
	
	// app settings
	private boolean changes = false;
	
	/** Only used for testing purposes */
	public JCards(boolean noGUI)
	{}
	
	/** Initialise the application GUI */
	public JCards()
	{
		// set up the cards
		cards = new Cards();
		// check for changes
		cards.addObserver(this);
		def.addObserver(cards);
		def.addObserver(this);
	}
	
	/** Finished initialising the application and loading prefs */
	private void finishedInit()
	{
		// create the GUI
		jcardsFrame = new JCardsFrame(this);
		// add observer for changes in the view menu
		def.addObserver(jcardsFrame);
		jcardsFrame.createToolbar();
		jcardsFrame.finishedInit();
	}
	
	/**
	 * Start the application
	 * @param args command-line arguments
	 */
	public static void main(final String[] args)
	{
		log.entering("JCards", "main");
		
		// set font anti-aliasing on
		System.setProperty("swing.aatext", "true");
		
		// invoke GUI actions on the EventQueue
		Runnable runner = new Runnable()
		{
			@Override
			public void run()
			{
				// initialise the app
				JCards jcards = new JCards();

				// see if a file was specified on the command line
				// or other options have been specified
				String cmdLineFile = null;
				for (String arg : args)
				{
					if (arg.contains("-help"))
					{
						System.out.println("JCards options");
						System.out.println("-debug   output debugging information to the console");
						System.out.println("-help    this information");
						System.out.println("-version print version information");
						System.out.println("<file>   load the specified file");
						System.exit(0);
					}
					else if (arg.contains("-version"))
					{
						System.out.println("JCards v"+VERSION);
						System.out.println("Copyright Nick Sydenham <nsydenham@yahoo.co.uk>");
						System.out.println("Licenced under the GPL v2. See Help | About for details.");
						System.exit(0);
					}
					else if (arg.contains("-debug"))
					{
						Logger log = Logger.getLogger("nrs.jcards");
						log.setLevel(Level.FINEST);
						Handler handler = new ConsoleHandler();
						handler.setLevel(Level.FINEST);
						log.addHandler(handler);
					}
					else
					{
						File f = new File(args[0]);
						if (f.exists()) {
							cmdLineFile = f.getAbsolutePath();
						} else {
							System.err.println(args[0]+" doesn't exist!");
						}
					}
				}

				String lastFile = jcards.readProperties();

				// show the main window
				jcards.finishedInit();
				
				// check to see if user opened a file via the Welcome screen
				if (jcards.getCreationDate() == 0)
				{
					// attempt to open command line file if specified
					if (cmdLineFile != null)
					{
						log.log(Level.FINE, "Reading command line file: {0}", cmdLineFile);
						jcards.openFile(cmdLineFile, true);
					}
					else if (lastFile != null)
					{
						log.log(Level.FINE, "Reading last opened file: {0}", lastFile);
						// attempt to read last opened file
						File f = new File(lastFile);
						if (f.exists())
						{
							jcards.openFile(lastFile, true);
						}
						else
						{
							log.warning("Last opened file doesn't exist");
						}
					}
				}
			}
		};
		
		EventQueue.invokeLater(runner);
		
		log.exiting("JCards", "main");
	}
	
	/**
	 * Check to see if any changes have been made
	 * @return true if the user cancels the operation
	 */
	private boolean checkForChanges()
	{
		log.entering("JCards", "checkForChanges");
		
		boolean cancelChange = false;
		if (changes)
		{
			String fileName = (cardFile == null ? getMessage("new.db") : cardFile.getName());
			int response = JOptionPane.showConfirmDialog(jcardsFrame, getFormattedMessage("data.save", fileName),
					getMessage("save.changes"), JOptionPane.YES_NO_CANCEL_OPTION);
			switch (response)
			{
				case JOptionPane.YES_OPTION:
					cancelChange = !saveFile();
					break;
					
				case JOptionPane.NO_OPTION:
					cancelChange = false;
					break;
					
				default:
					cancelChange = true;
					break;
			}
		}
		
		return cancelChange;
	}
	
	// start a new file
	public void newFile()
	{
		if (!checkForChanges()) {
			reset();
		}
	}
	
	/** Reset the app to defaults */
	private void reset()
	{
		log.entering("JCards", "reset");
		cards.reset();
		def.reset();
		jcardsFrame.reset();
		cardFile = null;
		version5 = false;
		changes = false;
		password = null;
		creationDate = (new Date()).getTime();
		modifiedDate = (new Date()).getTime();
		log.exiting("JCards", "reset");
	}
	
	/** Open a file from a URL */
	public void openFile(URL url)
	{
		log.entering("JCards", "openFile", url);
		try
		{
			// download to a temporary file
			URLConnection conn = url.openConnection();
			File file = File.createTempFile("jcardsurl", ".tmp");
			BufferedInputStream bis = new BufferedInputStream(conn.getInputStream());
			BufferedOutputStream bos = new BufferedOutputStream(new FileOutputStream(file));
			int read;
			byte[] data = new byte[128];
			while ((read = bis.read(data)) != -1)
			{
				bos.write(data, 0, read);
			}
			bis.close();
			bos.close();
			file.setReadOnly();
			file.deleteOnExit(); // in cae there's an error opening the file
			
			// open temporary file
			openFile(file.getAbsolutePath());
			
			jcardsFrame.updateRecentFilesMenu(url.toString());
			
			if (!file.delete()) {
				log.log(Level.WARNING, "Cannot delete temporary file {0}. File should be deleted on JCards exit", file.getAbsolutePath());
			}
		}
		catch (FileNotFoundException fnex)
		{
			log.log(Level.WARNING, "Cannot find file", fnex);
			jcardsFrame.showError(getMessage("cannot.find"), getMessage("error.open"), "JCards.openFile(URL)", fnex, Level.WARNING);
		}
		catch (IOException ioex)
		{
			log.log(Level.WARNING, "Cannot read from URL", ioex);
			jcardsFrame.showError(getMessage("error.open"), getMessage("error"), "JCards.openFile(URL)", ioex, Level.WARNING);
		}
		finally
		{
			log.exiting("JCards", "openFile");
		}
	}
	
	/**
	 * Load a card file
	 * @param loadFile the file to load. If null a dialog box will open
	 * @return true if the file was opened, false otherwise
	 * @see #openFile(String, boolean)
	 */
	public boolean openFile(String loadFile)
	{
		return openFile(loadFile, false);
	}
	
	/**
	 * Load a card file
	 * @param loadFile the file to load. If null a dialog box will open.
	 * @param init set to true if opening a file during initialisation
	 * @return true if the file was opened, false otherwise
	 */
	public boolean openFile(String loadFile, boolean init)
	{
		File fileToOpen;
		
		log.entering("JCards", "openFile", loadFile);
		
		if (!init && checkForChanges()) {
			return false;
		}
		
		if (loadFile == null)
		{
			fileToOpen = getJCAFile(getMessage("open.file"), false);
			if (fileToOpen == null)
			{
				return false;
			}
		}
		else
		{
			fileToOpen = new File(loadFile);
			// check it exists
			if (!JCAReader.isJCAFile(fileToOpen))
			{
				JOptionPane.showMessageDialog(jcardsFrame, getMessage("not.jcards.file"), getMessage("error"), JOptionPane.ERROR_MESSAGE);
				return false;
			}
		}
		log.log(Level.FINE, "Loading file {0}", fileToOpen.getAbsolutePath());
		
		if (!init)
		{
			// reset all data structures
			reset();
		}
		
		try
		{
			JCAReader reader = new JCAReader(this);
			if (password == null && JCAReader.isEncrypted(fileToOpen))
			{
				PasswordDialog pd = new PasswordDialog(this, fileToOpen, password, false);
				pd.pack();
				pd.setLocationRelativeTo(jcardsFrame);
				pd.setVisible(true);
				if (password == null)
				{
					return false;
				}
			}
			
			reader.openFile(fileToOpen, password);
			if (fileToOpen.getName().endsWith("tmp"))
			{
				jcardsFrame.setTitle("JCards");
			}
			else
			{
				jcardsFrame.setTitle("JCards - "+fileToOpen.getName());
			}
			changes = false;
			cards.finishedLoad();
			cardFile = fileToOpen;
			jcardsFrame.setSortedColumn(reader.getSortedColumn());
			jcardsFrame.setColWidths(reader.getColumnWidths());
			
			// check for version 5
			Map<String, String> header = JCAReader.readHeader(cardFile);
			if (header.get(JCAReader.VERSION) != null)
			{
				int ver = Integer.parseInt(header.get(JCAReader.VERSION).substring(0, 1));
				if (ver < 6) {
					version5 = true;
				}
			}
			
			// is password set
			if (password != null) {
				jcardsFrame.passwordSet(true);
			}
		}
		catch (final Exception ex)
		{
			log.log(Level.WARNING, "Can't read file", ex);
			jcardsFrame.showError(getMessage("error"), getMessage("errorfile"), "JCards.openFile", ex, Level.SEVERE);
			reset();
			return false;
		}
		finally
		{
			if (loadFile != null && loadFile.endsWith("tmp"))
			{
				cardFile = null;
			}
			
			if (cardFile == null)
			{
				jcardsFrame.updateRecentFilesMenu(null);
			}
			else
			{
				jcardsFrame.updateRecentFilesMenu(cardFile.getAbsolutePath());
				jcardsFrame.setStatus(getFormattedMessage("opened", cardFile.getName(), cards.size()));
			}
			jcardsFrame.updateViewMenu();
		}
		
		log.exiting("JCards", "openFile");
		
		return true;
	}
	
	/**
	 * Ask the user to select a JCards file
	 * @param title dialog title
	 * @param save if true open save dialog; set to false for open dialog
	 * @return either null or the selected file
	 */
	private File getJCAFile(String title, boolean save)
	{
		log.entering("JCards", "getJCAFile", save);
		
		// set up file chooser
		final JFileChooser fileDialog = new JFileChooser();
		fileDialog.setDialogTitle(title);
		if (cardFile != null)
		{
			String name = (version5 ? cardFile.getName().replace(".jca", ".jc5") : cardFile.getName());
			File f = new File(cardFile.getPath(), name);
			fileDialog.setSelectedFile(f);
		}
		else
		{
			File f = new File(FileUtils.getHomeDir(), "jcards.jca");
			fileDialog.setSelectedFile(f);
		}
		
		// file filters
		FileFilter jcardsFilter = new JCardsFileFilter();
		fileDialog.addChoosableFileFilter(jcardsFilter);
		JCardsV5FileFilter jcaFiveFilter = new JCardsV5FileFilter();
		fileDialog.addChoosableFileFilter(jcaFiveFilter);
		if (version5) {
			fileDialog.setFileFilter(jcaFiveFilter);
		} else {
			fileDialog.setFileFilter(jcardsFilter);
		}
		fileDialog.setAcceptAllFileFilterUsed(false);
		
		// preview panel
		JCAPreviewPanel jcaPreview = new JCAPreviewPanel(save);
		fileDialog.setAccessory(jcaPreview);
		if (save)
		{
			if (version5) {
				jcaPreview.setVersion(VERSION5);
			} else {
				jcaPreview.setVersion(VERSION);
			}
			Date cDate = (creationDate == 0 ? new Date() : new Date(creationDate));
			jcaPreview.setCreateDate(cDate);
			jcaPreview.setModifiedDate(new Date());
			jcaPreview.setEncrypted((password != null));
		}
		fileDialog.addPropertyChangeListener(jcaPreview);
		
		// get selected file
		int ret;
		if (save)
		{
			ret = fileDialog.showSaveDialog(jcardsFrame);
		}
		else
		{
			ret = fileDialog.showOpenDialog(jcardsFrame);
		}
		
		File selFile = null;
		if (ret == JFileChooser.APPROVE_OPTION)
		{
			selFile = fileDialog.getSelectedFile();
			if (save && selFile.exists())
			{
				ret = JOptionPane.showConfirmDialog(jcardsFrame, getMessage("overwrite.existing"),
						getFormattedMessage("overwrite", selFile.getName()),
						JOptionPane.YES_NO_OPTION, JOptionPane.WARNING_MESSAGE);
				if (ret == JOptionPane.NO_OPTION)
				{
					selFile = null;
				}
			}
			
			if (selFile != null && jcaFiveFilter.equals(fileDialog.getFileFilter()))
			{
				version5 = true;
			}
		}
		fileDialog.removePropertyChangeListener(jcaPreview);
		
		log.exiting("JCards", "getJCAFile", selFile);
		return selFile;
	}
	
	/**
	 * Save the data to an existing or new file
	 * @return true if saved, false otherwise
	 */
	public boolean saveFile()
	{
		log.entering("JCards", "saveFile");
		
		modifiedDate = (new Date()).getTime();
		
		if (cards.size() == 0 && !changes)
		{
			JOptionPane.showMessageDialog(jcardsFrame, getMessage("no.data.save"), getMessage("no.data"), JOptionPane.INFORMATION_MESSAGE);
			return false;
		}
		
		// see if the user really wants to save as v5
		log.log(Level.FINEST, "V5 file={0}", version5);
		if (version5)
		{
			Object[] options = { getButtonString("save.as.version5"), getButtonString("save.as.version6"), getButtonString("cancel") };
			int resp = JOptionPane.showOptionDialog(jcardsFrame, getMessage("version5.file"), getMessage("warning"),
				JOptionPane.YES_NO_CANCEL_OPTION, JOptionPane.WARNING_MESSAGE, null, options, options[0]);
			log.log(Level.FINEST, "Save as v5 option={0}", resp);
			
			if (resp == 1) {
				version5 = false;
			} else if (resp == 2) {
				return false;
			}
		}
		
		if (cardFile == null)
		{
			log.finer("No file specified");
			File f = getJCAFile(getMessage("save.to"), true);
			if (f == null)
			{
				return false;
			}
			else
			{
				cardFile = f;
			}
		}
		log.log(Level.FINE, "File to save to: {0}", cardFile.getAbsolutePath());
		
		boolean exitStatus = false;
		try
		{
			JCAReader reader = new JCAReader(this);
			reader.saveFile(cardFile, password, jcardsFrame.getColWidths(), version5);
			
			changes = false;

			MessageFormat msgFormat = new MessageFormat(getMessage("saved.to"));
			String msg;
			if (password != null)
			{
				String[] args = { cardFile.getName() };
				StringBuilder sb = new StringBuilder(msgFormat.format(args));
				sb.append(". ");
				sb.append(getMessage("file.enc"));
				msg = sb.toString();
			}
			else
			{
				String[] args = { cardFile.getName() };
				msg = msgFormat.format(args);
			}
			jcardsFrame.setStatus(msg);
			jcardsFrame.updateRecentFilesMenu(cardFile.getAbsolutePath());
			exitStatus = true;
		}
		catch (Exception ex)
		{
			log.log(Level.SEVERE, "saveFile", ex);
			String msg = (ex.getMessage() == null ? getMessage("unknown.error") : ex.getMessage());
			StringBuilder sb = new StringBuilder(getFormattedMessage("cannot.save", cardFile.getName()));
			sb.append("\n");
			sb.append(msg);
			jcardsFrame.showError(sb.toString(), getMessage("error"), "JCards.saveFile", ex, Level.WARNING);
		}
		
		jcardsFrame.setTitle("JCards - "+cardFile.getName());
		log.exiting("JCards", "saveFile", exitStatus);
		
		return exitStatus;
	}
	
	/** Save data to a new file */
	public void saveFileAs()
	{
		log.entering("JCards", "saveFileAs");
		
		if (cards.size() == 0 && def.getNumberOfFields() == 0)
		{
			JOptionPane.showMessageDialog(jcardsFrame, getMessage("no.data.save"), getMessage("error"), JOptionPane.INFORMATION_MESSAGE);
			return;
		}
		
		modifiedDate = (new Date()).getTime();
		File f = getJCAFile(getMessage("save.to"), true);
		if (f != null)
		{
			cardFile = f;
			saveFile();
		}
		
		log.exiting("JCards", "saveFileAs");
	}
	
	/** Exit the program */
	public void exit()
	{
		log.entering("JCards", "exit");
		
		if (checkForChanges()) {
			return;
		}
		
		try
		{
			saveProperties();
		}
		catch (Exception ex)
		{
			log.log(Level.WARNING, "Exception when saving properties", ex);
		}
		
		jcardsFrame.dispose();
		log.exiting("JCards", "exit");
		System.exit(0);
	}
	
	/** Save the properties to a file */
	private void saveProperties()
	{
		log.entering("JCards", "saveProperties");
		
		// get user settings
		settings.setViewToolbar(jcardsFrame.isToolbarVisible());
		Properties props = settings.getProperties();
		
		// save window position and size
		Dimension size = jcardsFrame.getSize();
		Point location = jcardsFrame.getLocationOnScreen();
		props.put(UserSettings.PREF_SIZE, size.width+"x"+size.height);
		props.put(UserSettings.PREF_LOCATION, location.x+","+location.y);
		props.put(UserSettings.PREF_TABLE_RESIZE, Integer.toString(jcardsFrame.getTableAutoResizeMode()));
		
		// save user icons
		if (!settings.getUserIconList().isEmpty()) {
			props.put(UserSettings.PREF_ICONS, settings.getUserIconList());
		}

		try
		{
			File home = new File(System.getProperty("user.home"));
			File saveDir = new File(home, UserSettings.PREF_FILE_DIR);
			if (!saveDir.exists()) {
				saveDir.mkdir();
			}
			File saveFile = new File(saveDir, UserSettings.PREF_FILE);
			FileOutputStream fos = new FileOutputStream(saveFile);
			props.store(fos, "JCards properties - do not edit");
			fos.close();
		}
		catch (IOException ioe)
		{
			log.log(Level.WARNING, "Cannot save properties", ioe);
			jcardsFrame.showError(getMessage("cannot.save"), getMessage("props.save.failed"), "JCards.saveProperties", ioe, Level.INFO);
		}
		log.exiting("JCards", "saveProperties");
	}
	
	/**
	 * Read any previously saved properties
	 * @return the last opened file or null if not specified or null if not required
	 */
	private String readProperties()
	{
		Properties props = new Properties();
		String lastFile = null;
		File f;
		
		log.entering("JCards", "readProperties");
		
		try
		{
			File home = new File(System.getProperty("user.home"));
			f = new File(home, UserSettings.PREF_FILE_PATH);
			log.log(Level.FINER, "Checking for properties file {0}", f.getAbsolutePath());
			if (f.exists())
			{
				log.finer("Reading properties");
				FileInputStream fis = new FileInputStream(f);
				props.load(fis);
				fis.close();
				settings.setProperties(props);

				// load the last referenced file
				if (settings.isOpenLastFile())
				{
					lastFile = props.getProperty(UserSettings.PREF_LASTFILE+0);
				}
				else
				{
					lastFile = null;
				}
			}
		}
		catch (IOException ioe)
		{
			log.log(Level.SEVERE, "Error loading jca.properties", ioe);
		}
		
		log.exiting("JCards", "readProperties", lastFile);
		
		return lastFile;
	}
	
	/**
	 * Get the cards
	 */
	public Cards getCards()
	{
		return cards;
	}
	
	/**
	 * Set the password used to encrypt the file
	 * @param newPassword new password
	 */
	public final void setPassword(String newPassword)
	{
		if (password == null && newPassword == null)
		{
			return;
		}
		
		if (newPassword != null && newPassword.equals(password))
		{
			return;
		}
		
		password = newPassword;
		
		if (newPassword == null)
		{
			jcardsFrame.passwordSet(false);
			setChanged(getMessage("pass.cleared"));
		}
		else
		{
			jcardsFrame.passwordSet(true);
			setChanged(getMessage("pass.updated"));
		}
	}
	
	/**
	 * Get the password
	 * @param return the possibly null password
	 */
	public final String getPassword()
	{
		return password;
	}
	
	/**
	 * This method is called whenever the DBDef or Cards objects are changed.
	 * JCards records the change
	 */
	@Override
	public void update(Observable o, Object arg)
	{
		log.entering("JCards", "update", o);
		
		if (o instanceof DBDef)
		{
			changes = true;
		}
		else if (o instanceof Cards)
		{
			CardUpdate cu = (CardUpdate)arg;
			switch (cu.cardOperation)
			{
				case FIND:
				case LOAD:
					break;
				
				default:
					changes = true;
			}
		}
		
		if (changes) {
			setChanged();
		}
		
		log.exiting("JCards", "update");
	}
	
	/**
	 * Show statistics for the selected field or for the database
	 * @param fieldName field selected by user
	 */
	public String getStatistics(String fieldName)
	{
		log.entering("JCards", "getStatistics", fieldName);
		
		StringBuilder stats = new StringBuilder();
		
		if (fieldName != null)
		{
			FieldDef fd = def.getFieldDef(fieldName);
			FieldType type = fd.getType();
			
			if (type == FieldType.DATE)
			{
				stats.append(cards.getDateStatistics(fieldName));
			}
			else if (type == FieldType.NUMBER)
			{
				stats.append(cards.getNumericStatistics(fieldName));
			}
			
			// check for no data
			if (stats == null)
			{
				stats.append(getMessage("no.data"));
			}
		}
		else
		{
			stats.append(getFormattedMessage("db.recs", cards.size()));
			stats.append("\n");
			stats.append(getFormattedMessage("db.fields", def.getNumberOfFields()));
			stats.append("\n");
			stats.append(getFormattedMessage("db.created", new Date(creationDate)));
			stats.append("\n");
			stats.append(getFormattedMessage("db.modified", new Date(modifiedDate)));
		}
		
		log.exiting("JCards", "getStatistics");
		return stats.toString();
	}
	
	/** Create an example DB */
	public void createExampleDB()
	{
		newFile();
		
		// define ID
		Counter counter = new Counter("idcounter");
		def.addCounter(counter);
		CounterFieldDef countFD = (CounterFieldDef)FieldDef.createFieldDef("ID", FieldType.COUNTER);
		countFD.setCounter("idcounter");
		def.addFieldDef(countFD);
		
		// define summary
		FieldDef fd = FieldDef.createFieldDef("Summary", FieldType.STRING);
		fd.setMandatory(true);
		def.addFieldDef(fd);
		
		// define category
		List<String> catItems = Arrays.asList("Personal", "Business");
		def.setCategory("Task Type", catItems);
		CategoryFieldDef catFD = (CategoryFieldDef)FieldDef.createFieldDef("Type", FieldType.CATEGORY);
		catFD.setCategory("Task Type");
		catFD.setDefaultValue("Personal");
		def.addFieldDef(catFD);
		
		// define start date
		DateFieldDef dateFD = (DateFieldDef)FieldDef.createFieldDef("Start Date", FieldType.DATE);
		dateFD.setDatePrepopulated(true);
		dateFD.setDateFormat("dd/MM/yyyy");
		def.addFieldDef(dateFD);
		
		// define end date
		dateFD = (DateFieldDef)FieldDef.createFieldDef("End Date", FieldType.DATE);
		dateFD.setDateFormat("dd/MM/yyyy");
		def.addFieldDef(dateFD);
		
		// define memo
		MemoFieldDef memoFD = (MemoFieldDef)FieldDef.createFieldDef("Description", FieldType.MEMO);
		memoFD.setMemoLines(10);
		def.addFieldDef(memoFD);
		
		// set displayed columns
		List<String> visFields = new ArrayList<String>(4);
		visFields.add("ID");
		visFields.add("Summary");
		visFields.add("Type");
		visFields.add("Start Date");
		def.setVisibleFields(visFields);
		
		// add some cards
		Map<String, Object> values = new HashMap<String, Object>(5);
		values.put("ID", counter.increment());
		values.put("Summary", "Book dentists appointment");
		values.put("Type", "Personal");
		values.put("Start Date", new Date());
		values.put("Description", "Remember not to book during holiday period");
		cards.addCard(new Card(values));
		
		values = new HashMap<String, Object>(4);
		values.put("ID", counter.increment());
		values.put("Summary", "Write report");
		values.put("Type", "Business");
		values.put("Description", "Refer to previous emails on the subject");
		cards.addCard(new Card(values));
	}
	
	/**
	 * Get the current file
	 * @param current file or null
	 */
	public File getCardFile()
	{
		return cardFile;
	}
	
	/** Get the current settings */
	public static UserSettings getUserSettings()
	{
		return settings;
	}
	
	/** Get a reference to JCardsFrame */
	public JCardsFrame getFrame()
	{
		return jcardsFrame;
	}
	
	/** Copy a Card's contents to the clipboard */
	public void copyToClipboard(Card card)
	{
		log.entering("JCards", "copyToClipboard", card);
		Clipboard clip = Toolkit.getDefaultToolkit().getSystemClipboard();
		clip.setContents(card, null);
		jcardsFrame.setStatus(getMessage("copy.clip"));
		log.exiting("JCards", "copyToClipboard");
	}
	
	/** Has the JCards file changed */
	public boolean hasChanged()
	{
		return changes;
	}
	
	/**
	 * Current JCards file has changed
	 * @see #setChanged(String)
	 */
	public void setChanged()
	{
		log.entering("JCards", "setChanged");
		changes = true;
		String title = jcardsFrame.getTitle();
		if (!title.endsWith("*"))
		{
			jcardsFrame.setTitle(title+"*");
		}
		log.exiting("JCards", "setChanged");
	}
	
	/**
	 * Current JCards file has changed. Display a message in the status bar as well
	 * @param message the text to display
	 * @see #setChanged()
	 */
	public void setChanged(String message)
	{
		setChanged();
		jcardsFrame.setStatus(message);
	}
	
	/** Get the date the database was created */
	public long getCreationDate()
	{
		return creationDate;
	}
	
	/** Set the creation date */
	public void setCreationDate(long cDate)
	{
		creationDate = cDate;
	}
	
	/** Get the date the database was modified */
	public long getModifiedDate()
	{
		return modifiedDate;
	}
	
	/** Set the time when the database was last modified */
	public void setModifiedDate(long modDate)
	{
		modifiedDate = modDate;
	}

	/**
	 * Check that a resource bundle has been initialised correctly
	 * @param rb resource bundle to init
	 * @return true if initialised, false otherwise
	 */
	private static boolean initResourceBundle(RESOURCE_BUNDLE rb)
	{
		ResourceBundle resBundle = null;

		switch (rb)
		{
			case BUTTONS:
				if (buttonsBundle == null) {
					buttonsBundle = ResourceBundle.getBundle("nrs.jcards.props.buttons");
				}
				resBundle = buttonsBundle;
				break;

			case LABELS:
				if (labelsBundle == null) {
					labelsBundle = ResourceBundle.getBundle("nrs.jcards.props.labels");
				}
				resBundle = labelsBundle;
				break;

			case MENUS:
				if (menusBundle == null) {
					menusBundle = ResourceBundle.getBundle("nrs.jcards.props.menu");
				}
				resBundle = menusBundle;
				break;

			case MESSAGES:
				if (msgsBundle == null) {
					msgsBundle = ResourceBundle.getBundle("nrs.jcards.props.msgs");
				}
				resBundle = msgsBundle;
				break;
		}

		return resBundle != null;
	}

	/**
	 * Get an internationalised message
	 * @param key the key to the message
	 * @return the localised message
	 * @see #getFormattedMessage(String, Object...)
	 */
	public static String getMessage(String key)
	{
		assert key != null : "Cannot get a null message";

		String msg = " ";

		if (initResourceBundle(RESOURCE_BUNDLE.MESSAGES) && msgsBundle.containsKey(key))
		{
			msg = msgsBundle.getString(key);
		}
		else
		{
			log.finest("Cannot get resource bundle entry for "+key);
			// try to return a default error message
			if (msgsBundle != null && msgsBundle.containsKey("error.message")) {
				msg = msgsBundle.getString("error.message");
			}
		}
		
		return msg;
	}
	
	/**
	 * Look up the message from the resource bundle
	 * @param key the message key in the resource bundle
	 * @param args any arguments for the message
	 * @return a formatted message
	 * @see #getMessage(String)
	 */
	public static String getFormattedMessage(String key, Object... args)
	{
		String msg = getMessage(key);
		return MessageFormat.format(msg, args);
	}
	
	/**
	 * Get a label value from a ResourceBundle
	 * @param key the key from the labels.properties ResourceBundle
	 * @param withMnemonicIndicator if true include the mnemonic indicator; if false remove it first
	 * @return the text which may or may not have a mnemonic indicator
	 * @see #getLabelString(String)
	 */
	public static String getLabelString(String key, boolean withMnemonicIndicator)
	{
		assert key != null : "Cannot lookup a null label";
		
		String val = null;

		if (initResourceBundle(RESOURCE_BUNDLE.LABELS) && labelsBundle.containsKey(key))
		{
			val = labelsBundle.getString(key);
			if (!withMnemonicIndicator)
			{
				val = JCardsAction.getTextWithoutMnemonic(val);
			}
		}
		else
		{
			log.finest("Cannot get resource bundle label entry for "+key);
		}
		
		return val;
	}
	
	/**
	 * Get a label value from a ResourceBundle removing any mnemonic indicator that may be present
	 * @param key the key from the labels.properties ResourceBundle
	 * @return the text
	 * @see #getLabelString(String, boolean)
	 */
	public static String getLabelString(String key)
	{
		return getLabelString(key, false);
	}
		
	/**
	 * Get a buttons value from a ResourceBundle
	 * @param key the key from the buttons.properties ResourceBundle
	 * @param withMnemonicIndicator set to true if the returned value should include the mnemonic indicator
	 * @return the text
	 * @see getButtonString(String)
	 */
	public static String getButtonString(String key, boolean withMnemonicIndicator)
	{
		assert key != null : "Cannot lookup a null button label";

		String val = null;
		
		if (initResourceBundle(RESOURCE_BUNDLE.BUTTONS) && buttonsBundle.containsKey(key))
		{
			val = buttonsBundle.getString(key);
			if (!withMnemonicIndicator)
			{
				val = JCardsAction.getTextWithoutMnemonic(val);
			}
		}
		else
		{
			log.finest("Cannot get resource bundle button entry for "+key);
		}
		
		return val;
	}
	
	/**
	 * Get a buttons value from a ResourceBundle removing any mnemonic indicator that may be present
	 * @key the key from the buttons.properties ResourceBundle
	 * @return the text
	 * @see #getButtonString(String, boolean)
	 */
	public static String getButtonString(String key)
	{
		return getButtonString(key, false);
	}
	
	/**
	 * Get a menu value from a ResourceBundle
	 * @key the key from the menu.properties ResourceBundle
	 * @return the text
	 */
	public static String getMenuString(String key)
	{
		assert key != null : "Cannot lookup a null menu label";

		String menuString = null;
		
		if (initResourceBundle(RESOURCE_BUNDLE.MENUS) && menusBundle.containsKey(key))
		{
			menuString = menusBundle.getString(key);
		}
		else
		{
			log.finest("Cannot find resource bundle menu entry for "+key);
		}
		
		return menuString;
	}

	/**
	 * Load an image from the current JCards file
	 * @param imgName internal image ref
	 * @return InputStream for the image
	 * @throws JCAException
	 */
	public InputStream loadImageFromDB(String imgName) throws JCAException
	{
		return JCAReader.getImageInputStream(cardFile, imgName);
	}
}
