/*
 * JCards
 * Copyright (C) 2002-2010 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package nrs.jcards.fields;

import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JLabel;
import nrs.jcards.DBDef;
import nrs.jcards.gui.JCardsFrame;
import nrs.jcards.io.Elements;
import org.jdesktop.swingx.autocomplete.AutoCompleteDecorator;

/**
 * Category field definition
 *
 * @author Copyright &#169; 2002-2010 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class CategoryFieldDef extends FieldDef
{
	protected String catName;

	/** Creates a new instance of CategoryFieldDef */
	public CategoryFieldDef(String name)
	{
		setName(name);
		type = FieldType.CATEGORY;
	}
	
	/** This field uses a category to get its values */
	public void setCategory(String newCat)
	{
		catName = newCat;
	}
	
	/**
	 * Get the field's category
	 * @return the category or null if not set
	 */
	public String getCategory()
	{
		return catName;
	}
	
	/** Set the default value of this field */
	@Override
	public boolean setDefaultValue(String value)
	{
		log.entering("CategoryFieldDef", "setDefaultValue", value);
		
		boolean ret = false;
		if (catName != null && value != null)
		{
			List<String> cats = DBDef.getInstance().getCategory(catName);
			if (cats.contains(value))
			{
				defaultValue = value;
				ret = true;
			}
			else
			{
				ret = false;
			}
		}
		
		log.exiting("CategoryFieldDef", "setDefaultValue", ret);
		return ret;
	}
	
	@Override
	public void setFieldDetails(Map<String, String> fieldDetails)
	{
		catName = fieldDetails.get(Elements.FIELD_CAT);
		super.setFieldDetails(fieldDetails);
	}
	
	@Override
	public String toString()
	{
		StringBuilder sb = new StringBuilder(super.toString());
		sb.append(",");
		sb.append(catName);
		
		return sb.toString();
	}
	
	/**
	 * Get the field definition
	 * @return Map of name/value pairs
	 */
	@Override
	public Map<String, String> getFieldDetails()
	{
		Map<String, String> def = super.getFieldDetails();
		def.put(Elements.FIELD_CAT, catName);
		return def;
	}
	
	/**
	 * Parse the value and return the correct type
	 * @param value the value to parse
	 * @return the parsed value
	 */
	@Override
	public Object parseValue(String value)
	{
		log.entering("CategoryFieldDef", "parseValue", value);
		
		if (value != null && value.trim().length() == 0)
		{
			value = null;
		}
		
		log.exiting("CategoryFieldDef", "parseValue", value);
		return value;
	}
	
	@Override
	public Class getFieldClass()
	{
		return JComboBox.class;
	}
	
	/**
	 * Set the value of the component
	 * @param comp the component to be updated
	 * @param value the value to set
	 * @param editing true if editing
	 */
	@Override
	public void setGUIValue(JComponent comp, Object value, boolean editing, boolean newCard)
	{
		log.entering("CategoryFieldDef", "setGUIValue", value);
		
		if (editing)
		{
			JComboBox cb = (JComboBox)comp;
			if (value != null)
			{
				cb.setSelectedItem(value);
			}
			else if (newCard && defaultValue != null)
			{
				cb.setSelectedItem(defaultValue);
			}
		}
		else
		{
			if (value != null)
			{
				JLabel label = (JLabel)comp;
				label.setText(value.toString());
			}
		}
		
		log.exiting("CategoryFieldDef", "setGUIValue");
	}
	
	private void addComboItems(JComboBox combo)
	{
		log.entering("CategoryFieldDef", "addComboItems");
		
		// add a blank item first if optional
		if (!isMandatory()) {
			combo.addItem("");
		}
		
		// add category item
		List<String> catItems = DBDef.getInstance().getCategory(catName);
		if (catItems != null)
		{
			for (String item : catItems)
			{
				combo.addItem(item);
			}
		}
		else
		{
			log.log(Level.WARNING, "User has deleted category {0}", name);
		}
		
		log.exiting("CategoryFieldDef", "addComboItems");
	}

	/**
	 * Get the GUI component that represents the field
	 * @param editing true if editing
	 * @return returns a JPanel for editing or a JLabel for viewing
	 */
	@Override
	public JComponent getGUIComponent(boolean newCard, boolean editCard, boolean copyCard, JCardsFrame frame)
	{
		log.entering("CategoryFieldDef", "getGUIComponent");
		
		JComponent jcomp;
		if (editCard)
		{
			JComboBox combo = new JComboBox();
			combo.setEditable(true);
			combo.setName(name);
			addComboItems(combo);
			AutoCompleteDecorator.decorate(combo);
			jcomp = combo;
		}
		else
		{
			JLabel label = new JLabel();
			label.setName(name);
			addCopyPopupMenu(label, editCard);
			addCopyDnD(label);
			jcomp = label;
		}
		
		log.exiting("CategoryFieldDef", "getGUIComponent");
		return jcomp;
	}
	
	/**
	 * Check the value entered by a user. This is different from parseValue in that
	 * it also takes account of the mandatory flag
	 * @param comp the component to be checked
	 * @return InputCheck
	 */
	@Override
	public InputCheck checkInputValue(JComponent comp)
	{
		log.entering("CategoryFieldDef", "checkInputValue", comp);

		InputCheck ic;
		JComboBox cb = (JComboBox)comp;
		String val = (String)cb.getSelectedItem();
		if (mandatory && (val == null || val.trim().length() == 0))
		{
			cb.requestFocus();
			ic = InputCheck.mandatoryInputCheck(name);
		}
		else
		{
			// a blank value in a combo box doesn't make much sense
			if (val != null && val.trim().length() == 0)
			{
				val = null;
			}

			ic = InputCheck.validInputCheck(val);
			// check that the value is in the list. If not, add it.
			if (val != null)
			{
				List<String> catItems = DBDef.getInstance().getCategory(catName);
				assert catItems != null : "Cannot get category items for category "+catName;
				if (!catItems.contains((String)cb.getSelectedItem()))
				{
					DBDef.getInstance().addCategoryItem(catName, cb.getSelectedItem().toString());
				}
			}
		}
		
		log.exiting("CategoryFieldDef", "checkInputValue", ic);
		return ic;
	}
}
