/*
 * JCards
 * Copyright (C) 2002-2009 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package nrs.jcards.fields;

import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Map;
import java.util.logging.Level;
import javax.swing.JComponent;
import javax.swing.JLabel;
import nrs.jcards.UserSettings;
import nrs.jcards.gui.JCardsFrame;
import nrs.jcards.io.Elements;
import org.jdesktop.swingx.JXDatePicker;
import org.jdesktop.swingx.calendar.SingleDaySelectionModel;

/**
 * Date field definition
 * 
 * @author Copyright &#169; 2002-2009 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class DateFieldDef extends FieldDef
{
	private String dateFormat = UserSettings.DATE_FORMATS[0];
	private SimpleDateFormat sdf = new SimpleDateFormat(dateFormat);
	private boolean prepopulateDate = false;
	
	/** Creates a new instance of DateFieldDef */
	public DateFieldDef(String name)
	{
		setName(name);
		type = FieldType.DATE;
	}
	
	/**
	 * Set the default value which is limited by the type,
	 * i.e. you can't set it to 'A' if the type is Integer
	 * @param value the value to set
	 * @return true if the default value has been set; false if it is of the incorrect type
	 */
	@Override public boolean setDefaultValue(String value)
	{
		boolean setDefValue = false;
		
		if (value != null)
		{
			try
			{
				parseDate(value);
				defaultValue = value;
				setDefValue = true;
			}
			catch (ParseException pex)
			{
				log.log(Level.FINER, "Cannot set default date value: "+value, pex);
				setDefValue = false;
			}
		}
		return setDefValue;
	}
	
	/**
	 * Parse the string and return a Date
	 * @param value of the field
	 * @return an object representing the field value, or null if an unknown field type
	 */
	@Override public String formatValue(Object value)
	{
		String formatted = null;
		if (value != null) {
			formatted = formatDate((Date)value);
		}
		return formatted;
	}
	
	/** Parse the value and return a Date */
	@Override public Object parseValue(String value)
	{
		try
		{
			return parseDate(value);
		}
		catch (ParseException pex)
		{
			log.log(Level.FINER, "Incorrect date value "+value, pex);
		}
		
		return null;
	}
	
	@Override public void setFieldDetails(Map<String, String> fieldDetails)
	{
		log.entering("DateFieldDef", "setFieldDetails");
		
		super.setFieldDetails(fieldDetails);
		try
		{
			String df = fieldDetails.get(Elements.DATE_FORMAT);
			if (df != null)
			{
				sdf.applyPattern(df);
				dateFormat = df;
			}
		}
		catch (IllegalArgumentException ex)
		{
			log.log(Level.WARNING, "Invalid date format", ex);
		}

		prepopulateDate = Boolean.parseBoolean(fieldDetails.get(Elements.PREPOPULATE_DATE));
		
		log.exiting("DateFieldDef", "setFieldDetails");
	}
	
	/**
	 * Get the field definition
	 * @return Map of name/value pairs
	 */
	@Override public Map<String, String> getFieldDetails()
	{
		Map<String, String> def = super.getFieldDetails();
		def.put(Elements.DATE_FORMAT, dateFormat);
		def.put(Elements.PREPOPULATE_DATE, Boolean.toString(prepopulateDate));
		
		return def;
	}
	
	@Override public Class getFieldClass()
	{
		return Date.class;
	}
	
	/**
	 * Set the date format. This cannot be null.
	 * @param newDateFormat the new date format
	 * @return false if newDateFormat is not legal
	 */
	public boolean setDateFormat(String newDateFormat)
	{
		log.entering("DateFieldDef", "setDateFormat", newDateFormat);
		
		boolean setPattern = false;
		if (newDateFormat != null && newDateFormat.length() > 0)
		{
			try
			{
				sdf.applyPattern(newDateFormat);
				dateFormat = newDateFormat;
				setPattern = true;
				log.fine("Set date format to "+dateFormat);
			}
			catch (IllegalArgumentException ex)
			{
				log.log(Level.WARNING, "Incorrect date format: "+newDateFormat, ex);
			}
		}
		
		log.exiting("DateFieldDef", "setDateFormat", setPattern);
		return setPattern;
	}
	
	/**
	 * Get the date format
	 * @return the current date format
	 */
	public String getDateFormat()
	{
		return dateFormat;
	}
	
	/**
	 * Convenience method for parsing a Date into a String which means that
	 * other classes don't have to keep setting up SimpleDateFormat objects.
	 * @param d the Date to format
	 * @return the Date parsed into a String using the current date format
	 */
	public String formatDate(Date d)
	{
		String formattedDate = null;
		if (d != null) {
			formattedDate = sdf.format(d);
		}
		return formattedDate;
	}
	
	/**
	 * Convenience method for parsing a String into a Date which means that
	 * other classes don't have to keep setting up SimpleDateFormat objects.
	 * @param s the date to parse
	 * @return a Date object
	 * @exception ParseException thrown if the date cannot be parsed
	 */
	public Date parseDate(String s) throws ParseException
	{
		Date d = null;
		if (s != null) {
			d = sdf.parse(s);
		}
		return d;
	}
	
	/**
	 * Is the today's date prepopulated in date fields
	 * @return true if prepopulation is on
	 */
	public boolean isDatePrepopulated()
	{
		return prepopulateDate;
	}
	
	/**
	 * Set whether to prepopulate date fields
	 * @param prepopulate set to true to prepopulate date fields
	 */
	public void setDatePrepopulated(boolean prepopulate)
	{
		prepopulateDate = prepopulate;
	}
	
	@Override public void setGUIValue(JComponent comp, Object value, boolean editing, boolean newCard)
	{
		log.entering("DateFieldDef", "setGUIValue");
		
		if (editing)
		{
			JXDatePicker datePicker = (JXDatePicker)comp;
			if (value != null)
			{
				datePicker.setDate((Date)value);
			}
			else if (newCard)
			{
				if (defaultValue != null)
				{
					try
					{
						datePicker.setDate(parseDate(getDefaultValue()));
					}
					catch (ParseException pex)
					{
						log.log(Level.WARNING, "Cannot set date to default value: "+getDefaultValue(), pex);
					}
				}
				else if (isDatePrepopulated())
				{
					datePicker.setDate(new Date());
				}
			}
		}
		else if (value != null)
		{
			setTextValue(comp, formatValue(value));
		}
		
		log.exiting("DateFieldDef", "setGUIValue");
	}
	
	/**
	 * Get the GUI component that represents the field
	 */
	@Override public JComponent getGUIComponent(boolean newCard, boolean editCard, boolean copyCard, JCardsFrame frame)
	{
		log.entering("DateFieldDef", "getGUIComponent");
		
		JComponent comp;
		if (editCard)
		{
			// @todo extend to handle time picker
			Date d = (prepopulateDate ? new Date() : null);
			JXDatePicker datePicker = new JXDatePicker(d);
			datePicker.setFormats(dateFormat);
			datePicker.getMonthView().setSelectionModel(new SingleDaySelectionModel());
			datePicker.setName(getName());
			comp = datePicker;
		}
		else
		{
			JLabel l = new JLabel();
			l.setName(name);
			addCopyPopupMenu(l, editCard);
			addCopyDnD(l);
			comp = l;
		}
		
		log.exiting("DateFieldDef", "getGUIComponent");
		
		return comp;
	}
	
	/**
	 * Check the value entered by a user. This is different from parseValue in that
	 * it also takes account of the mandatory flag
	 * @param comp the component to be checked
	 * @return InputCheck
	 */
	@Override public InputCheck checkInputValue(JComponent comp)
	{
		log.entering("DateFieldDef", "checkInputValue");
		
		InputCheck ic = null;
		JXDatePicker datePicker = (JXDatePicker)comp;
		Date dateVal = datePicker.getDate();
		if (mandatory && dateVal == null)
		{
			datePicker.requestFocus();
			ic = InputCheck.mandatoryInputCheck(name);
		}
		else
		{
			ic = InputCheck.validInputCheck(dateVal);
		}
		
		log.exiting("DateFieldDef", "checkInputValue", ic);
		
		return ic;
	}
}
