/*
 * JCards
 * Copyright (C) 2002-2010 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package nrs.jcards.fields;

import java.awt.Toolkit;
import java.awt.datatransfer.DataFlavor;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Logger;
import javax.swing.*;
import javax.swing.text.DefaultEditorKit;
import javax.swing.text.JTextComponent;
import nrs.jcards.gui.JCardsFrame;
import nrs.jcards.gui.JCardsMenuItem;
import nrs.jcards.io.Elements;

/**
 * A field definition.
 *
 * @author Copyright &#169; 2002-2010 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public abstract class FieldDef
{
	protected final static Logger log = Logger.getLogger(FieldDef.class.getName());
	
	protected FieldType type = FieldType.STRING;
	protected String name = null;
	protected String defaultValue = null;
	protected boolean mandatory = false;
	
	protected FieldDef()
	{}
	
	/**
	 * Set the default value which is limited by the type,
	 * i.e. you can't set it to 'A' if the type is Integer
	 * @param value the value to set
	 * @return true if the default value has been set; false if it is of the incorrect type
	 */
	public abstract boolean setDefaultValue(String value);
	
	/**
	 * Set the value of the GUI component programmatically
	 *
	 * @param comp the component to be updated
	 * @param value the value to set
	 * @param editing true if edit component; false for read-only component
	 * @param newCard true if opened a new card
	 */
	public abstract void setGUIValue(JComponent comp, Object value, boolean editing, boolean newCard);
	
	/**
	 * Check the value entered by a user. This is different from parseValue in that
	 * it also takes account of the mandatory flag
	 * @param comp the component to be checked
	 * @return InputCheck
	 */
	public abstract InputCheck checkInputValue(JComponent comp);
	
	protected void setTextValue(JComponent comp, Object value)
	{
		if (value != null)
		{
			JLabel label = (JLabel)comp;
			label.setText(value.toString());
		}
	}
	
	/**
	 * When loading a file set any field specific info
	 * @param fieldDetails Map of name/value pairs
	 */
	public void setFieldDetails(Map<String, String> fieldDetails)
	{
		mandatory = (Boolean.parseBoolean(fieldDetails.get(Elements.MANDATORY)));
		setDefaultValue(fieldDetails.get(Elements.DEFAULT));
	};
	
	/**
	 * Parse the value and return the correct type
	 * @param value the value to parse
	 * @return the parsed value
	 */
	public abstract Object parseValue(String value);
	
	/**
	 * Get the Class that represents the data value. This method
	 * makes it easier to group FieldDefs according to their Class.
	 */
	public Class getFieldClass()
	{
		return String.class;
	}
	
	/** Get the GUI component that represents the field */
	public abstract JComponent getGUIComponent(boolean newCard, boolean editCard, boolean copyCard, JCardsFrame frame);
	
	/**
	 * Create a new FieldDef based on the type
	 * @param name name of the field
	 * @param type type of the field
	 */
	public static FieldDef createFieldDef(String name, FieldType type)
	{
		log.entering("FieldDef", "createFieldDef", name);
		
		assert type != null : "createFieldDef: null FieldType";
		return type.createFieldDef(name);
	}
	
	public FieldType getType()
	{
		return type;
	}
	
	/** Get a String to represent the type */
	public String getTypeName()
	{
		return type.getName();
	}
	
	/** Get the default value */
	public String getDefaultValue()
	{
		return defaultValue;
	}
	
	/**
	 * Synonym for getDefaultValue
	 * @see #getDefaultValue
	 */
	public String getCalculation()
	{
		return getDefaultValue();
	}
	
	/**
	 * Get the name of the field
	 * @return name of the field
	 */
	public String getName()
	{
		return name;
	}
	
	/**
	 * Set the name of the field
	 * @param newName the name of the field
	 */
	public void setName(String newName)
	{
		name = newName;
	}
	
	/**
	 * Set this field to be mandatory or not
	 * @param mand set to true if you want the field to be mandatory
	 */
	public void setMandatory(boolean mand)
	{
		mandatory = mand;
	}
	
	/**
	 * Is this field mandatory
	 * @return true if mandatory
	 */
	public boolean isMandatory()
	{
		return mandatory;
	}
	
	/**
	 * Get a String representation of this class
	 * @return a String representation of this class
	 */
	@Override
	public String toString()
	{
		StringBuilder sb = new StringBuilder(name);
		sb.append(",");
		sb.append(type.getName());
		if (type != FieldType.BOOLEAN)
		{
			sb.append(",");
			sb.append(mandatory);
		}
		
		return sb.toString();
	}
	
	/**
	 * Get a representation of the field definition
	 * @return Map of name/value pairs
	 */
	public Map<String, String> getFieldDetails()
	{
		log.entering("FieldDef", "getFieldDetails");
		
		Map<String, String> def = new HashMap<String, String>();
		def.put(Elements.NAME, name);
		def.put(Elements.TYPE, getTypeName());
		def.put(Elements.MANDATORY, Boolean.toString(mandatory));
		if (defaultValue != null && defaultValue.length() > 0)
		{
			def.put(Elements.DEFAULT, defaultValue);
		}
		
		log.exiting("FieldDef", "getFieldDetails");
		
		return def;
	}
	
	/**
	 * Get a String representation of the Object
	 * @param obj the Object to format
	 * @return a String representation of the Object
	 */
	public String formatValue(Object obj)
	{
		return (obj == null ? null : obj.toString());
	}
	
	/**
	 * Add copy functionality to a JLabel via drag n drop
	 * @param comp the component to update
	 * @param editing if editing enable dropping text onto the component, false makes the component effectively read-only
	 */
	protected void addCopyDnD(JLabel comp)
	{
		comp.setTransferHandler(new TransferHandler("text")
		{
			@Override
			public boolean canImport(JComponent comp, DataFlavor[] transferFlavors)
			{
				return false;
			}
			
			@Override
			public int getSourceActions(JComponent c)
			{
				return TransferHandler.COPY;
			}
		});
		
		comp.addMouseListener(new MouseAdapter()
		{
			@Override
			public void mousePressed(MouseEvent e)
			{
				// ensure we don't conflict with any popup menu
				if (!e.isPopupTrigger())
				{
					JComponent c = (JComponent)e.getSource();
//					c.requestFocusInWindow();
					TransferHandler handler = c.getTransferHandler();
					handler.exportAsDrag(c, e, TransferHandler.COPY);
				}
			}
		});
	}
	
	/**
	 * Add a pop-up menu to a component.
	 * @param comp the component to decorate
	 */
	protected void addCopyPopupMenu(final JComponent comp, final boolean editing)
	{
		comp.addMouseListener(new MouseAdapter()
		{
			@Override
			public void mouseClicked(MouseEvent e)
			{
				if (e.isPopupTrigger())	{
					checkPopup(e);
				}
			}

			@Override
			public void mousePressed(MouseEvent e)
			{
				if (e.isPopupTrigger())	{
					checkPopup(e);
				}
			}

			@Override
			public void mouseReleased(MouseEvent e)
			{
				if (e.isPopupTrigger())	{
					checkPopup(e);
				}
			}

			public void checkPopup(MouseEvent e)
			{
				JPopupMenu copyMenu = new JPopupMenu();
				JMenuItem menuItem;
				
				if (comp instanceof JTextComponent)
				{
					// create a popup menu for copying/pasting values from the clipboard using default actions
					if (editing) {
						menuItem = new JCardsMenuItem(new DefaultEditorKit.CutAction(), "cut");
						copyMenu.add(menuItem);
					}
					
					menuItem = new JCardsMenuItem(new DefaultEditorKit.CopyAction(), "copy");
					copyMenu.add(menuItem);
					
					if (editing)
					{
						menuItem = new JCardsMenuItem(new DefaultEditorKit.PasteAction(), "paste");
						copyMenu.add(menuItem);
					}
				}
				else if (comp instanceof JLabel)
				{
					final JLabel label = (JLabel)comp;
					comp.setTransferHandler(new TransferHandler("text"));
					menuItem = new JCardsMenuItem(TransferHandler.getCopyAction(), "copy");
					menuItem.addActionListener(new ActionListener() {
						@Override
						public void actionPerformed(ActionEvent e)
						{
							TransferHandler handler = label.getTransferHandler();
							handler.exportToClipboard(label, Toolkit.getDefaultToolkit().getSystemClipboard(), TransferHandler.COPY);
						}
					});
					copyMenu.add(menuItem);
				}
				copyMenu.show(e.getComponent(), e.getX(), e.getY());
			}
		});
	}
}