/*
 * JCards
 * Copyright (C) 2002-2009 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package nrs.jcards.fields;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Image;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.util.Arrays;
import java.util.Map;
import java.util.logging.Level;
import javax.imageio.ImageIO;
import javax.swing.*;
import javax.swing.border.EmptyBorder;
import nrs.jcards.JCards;
import nrs.jcards.UserSettings;
import nrs.jcards.gui.FoldComponent;
import nrs.jcards.gui.JCardsButton;
import nrs.jcards.gui.JCardsFrame;
import nrs.jcards.gui.ScrollablePicture;
import nrs.jcards.io.Elements;
import nrs.jcards.io.JCAReader;
import nrs.jcards.io.JCAReader.SplitFilename;
import nrs.jcards.io.filter.ImageFileFilter;

/**
 * Image field definition
 *
 * @author Copyright &#169; 2002-2009 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class ImageFieldDef extends FieldDef implements ActionListener
{
	private JCardsFrame jcardsFrame;
	private JTextField imageTF;
	private boolean autoFold;
	private boolean storeInDB;
	private boolean maintainAspectRatio;
	private Dimension resizeDimension;

	public static final String PIC = "pic:";
	private static final String BROWSE = "browse";
	private static final String CLEAR = "clear.image";

	// used when selecting an image from the file system
	private class ImagePreviewPanel extends JLabel implements PropertyChangeListener
	{
		private String[] imgSuffixes = ImageIO.getReaderFileSuffixes();
		private int SIZE = 200;

		public ImagePreviewPanel()
		{
			Arrays.sort(imgSuffixes);
			setBackground(Color.WHITE);
			setBorder(new EmptyBorder(5, 20, 5, 20));
		}

		@Override
		public Dimension getPreferredSize()
		{
			return new Dimension(SIZE+20, SIZE+20);
		}

		@Override
		public Dimension getMinimumSize()
		{
			return new Dimension(SIZE, SIZE);
		}

		@Override
		public void propertyChange(PropertyChangeEvent evt)
		{
			String prop = evt.getPropertyName();
			if (JFileChooser.SELECTED_FILE_CHANGED_PROPERTY.equals(prop))
			{
				reset();
				File file = (File)evt.getNewValue();
				if (file == null || file.isDirectory())	{
					return;
				}

				SplitFilename sf = JCAReader.splitFilename(file.getName());
				if (Arrays.binarySearch(imgSuffixes, sf.extension) > -1) {
					showImage(file);
				}
			}
		}

		private void reset()
		{
			setIcon(null);
		}

		private void showImage(File imgFile)
		{
			ImageIcon icon = new ImageIcon(imgFile.getPath());

			// resize to a size appropriate for a preview
			float h = icon.getIconHeight();
			float w = icon.getIconWidth();

			if (w > SIZE && w >= h)
			{
				h = (SIZE/w) * h;
				w = SIZE;
			}
			else if (h > SIZE && h >= w)
			{
				w = (SIZE/h) * w;
				h = SIZE;
			}

			Image img = icon.getImage().getScaledInstance((int)w, (int)h, Image.SCALE_FAST);
			setIcon(new ImageIcon(img));
		}
	}

	/** Creates a new instance of ImageFieldDef */
	public ImageFieldDef(String name)
	{
		setName(name);
		type = FieldType.IMAGE;
	}
	
	/**
	 * Change the auto fold option
	 * @param autoFold if true the text area is initially hidden, if false it will be shown (default)
	 */
	public void setAutoFold(boolean autoFold)
	{
		this.autoFold = autoFold;
	}
	
	/**
	 * See if the text area is folded
	 * @return true if the text area is folded
	 */
	public boolean isAutoFold()
	{
		return autoFold;
	}
	
	/**
	 * Set the default value which is limited by the type,
	 * i.e. you can't set it to 'A' if the type is Integer
	 * @param value the value to set
	 * @return true if the default value has been set; false if it is of the incorrect type
	 */
	@Override
	public boolean setDefaultValue(String value)
	{
		boolean setDefaultValue = false;
		
		if (value == null)
		{
			defaultValue = null;
		}
		else
		{
			File f = new File(value);
			if (f.exists() && f.isFile())
			{
				defaultValue = value;
				setDefaultValue = true;
			}
		}
		
		return setDefaultValue;
	}
	
	/**
	 * Parse the value and return the correct type
	 * @param value the value to parse
	 * @return the parsed value
	 */
	@Override
	public Object parseValue(String value)
	{
		return value;
	}
	
	/**
	 * Set the value of the component programmatically
	 *
	 * @param comp the component to be updated
	 * @param value the value to set
	 * @param editing true if edit component; false for read-only component
	 * @param newCard true if opened a new card
	 */
	@Override
	public void setGUIValue(JComponent comp, Object value, boolean editing, boolean newCard)
	{
		log.entering("ImageFieldDef", "setGUIValue");
		
		if (editing)
		{
			JTextField label = (JTextField)comp;
			if (value != null)
			{
				label.setText((String)value);
			}
			else if (newCard && defaultValue != null)
			{
				label.setText(defaultValue);
			}
		}
		else if (value != null)
		{
			ScrollablePicture pic = (ScrollablePicture)comp;
			String imgName = value.toString();
			if (imgName.startsWith(PIC))
			{
				pic.setImage(jcardsFrame.loadImageFromDB(imgName), imgName);
			}
			else
			{
				try
				{
					pic.setImage(new FileInputStream(value.toString()), value.toString());
				}
				catch (FileNotFoundException fnex)
				{
					log.log(Level.WARNING, "Cannot load image {0}", value.toString());
					pic.displayError();
				}
			}
		}
		
		log.exiting("ImageFieldDef", "setGUIValue");
	}
	
	/**
	 * Check the value entered by a user. This is different from parseValue in that
	 * it also takes account of the mandatory flag
	 * @param comp the component to be checked
	 * @return InputCheck
	 */
	@Override
	public InputCheck checkInputValue(JComponent comp)
	{
		log.entering("ImageFieldDef", "checkInputValue");
		
		InputCheck ic = null;
		JTextField tf = (JTextField)comp;
		String text = tf.getText().trim();
		if (mandatory && text.length() == 0)
		{
			tf.requestFocus();
			ic = InputCheck.mandatoryInputCheck(name);
		}
		else if (text.length() == 0)
		{
			ic = InputCheck.validInputCheck(null);
		}
		
		if (ic == null)
		{
			if (text.startsWith(PIC))
			{
				ic = InputCheck.validInputCheck(text);
			}
			else
			{
				File f = new File(text);
				if (f.exists() && f.isFile())
				{
					ic = InputCheck.validInputCheck(text);
				}
				else
				{
					tf.requestFocus();
					ic = InputCheck.invalidInputCheck(JCards.getMessage("invalid.image"));
				}
			}
		}
		
		log.exiting("ImageFieldDef", "checkInputValue", ic);
		
		return ic;
	}
	
	/**
	 * Get the GUI component that represents the field
	 */
	@Override
	public JComponent getGUIComponent(boolean newCard, boolean editCard, boolean copyCard, JCardsFrame frame)
	{
		log.entering("ImageFieldDef", "getGUIComponent");

		jcardsFrame = frame;
		
		JComponent comp;
		if (editCard)
		{
			JPanel panel = new JPanel(new FlowLayout(FlowLayout.LEFT, 2, 0));
			imageTF = new JTextField(20);
			imageTF.setName(name);
			imageTF.setEditable(false);
			panel.add(imageTF);
			
			JButton browse = new JCardsButton(BROWSE);
			browse.addActionListener(this);
			panel.add(browse);

			JButton clear = new JCardsButton(CLEAR, JCardsFrame.getImage(UserSettings.ICON_CANCEL));
			clear.setText(null);
			clear.addActionListener(this);
			panel.add(clear);
			comp = panel;
		}
		else
		{
			ScrollablePicture pic = new ScrollablePicture(maintainAspectRatio, resizeDimension, jcardsFrame);
			pic.setName(name);
			FoldComponent fold = new FoldComponent(name, pic, !autoFold);
			comp = fold;
		}
		
		log.exiting("ImageFieldDef", "getGUIComponent");
			
		return comp;
	}
	
	/** User is browsing the file system for an image */
	@Override
	public void actionPerformed(ActionEvent e)
	{
		String cmd = e.getActionCommand();

		if (cmd.equals(BROWSE))
		{
			JFileChooser fc = new JFileChooser();
			fc.setDialogTitle(JCards.getLabelString("select.image"));
			fc.setApproveButtonText(JCards.getButtonString("select"));
			fc.setApproveButtonToolTipText(JCards.getLabelString("select.image"));
			fc.setApproveButtonMnemonic(KeyEvent.VK_S);
			fc.setFileFilter(new ImageFileFilter());
			String fileName = imageTF.getText();
			if (!fileName.startsWith(PIC)) {
				fc.setSelectedFile(new File(fileName));
			}

			ImagePreviewPanel ipp = new ImagePreviewPanel();
			fc.addPropertyChangeListener(ipp);
			fc.setAccessory(ipp);

			if (fc.showOpenDialog(jcardsFrame) == JFileChooser.APPROVE_OPTION)
			{
				File f = fc.getSelectedFile();
				if (f.exists() && f.isFile())
				{
					imageTF.setText(f.getAbsolutePath());
				}
			}
		}
		else if (cmd.equals(CLEAR))
		{
			imageTF.setText(null);
		}
	}
	
	/**
	 * Get the field definition
	 * @return Map of name/value pairs
	 */
	@Override
	public Map<String, String> getFieldDetails()
	{
		Map<String, String> def = super.getFieldDetails();
		def.put(Elements.AUTO_FOLD, Boolean.toString(autoFold));
		def.put(Elements.IMAGE_STORE_IN_DB, Boolean.toString(storeInDB));
		def.put(Elements.IMAGE_MAINTAIN_ASPECT_RATIO, Boolean.toString(maintainAspectRatio));
		if (resizeDimension != null) {
			def.put(Elements.IMAGE_RESIZE_DIM, resizeDimension.width+"x"+resizeDimension.height);
		}
		
		return def;
	}
	
	@Override
	public void setFieldDetails(Map<String, String> fieldDetails)
	{
		log.entering("ImageFieldDef", "setFieldDetails");
		
		super.setFieldDetails(fieldDetails);
		autoFold = Boolean.parseBoolean(fieldDetails.get(Elements.AUTO_FOLD));
		storeInDB = Boolean.parseBoolean(fieldDetails.get(Elements.IMAGE_STORE_IN_DB));
		maintainAspectRatio = Boolean.parseBoolean(fieldDetails.get(Elements.IMAGE_MAINTAIN_ASPECT_RATIO));

		String dim = fieldDetails.get(Elements.IMAGE_RESIZE_DIM);
		if (dim != null)
		{
			String[] dims = fieldDetails.get(Elements.IMAGE_RESIZE_DIM).split("x");
			if (dims.length == 2)
			{
				try {
					resizeDimension = new Dimension(Integer.parseInt(dims[0]), Integer.parseInt(dims[1]));
				} catch (NumberFormatException nfex) {
					log.log(Level.WARNING, "Invalid image resize dimension specified - {0}", dim);
				}
			}
		}
		
		log.exiting("ImageFieldDef", "setFieldDetails");
	}

	/**
	 * Should images be stored in the DB or left on the file system
	 * @return true/false
	 */
	public boolean isStoreInDB()
	{
		return storeInDB;
	}

	/**
	 * Set whether images should be stored in the DB or left on the file system
	 * @param storeInDB true/false
	 */
	public void setStoreInDB(boolean storeInDB)
	{
		this.storeInDB = storeInDB;
	}

	/**
	 * Auto resize of images
	 * @return the resizeDimension or null if not set
	 */
	public Dimension getResizeDimension()
	{
		return resizeDimension;
	}

	/**
	 * Set whether an image should be auto resized
	 * @param resizeDimension the resizeDimension to set or null
	 */
	public void setResizeDimension(Dimension resizeDimension)
	{
		this.resizeDimension = resizeDimension;
	}

	/**
	 * @return the maintainAspectRatio
	 */
	public boolean isMaintainAspectRatio()
	{
		return maintainAspectRatio;
	}

	/**
	 * @param maintainAspectRatio the maintainAspectRatio to set
	 */
	public void setMaintainAspectRatio(boolean maintainAspectRatio)
	{
		this.maintainAspectRatio = maintainAspectRatio;
	}
}
