/*
 * JCards
 * Copyright (C) 2002-2009 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package nrs.jcards.fields;

import java.util.Map;
import java.util.logging.Level;
import javax.swing.*;
import javax.swing.text.JTextComponent;
import nrs.jcards.JCards;
import nrs.jcards.gui.FoldComponent;
import nrs.jcards.gui.JCardsFrame;
import nrs.jcards.io.Elements;

/**
 * Memo field definition
 *
 * @author Copyright &#169; 2002-2009 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class MemoFieldDef extends StringFieldDef
{
	protected int memoLines = 3;
	protected boolean displayAsHTML;
	protected boolean autoFold = false;

	public MemoFieldDef(String name)
	{
		super(name);
		type = FieldType.MEMO;
	}

	/** Creates a new instance of MemoFieldDef */
	public MemoFieldDef(String name, int memoLines)
	{
		this(name);
		this.memoLines = memoLines;
	}

	/** Set the number of lines used for this memo field */
	public void setMemoLines(int ml)
	{
		memoLines = ml;
	}

	/** Get the number of lines used for a memo field */
	public int getMemoLines()
	{
		return memoLines;
	}

	/**
	 * Sets whether the content of the Memo field is displayed as HTML or plain text
	 * @param displayAsHTML
	 */
	public void setDisplayAsHTML(boolean displayAsHTML)
	{
		this.displayAsHTML = displayAsHTML;
	}

	/**
	 * See if this field displays the contents as HTML
	 * @return whether text is displayed as HTML or plain text
	 */
	public boolean isDisplayAsHTML()
	{
		return displayAsHTML;
	}
	
	/**
	 * Change the auto fold option
	 * @param autoFold if true the text area is initially hidden, if false it will be shown (default)
	 */
	public void setAutoFold(boolean autoFold)
	{
		this.autoFold = autoFold;
	}
	
	/**
	 * See if the text area is folded
	 * @return true if the text area is folded
	 */
	public boolean isAutoFold()
	{
		return autoFold;
	}

	@Override public void setFieldDetails(Map<String, String> fieldDetails)
	{
		log.entering("MemoFieldDef", "setFieldDetails");
		
		super.setFieldDetails(fieldDetails);
		if (fieldDetails.containsKey(Elements.MEMO_LINES))
		{
			try
			{
				memoLines = Integer.parseInt(fieldDetails.get(Elements.MEMO_LINES));
			}
			catch (NumberFormatException nfex)
			{
				log.log(Level.WARNING, "Error setting memo lines {0}", nfex.getMessage());
			}
		}

		autoFold = Boolean.parseBoolean(fieldDetails.get(Elements.AUTO_FOLD));
		displayAsHTML = Boolean.parseBoolean(fieldDetails.get(Elements.DISPLAY_AS_HTML));
		
		log.exiting("MemoFieldDef", "setFieldDetails");
	}

	/**
	 * Get the field definition
	 * @return Map of name/value pairs
	 */
	@Override public Map<String, String> getFieldDetails()
	{
		Map<String, String> def = super.getFieldDetails();
		def.put(Elements.MEMO_LINES, Integer.toString(memoLines));
		def.put(Elements.DISPLAY_AS_HTML, Boolean.toString(displayAsHTML));
		def.put(Elements.AUTO_FOLD, Boolean.toString(autoFold));

		return def;
	}

	@Override public JComponent getGUIComponent(boolean newCard, boolean editCard, boolean copyCard, JCardsFrame frame)
	{
		log.entering("MemoFieldDef", "getGUIComponent");
		
		JTextComponent textComp;
		// @todo check layout is working
		if (editCard || !displayAsHTML)
		{
			JTextArea tc = new JTextArea(null, memoLines, 20);
			tc.setName(name);
			tc.setLineWrap(JCards.getUserSettings().isWordWrap());
			tc.setWrapStyleWord(true);
			tc.setEditable(editCard);
			tc.setBorder(BorderFactory.createEmptyBorder(4, 4, 4, 4));
			if (!editCard) {
				tc.setDropTarget(null);
			}
			textComp = tc;
		}
		else
		{
			JEditorPane tp = new JEditorPane();
			tp.setContentType("text/html");
			tp.setEditable(false);
			tp.setName(name);
			textComp = tp;
		}
		addCopyPopupMenu(textComp, editCard);
		textComp.setDragEnabled(true);
		
		FoldComponent foldComp = new FoldComponent(name, new JScrollPane(textComp), !autoFold);

		log.exiting("MemoFieldDef", "getGUIComponent");
			
		return foldComp;
	}

	/**
	 * Set the value of the component programmatically
	 *
	 * @param comp the component to be updated
	 * @param value the value to set
	 * @param editing true if edit component; false for read-only component
	 * @param newCard true if opened a new card
	 */
	@Override public void setGUIValue(JComponent comp, Object value, boolean editing, boolean newCard)
	{
		log.entering("MemoFieldDef", "setGUIValue");
		
		JTextComponent textComp = (JTextComponent) comp;
		if (value != null)
		{
			textComp.setText(value.toString());
		}
		else if (newCard && defaultValue != null)
		{
			textComp.setText(defaultValue);
		}
		textComp.setCaretPosition(0);
		
		log.exiting("MemoFieldDef", "setGUIValue");
	}
}
