/*
 * JCards
 * Copyright (C) 2002-2009 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package nrs.jcards.fields;

import java.text.NumberFormat;
import java.text.ParseException;
import java.util.Locale;
import java.util.Map;
import java.util.logging.Level;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.text.JTextComponent;
import nrs.jcards.JCards;
import nrs.jcards.io.Elements;

/**
 * Number field definition
 * 
 * @author Copyright &#169; 2002-2009 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class NumberFieldDef extends StringFieldDef
{
	/** Number type used for formatting */
    public enum NUMBER_TYPE { NUMBER, INTEGER, CURRENCY, PERCENTAGE };
	
	private NUMBER_TYPE numType = NUMBER_TYPE.NUMBER;
	
	/** Formatter for numbers */
	protected NumberFormat numFormat;
	protected Locale locale;
	
	protected final static String ERROR = "<ERROR>";
	public final static double ERROR_NUM = -99999;
	
	/** Creates a new instance of DoubleFieldDef */
	public NumberFieldDef(String name)
	{
		super(name);
		type = FieldType.NUMBER;
		
		numFormat = NumberFormat.getNumberInstance();
		locale = Locale.getDefault();
	}
	
	@Override public boolean setDefaultValue(String value)
	{
		log.entering("NumberFieldDef", "setDefaultValue", value);
		
		boolean set = false;
		if (value != null)
		{
			try
			{
				numFormat.parse(value);
				defaultValue = value;
				set = true;
			}
			catch (ParseException pex)
			{
				set = false;
			}
		}
		
		log.exiting("NumberFieldDef", "setDefaultValue", defaultValue);
		
		return set;
	}
	
	/**
	 * Get a String representation of the Number
	 * @param obj the Object to format
	 * @return a formatted representation of the Number
	 */
	@Override public String formatValue(Object value)
	{
		return numFormat.format(value);
	}
	
	/**
	 * Parse the value and return the correct type
	 * @param value the value to parse
	 * @return the parsed value
	 */
	@Override public Object parseValue(String value)
	{
		log.entering("NumberFieldDef", "parseValue", value);
		
		Object parsedValue = null;
		if (value != null && value.trim().length() > 0)
		{
			try
			{
				switch (numType)
				{
					case INTEGER:
						parsedValue = numFormat.parse(value).intValue();
						break;

					default:
						// get a default number format that doesn't worry about any percentage or currency symbols
						NumberFormat tempNF = NumberFormat.getNumberInstance(locale);
						parsedValue = tempNF.parse(value).doubleValue();
						break;
				}
			}
			catch (ParseException pex)
			{
				log.log(Level.WARNING, "NumType is {0}", numType.toString());
				log.log(Level.WARNING, "Invalid double value "+value, pex);
			}
		}

		log.exiting("NumberFieldDef", "parseValue", parsedValue);
		
		return parsedValue;
	}
	
	@Override public Class getFieldClass()
	{
		return Number.class;
	}
	
	/**
	 * Check the value entered by a user. This is different from parseValue in that
	 * it also takes account of the mandatory flag
	 * @param comp the component to be checked
	 * @return InputCheck
	 */
	@Override public InputCheck checkInputValue(JComponent comp)
	{
		log.entering("NumberFieldDef", "checkInputValue", comp);
		
		InputCheck val = super.checkInputValue(comp);
		if (val.isValid())
		{
			JTextComponent text = (JTextComponent)comp;
			String textVal = text.getText().trim();
			Object obj = parseValue(textVal);
			if (obj == null && textVal != null && textVal.length() > 0)
			{
				text.requestFocus();
				val = InputCheck.invalidInputCheck(JCards.getFormattedMessage("incorrect.value.double", textVal));
			}
			else
			{
				val = InputCheck.validInputCheck(obj);
			}
		}

		return val;
	}
	
	/**
	 * Set the value of the component programmatically
	 *
	 * @param comp the component to be updated
	 * @param value the value to set
	 * @param editing true if edit component; false for read-only component
	 * @param newCard true if opened a new card
	 */
	@Override public void setGUIValue(JComponent comp, Object value, boolean editing, boolean newCard)
	{
		log.entering("NumberFieldDef", "setGUIValue", value);
		
		if (editing && comp instanceof JTextComponent)
		{
			JTextComponent text = (JTextComponent)comp;
			if (value != null)
			{
				text.setText(value.toString());
			}
			else if (newCard && defaultValue != null)
			{
				text.setText(defaultValue);
			}
			text.setCaretPosition(0);
		}
		else
		{
			if (value != null)
			{
				JLabel label = (JLabel)comp;
				Number num = (Number)value;
				label.setText(numFormat.format(num));
			}
		}
		
		log.exiting("NumberFieldDef", "setGUIValue");
	}
	
	/**
	 * Get the number of displayed decimal places
	 * @return number of displayed decimal places
	 */
	public int getDecimalPlaces()
	{
		return numFormat.getMaximumFractionDigits();
	}
	
	/**
	 * Set the number of digits after the decimal place
	 * @param decimalPlaces number of decimal places to display
	 */
	public void setDecimalPlaces(int decimalPlaces)
	{
		numFormat.setMaximumFractionDigits(decimalPlaces);
	}
	
	/**
	 * Get the locale used by this number formatter
	 * @return Locale
	 */
	public Locale getLocale()
	{
		return locale;
	}
	
	public NUMBER_TYPE getNumType()
	{
		return numType;
	}
	
	public void setNumType(NUMBER_TYPE numType)
	{
		setNumType(numType, Locale.getDefault());
	}
	
	/**
	 * Set the numeric type. If set to CURRENCY decimalPlaces automatically gets
	 * set to 2. If set to INTEGER decimalPlaces automatically gets set to 0.
	 * @param numType the new type
	 * @param locale used for formatting
	 * @see #setNumType(INumberField.NUMBER_TYPE)
	 */
	public void setNumType(NUMBER_TYPE numType, Locale locale)
	{
		log.entering("NumberFieldDef", "setNumType", numType);
		
		this.numType = numType;
		this.locale = locale;
		
		switch (numType)
		{
			case CURRENCY:
				numFormat = NumberFormat.getCurrencyInstance(locale);
				setDecimalPlaces(2);
				break;
				
			case PERCENTAGE:
				numFormat = NumberFormat.getPercentInstance(locale);
				break;
				
			case INTEGER:
				numFormat = NumberFormat.getIntegerInstance(locale);
				setDecimalPlaces(0);
				break;
				
			default:
				numFormat = NumberFormat.getNumberInstance(locale);
				break;
		}
		
		log.exiting("NumberFieldDef", "setNumType");
	}
	
	/**
	 * When loading a file get any field specific info
	 * @param fieldDetails map of name/value pairs
	 */
	@Override public void setFieldDetails(Map<String, String> fieldDetails)
	{
		log.entering("NumberFieldDef", "setFieldDetails");

		super.setFieldDetails(fieldDetails);
		if (fieldDetails.containsKey(Elements.NUM_TYPE))
		{
			Locale l = Locale.getDefault();
			if (fieldDetails.containsKey(Elements.LOCALE))
			{
				String[] countryVariant = fieldDetails.get(Elements.LOCALE).split("_");
				if (countryVariant.length == 2) {
					l = new Locale(countryVariant[0], countryVariant[1]);
				}
			}
			setNumType(NUMBER_TYPE.valueOf(fieldDetails.get(Elements.NUM_TYPE).toUpperCase()), l);
		}
		
		if (fieldDetails.containsKey(Elements.DECIMAL_PLACES))
		{
			try
			{
				int decimalPlaces = Integer.parseInt(fieldDetails.get(Elements.DECIMAL_PLACES));
				setDecimalPlaces(decimalPlaces);
			}
			catch (NumberFormatException nfex)
			{
				log.log(Level.SEVERE, "Incorrect value specified for decimal places", nfex);
			}
		}
		
		log.exiting("NumberFieldDef", "setFieldDetails");
	}
	
	/**
	 * Get the field definition
	 * @return Map of name/value pairs
	 */
	@Override public Map<String, String> getFieldDetails()
	{
		log.entering("NumberFieldDef", "getFieldDetails");
		
		Map<String, String> def = super.getFieldDetails();
		def.put(Elements.DECIMAL_PLACES, Integer.toString(numFormat.getMaximumFractionDigits()));
		def.put(Elements.NUM_TYPE, numType.toString());
		def.put(Elements.LOCALE, locale.toString());
		
		log.exiting("NumberFieldDef", "getFieldDetails");
		
		return def;
	}
}
