/*
 * JCards
 * Copyright (C) 2002-2009 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package nrs.jcards.fields;

import java.awt.Color;
import java.awt.Cursor;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.logging.Level;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JTextField;
import javax.swing.text.JTextComponent;
import nrs.jcards.JCards;
import nrs.jcards.gui.JCardsFrame;

/**
 * String field definition
 *
 * @author Copyright &#169; 2002-2009 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class StringFieldDef extends FieldDef
{
	private JCardsFrame jcardsFrame;

	/** Holds a URL as well */
	private class URLLabel extends JLabel implements MouseListener
	{
		private URL url;

		public void setURL(URL url)
		{
			assert url != null : "Cannot set url to null";
			
			this.url = url;
			super.setText("<html><u>" + getText() + "</u></html>");
			setForeground(Color.blue);
			setToolTipText(JCards.getLabelString("url.tt"));
			setCursor(new Cursor(Cursor.HAND_CURSOR));
			
			addMouseListener(this);
		}

		public URL getURL()
		{
			return url;
		}

		@Override public void setText(String text)
		{
			super.setText(text);
			
			// check for a URL
			String lval = text.toLowerCase();
			if (lval.startsWith("http") || lval.startsWith("ftp"))
			{
				try
				{
					URL tryurl = new URL(lval); // check for a valid URL
					setURL(tryurl);
				}
				catch (MalformedURLException mfex)
				{
					log.log(Level.WARNING, "Invalid URL", mfex);
				}
			}
		}
		
		@Override public void mousePressed(MouseEvent e)
		{
			if (url != null) {
				jcardsFrame.displayURL(url);
			}
		}
		
		@Override public void mouseReleased(MouseEvent e) {}
		@Override public void mouseExited(MouseEvent e) {}
		@Override public void mouseEntered(MouseEvent e) {}
		@Override public void mouseClicked(MouseEvent e) {}
	}

	/** Creates a new instance of StringFieldDef */
	public StringFieldDef(String name)
	{
		setName(name);
		type = FieldType.STRING;
	}

	/**
	 * Set the default value which is limited by the type,
	 * i.e. you can't set it to 'A' if the type is Integer
	 * @param value the value to set
	 * @return true if the default value has been set; false if it is of the incorrect type
	 */
	@Override public boolean setDefaultValue(String value)
	{
		defaultValue = value;
		return true;
	}

	/**
	 * Based on the field type, return an Object of the right type.
	 * @param value of the field
	 * @return exactly what is passed in
	 */
	@Override public Object parseValue(String value)
	{
		return value;
	}

	/**
	 * Check the value entered by a user. This is different from parseValue in that
	 * it also takes account of the mandatory flag
	 * @param comp the component to be checked
	 * @return InputCheck
	 */
	@Override public InputCheck checkInputValue(JComponent comp)
	{
		log.entering("StringFieldDef", "checkInputValue");
		
		InputCheck ic = null;
		JTextComponent text = (JTextComponent)comp;
		String val = text.getText();
		if (mandatory && (val == null || val.length() == 0))
		{
			text.requestFocus();
			ic = InputCheck.mandatoryInputCheck(name);
		}
		else
		{
			if (val.length() == 0)
			{
				val = null;
			}
			ic = InputCheck.validInputCheck(val);
		}
		
		log.exiting("StringFieldDef", "checkInputValue", ic);
		
		return ic;
	}

	/**
	 * Set the value of the component programmatically
	 *
	 * @param comp the component to be updated
	 * @param value the value to set
	 * @param editing true if edit component; false for read-only component
	 * @param newCard true if opened a new card
	 */
	@Override public void setGUIValue(JComponent comp, Object value, boolean editing, boolean newCard)
	{
		log.entering("StringFieldDef", "setGUIValue");
		
		if (editing)
		{
			JTextComponent text = (JTextComponent) comp;
			if (value != null)
			{
				text.setText(value.toString());
			}
			else if (newCard && defaultValue != null)
			{
				text.setText(defaultValue);
			}
			text.setCaretPosition(0);
		}
		else
		{
			if (value != null)
			{
				URLLabel label = (URLLabel)comp;
				String sval = value.toString();
				label.setText(sval);
			}
		}
		
		log.exiting("StringFieldDef", "setGUIValue");
	}

	/**
	 * Get the GUI component that represents the field
	 */
	@Override public JComponent getGUIComponent(boolean newCard, boolean editCard, boolean copyCard, JCardsFrame frame)
	{
		log.entering("StringFieldDef", "getGUIComponent");
		
		JComponent comp;
		if (editCard)
		{
			JTextField tf = new JTextField(10);
			tf.setDragEnabled(true);
			comp = tf;
		}
		else
		{
			URLLabel label = new URLLabel();
			jcardsFrame = frame;
			addCopyDnD(label);
			comp = label;
		}
		addCopyPopupMenu(comp, editCard);

		comp.setName(name);
		
		log.exiting("StringFieldDef", "getGUIComponent");
		
		return comp;
	}
}
