/*
 * JCards
 * Copyright (C) 2002-2009 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package nrs.jcards.gui;

import java.awt.Color;
import java.awt.Component;
import java.awt.Toolkit;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.text.ParseException;
import java.util.Date;
import javax.swing.InputVerifier;
import javax.swing.JComponent;
import javax.swing.JTextField;
import nrs.jcards.fields.DateFieldDef;

/**
 * Extension of JTextField that knows about the date format and can display a hint
 *
 * @author Copyright &#169; 2002-2009 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class DateField extends JTextField implements FocusListener
{
	private DateFieldDef dfd;

	/** Check that the user entered the correct date format */
	private class DateVerifier extends InputVerifier
	{
		@Override public boolean verify(JComponent input)
		{
			String text = getText();
			String dateFormat = dfd.getDateFormat();
			boolean valid = false;
			
			// check for no entry
			if (text.length() == 0 || dateFormat.equals(getText()))
			{
				valid = true;
			}
			else
			{
				// is date valid?
				try
				{
					Date d = dfd.parseDate(getText());
					if (d != null) {
						valid = true;
					}
				}
				catch (ParseException pex)
				{
					Toolkit.getDefaultToolkit().beep();
				}
			}

			return valid;
		}
	}

	/** Creates a new instance of DateField */
	public DateField(DateFieldDef dfd)
	{
		this.dfd = dfd;

		setText(dfd.getDateFormat());
		setForeground(Color.darkGray);

		setToolTipText(dfd.getDateFormat());
		addFocusListener(this);
		setInputVerifier(new DateVerifier());
	}

	/** Get the DateFieldDef held by this field */
	public DateFieldDef getDateFieldDef()
	{
		return dfd;
	}

	/**
	 * This method is called whenever a date field gets the focus.
	 * The text field will be set to blank if it contains the date hint.
	 */
	@Override public void focusGained(FocusEvent e)
	{
		Component com = e.getComponent();
		DateField df = (DateField) com;
		if (df.getText().equals(df.getDateFieldDef().getDateFormat()))
		{
			df.setText(null);
			df.setForeground(Color.black);
		}
	}

	/**
	 * This method is called whenever a date field loses the focus.
	 * If the text field is blank the date hint will be reapplied.
	 */
	@Override public void focusLost(FocusEvent e)
	{
		DateField df = (DateField) e.getComponent();

		if (df.getText().length() == 0)
		{
			if (df.getDateFieldDef().isDatePrepopulated())
			{
				df.setText(df.getDateFieldDef().formatDate(new Date()));
			}
			else
			{
				df.setText(df.getDateFieldDef().getDateFormat());
			}
			df.setForeground(Color.darkGray);
		}
	}
}
