/*
 * JCards
 * Copyright (C) 2002-2009 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * JCards is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * JCards is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JCards; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package nrs.jcards.gui;

import java.awt.*;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;
import javax.swing.*;
import javax.swing.text.JTextComponent;
import nrs.jcards.Card;
import nrs.jcards.Counter;
import nrs.jcards.DBDef;
import nrs.jcards.JCards;
import nrs.jcards.fields.CalculatedFieldDef;
import nrs.jcards.fields.CounterFieldDef;
import nrs.jcards.fields.FieldDef;
import nrs.jcards.fields.FieldType;
import nrs.jcards.gui.dialog.CardDialog;

/**
 * Class which displays and edits Cards
 * @todo update calculations automatically - add listener to affected fields
 * @author Copyright &#169; 2002-2009 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class DisplayCardPanel extends JPanel
{
	private final static Logger log = Logger.getLogger(DisplayCardPanel.class.getName());
	
	private CardDialog cardDialog = null;
	private Card originalCard;
	private Map<String, JComponent> entries = null;
	private boolean newCard, editCard;
	private DBDef def = null;
	private Font componentFont;
	private Font optionalLabelFont;
	private Font mandatoryLabelFont;
	
	/**
	 * Initialise the DisplayCardPanel
	 * @param cardDialog parent dialog
	 * @param newCard set to true if editing a new card
	 * @param editCard set to true to auto-edit
	 */
	public DisplayCardPanel(CardDialog cardDialog, boolean newCard, boolean editCard)
	{
		this.cardDialog = cardDialog;
		this.newCard = newCard;
		this.editCard = editCard;
		
		add(new JLabel("NO SELECTION"));
		
		componentFont = UIManager.getFont("Label.font");
		optionalLabelFont = componentFont.deriveFont(Font.PLAIN);
		mandatoryLabelFont = componentFont.deriveFont(Font.BOLD);
		
		// cache some of the settings from DBDef
		def = DBDef.getInstance();
		
		setBackground(Color.white);
	}
	
	/**
	 * Display a card
	 * @param card the Card definition
	 * @param jcardsFrame parent frame
	 * @param cDialog the CardDialog so we can edit categories
	 * @param copyCard set to true if copying an existing card
	 */
	public void display(Card card, JCardsFrame jcardsFrame, CardDialog cDialog, boolean copyCard)
	{
		log.entering("DisplayCardPanel", "display");
		
		originalCard = card;
		
		FieldDef fd;
		JLabel l;
		JComponent c;
		Object objValue = null;
		boolean displayFieldLabels = def.displayFieldNames();
		
		// don't bother if not displayed
		if (!isVisible()) {
			return;
		}
		
		int numberOfFields = def.getNumberOfFields();
		entries = new HashMap<String, JComponent>(numberOfFields);
		
		// remove any existing components
		removeAll();
		
		GridBagLayout gbl = new GridBagLayout();
		GridBagConstraints gbc = new GridBagConstraints();
		
		JPanel fieldPanel = new JPanel(gbl);
		fieldPanel.setBorder(BorderFactory.createEmptyBorder(10, 10, 10, 10));
		
		gbc.insets = new Insets(2, 2, 2, 2);
		gbc.anchor = GridBagConstraints.NORTHWEST;
		gbc.weightx = 0.0;
		gbc.weighty = 0.0;
		
		// set up and optionally populate the fields
		// don't display the line if not editing and the field has no value
		for (int i=0; i<numberOfFields; i++)
		{
			fd = def.getFieldDef(i);
			if (!newCard)
			{
				objValue = card.getValue(fd.getName());
				// don't display lines with no values
				if (objValue == null && !editCard)
				{
					continue;
				}
			}
			
			// get the edit component
			c = fd.getGUIComponent(newCard, editCard, copyCard, jcardsFrame);
			
			// add label
			if ((displayFieldLabels || editCard) && !(c instanceof FoldComponent))
			{
				gbc.gridwidth = 1;
				gbc.fill = GridBagConstraints.NONE;
				gbc.weightx = 0.0;
				gbc.anchor = GridBagConstraints.NORTHWEST;
				String lname = fd.getName();
				if (def.displayColons()) {
					lname += ":";
				}
				l = new JLabel(lname);
				if (fd.isMandatory())
				{
					l.setFont(mandatoryLabelFont);
				}
				else
				{
					l.setFont(optionalLabelFont);
				}
				fieldPanel.add(l, gbc);
			}
			
			// see if the FoldComponent needs to have it's font set
			if (c instanceof FoldComponent && fd.isMandatory())
			{
				FoldComponent fc = (FoldComponent)c;
				fc.setFont(mandatoryLabelFont);
			}
			else
			{
				c.setFont(componentFont);
			}
			
			// add input/display field			
			addComponent(fieldPanel, c);
			// find the editable GUI component as some of the displayable
			// items may be composite items
			c = getNamedGUIComponent(fd.getName(), c);
			
			entries.put(fd.getName(), c);
		}
		setValues(false);
		
		// add a dummy component to take up any remaining vertical space
		l = new JLabel();
		gbc = new GridBagConstraints();
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		gbc.fill = GridBagConstraints.BOTH;
		gbc.weightx = 1.0;
		gbc.weighty = 1.0;
		fieldPanel.add(l, gbc);
		
		JScrollPane scroll = new JScrollPane(fieldPanel);
		setLayout(new BorderLayout());
		add(scroll, BorderLayout.CENTER);
		validate();
		
		log.exiting("DisplayCardPanel", "display");
	}
	
	/**
	 * Examine the GUI components looking for one with the right name
	 * @param fieldName the field to look for
	 * @param c the component to start searching from
	 * @return the looked for component
	 */
	private JComponent getNamedGUIComponent(String fieldName, JComponent c)
	{
		JComponent jcomp = null;
		
		if (fieldName.equals(c.getName()))
		{
			jcomp = c;
		}
		else
		{
			for (Component comp : c.getComponents())
			{
				if (comp instanceof JComponent)
				{
					jcomp = (JComponent)comp;
					if (fieldName.equals(jcomp.getName()))
					{
						break;
					}

					if (jcomp.getComponentCount() > 0)
					{
						jcomp = getNamedGUIComponent(fieldName, jcomp);
						break;
					}
				}
			}
		}
		
		return jcomp;
	}
	
	/** 
	 * Set the field values
	 * @param reset set to true if resetting values
	 */
	public void setValues(boolean reset)
	{
		log.entering("DisplayCardPanel", "setValues", reset);
		
		FieldDef fd;
		Object value;
		JComponent comp;
		
		// needs to always be applied to handle default values in new cards
		for (int i=0; i<def.getNumberOfFields(); i++)
		{
			fd = def.getFieldDef(i);
			comp = entries.get(fd.getName());
			if (comp != null)
			{
				value = originalCard == null ? null : originalCard.getValue(fd.getName());
				if (fd instanceof CounterFieldDef)
				{
					if (!newCard) {
						fd.setGUIValue(comp, value, editCard, newCard);
					}
				}
				else
				{
					fd.setGUIValue(comp, value, editCard, newCard);
				}
			}
		}
		
		log.exiting("DisplayCardPanel", "setValues");
	}
	
	/** Set the layout proerties and add the component */
	private void addComponent(JPanel p, JComponent c)
	{
		GridBagConstraints gbc = new GridBagConstraints();
		gbc.weightx = 1.0;
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		gbc.anchor = GridBagConstraints.NORTHWEST;
		gbc.fill = (c instanceof FoldComponent) ? GridBagConstraints.BOTH : GridBagConstraints.HORIZONTAL;
		gbc.insets = new Insets(2, 2, 2, 2);
		p.add(c, gbc);
	}
	
	/**
	 * Update the calculated fields.
	 * <p>Formulae are nominally correct as they are
	 * checked when entered, however due to field
	 * changes they may no longer be correct.</p>
	 */
	public void updateCalculations()
	{
		log.entering("DisplayCardPanel", "updateCalculations");
		
		FieldDef fd;
		CalculatedFieldDef cfd;
		CalculatedLabel calcLabel;
		JTextComponent tcomp;
		double result;
		String text;
		Object value;
		Map<String, Object>values = new HashMap<String, Object>(entries.size());

		// first get the values of all the numerical fields
		for (int i=0; i<def.getNumberOfFields(); i++)
		{
			fd = def.getFieldDef(i);
			switch (fd.getType())
			{
				case NUMBER:
				case COUNTER:
					tcomp = (JTextComponent)entries.get(fd.getName());
					text = tcomp.getText();
					value = fd.parseValue(text);
					if (value == null && text.trim().length() > 0)
					{
						JOptionPane.showMessageDialog(cardDialog,
								JCards.getFormattedMessage("incorrect.value.double", fd.getName()),
								JCards.getMessage("error"), JOptionPane.ERROR_MESSAGE);
						tcomp.requestFocus();
						return;
					}
					values.put(fd.getName(), value);
					break;
					
				case CALCULATED:
					calcLabel = (CalculatedLabel)entries.get(fd.getName());
					values.put(fd.getName(), calcLabel.getValue());
					break;
			}
		}
		
		for (int i=0; i<def.getNumberOfFields(); i++)
		{
			fd = def.getFieldDef(i);
			if (fd.getType() == FieldType.CALCULATED)
			{
				calcLabel = (CalculatedLabel)entries.get(fd.getName());
				assert calcLabel != null : "DisplayCardPanel.updateCalculations: Cannot get calcLabel for "+fd.getName();
				try
				{
					result = def.updateCalculation(fd, values);
					cfd = (CalculatedFieldDef)fd;
					cfd.setGUIValue(calcLabel, result, editCard, newCard);
				}
				catch (IllegalArgumentException iex)
				{
					calcLabel.setText(JCards.getMessage("error").toUpperCase());
					JOptionPane.showMessageDialog(cardDialog, iex.getMessage(), JCards.getMessage("error"), JOptionPane.ERROR_MESSAGE);
				}
			}
		}
		
		log.exiting("DisplayCardPanel", "updateCalculations");
	}
	
	/** Rollback the counter fields */
	public void rollbackCounters()
	{
		log.entering("DisplayCardPanel", "rollbackCounters");
		
		FieldDef fd;
		CounterFieldDef cfd;
		Counter c;
		List<String> fds = def.getFieldNames();
		for (String fieldName : fds)
		{
			fd = def.getFieldDef(fieldName);
			if (fd.getType() == FieldType.COUNTER)
			{
				cfd = (CounterFieldDef)fd;
				c = def.getCounter(cfd.getCounter());
				c.rollback();
			}
		}
		
		log.exiting("DisplayCardPanel", "rollbackCounters");
	}
	
	/**
	 * Get all the JComponents being edited.
	 * @return a Map of JComponents
	 */
	public Map<String, JComponent> getEntries()
	{
		return entries;
	}
}
