/*
 * JCards
 * Copyright (C) 2002-2010 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package nrs.jcards.gui;

import java.awt.*;
import java.awt.event.*;
import java.text.NumberFormat;
import java.util.List;
import java.util.*;
import java.util.logging.Logger;
import javax.swing.*;
import javax.swing.border.TitledBorder;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import nrs.jcards.Cards;
import nrs.jcards.DBDef;
import nrs.jcards.JCards;
import nrs.jcards.UserSettings;
import nrs.jcards.fields.*;
import nrs.util.PostfixCalculator;

/**
 * This class displays the fields and allows additions, modifications and deletion
 * of fields.
 * 
 * @author Copyright &#169; 2002-2010 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class EditFieldPanel extends JPanel implements ActionListener, ItemListener
{
	private int fieldIndex = -1;
	private JDialog parent;
	private Cards cards;
	private JCheckBox manCB, todayCB, displayHTMLCB, autoFoldCB, imageCopyCB, maintainAspectRatioCB;
	private JComboBox fieldType, dateFormat, catName, countName, numType, localeCB;
	private JLabel numberExample;
	private JSpinner memoLines, decimalPlaces;
	private JTextField fieldNameTF, defaultValueTF, resizeWidthTF, resizeHeightTF;
	private JPanel imageResizePanel;
	private JButton updateButton;
	private Map<JComponent, JLabel> labels = new HashMap<JComponent, JLabel>(10);
	private final static Logger log = Logger.getLogger(EditFieldPanel.class.getName());
	
	private final static double EXAMPLE_NUMBER = 123456.789;
	
	/** Listener for number type changes */
	private class NumberTypeListener implements ItemListener
	{
		/** Method called when item in type choice list is changed */
		@Override
		public void itemStateChanged(ItemEvent e)
		{
			switch (numType.getSelectedIndex())
			{
				case 1:
					decimalPlaces.setValue(0);
					break;
					
				default:
					decimalPlaces.setValue(2);
					break;
			}
		}
	}
	
	/**
	 * Listens for changes in the locale, num type and decimal places fields to
	 * enable updates to the example number field
	 */
	private class LocaleListener implements ChangeListener, ItemListener
	{
		@Override
		public void stateChanged(ChangeEvent e)
		{
			updateExampleNumber();
		}
		
		@Override
		public void itemStateChanged(ItemEvent e)
		{
			updateExampleNumber();
		}
	}
			
	/** Creates a new instance of EditFieldPanel */
	public EditFieldPanel(JDialog parent, int fieldIndex, Cards cards)
	{
		this.fieldIndex = fieldIndex;
		this.parent = parent;
		this.cards = cards;
		
		FieldDef fdef = null;
		DBDef dbDef = DBDef.getInstance();
		if (fieldIndex > -1)
		{
			fdef = dbDef.getFieldDef(fieldIndex);
		}
		
		Font labelFont = UIManager.getFont("Label.font").deriveFont(Font.BOLD);
		setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(),
				JCards.getLabelString("edit.field"), TitledBorder.CENTER, TitledBorder.ABOVE_TOP));
		
		GridBagLayout gbl = new GridBagLayout();
		setLayout(gbl);
		
		// add fields
		JLabel l = new JCardsLabel("name");
		l.setFont(labelFont);
		fieldNameTF = new JTextField(20);
		l.setLabelFor(fieldNameTF);
		addElements(l, fieldNameTF);
		
		l = new JCardsLabel("type");
		l.setFont(labelFont);
		fieldType = new JComboBox();
		fieldType.setToolTipText(JCards.getButtonString("type.tt"));
		l.setLabelFor(fieldType);
		addElements(l, fieldType);
		boolean addedCategory = false;
		boolean addedCounter = false;
		boolean addedCalculated = false;
		boolean addedDate = false;
		
		if (fdef != null)
		{
			FieldType dType = fdef.getType();
			for (FieldType fType : FieldType.values())
			{
				if (dType.equals(fType) || fType.equals(FieldType.STRING) || fType.equals(FieldType.MEMO))
				{
					fieldType.addItem(fType.getName());
				}
			}
			
			switch (dType)
			{
				case CATEGORY:
				case LIST:
					addedCategory = true;
					break;
					
				case COUNTER:
					addedCounter = true;
					break;
					
				case CALCULATED:
					addedCalculated = true;
					break;
					
				case DATE:
					addedDate = true;
					break;
			}
		}
		else
		{
			for (FieldType fType : FieldType.values())
			{
				if (fType.isShowInFieldList()) {
					fieldType.addItem(fType.getName());
				}
			}
			addedCategory = addedCounter = addedCalculated = addedDate = true;
		}
		
		// only allow Categories and Counters if one or more have been defined
		if (addedCategory && !dbDef.hasCategories())
		{
			fieldType.removeItem(FieldType.CATEGORY.getName());
			fieldType.removeItem(FieldType.LIST.getName());
			addedCategory = false;
		}
		if (addedCounter && !dbDef.hasCounters())
		{
			fieldType.removeItem(FieldType.COUNTER.getName());
			addedCounter = false;
		}
		
		// only allow Calculated if other numeric field defined
		if (addedCalculated && !dbDef.areCalculationsAllowed())
		{
			fieldType.removeItem(FieldType.CALCULATED.getName());
		}
		fieldType.addItemListener(this);
		
		if (addedCategory)
		{
			// category
			l = new JCardsLabel("category");
			l.setFont(labelFont);
			catName = new JComboBox();
			// add categories
			Map<String, List<String>> cats = dbDef.getCategories();
			for (String cn : cats.keySet())
			{
				catName.addItem(cn);
			}
			catName.setEnabled(false);
			l.setLabelFor(catName);
			addElements(l, catName);
		}
		
		if (addedCounter)
		{
			// automatic counter
			l = new JCardsLabel("counter");
			l.setFont(labelFont);
			countName = new JComboBox();
			// add counters
			Map counts = dbDef.getCounters();
			Iterator it = counts.keySet().iterator();
			while (it.hasNext())
			{
				countName.addItem((String)it.next());
			}
			countName.setEnabled(false);
			l.setLabelFor(countName);
			addElements(l, countName);
		}
		
		// default value
		l = new JCardsLabel("default");
		l.setFont(labelFont);
		defaultValueTF = new JTextField(20);
		l.setLabelFor(defaultValueTF);
		addElements(l, defaultValueTF);
		
		// is the field mandatory
		l = new JCardsLabel("mandatory");
		l.setFont(labelFont);
		manCB = new JCheckBox();
		l.setLabelFor(manCB);
		addElements(l, manCB);
		
		if (addedDate)
		{
			// date
			l = new JCardsLabel("date.format");
			l.setFont(labelFont);
			dateFormat = new JComboBox();
			dateFormat.setEditable(true);
			dateFormat.setEnabled(false);
			dateFormat.setToolTipText(JCards.getLabelString("date.format.tt"));
			for (int i=0; i<UserSettings.DATE_FORMATS.length; i++)
			{
				dateFormat.addItem(UserSettings.DATE_FORMATS[i]);
			}
			l.setDisplayedMnemonic(KeyEvent.VK_F);
			l.setLabelFor(dateFormat);
			addElements(l, dateFormat);
			
			l = new JCardsLabel("use.today");
			l.setFont(labelFont);
			todayCB = new JCheckBox();
			todayCB.setEnabled(false);
			l.setDisplayedMnemonic(KeyEvent.VK_S);
			l.setLabelFor(todayCB);
			addElements(l, todayCB);
		}
		
		// number of memo lines
		l = new JCardsLabel("memo.lines");
		l.setFont(labelFont);
		SpinnerNumberModel memoModel = new SpinnerNumberModel(2, 2, 30, 1);
		memoLines = new JSpinner(memoModel);
		l.setLabelFor(memoLines);
		addElements(l, memoLines);
		flipComponent(false, memoLines);
		
		// memo HTML display
		l = new JCardsLabel("memo.display.as.html");
		l.setFont(labelFont);
		displayHTMLCB = new JCheckBox();
		displayHTMLCB.setToolTipText(JCards.getLabelString("memo.display.as.html.tt"));
		l.setLabelFor(displayHTMLCB);
		addElements(l, displayHTMLCB);
		flipComponent(false, displayHTMLCB);
		
		// integer type
		l = new JCardsLabel("num.type");
		l.setFont(labelFont);
		numType = new JComboBox();
		numType.addItem(JCards.getLabelString("number"));
		numType.addItem(JCards.getLabelString("integer"));
		numType.addItem(JCards.getLabelString("currency"));
		numType.addItem(JCards.getLabelString("percentage"));
		
		LocaleListener localeListener = new LocaleListener();
		numType.addItemListener(new NumberTypeListener());
		numType.addItemListener(localeListener);
		l.setLabelFor(numType);
		addElements(l, numType);
		flipComponent(false, numType);
		
		// locale
		l = new JCardsLabel("locale");
		l.setFont(labelFont);
		Vector<String> localeList  = new Vector<String>();
		String s;
		for (Locale loc : NumberFormat.getAvailableLocales())
		{
			s = loc.toString();
			if (s.length() == 5) {
				localeList.add(s);
			}
		}
		Collections.sort(localeList);
		localeCB = new JComboBox(localeList);
		localeCB.setSelectedItem(Locale.getDefault().toString());
		localeCB.addItemListener(localeListener);
		l.setLabelFor(localeCB);
		addElements(l, localeCB);
		flipComponent(false, localeCB);
		
		// number of decimal places
		l = new JCardsLabel("decimal.places");
		l.setFont(labelFont);
		SpinnerNumberModel dpModel = new SpinnerNumberModel(2, 0, 10, 1);
		dpModel.addChangeListener(localeListener);
		decimalPlaces = new JSpinner(dpModel);
		l.setLabelFor(decimalPlaces);
		addElements(l, decimalPlaces);
		flipComponent(false, decimalPlaces);
		
		// example formatted number
		l = new JCardsLabel("example");
		l.setFont(labelFont);
		numberExample = new JLabel(NumberFormat.getNumberInstance().format(EXAMPLE_NUMBER));
		l.setLabelFor(numberExample);
		addElements(l, numberExample);
		flipComponent(false, numberExample);
		
		// auto fold
		l = new JCardsLabel("auto.fold");
		l.setFont(labelFont);
		autoFoldCB = new JCheckBox();
		autoFoldCB.setToolTipText(JCards.getLabelString("auto.fold.tt"));
		l.setLabelFor(autoFoldCB);
		addElements(l, autoFoldCB);
		flipComponent(false, autoFoldCB);

		// image options
		l = new JCardsLabel("store.in.db");
		l.setFont(labelFont);
		imageCopyCB = new JCheckBox();
		imageCopyCB.setToolTipText(JCards.getLabelString("store.in.db.tt"));
		imageCopyCB.addItemListener(new ItemListener() {
			@Override
			public void itemStateChanged(ItemEvent e)
			{
				confirmImageStoreDeselection(e);
			}
		});
		l.setLabelFor(imageCopyCB);
		addElements(l, imageCopyCB);
		flipComponent(false, imageCopyCB);

		l = new JCardsLabel("auto.resize");
		l.setFont(labelFont);
		imageResizePanel = getImageResizePanel();
		l.setLabelFor(resizeWidthTF);
		addElements(l, imageResizePanel);
		flipComponent(false, imageResizePanel);

		l = new JCardsLabel("maintain.aspect.ratio");
		l.setFont(labelFont);
		maintainAspectRatioCB = new JCheckBox();
		l.setLabelFor(maintainAspectRatioCB);
		addElements(l, maintainAspectRatioCB);
		flipComponent(false, maintainAspectRatioCB);
		
		// add update button
		updateButton = new JCardsButton("update", JCardsFrame.getImage(UserSettings.ICON_SAVE));
		updateButton.addActionListener(this);
		GridBagConstraints gbc = new GridBagConstraints();
		gbc.weightx = 1.0;
		gbc.weighty = 1.0;
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		gbc.anchor = GridBagConstraints.SOUTH;
		add(updateButton, gbc);
		
		// fill in the fields if editing
		if (fdef == null)
		{
			fieldType.setSelectedItem(FieldType.STRING.getName());
		}
		else
		{
			fieldNameTF.setText(fdef.getName());
			fieldType.setSelectedItem(fdef.getTypeName());
			switch (fdef.getType())
			{
				case CATEGORY:
				case LIST:
					CategoryFieldDef cfd = (CategoryFieldDef)fdef;
					catName.setSelectedItem(cfd.getCategory());
					flipComponents(true, catName);
					break;
					
				case COUNTER:
					defaultValueTF.setEnabled(false);
					defaultValueTF.setText(null);
					CounterFieldDef cofd = (CounterFieldDef)fdef;
					countName.setSelectedItem(cofd.getCounter());
					numType.setSelectedIndex(cofd.getNumType().ordinal());
					dpModel.setValue(cofd.getDecimalPlaces());
					localeCB.setSelectedItem(cofd.getLocale().toString());
					updateExampleNumber();
					flipComponents(true, countName, numType, decimalPlaces, localeCB, numberExample);
					break;
					
				case MEMO:
					MemoFieldDef mfd = (MemoFieldDef)fdef;
					memoModel.setValue(mfd.getMemoLines());
					displayHTMLCB.setSelected(mfd.isDisplayAsHTML());
					autoFoldCB.setSelected(mfd.isAutoFold());
					flipComponents(true, memoLines, displayHTMLCB, autoFoldCB);
					break;
					
				case DATE:
					DateFieldDef dfd = (DateFieldDef)fdef;
					dateFormat.setSelectedItem(dfd.getDateFormat());
					todayCB.setSelected(dfd.isDatePrepopulated());
					flipComponents(true, dateFormat, todayCB);
					break;
					
				case NUMBER:
					NumberFieldDef nfd = (NumberFieldDef)fdef;
					numType.setSelectedIndex(nfd.getNumType().ordinal());
					dpModel.setValue(nfd.getDecimalPlaces());
					localeCB.setSelectedItem(nfd.getLocale().toString());
					updateExampleNumber();
					flipComponents(true, numType, decimalPlaces, localeCB, numberExample);
					break;

				case CALCULATED:
					CalculatedFieldDef clfd = (CalculatedFieldDef)fdef;
					numType.setSelectedIndex(clfd.getNumType().ordinal());
					dpModel.setValue(clfd.getDecimalPlaces());
					localeCB.setSelectedItem(clfd.getLocale().toString());
					updateExampleNumber();
					flipComponents(true, numType, decimalPlaces, localeCB, numberExample);
					break;
					
				case IMAGE:
					ImageFieldDef ifd = (ImageFieldDef)fdef;
					autoFoldCB.setSelected(ifd.isAutoFold());
					imageCopyCB.setSelected(ifd.isStoreInDB());
					maintainAspectRatioCB.setSelected(ifd.isMaintainAspectRatio());
					if (ifd.getResizeDimension() != null)
					{
						resizeWidthTF.setText(Integer.toString(ifd.getResizeDimension().width));
						resizeHeightTF.setText(Integer.toString(ifd.getResizeDimension().height));
					}
					flipComponents(true, autoFoldCB, imageCopyCB, imageResizePanel, maintainAspectRatioCB);
					break;
			}
			manCB.setSelected(fdef.isMandatory());
			
			String defVal = fdef.getDefaultValue();
			if (defVal != null)
			{
				defaultValueTF.setText(defVal);
			}
		}
		
		parent.getRootPane().setDefaultButton(updateButton);
	}

	/**
	 * Ask the user to confirm that turning off storing images in the DB
	 * will result in their loss
	 * @param e
	 */
	private void confirmImageStoreDeselection(ItemEvent e)
	{
		if (e.getStateChange() == ItemEvent.DESELECTED)
		{
			if (JOptionPane.showConfirmDialog(parent, JCards.getMessage("img.store.deselect"),
				JCards.getMessage("confirm"), JOptionPane.YES_NO_OPTION) == JOptionPane.NO_OPTION)
			{
				imageCopyCB.setSelected(true);
			}
		}
	}

	private JPanel getImageResizePanel()
	{
		JPanel p = new JPanel();
		resizeWidthTF = new JTextField(3);
		resizeWidthTF.setToolTipText(JCards.getLabelString("auto.resize.tt"));
		p.add(resizeWidthTF);
		p.add(new JLabel(" x "));
		resizeHeightTF = new JTextField(3);
		resizeHeightTF.setToolTipText(JCards.getLabelString("auto.resize.tt"));
		p.add(resizeHeightTF);

		class IntegerVerifier extends InputVerifier
		{
			@Override
			public boolean shouldYieldFocus(JComponent input)
			{
				return verify(input);
			}

			@Override
			public boolean verify(JComponent input)
			{
				JTextField tf = (JTextField)input;
				if (tf.getText().length() == 0 || tf.getText().equals("0"))
				{
					return true;
				}
				else
				{
					try {
						int i = Integer.parseInt(tf.getText());
						if (i < 1 || i > 5000) {
							return false;
						}
						return true;
					}
					catch (NumberFormatException nfex)
					{
						Toolkit.getDefaultToolkit().beep();
						return false;
					}
				}
			}
		}

		IntegerVerifier iv = new IntegerVerifier();
		resizeWidthTF.setInputVerifier(iv);
		resizeHeightTF.setInputVerifier(iv);

		return p;
	}

	/** Set the focus to the field name field */
	public void focusFieldName()
	{
		fieldNameTF.requestFocus();
	}
	
	/** Set the GridBagConstraints and add the label and component to the panel */
	private void addElements(JLabel label, JComponent comp)
	{
		GridBagConstraints gbc = new GridBagConstraints();
		gbc.gridwidth = 1;
		gbc.anchor = GridBagConstraints.WEST;
		gbc.insets = new Insets(2, 5, 2, 10);
		add(label, gbc);
		
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		gbc.weightx = 1;
		add(comp, gbc);
		
		// add to map of labels
		labels.put(comp, label);
	}
	
	/** User clicked update button */
	@Override
	public void actionPerformed(ActionEvent e)
	{
		updateField();
	}
	
	/**
	 * Sets the components to be visible or not
	 * @param state
	 * @param comps
	 */
	private void flipComponents(boolean state, JComponent... comps)
	{
		for (JComponent comp : comps)
		{
			flipComponent(state, comp);
		}
	}

	/**
	 * Set whether a component is displayed or not
	 * @param comp the component to be affected
	 * @param state if true make the field editable and visible
	 */
	private void flipComponent(boolean state, JComponent comp)
	{
	    if (comp == null) {
		log.warning(EditFieldPanel.class.getName()+".flipComponent null component");
	    }
	    else
	    {
		comp.setEnabled(state);
		comp.setVisible(state);
		JLabel label = labels.get(comp);
		assert label != null : "Cannot get label for comp "+comp;
		label.setVisible(state);
	    }
	}
	
	/** Method called when item in type choice list is changed */
	@Override
	public void itemStateChanged(ItemEvent e)
	{
		String type = (String)fieldType.getSelectedItem();
		
		if (type != null)
		{
			flipComponent(true, manCB);
			if (catName != null)
			{
				flipComponent(false, catName);
			}
			if (countName != null)
			{
				flipComponent(false, countName);
			}
			if (dateFormat != null)
			{
				flipComponent(false, dateFormat);
				flipComponent(false, todayCB);
			}
			flipComponents(false, memoLines, displayHTMLCB, autoFoldCB, numType, localeCB, decimalPlaces,
				numberExample, imageCopyCB, imageResizePanel, maintainAspectRatioCB);
			flipComponent(true, defaultValueTF);
			
			FieldType fType = FieldType.valueOf(type.toUpperCase());
			switch (fType)
			{
				case BOOLEAN:
					// boolean fields cannot be mandatory
					flipComponent(false, manCB);
					break;
					
				case CALCULATED:
					flipComponent(false, manCB);
					manCB.setSelected(false);
					defaultValueTF.requestFocus();
					flipComponents(true, numType, localeCB, decimalPlaces, numberExample);
					break;
					
				case CATEGORY:
				case LIST:
					flipComponents(true, catName);
					break;
					
				case COUNTER:
					defaultValueTF.setText(null);
					manCB.setSelected(false);

					flipComponents(false, defaultValueTF, manCB);

					flipComponents(true, countName, numType, localeCB, decimalPlaces, numberExample);
					countName.requestFocus();
					break;
					
				case DATE:
					flipComponents(true, todayCB, dateFormat);
					dateFormat.requestFocus();
					break;
					
				case NUMBER:
					flipComponents(true, numType, localeCB, decimalPlaces, numberExample);
					break;
					
				case MEMO:
					flipComponents(true, memoLines, displayHTMLCB, autoFoldCB);
					memoLines.requestFocus();
					break;
					
				case IMAGE:
					flipComponents(true, autoFoldCB, imageCopyCB, imageResizePanel, maintainAspectRatioCB);
					break;
			}
		}
	}
	
	/**
	 * Show an error message
	 * @param message error message
	 * @param title error title
	 * @param comp if not null request focus to this component
	 */
	private void showError(String message, String title, JComponent comp)
	{
		Toolkit.getDefaultToolkit().beep();
		JOptionPane.showMessageDialog(parent, message, title, JOptionPane.ERROR_MESSAGE);
		if (comp != null)
		{
			comp.requestFocus();
		}
	}
	
	/**
	 * Update a calculated field
	 * @param fieldDef FieldDef
	 * @return true if updated
	 */
	private boolean updateCalculated(FieldDef fieldDef)
	{
		log.entering("EditFieldPanel", "updateCalculated", fieldDef.getName());
		
		boolean updated = false;
		String defaultText = defaultValueTF.getText();
		if (defaultText.length() == 0)
		{
			showError(JCards.getMessage("enter.calc"), JCards.getMessage("no.calc"), defaultValueTF);
			updated = false;
		}
		else
		{
			String msg = checkCalculation(defaultText);
			if (msg != null)
			{
				showError(msg, JCards.getMessage("incorrect.calc"), defaultValueTF);
				updated = false;
			}
			else
			{
				CalculatedFieldDef cfd = (CalculatedFieldDef)fieldDef;
				cfd.setCalculation(defaultText);
				updateNumber(cfd);
				updated = true;
			}
		}
		
		log.exiting("EditFieldPanel", "updateCalculated", updated);
		
		return updated;
	}
	
	/**
	 * Update a category field
	 * @param fieldDef FieldDef
	 * @return true if updated
	 */
	private boolean updateCategory(FieldDef fieldDef)
	{
		log.entering("EditFieldPanel", "updateCategory", fieldDef.getName());
		
		CategoryFieldDef cfd = (CategoryFieldDef)fieldDef;
		cfd.setCategory((String)catName.getSelectedItem());
		
		log.exiting("EditFieldPanel", "updateCategory", true);
		
		return true;
	}
	
	/**
	 * Update a counter field
	 * @param fieldDef FieldDef
	 * @return true if updated
	 */
	private boolean updateCounter(FieldDef fieldDef)
	{
		log.entering("EditFieldPanel", "updateCounter", fieldDef.getName());
		
		CounterFieldDef cfd = (CounterFieldDef)fieldDef;
		cfd.setCounter((String)countName.getSelectedItem());
		updateNumber(cfd);
		
		log.entering("EditFieldPanel", "updateCounter", true);
		
		return true;
	}
	
	/**
	 * Update a date field
	 * @param fieldDef FieldDef
	 * @return true if updated
	 */
	private boolean updateDate(FieldDef fieldDef)
	{
		log.entering("EditFieldPanel", "updateDate", fieldDef.getName());
		
		boolean updated = false;
		DateFieldDef dfd = (DateFieldDef)fieldDef;
		// check date format entered
		if (!dfd.setDateFormat((String)dateFormat.getSelectedItem()))
		{
			showError(JCards.getMessage("invalid.date.format"), JCards.getMessage("incorrect.date"), dateFormat);
			updated = false;
		}
		dfd.setDatePrepopulated(todayCB.isSelected());
		updated = true;
		
		log.exiting("EditFieldPanel", "updateDate", updated);
		
		return updated;
	}
	
	/**
	 * Update a memo field
	 * @param fieldDef FieldDef
	 * @return true if updated
	 */
	private boolean updateMemo(FieldDef fieldDef)
	{
		log.entering("EditFieldPanel", "updateMemo", fieldDef.getName());
		
		MemoFieldDef mfd = (MemoFieldDef)fieldDef;
		SpinnerNumberModel numModel = (SpinnerNumberModel)memoLines.getModel();
		mfd.setMemoLines(numModel.getNumber().intValue());
		mfd.setDisplayAsHTML(displayHTMLCB.isSelected());
		mfd.setAutoFold(autoFoldCB.isSelected());
		
		log.exiting("EditFieldPanel", "updateMemo", true);
		
		return true;
	}
	
	/**
	 * Update an image field
	 * @param fieldDef FieldDef
	 * @return true if updated
	 */
	private boolean updateImage(FieldDef fieldDef)
	{
		log.entering("EditFieldPanel", "updateImage", fieldDef.getName());
		ImageFieldDef ifd = (ImageFieldDef)fieldDef;
		ifd.setAutoFold(autoFoldCB.isSelected());
		ifd.setStoreInDB(imageCopyCB.isSelected());
		ifd.setMaintainAspectRatio(maintainAspectRatioCB.isSelected());
		if (resizeWidthTF.getText().length() > 0 && resizeHeightTF.getText().length() > 0)
		{
			try {
				Dimension dim = new Dimension(Integer.parseInt(resizeWidthTF.getText()), Integer.parseInt(resizeHeightTF.getText()));
				ifd.setResizeDimension(dim);
			} catch (NumberFormatException nfex) {
				showError(JCards.getMessage("check.dim"), JCards.getMessage("invalid.dim"), resizeWidthTF);
			}
		}

		// if turned off "Store in DB" need to remove image field values that refer to DB images
		if (!imageCopyCB.isSelected())
		{
			cards.removeImageDBFileRefs(ifd.getName());
		}
		
		log.exiting("EditFieldPanel", "updateImage", fieldDef.getName());
		return true;
	}
	
	/**
	 * Update a number field
	 * @param fieldDef FieldDef
	 * @return true if updated
	 */
	private boolean updateNumber(FieldDef fieldDef)
	{
		log.entering("EditFieldPanel", "updateNumber", fieldDef.getName());
		String selectedLocale = (String)localeCB.getSelectedItem();
		String[] localeSplit = selectedLocale.split("_");
		Locale loc = new Locale(localeSplit[0], localeSplit[1]);
		NumberFieldDef nfd = (NumberFieldDef)fieldDef;
		switch (numType.getSelectedIndex())
		{
			case 0:
				nfd.setNumType(NumberFieldDef.NUMBER_TYPE.NUMBER, loc);
				break;
				
			case 1:
				nfd.setNumType(NumberFieldDef.NUMBER_TYPE.INTEGER, loc);
				break;
				
			case 2:
				nfd.setNumType(NumberFieldDef.NUMBER_TYPE.CURRENCY, loc);
				break;
				
			case 3:
				nfd.setNumType(NumberFieldDef.NUMBER_TYPE.PERCENTAGE, loc);
				break;
				
			default: return false;
		}
		SpinnerNumberModel snm = (SpinnerNumberModel)decimalPlaces.getModel();
		nfd.setDecimalPlaces(snm.getNumber().intValue());
		
		log.exiting("EditFieldPanel", "updateNumber", true);
		
		return true;
	}
	
	/**
	 * Create or update a field
	 * @return true if field updated, false otherwise
	 */
	private boolean updateField()
	{
		log.entering("EditFieldPanel", "updateField");
		
		DBDef dbDef = DBDef.getInstance();
		String defaultText;
		
		String fName = fieldNameTF.getText();
		if (fName == null || fName.length() == 0)
		{
			showError(JCards.getMessage("field.name.null"), JCards.getMessage("null.entry"), fieldNameTF);
			return false;
		}
		
		String fTypeS = (String)fieldType.getSelectedItem();
		FieldType fType = FieldType.valueOf(fTypeS.toUpperCase());
		FieldDef fieldDef = FieldDef.createFieldDef(fName, fType);
		boolean updated = false;
		
		switch (fType)
		{
			case CALCULATED:
				updated = updateCalculated(fieldDef);
				updated = updated & updateNumber(fieldDef);
				break;
				
			case CATEGORY:
			case LIST:
				updated = updateCategory(fieldDef);
				break;
				
			case COUNTER:
				updated = updateCounter(fieldDef);
				break;
				
			case DATE:
				updated = updateDate(fieldDef);
				break;
				
			case MEMO:
				updated = updateMemo(fieldDef);
				break;
				
			case NUMBER:
				updated = updateNumber(fieldDef);
				break;
				
			case IMAGE:
				updated = updateImage(fieldDef);
				break;
				
			default:
				updated = true;
		}
		
		if (!updated)
		{
			return false;
		}
		
		// first field is always mandatory
		if (dbDef.getNumberOfFields() == 0 || fieldIndex == 0)
		{
			fieldDef.setMandatory(true);
		}
		else
		{
			fieldDef.setMandatory(manCB.isSelected());
		}
		
		// set the optional default
		defaultText = defaultValueTF.getText();
		if (defaultText.length() > 0 && !(fType == FieldType.COUNTER || fType == FieldType.CALCULATED))
		{
			if (!fieldDef.setDefaultValue(defaultText))
			{
				showError(JCards.getMessage("default.type.mismatch"), JCards.getMessage("incorrect.default"), defaultValueTF);
				return false;
			}
		}
		
		if (fieldIndex < 0)
		{
			if (dbDef.addFieldDef(fieldDef))
			{
				fieldIndex = dbDef.getFieldNames().indexOf(fieldDef.getName());
			}
			else
			{
				showError(JCards.getMessage("field.already.exists"), JCards.getMessage("not.unique"), fieldNameTF);
				return false;
			}
		}
		else
		{
			dbDef.replaceFieldDef(fieldIndex, fieldDef);
		}
		
		JOptionPane.showMessageDialog(parent, JCards.getMessage("field.updated"), JCards.getMessage("saved"), JOptionPane.INFORMATION_MESSAGE);
		
		log.exiting("EditFieldPanel", "updateField");
		
		return true;
	}
	
	/**
	 * Check that the calculation is valid
	 * @param calculation the calculation to check
	 * @return null if the calculation is OK, or a message indicating the error
	 */
	private String checkCalculation(String calculation)
	{
		log.entering("EditFieldPanel", "checkCalculation", calculation);
		
		DBDef dbDef = DBDef.getInstance();
		String token;
		FieldDef fd;
		StringTokenizer st = new StringTokenizer(calculation, " ");
		StringBuilder calcBuilder;
		String msg = null;
		
		try
		{
			while (st.hasMoreTokens())
			{
				token = st.nextToken();
				
				// field
				if (token.startsWith("'"))
				{
					calcBuilder = new StringBuilder(token);
					while (!token.endsWith("'"))
					{
						calcBuilder.append(" ");
						token = st.nextToken();
						calcBuilder.append(token);
					}
					// remove quotes
					token = calcBuilder.substring(1, calcBuilder.length()-1);
					
					// check field exists
					fd = dbDef.getFieldDef(token);
					if (fd == null)
					{
						msg = JCards.getFormattedMessage("field.no.exist", token);
						break;
					}
					
					// check field is of the right type
					switch (fd.getType())
					{
						case NUMBER:
						case CALCULATED:
						case COUNTER:
							break;
							
						default:
							msg = JCards.getFormattedMessage("field.no.num", token);
							break;
					}
				}
				else if (!(PostfixCalculator.isOperand(token) || isNumber(token) || PostfixCalculator.isMathFunction(token)))
				{
					msg = JCards.getFormattedMessage("field.unknown", token);
					break;
				}
			}
		}
		catch (NoSuchElementException nsex)
		{
			msg = JCards.getMessage("field.end");
		}
		
		log.exiting("EditFieldPanel", "checkCalculation", msg);
		
		return msg;
	}
	
	/**
	 * Checks to see if a String is a number
	 * @param num the number to check
	 * @return true if number, false otherwise
	 */
	private static boolean isNumber(String num)
	{
		boolean isNumber = false;
		try
		{
			Double.valueOf(num);
			isNumber = true;
		}
		catch (NumberFormatException ex)
		{
			log.finest("isNumber - invalid number "+num);
		}
		
		return isNumber;
	}
	
	/**
	 * Update the example number
	 */
	private void updateExampleNumber()
	{
		String locCode = (String)localeCB.getSelectedItem();
		String[] locCodes = locCode.split("_");
		Locale loc = new Locale(locCodes[0], locCodes[1]);
		NumberFormat nf;
		switch (numType.getSelectedIndex())
		{
			case 1:
				nf = NumberFormat.getIntegerInstance(loc);
				break;

			case 2:
				nf = NumberFormat.getCurrencyInstance(loc);
				break;

			case 3:
				nf = NumberFormat.getPercentInstance(loc);
				break;

			default:
				nf = NumberFormat.getNumberInstance(loc);
				break;
		}
		SpinnerNumberModel snm = (SpinnerNumberModel)decimalPlaces.getModel();
		nf.setMaximumFractionDigits(snm.getNumber().intValue());
		numberExample.setText(nf.format(EXAMPLE_NUMBER));
	}
}
