/*
 * JCards
 * Copyright (C) 2002-2009 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package nrs.jcards.gui;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.text.MessageFormat;
import java.util.logging.Logger;
import javax.swing.*;
import javax.swing.event.ListDataEvent;
import javax.swing.event.ListDataListener;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import nrs.jcards.Cards;
import nrs.jcards.DBDef;
import nrs.jcards.JCards;
import nrs.jcards.UserSettings;
import nrs.jcards.fields.FieldDef;
import nrs.jcards.fields.FieldType;
import nrs.jcards.gui.dialog.DBDialog;

/**
 * This class displays the fields and allows additions, modifications and deletion
 * of fields.
 *
 * @author Copyright &#169; 2002-2009 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class FieldsPanel extends JPanel implements DBDialogActionsI, ListDataListener
{
	private DBDialog dbDialog;
	private Cards cards;
	private JList fieldsList;
	private FieldsListModel fieldModel;
	private final static Logger log = Logger.getLogger(FieldsPanel.class.getName());
	
	/** Initialise the FieldsDialog
	 * @param dbDialog DBDialog
	 */
	public FieldsPanel(DBDialog dbDialog, Cards cards)
	{
		this.dbDialog = dbDialog;
		this.cards = cards;
		
		setLayout(new BorderLayout());
		setBorder(BorderFactory.createEmptyBorder(2, 2, 2, 2));
		
		// create toolbar
		JToolBar rToolbar = new JToolBar(JToolBar.VERTICAL);
		rToolbar.setFloatable(false);
		rToolbar.setRollover(true);
		rToolbar.setBorderPainted(false);
		
		JButton up = new JButton(JCardsFrame.getImage(UserSettings.ICON_UP));
		up.setToolTipText(JCards.getButtonString("field.up.tt"));
		up.addActionListener(new ActionListener()
		{
			@Override
			public void actionPerformed(ActionEvent e)
			{
				fieldMove(-1);
			}
		});
		rToolbar.add(up);
		
		JButton down = new JButton(JCardsFrame.getImage(UserSettings.ICON_DOWN));
		down.setToolTipText(JCards.getButtonString("field.down.tt"));
		down.addActionListener(new ActionListener()
		{
			@Override
			public void actionPerformed(ActionEvent e)
			{
				fieldMove(1);
			}
		});
		rToolbar.add(down);
		
		// create fields list
		fieldModel = new FieldsListModel();
		fieldModel.addListDataListener(this);
		DBDef.getInstance().addObserver(fieldModel);
		fieldsList = new JList(fieldModel);
		fieldsList.setBorder(BorderFactory.createEmptyBorder(4, 4, 4, 4));
		fieldsList.setCellRenderer(new FieldListCellRenderer());
		fieldsList.addListSelectionListener(new ListSelectionListener()
		{
			@Override
			public void valueChanged(ListSelectionEvent e)
			{
				selectField();
			}
		});
		JScrollPane fScroll = new JScrollPane(fieldsList);
		
		add(fScroll, BorderLayout.CENTER);
		add(rToolbar, BorderLayout.EAST);
	}
	
	/** User selected a field */
	private void selectField()
	{
		log.entering("FieldsPanel", "selectField");
		
		int fieldIndex = fieldsList.getSelectedIndex();
		if (fieldIndex > -1)
		{
			EditFieldPanel efp = new EditFieldPanel(dbDialog, fieldIndex, cards);
			dbDialog.setContentPanel(new JScrollPane(efp));
			efp.focusFieldName();
		}
		
		log.exiting("FieldsPanel", "selectField");
	}
	
	/**
	 * Move a field up or down in the list
	 * @param incr either +1 or -1
	 */
	private void fieldMove(int incr)
	{
		log.entering("FieldsPanel", "fieldMove", incr);
		
		int selection = fieldsList.getSelectedIndex();
		DBDef dbdef = DBDef.getInstance();
		
		if (selection > -1)
		{
			// cannot move a boolean field to first in the list
			if ((dbdef.getNumberOfFields() > 1) && (selection == 0 && incr == 1) || (selection == 1 && incr == -1))
			{
				FieldDef fd = dbdef.getFieldDef(1);
				
				if (fd.getType() == FieldType.BOOLEAN)
				{
					Toolkit.getDefaultToolkit().beep();
					JOptionPane.showMessageDialog(dbDialog, JCards.getMessage("cannot.move.msg"), JCards.getMessage("cannot.move"), JOptionPane.WARNING_MESSAGE);
					return;
				}
			}
			int newpos = dbdef.moveField(selection, incr);
			fieldsList.setSelectedIndex(newpos);
		}
		
		log.exiting("FieldsPanel", "fieldMove");
	}
	
	/** Create a new field */
	@Override
	public void newItem()
	{
		log.entering("FieldsPanel", "newItem");
		
		EditFieldPanel edit = new EditFieldPanel(dbDialog, -1, cards);
		dbDialog.setContentPanel(new JScrollPane(edit));
		edit.focusFieldName();
		
		log.exiting("FieldsPanel", "newItem");
	}
	
	/** Delete an item */
	@Override
	public boolean deleteItem()
	{
		log.entering("FieldsPanel", "deleteItem");
		
		boolean deleted = true;
		DBDef dbdef = DBDef.getInstance();
		
		int index = fieldsList.getSelectionModel().getMinSelectionIndex();
		if (index > -1)
		{
			FieldDef deleteFD = dbdef.getFieldDef(index);

			// check if field is part of a calculation
			String calcFieldDefName = dbdef.isPartOfCalculation(deleteFD);
			if (calcFieldDefName != null)
			{
				MessageFormat mf = new MessageFormat(JCards.getMessage("field.in.calc"));
				String[] args = { calcFieldDefName };
				JOptionPane.showMessageDialog(dbDialog, mf.format(args), JCards.getMessage("cannot.delete"), JOptionPane.WARNING_MESSAGE);
				deleted = false;
			}
			else if (index == 0 && dbdef.getNumberOfFields() > 1)
			{
				// next field automatically becomes mandatory
				FieldDef nextFD = dbdef.getFieldDef(1);
				nextFD.setMandatory(true);
			}

			if (deleted) {
				dbdef.removeFieldDef(deleteFD.getName());
			}
		}
		else {
			deleted = false;
		}
		
		log.exiting("FieldsPanel", "deleteItem", deleted);
		return deleted;
	}

	@Override
	public boolean isItemSelected()
	{
		return !fieldsList.isSelectionEmpty();
	}
	
	
	/** Remove the current selection */
	@Override
	public void clearSelection()
	{
		fieldsList.clearSelection();
	}
	
	@Override
	public Dimension getPreferredSize()
	{
		return new Dimension(100, 300);
	}

	/**
	 * Sent after the indices in the index0,index1 interval
	 * have been removed from the data model.  The interval 
	 * includes both index0 and index1.
	 * 
	 * @param e  a <code>ListDataEvent</code> encapsulating the
	 *    event information
	 */
	@Override
	public void intervalRemoved(ListDataEvent e)
	{
	}

	/**
	 * Sent after the indices in the index0,index1 
	 * interval have been inserted in the data model.
	 * The new interval includes both index0 and index1.
	 * 
	 * @param e  a <code>ListDataEvent</code> encapsulating the
	 *    event information
	 */
	@Override
	public void intervalAdded(ListDataEvent e)
	{
		log.entering("FieldsPanel", "intervalAdded");
		fieldsList.setSelectionInterval(e.getIndex0(), e.getIndex1());
	}

	/**
	 * Sent when the contents of the list has changed in a way 
	 * that's too complex to characterize with the previous 
	 * methods. For example, this is sent when an item has been
	 * replaced. Index0 and index1 bracket the change.
	 * 
	 * @param e  a <code>ListDataEvent</code> encapsulating the
	 *    event information
	 */
	@Override
	public void contentsChanged(ListDataEvent e)
	{
		log.entering("FieldsPanel", "contentsChanged");
		fieldsList.setSelectionInterval(e.getIndex0(), e.getIndex1());
	}

	@Override
	public void cleanup()
	{
		DBDef.getInstance().deleteObserver(fieldModel);
	}
}