/*
 * JCards
 * Copyright (C) 2002-2009 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package nrs.jcards.gui;

import java.awt.Font;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import javax.swing.*;
import nrs.jcards.UserSettings;

/**
 * Component to handle hiding and showing (e.g. code folding) another component. Displays text plus an arrow to indicate fold status.
 * 
 * @author Copyright &#169; 2002-2009 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class FoldComponent extends JPanel implements MouseListener
{
	private JLabel foldLabel;
	private JComponent foldableComponent;
	private boolean foldedStatus = false;
	
	/**
	 * Create a new FoldComponent
	 * @param text text to display
	 * @param comp the component to fold
	 * @see FoldComponent(String, JComponent, boolean)
	 */
	public FoldComponent(String text, JComponent foldComp)
	{
		this(text, foldComp, false);
	}
	
	/**
	 * Create a new FoldComponent
	 * @param text text to display
	 * @param foldComp the component to fold
	 * @param folded initial folded status, true for folded false for unfolded
	 * @see FoldComponent(String, JComponent)
	 */
	public FoldComponent(String text, JComponent foldComp, boolean folded)
	{
		foldableComponent = foldComp;
		foldedStatus = folded;
		
		foldLabel = new JLabel(text, JCardsFrame.getImage(UserSettings.ICON_ARROW_DOWN), SwingConstants.LEFT);
		foldLabel.setHorizontalTextPosition(SwingConstants.LEADING);
		foldLabel.addMouseListener(this);
		foldLabel.setBorder(BorderFactory.createEmptyBorder(2, 0, 2, 2));
		
		setLayout(new GridBagLayout());
		GridBagConstraints gbc = new GridBagConstraints();
		
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		gbc.fill = GridBagConstraints.HORIZONTAL;
		gbc.anchor = GridBagConstraints.WEST;
		gbc.weightx = 1.0;
		add(foldLabel, gbc);
		
		gbc.weighty = 1.0;
		gbc.fill = GridBagConstraints.BOTH;
		add(foldComp, gbc);
		
		if (!folded) {
			mouseClicked(null);
		}
	}
	
	/**
	 * Indicate whether the foldable component is folded or not
	 * @param folded set to true to hide; false to make visible
	 */
	public void setFolded(boolean folded)
	{
		foldedStatus = folded;
	}
	
	/**
	 * Set the font of the label
	 * @param labelFont
	 */
	@Override
	public void setFont(Font labelFont)
	{
		if (foldLabel != null) {
			foldLabel.setFont(labelFont);
		}
	}
	
	@Override public final void mouseClicked(MouseEvent e)
	{
		foldableComponent.setVisible(foldedStatus);
		if (foldedStatus)
		{
			foldLabel.setIcon(JCardsFrame.getImage(UserSettings.ICON_ARROW_DOWN));
		}
		else
		{
			foldLabel.setIcon(JCardsFrame.getImage(UserSettings.ICON_ARROW_RIGHT));
		}
		foldedStatus = !foldedStatus;
	}

	@Override public void mouseEntered(MouseEvent e)
	{
		foldLabel.setBorder(BorderFactory.createEtchedBorder());
	}

	@Override public void mouseExited(MouseEvent e)
	{
		foldLabel.setBorder(BorderFactory.createEmptyBorder(2, 0, 2, 2));
	}

	@Override public void mousePressed(MouseEvent e)
	{}

	@Override public void mouseReleased(MouseEvent e)
	{}
}
