/*
 * JCards
 * Copyright (C) 2002-2009 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package nrs.jcards.gui;

import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.font.TextAttribute;
import java.util.List;
import java.util.*;
import javax.swing.*;
import javax.swing.border.TitledBorder;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import nrs.jcards.JCards;
import nrs.jcards.UserSettings;
import nrs.jcards.gui.action.CancelAction;
import nrs.jcards.gui.action.OKAction;
import nrs.jcards.gui.action.OKActionI;
import nrs.jcards.gui.dialog.BaseDialog;

/**
 * Font chooser
 * 
 * @author Copyright &#169; 2002-2009 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class FontChooser extends JPanel
{
	private Font initialFont;
    private JTextArea fontExample;
    private JComboBox fontNameCB, fontStyleCB;
    private JSpinner fontSizeSpinner;
	
	private class FontCBModel extends DefaultComboBoxModel
	{
		private List<String> fontNames = new ArrayList<String>(20);

		private FontCBModel()
		{
			String[] fontNamesArray = GraphicsEnvironment.getLocalGraphicsEnvironment().getAvailableFontFamilyNames();
			Collections.addAll(fontNames, fontNamesArray);
		}

		@Override public int getSize()
		{
			return fontNames.size();
		}

		@Override public Object getElementAt(int index)
		{
			return fontNames.get(index);
		}

		@Override public int getIndexOf(Object anObject)
		{
			return fontNames.indexOf(anObject);
		}
	}

	private class FontStyleCBModel extends DefaultComboBoxModel
	{
		private List<String> fontStyles = new ArrayList<String>(4);

		private FontStyleCBModel()
		{
			Collections.addAll(fontStyles,
				JCards.getLabelString("font.plain"), JCards.getLabelString("font.bold"),
				JCards.getLabelString("font.italic"), JCards.getLabelString("font.bold.italic")
				);
		}

		@Override public int getSize()
		{
			return fontStyles.size();
		}

		@Override public Object getElementAt(int index)
		{
			return fontStyles.get(index);
		}

		@Override public int getIndexOf(Object anObject)
		{
			return fontStyles.indexOf(anObject);
		}
	}

	private static class OKDialog extends BaseDialog implements OKActionI
	{
		public OKDialog(Dialog parent, String title, boolean modal)
		{
			super(parent, title, modal);
		}
		
		@Override public void okClicked()
		{
			setVisible(false);
		}
	}
	
    /** Creates new form FontChoosee */
    public FontChooser()
	{
        initComponents();
		setUserFont(new Font("Monospaced", Font.PLAIN, 13));
    }

	public FontChooser(Font initialFont)
	{
		initComponents();
		setUserFont(initialFont);
	}
	
	/**
	 * Wrap a font chooser component in a dialog box
	 * @param parent owning dialog box
	 * @param title title of the dialog box
	 * @param modal if the dialog box should be modal or not
	 * @param fontChooser font chooser to display
	 * @param okListener listener for ok
	 * @param cancelListener listener for cancel, may be null
	 * @return a JDialog ready to be display
	 * @throws NullPointerException thrown if okListener is null
	 */
	public static JDialog createDialog(Dialog parent, String title, boolean modal, final FontChooser fontChooser,
		ActionListener okListener, ActionListener cancelListener)
	{
		final JDialog fontDialog = new OKDialog(parent, title, modal);
		fontDialog.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
		fontDialog.setLayout(new BorderLayout(5, 5));
		fontDialog.add(fontChooser, BorderLayout.CENTER);
		
		// create buttons
		JPanel buttonPanel = new JPanel(new FlowLayout(FlowLayout.RIGHT));
		fontDialog.add(buttonPanel, BorderLayout.SOUTH);
		
		JButton resetButton = new JCardsButton("reset", JCardsFrame.getImage(UserSettings.ICON_RESET));
		resetButton.addActionListener(new ActionListener() {
			@Override public void actionPerformed(ActionEvent e)
			{
				fontChooser.setUserFont(fontChooser.getInitialFont());
			}
		});
		buttonPanel.add(resetButton);
		
		JButton cancelButton = new JButton(new CancelAction(fontDialog));
		if (cancelListener != null) {
			cancelButton.addActionListener(cancelListener);
		}
		buttonPanel.add(cancelButton);
		
		JButton okButton = new JButton(new OKAction((OKActionI)fontDialog));
		fontDialog.getRootPane().setDefaultButton(okButton);
		okButton.addActionListener(new ActionListener() {
			@Override public void actionPerformed(ActionEvent e)
			{
				fontDialog.setVisible(false);
			}
		});
		if (okListener == null) {
			throw new NullPointerException("Missing ok listener");
		} else {
			okButton.addActionListener(okListener);
		}
		buttonPanel.add(okButton);
		
		return fontDialog;
	}
	
	/**
	 * Set the font to display
	 * @param font
	 */
	public final void setUserFont(Font font)
	{
		if (font != null)
		{
			initialFont = font;
			fontNameCB.getModel().setSelectedItem(font.getFamily());
			fontSizeSpinner.getModel().setValue(font.getSize());
			switch (font.getStyle())
			{
				case Font.PLAIN:
					fontStyleCB.setSelectedIndex(0);
					break;
					
				case Font.BOLD:
					fontStyleCB.setSelectedIndex(1);
					break;
					
				case Font.ITALIC:
					fontStyleCB.setSelectedIndex(2);
					break;
					
				case Font.BOLD | Font.ITALIC:
					fontStyleCB.setSelectedIndex(3);
					break;
			}
		}
	}
	
	/**
	 * Gets the font that was last used in the constuctor or setUserFont method
	 * @return initial font
	 */
	public Font getInitialFont()
	{
		return initialFont;
	}
	
	/**
	 * Get the font selected by the user
	 * @return user selected font
	 */
	public Font getUserFont()
	{
		Map<TextAttribute, Object> attrs = new HashMap<TextAttribute, Object>(4);
		attrs.put(TextAttribute.FAMILY, fontNameCB.getSelectedItem());

		int fsname = fontStyleCB.getSelectedIndex();
		switch (fsname)
		{
			case 0:
				attrs.put(TextAttribute.WEIGHT, TextAttribute.WEIGHT_REGULAR);
				break;

			case 1:
				attrs.put(TextAttribute.WEIGHT, TextAttribute.WEIGHT_BOLD);
				break;

			case 2:
				attrs.put(TextAttribute.POSTURE, TextAttribute.POSTURE_OBLIQUE);
				break;

			case 3:
				attrs.put(TextAttribute.POSTURE, TextAttribute.POSTURE_OBLIQUE);
				attrs.put(TextAttribute.WEIGHT, TextAttribute.WEIGHT_BOLD);
				break;

			default:
				attrs.put(TextAttribute.WEIGHT, TextAttribute.WEIGHT_REGULAR);
				break;
		}

		SpinnerNumberModel snm = (SpinnerNumberModel)fontSizeSpinner.getModel();
		attrs.put(TextAttribute.SIZE, snm.getNumber().floatValue());

		return new Font(attrs);
	}
	
	private void updateFontExample()
	{
		fontExample.setFont(getUserFont());
		fontExample.setCaretPosition(0);
	}

    private void initComponents()
	{
        GridBagConstraints gridBagConstraints;

        fontNameCB = new JComboBox();        
        fontExample = new JTextArea();
        fontStyleCB = new JComboBox();
        fontSizeSpinner = new JSpinner();

        setLayout(new GridBagLayout());

		JLabel fontNameLabel = new JCardsLabel("font.name");
        fontNameLabel.setLabelFor(fontNameCB);
        gridBagConstraints = new GridBagConstraints();
        gridBagConstraints.gridy = 0;
        gridBagConstraints.anchor = GridBagConstraints.WEST;
        gridBagConstraints.insets = new Insets(15, 5, 5, 5);
        add(fontNameLabel, gridBagConstraints);

        fontNameCB.setModel(new FontCBModel());
        fontNameCB.addItemListener(new ItemListener() {
            @Override public void itemStateChanged(ItemEvent evt) {
                updateFontExample();
            }
        });
        gridBagConstraints = new GridBagConstraints();
		gridBagConstraints.gridy = 0;
        gridBagConstraints.anchor = GridBagConstraints.WEST;
        gridBagConstraints.insets = new java.awt.Insets(15, 5, 5, 5);
        add(fontNameCB, gridBagConstraints);

        fontExample.setColumns(35);
        fontExample.setEditable(false);
        fontExample.setRows(6);
        fontExample.setTabSize(4);
        fontExample.setText("abcdefghijklmnopqrstuvwxyz\nABCDEFGHIJKLMNOPQRSTUVWXYZ\n0123456789\nÀÁÂÃÄÅÆßÇÈÉÊËÌÍÎÏÐÑÒÓÔÕÖØÙÚÛÜÝ\nàáâãäåæçèéêëìíîïðñòóôõöøùúûüý\n€!\\\"£$%^&*()-=_+{}[]:@~;'#|\\\\<>?,./");
		JScrollPane fontExampleScroll = new JScrollPane(fontExample);
		fontExampleScroll.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(),
			JCards.getLabelString("example"), TitledBorder.CENTER, TitledBorder.DEFAULT_POSITION));

        gridBagConstraints = new GridBagConstraints();
        gridBagConstraints.gridx = 2;
        gridBagConstraints.gridy = 0;
        gridBagConstraints.gridwidth = GridBagConstraints.REMAINDER;
        gridBagConstraints.gridheight = 3;
        gridBagConstraints.fill = GridBagConstraints.BOTH;
        gridBagConstraints.insets = new Insets(5, 5, 5, 5);
        add(fontExampleScroll, gridBagConstraints);

		JLabel fontStyleLabel = new JCardsLabel("font.style");
        fontStyleLabel.setLabelFor(fontStyleCB);
        gridBagConstraints = new GridBagConstraints();
        gridBagConstraints.gridx = 0;
        gridBagConstraints.gridy = 1;
        gridBagConstraints.anchor = GridBagConstraints.WEST;
        gridBagConstraints.insets = new Insets(5, 5, 5, 5);
        add(fontStyleLabel, gridBagConstraints);

        fontStyleCB.setModel(new FontStyleCBModel());
        fontStyleCB.addItemListener(new ItemListener() {
            @Override public void itemStateChanged(ItemEvent evt) {
                updateFontExample();
            }
        });
        gridBagConstraints = new GridBagConstraints();
        gridBagConstraints.gridx = 1;
        gridBagConstraints.gridy = 1;
        gridBagConstraints.anchor = GridBagConstraints.WEST;
        gridBagConstraints.insets = new Insets(5, 5, 5, 5);
        add(fontStyleCB, gridBagConstraints);

		JLabel fontSizeLabel = new JCardsLabel("font.size");
        fontSizeLabel.setLabelFor(fontSizeSpinner);
        gridBagConstraints = new GridBagConstraints();
        gridBagConstraints.gridx = 0;
        gridBagConstraints.gridy = 2;
        gridBagConstraints.anchor = GridBagConstraints.NORTHWEST;
        gridBagConstraints.weighty = 1.0;
        gridBagConstraints.insets = new Insets(5, 5, 5, 5);
        add(fontSizeLabel, gridBagConstraints);

        fontSizeSpinner.setModel(new SpinnerNumberModel(10, 8, 30, 1));
        fontSizeSpinner.addChangeListener(new ChangeListener() {
            @Override public void stateChanged(ChangeEvent evt) {
                updateFontExample();
            }
        });
        gridBagConstraints = new GridBagConstraints();
        gridBagConstraints.gridx = 1;
        gridBagConstraints.gridy = 2;
        gridBagConstraints.gridwidth = GridBagConstraints.RELATIVE;
        gridBagConstraints.anchor = GridBagConstraints.NORTHWEST;
        gridBagConstraints.weightx = 1.0;
        gridBagConstraints.weighty = 1.0;
        gridBagConstraints.insets = new Insets(5, 5, 5, 5);
        add(fontSizeSpinner, gridBagConstraints);
    }
}
