/*
 * JCards
 * Copyright (C) 2002-2011 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package nrs.jcards.gui;

import java.awt.*;
import java.awt.event.*;
import java.awt.print.PrinterException;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URL;
import java.text.MessageFormat;
import java.util.List;
import java.util.*;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.print.attribute.HashPrintRequestAttributeSet;
import javax.print.attribute.PrintRequestAttributeSet;
import javax.print.attribute.standard.JobName;
import javax.print.attribute.standard.OrientationRequested;
import javax.print.attribute.standard.RequestingUserName;
import javax.swing.*;
import javax.swing.plaf.FontUIResource;
import javax.swing.table.TableColumn;
import javax.swing.table.TableColumnModel;
import nrs.jcards.*;
import nrs.jcards.fields.CategoryFieldDef;
import nrs.jcards.fields.FieldDef;
import nrs.jcards.fields.FieldType;
import nrs.jcards.gui.action.ActionEnum;
import nrs.jcards.gui.action.ActionManager;
import nrs.jcards.gui.action.JCardsAction;
import nrs.jcards.gui.dialog.CardDialog;
import nrs.jcards.gui.dialog.CommentDialog;
import nrs.jcards.gui.dialog.PasswordDialog;
import nrs.jcards.gui.table.*;
import nrs.jcards.io.Elements;
import nrs.jcards.io.JCAException;
import org.jdesktop.swingx.JXErrorPane;
import org.jdesktop.swingx.JXTable;
import org.jdesktop.swingx.error.ErrorInfo;
import org.jdesktop.swingx.error.ErrorReporter;
import org.jdom.Element;

/**
 * GUI for JCards. Handles all of the GUI objects and events.
 * 
 * @author Copyright &#169; 2002-2011 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class JCardsFrame extends JFrame implements Observer, WindowListener
{
	private JCards jcards;
	private final static Logger log = Logger.getLogger(JCardsFrame.class.getName());
	
	// GUI elements
	private JToolBar toolbar;
	private StatusBar statusBar;
	private JMenu viewMenu;
	private JMenu recentFilesMenu;
	private JScrollPane tableScroll;
	private JXTable cardsTable;
	private CardsTableModel ctm;
	private JCheckBoxMenuItem toolbarMI;
	private ActionManager actionManager;
	private Dimension cardDialogSize;
	private Point cardDialogPosition;
	
	/** Listener class for recent files */
	private class RecentFilesListener implements ActionListener
	{
		@Override
		public void actionPerformed(ActionEvent e)
		{
			log.entering("RecentFilesListener", "actionPerformed");
			
			String file = e.getActionCommand();
			if (file != null)
			{
				File f = new File(file);
				if (f.isFile())
				{
					jcards.openFile(file);
				}
				else
				{
					try
					{
						URL url = new URL(file);
						jcards.openFile(url);
					}
					catch (MalformedURLException ex)
					{
						log.log(Level.SEVERE, "actionPerformed", ex);
						showError(JCards.getMessage("error.open"), ex.getMessage(), 
							"JCardsFrame.RecentFilesListener", ex, Level.WARNING);
					}

				}
			}
			
			log.exiting("RecentFilesListener", "actionPerformed");
		}
	}
	
	/** Listener class for categories */
	private class CategoryListener implements ActionListener
	{
		@Override
		public void actionPerformed(ActionEvent e)
		{
			log.entering("CategoryListener", "actionPerformed");
			
			JMenuItem mi = (JMenuItem)e.getSource();
			String field = mi.getName();
			String item = mi.getActionCommand();
			FindReplaceDetails frDetails = new FindReplaceDetails();
			frDetails.searchText = item;
			frDetails.field = field;
			frDetails.matchCase = true;
			frDetails.wholeWords = true;
			findCards(frDetails);
			
			log.exiting("CategoryListener", "actionPerformed");
		}
	}
	
	/** Creates a new instance of JCardsFrame */
	public JCardsFrame(JCards jcards)
	{
		super("JCards");
		
		log.entering("JCardsFrame", "()");
		this.jcards = jcards;
		
		// set icon
		setIconImage(getImage(UserSettings.ICON).getImage());
		
		// when user clicks close window decoration
		setDefaultCloseOperation(DO_NOTHING_ON_CLOSE);
		addWindowListener(this);
		
		// create ActionManager to hold actions
		actionManager = new ActionManager(jcards);
		
		// set up table
		log.fine("Creating table");
		createTable();
		
		// Create the scroll pane and add the table to it.
		tableScroll = new JScrollPane(cardsTable, JScrollPane.VERTICAL_SCROLLBAR_ALWAYS, JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
		setScrollCorners(tableScroll);
		
		// set up last opened menu
		recentFilesMenu = new JCardsMenu("open.recent");
		recentFilesMenu.setEnabled(false);
		
		getContentPane().setLayout(new BorderLayout());
		getContentPane().add(tableScroll, BorderLayout.CENTER);
		createMenuBar();
		createToolbar();
		getContentPane().add(toolbar, BorderLayout.NORTH);
		
		statusBar = new StatusBar(this);
		jcards.getCards().addObserver(statusBar);
		getContentPane().add(statusBar, BorderLayout.SOUTH);
		setStatus(JCards.getFormattedMessage("welcome", JCards.VERSION));
		
		// JCards drag and drop support
		setTransferHandler(new JCardsDropHandler(jcards));
		
		log.exiting("JCardsFrame", "()");
	}
	
	/** Set the controls in the corners of the scroll panel */
	private void setScrollCorners(JScrollPane scroll)
	{
		log.entering("JCardsFrame", "setScrollCorners");
		SelectVisibleColumnsButton svcb = new SelectVisibleColumnsButton(cardsTable);
		scroll.setCorner(JScrollPane.UPPER_RIGHT_CORNER, svcb);
		log.exiting("JCardsFrame", "setScrollCorners");
	}

	public JXTable getTable()
	{
		return cardsTable;
	}

	/** Set up the table */
	private void createTable()
	{
		log.entering("JCardsFrame", "createTable");
		ctm = new CardsTableModel(jcards);
		cardsTable = new JXTable(ctm);

		cardsTable.setAutoResizeMode(JTable.AUTO_RESIZE_SUBSEQUENT_COLUMNS);
		cardsTable.setShowHorizontalLines(false);
		cardsTable.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		cardsTable.getTableHeader().setReorderingAllowed(false);
		cardsTable.addMouseListener(new TableMouseListener(this, cardsTable, jcards.getCards(), actionManager));
		cardsTable.addKeyListener(new TableKeyListener(cardsTable, this));
		cardsTable.setEditable(JCards.getUserSettings().isTableEdit());
		cardsTable.getActionMap().remove("find"); // remove default find binding
		
		// renderers and editors
		cardsTable.setDefaultEditor(JComboBox.class, new CategoryEditor());
		cardsTable.setDefaultRenderer(Date.class, new DateRenderer());
		cardsTable.setDefaultEditor(Date.class, new DateEditor());
		NumberRenderer numRender = new NumberRenderer();
		cardsTable.setDefaultRenderer(Number.class, numRender);
		JTextField tf = new JTextField();
		tf.setBorder(null);
		cardsTable.setDefaultEditor(Number.class, new DefaultCellEditor(tf));
		
		// Set up tool tips
		cardsTable.getTableHeader().setToolTipText(JCards.getLabelString("table.tt"));
		
		// set up column header listener
		cardsTable.getTableHeader().addMouseListener(new TableColumnMouseListener(jcards));
		
		log.exiting("JCardsFrame", "createTable");
	}
	
	/**
	 * Get the table auto resize mode
	 * @return one of JTable.AUTO_RESIZE_ALL_COLUMNS, JTable.AUTO_RESIZE_LAST_COLUMN, JTable.AUTO_RESIZE_NEXT_COLUMN,
	 * JTable.AUTO_RESIZE_OFF, JTable.AUTO_RESIZE_SUBSEQUENT_COLUMNS
	 */
	public int getTableAutoResizeMode()
	{
		return cardsTable.getAutoResizeMode();
	}
	
	/** Finished initialising and reading properties */
	public void finishedInit()
	{
		log.entering("JCardsFrame", "finishedInit");
		
		fontChanged();
		updateRecentFilesMenu(null);
		
		try {
			UIManager.setLookAndFeel(JCards.getUserSettings().getTheme());
			SwingUtilities.updateComponentTreeUI(this);
		} catch (Exception ex) {
			log.log(Level.WARNING, "Cannot set LnF to "+JCards.getUserSettings().getTheme(), ex);
		}

		pack();
		setInitialLocation();
		viewToolbar(JCards.getUserSettings().isViewToolbar());
		cardsTable.setAutoResizeMode(JCards.getUserSettings().getTableAutoResizeMode());

		setVisible(true);
		
		// see if the welcome screen should be displayed
		if (JCards.getUserSettings().isShowWelcome())
		{
			log.finer("Showing welcome screen");
			Action action = actionManager.getAction(ActionEnum.WELCOME);
			action.actionPerformed(null);
		}
		
		log.entering("JCardsFrame", "finishedInit");
	}
	
	/** Set the initial location of the window */
	public void setInitialLocation()
	{
		log.entering("JCardsFrame", "setInitialLocation");
		
		UserSettings settings = JCards.getUserSettings();
		if (settings.getInitLocation() != null)
		{
			setLocation(settings.getInitLocation());
		}
		else
		{
			centreWindow(this, this);
		}
		
		log.exiting("JCardsFrame", "setInitialLocation");
	}
	
	/**
	 * Get the preferred size of the app
	 * @return the preferred size
	 */
	@Override
	public Dimension getPreferredSize()
	{
		int width, height;
		
		UserSettings settings = JCards.getUserSettings();
		Dimension prefSize = (settings.getOverrideSize() != null ? settings.getOverrideSize() : new Dimension(600, 400));
		
		return prefSize;
	}
	
	/** Set whether the toolbar is visible */
	public void viewToolbar(boolean view)
	{
		log.entering("JCardsFrame", "viewToolbar", view);
		toolbar.setVisible(view);
		if (view)
		{
			toolbar.validate();
		}
		// this shouldn't happen unless the GUI setup is really, really slow
		if (toolbarMI != null)
		{
			toolbarMI.setSelected(view);
		}
		log.exiting("JCardsFrame", "viewToolbar");
	}
	
	public Action getAction(ActionEnum ae)
	{
		return actionManager.getAction(ae);
	}
	
	/** Create a toolbar */
	public final void createToolbar()
	{
		log.entering("JCardsFrame", "createToolbar");
		if (toolbar == null)
		{
			toolbar = new JToolBar();
			toolbar.setFloatable(false);
		}
		else
		{
			toolbar.removeAll();
		}
		
		// add buttons
		Action action;
		JButton button;
		List<I18NIcon> buttons = JCards.getUserSettings().getUserIcons();
		for (I18NIcon icon : buttons)
		{
			action = actionManager.getAction(icon.getKey());
			if (action != null)
			{
				button = new JButton(action);
				button.setText(null);
				toolbar.add(button);
			}
		}
		
		if (toolbar.isVisible())
		{
			toolbar.repaint();
		}
		log.exiting("JCardsFrame", "createToolbar");
	}
	
	/**
	 * Add menu items
	 * @param menu Menu object to add the items to
	 * @param miObjs array of Strings, MenuShortcuts and other Menus
	 */
	private void addMenuItems(JMenu menu, Object[] miObjs)
	{
		Object miObj;
		String miName;
		JMenuItem mi = null;
		JCardsAction jcAction;
		
		for (int i=0; i<miObjs.length; i++)
		{
			miObj = miObjs[i];
			if (miObj instanceof JCardsAction)
			{
				jcAction = (JCardsAction)miObj;
				mi = new JMenuItem(jcAction);
				if (jcAction.getMnemonicIndex() > -1)
				{
					mi.setMnemonic(jcAction.getMnemonic());
					mi.setDisplayedMnemonicIndex(jcAction.getMnemonicIndex());
				}
				menu.add(mi);
			}
			else if (miObj instanceof String)
			{
				miName = (String)miObjs[i];
				if (miName.length() == 0)
				{
					menu.addSeparator();
				}
			}
			else if (miObj instanceof JMenu)
			{
				JMenu subMenu = (JMenu)miObj;
				menu.add(subMenu);
			}
		}
	}
	
	/** Create the menus */
	private void createMenuBar()
	{
		log.entering("JCardsFrame", "createMenuBar");
		JMenuBar bar = new JMenuBar();
		
		// export menu
		JMenu exportMenu = new JCardsMenu("export");
		exportMenu.setIcon(getImage(UserSettings.ICON_EXPORT));
		Object[] exportMenuItems = {
			actionManager.getAction(ActionEnum.CSV_EXPORT),
			actionManager.getAction(ActionEnum.CATEGORY_EXPORT),
			actionManager.getAction(ActionEnum.XML_EXPORT),
			actionManager.getAction(ActionEnum.HTML_CARD),
			actionManager.getAction(ActionEnum.HTML_TABLE)
		};
		addMenuItems(exportMenu, exportMenuItems);
		
		// file menu
		JMenu fileMenu = new JCardsMenu("file");
		Object[] fileMenuItems = {
			actionManager.getAction(ActionEnum.NEW),
			actionManager.getAction(ActionEnum.OPEN),
			actionManager.getAction(ActionEnum.OPEN_URL),
			recentFilesMenu, KeyEvent.VK_R,
			"",
			actionManager.getAction(ActionEnum.SAVE),
			actionManager.getAction(ActionEnum.SAVE_AS),
			actionManager.getAction(ActionEnum.PASSWORD),
			"",
			actionManager.getAction(ActionEnum.PRINT),
			"",
			actionManager.getAction(ActionEnum.CSV_IMPORT),
			exportMenu,
			"",
			actionManager.getAction(ActionEnum.EXIT)
		};
		addMenuItems(fileMenu, fileMenuItems);
		bar.add(fileMenu);
		fileMenuItems = null;
		
		// Edit menu
		JMenu editMenu = new JCardsMenu("edit");
		Object[] editMenuItems = {
			actionManager.getAction(ActionEnum.NEW_CARD),
			actionManager.getAction(ActionEnum.EDIT_CARD),
			actionManager.getAction(ActionEnum.DELETE_CARD),
			actionManager.getAction(ActionEnum.UNDELETE),
			"",
			actionManager.getAction(ActionEnum.COPY_CARD),
			actionManager.getAction(ActionEnum.COPY_CLIPBOARD),
			"",
			actionManager.getAction(ActionEnum.FIND),
			actionManager.getAction(ActionEnum.FIND_RESET),
			actionManager.getAction(ActionEnum.FIND_REPLACE)
		};
		addMenuItems(editMenu, editMenuItems);
		bar.add(editMenu);
		editMenuItems = null;
		
		bar.add(editMenu);
		
		// DB menu
		JMenu dbMenu = new JCardsMenu("db");
		
		Object[] dbMenuItems = {
			actionManager.getAction(ActionEnum.DB_DIALOG),
			actionManager.getAction(ActionEnum.COMMENTS),
			"",
			actionManager.getAction(ActionEnum.DB_STATS),
			actionManager.getAction(ActionEnum.FIELD_STATS),
			"",
			actionManager.getAction(ActionEnum.PREFS)
		};
		addMenuItems(dbMenu, dbMenuItems);
		bar.add(dbMenu);
		dbMenuItems = null;
		
		// View menu
		viewMenu = new JCardsMenu("view");
		
		// create toolbar menu item
		JCardsAction jcAction = actionManager.getAction(ActionEnum.TOOLBAR);
		toolbarMI = new JCheckBoxMenuItem(jcAction);
		int mnIndex = jcAction.getMnemonicIndex();
		if (mnIndex > -1)
		{
			toolbarMI.setDisplayedMnemonicIndex(mnIndex);
			toolbarMI.setMnemonic(jcAction.getMnemonic());
			toolbarMI.setText(JCardsAction.getTextWithoutMnemonic(toolbarMI.getText()));
		}
		viewMenu.add(toolbarMI);
		viewMenu.addSeparator();
		bar.add(viewMenu);
		
		JMenu helpMenu = new JCardsMenu("help");
		Object[] helpMenuItems = {
			actionManager.getAction(ActionEnum.HELP),
			actionManager.getAction(ActionEnum.TUTORIAL),
			actionManager.getAction(ActionEnum.WELCOME),
			actionManager.getAction(ActionEnum.ABOUT)
		};
		addMenuItems(helpMenu, helpMenuItems);
		bar.add(helpMenu);
		
		setJMenuBar(bar);
		log.exiting("JCardsFrame", "createMenuBar");
	}
	
	/**
	 * Load an image from either the JAR file or filesystem
	 * @param imgName name of the image
	 * @return the loaded image
	 */
	public static ImageIcon getImage(String imgName)
	{
		log.entering("JCardsFrame", "getImage", imgName);
		
		ImageIcon i;
		URL imgURL = JCards.class.getResource(imgName);
		if (imgURL != null)
		{
			i = new ImageIcon(imgURL);
		}
		else
		{
			// remove leading /
			i = new ImageIcon(imgName.substring(1));
		}
		
		log.exiting("JCardsFrame", "getImage");
		
		return i;
	}
	
	/** Reset the find results */
	public void resetFind()
	{
		log.entering("JCardsFrame", "resetFind");
		jcards.getCards().resetFind();
		actionManager.toggleActionEnabled(ActionEnum.FIND_RESET, false);
		log.exiting("JCardsFrame", "resetFind");
	}
	
	/** Refresh the table display */
	public void refreshTable()
	{
		log.entering("JCardsFrame", "refreshTable");
		ctm.fireTableDataChanged();
		log.exiting("JCardsFrame", "refreshTable");
	}
	
	/**
	 * Enable/disable table editing
	 * @param edit true to enable in-table editing, false otherwise
	 */
	public void setTableEditable(boolean edit)
	{
		cardsTable.setEditable(edit);
	}
	
	/**
	 * Calculate the statistics for date and numeric columns
	 * @param dbStats if true calculate DB statistics, if false ask for a field
	 */
	public void showStatisticsDialog(boolean dbStats)
	{
		log.entering("JCardsFrame", "showStatisticsDialog");
		String title;
		String field = null;
		
		// check that there is some data
		Cards cards = jcards.getCards();
		if (cards.size() == 0)
		{
			JOptionPane.showMessageDialog(this, JCards.getMessage("no.cards.def"), JCards.getMessage("no.data"), JOptionPane.INFORMATION_MESSAGE);
			return;
		}
		
		if (!dbStats)
		{
			DBDef def = DBDef.getInstance();
			List statFields = def.getDateOrNumericFields();
			if (statFields.isEmpty())
			{
				JOptionPane.showMessageDialog(this, JCards.getMessage("no.date.numeric"), JCards.getMessage("no.fields"), JOptionPane.INFORMATION_MESSAGE);
				return;
			}
			
			// get the user to select one
			JLabel msg = new JCardsLabel("field", true);
			title = JCards.getLabelString("select.field");
			Object[] fieldNames = def.getDateOrNumericFields().toArray();
			field = (String)JOptionPane.showInputDialog(this, msg, title, JOptionPane.QUESTION_MESSAGE, null, fieldNames, null);
			if (field != null)
			{
				title = JCards.getFormattedMessage("stats", field);
			}
			else
			{
				return;
			}
		}
		else
		{
			title = JCards.getMessage("stats.db");
		}
		String stats = jcards.getStatistics(field);
		
		JOptionPane.showMessageDialog(this, stats, title, JOptionPane.INFORMATION_MESSAGE);
		
		log.exiting("JCardsFrame", "showStatisticsDialog");
	}
	
	/**
	 * Search for matching cards
	 * @param FindReplaceDetails find/replace details
	 * @return number of cards found
	 */
	public int findCards(FindReplaceDetails frDetails)
	{
		log.entering("JCardsFrame", "findCards", frDetails.searchText);
		
		Cards cards = jcards.getCards();
		int found = cards.findCards(frDetails);
		if (found > 0 && !frDetails.replaceOp)
		{
			if (found < cards.size())
			{
				setStatus(JCards.getFormattedMessage("found", found));
				actionManager.toggleActionEnabled(ActionEnum.FIND_RESET, true);
			}
			else {
				setStatus(JCards.getFormattedMessage("found.all"));
			}
		}
		
		log.exiting("JCardsFrame", "findCards", found);
		return found;
	}
	
	/**
	 * Gets the selected table row taking into account any sorting or filtering
	 * @return selected table row
	 */
	private int getSelectedTableRow()
	{
		return cardsTable.convertRowIndexToModel(cardsTable.getSelectedRow());
	}
	
	/**
	 * Add or edit a card
	 * @param newCard set to true for a new card
	 * @param editing should be set to true if autoEdit is on or creating a new card
	 * @param copy should be set to true if using an existing card as a template
	 */
	public void editCard(boolean newCard, boolean editCard, boolean copyCard)
	{
		log.entering("JCardsFrame", "editCard");
		
		// see if there are any fields defined
		if (DBDef.getInstance().getNumberOfFields() == 0)
		{
			JOptionPane.showMessageDialog(this, JCards.getMessage("no.fields.def"), JCards.getMessage("error"), JOptionPane.INFORMATION_MESSAGE);
		}
		else
		{
			CardDialog cDialog = null;
			if (newCard)
			{
				cDialog = new CardDialog(this, null, false, true, -1);
			}
			else
			{
				int row = getSelectedTableRow();
				if (row > -1)
				{
					Cards cards = jcards.getCards();
					Card card = cards.getCard(row);
					cDialog = new CardDialog(this, card, copyCard, editCard, row);
				}
			}
			
			if (cDialog != null)
			{
				cDialog.setPreferredSize(cardDialogSize);
				cDialog.pack();
				if (cardDialogPosition == null) {
					cDialog.setLocationRelativeTo(this);
				} else {
					cDialog.setLocation(cardDialogPosition);
				}
				cDialog.setVisible(true);
			}
		}
		
		log.exiting("JCardsFrame", "editCard");
	}
	
	/**
	 * Centers any Window in the middle of the app or screen
	 * @param parent parent Window
	 * @param win the Window to centre
	 */
	public void centreWindow(Window parent, Window win)
	{
		log.entering("JCardsFrame", "centreWindow");
		
		Dimension outer = parent.getSize();
		if (outer.height == 0 || outer.width == 0)
		{
			// main window hasn't been shown yet
			UserSettings settings = JCards.getUserSettings();
			outer = settings.getOverrideSize();
		}
		Dimension window = win.getSize();
		int x = (outer.width - window.width) / 2;
		int y = (outer.height - window.height) / 2;
		win.setLocation(x, y);
		
		log.exiting("JCardsFrame", "centreWindow");
	}
	
	/** Copy card contents to clipboard */
	public void copyToClipboard()
	{
		log.entering("JCardsFrame", "copyToClipboard");
		
		int selIndex = getSelectedTableRow();
		if (selIndex > -1)
		{
			Card card = jcards.getCards().getCard(selIndex);
			if (card != null)
			{
				jcards.copyToClipboard(card);
			}
		}
		
		log.exiting("JCardsFrame", "copyToClipboard");
	}
	
	/** Update the View menu */
	public void updateViewMenu()
	{
		log.entering("JCardsFrame", "updateViewMenu");
		
		JMenuItem cMI;
		JMenu catMenu;
		
		viewMenu.removeAll();
		viewMenu.add(toolbarMI);
		viewMenu.addSeparator();
		
		JCardsAction jcAction = actionManager.getAction(ActionEnum.VIEW_ALL);
		assert jcAction != null : "Cannot get ViewAllAction";
		cMI = new JMenuItem(jcAction);
		int mnIndex = jcAction.getMnemonicIndex();
		if (mnIndex > -1)
		{
			cMI.setDisplayedMnemonicIndex(mnIndex);
			cMI.setMnemonic(jcAction.getMnemonic());
			cMI.setText(JCardsAction.getTextWithoutMnemonic(cMI.getText()));
		}
		viewMenu.add(cMI);
		
		DBDef def = DBDef.getInstance();
		if (def.hasCategories())
		{
			viewMenu.addSeparator();		
			FieldDef fd;
			CategoryFieldDef cfd;
			String field, cat;
			List<String>catItems;
			List<String> fields = def.getFieldNames();
			CategoryListener catListener = new CategoryListener();
			for (String fieldName: fields)
			{
				fd = def.getFieldDef(fieldName);
				if (fd.getType() == FieldType.CATEGORY)
				{
					cfd = (CategoryFieldDef)fd;
					cat = cfd.getCategory();
					field = cfd.getName();
					catMenu = new JMenu(field);
					catMenu.setDisplayedMnemonicIndex(0);
					catItems = def.getCategory(cat);
					if (catItems != null)
					{
						for (String item : catItems)
						{
							cMI = new JMenuItem(item);
							cMI.setName(field);
							cMI.setActionCommand(item);
							cMI.addActionListener(catListener);
							catMenu.add(cMI);
						}
						viewMenu.add(catMenu);
					}
				}
			}
		}
		
		log.exiting("JCardsFrame", "updateViewMenu");
	}
	
	/**
	 * Update the recent files menu
	 * @param resource the last opened file which may be on the local filesystem or a URL
	 */
	public void updateRecentFilesMenu(String resource)
	{
		log.entering("JCardsFrame", "updateRecentFilesMenu");
		
		UserSettings settings = JCards.getUserSettings();
		if (resource != null)
		{
			settings.addRecentFile(resource);
		}
		List<String> recentFiles = settings.getRecentFiles();
		recentFilesMenu.setEnabled(!recentFiles.isEmpty());
		
		// update menu
		recentFilesMenu.removeAll();
		JMenuItem lfMI;
		RecentFilesListener rcfListener = new RecentFilesListener();
		int i = 1;
		File f;
		int mn = KeyEvent.VK_1;
		for (String fileName : recentFiles)
		{
			f = new File(fileName);
			if (f.isFile())
			{
				lfMI = new JMenuItem(i + ". "+ f.getName());
			}
			else
			{
				try
				{
					URL url = new URL(fileName);
					lfMI = new JMenuItem(i + ". " + url.toString());
				}
				catch (MalformedURLException mex)
				{
					continue;
				}
			}
			lfMI.setMnemonic(mn);
			lfMI.setToolTipText(fileName);
			lfMI.setActionCommand(fileName);
			lfMI.addActionListener(rcfListener);
			recentFilesMenu.add(lfMI);
			i++;
			mn++;
		}
		
		log.exiting("JCardsFrame", "updateRecentFilesMenu");
	}
	
	/** User clicked close window decoration */
	@Override
	public void windowClosing(WindowEvent e)
	{
		jcards.exit();
	}
	
	@Override public void windowActivated(WindowEvent e) {}
	@Override public void windowClosed(WindowEvent e) {}
	@Override public void windowDeactivated(WindowEvent e) {}
	@Override public void windowDeiconified(WindowEvent e) {}
	@Override public void windowIconified(WindowEvent e) {}
	@Override public void windowOpened(WindowEvent e) {}
	
	/** Remove any DB specific settings */
	public void reset()
	{
		log.entering("JCardsFrame", "reset");
		setTitle("JCards");
		passwordSet(false);
		cardsTable.getSelectionModel().clearSelection();
		log.exiting("JCardsFrame", "reset");
	}
	
	/**
	 * Listen for updates to DBDef
	 * @param o DBDef
	 * @param arg DBDefUpdate
	 */
	@Override public void update(Observable o, Object arg)
	{
		log.entering("JCardsFrame", "update");
		
		assert arg != null : "JCardsFrame update arg is null";
		
		DBDefUpdate event = (DBDefUpdate)arg;
		switch (event.getEvent())
		{
			case CATEGORY_REMOVED:
			case CATEGORY_RENAMED:
			case CATEGORY_UPDATED:
			case CATEGORY_ITEM_ADDED:
			case CATEGORY_ITEM_MOVED:
			case CATEGORY_ITEM_REMOVED:
			case CATEGORY_ITEM_RENAMED:
			case FIELD_ADD:
			case FIELD_CHANGE:
			case FIELD_DELETE:
				updateViewMenu();
				break;
		}
		
		log.exiting("JCardsFrame", "update");
	}
	
	public boolean isToolbarVisible()
	{
		return toolbar.isVisible();
	}
	
	/**
	 * While viewing a card the user wants to browse to the next/previous card
	 * @param direction either 1 or -1 for next or previous respectively
	 * @return the new selected index
	 */
	public Card editOtherCard(int direction)
	{
		log.entering("JCardsFrame", "editOtherCard", direction);
		
		// find the next card
		Card c = null;
		int rowIndex = cardsTable.getSelectedRow();
		if (rowIndex > -1)
		{
			rowIndex += direction;
			int numRows = jcards.getCards().getRowCount();
			if (rowIndex < 0) {
				rowIndex = numRows - 1;
			} else if (rowIndex >= numRows) {
				rowIndex = 0;
			}
			cardsTable.getSelectionModel().setSelectionInterval(rowIndex, rowIndex);
			c = jcards.getCards().getCard(cardsTable.convertRowIndexToModel(rowIndex));
		}
		
		log.exiting("JCardsFrame", "editOtherCard", c);
		
		return c;
	}
	
	/**
	 * Add a card to the list
	 * @param card the new card to add
	 * @return the index of the new card
	 */
	public int addCard(Card card)
	{
		log.entering("JCardsFrame", "addCard");
		
		int index = jcards.getCards().addCard(card);
		int selindex = cardsTable.convertRowIndexToView(index);
		cardsTable.getSelectionModel().setSelectionInterval(selindex, selindex);
		cardsTable.scrollRowToVisible(selindex);
		
		log.exiting("JCardsFrame", "addCard", index);
		
		return index;
	}
	
	/** Delete one or more cards */
	public void deleteCard()
	{
		log.entering("JCardsFrame", "deleteCard");
		
		int selection = getSelectedTableRow();
		if (selection > -1)
		{
			Cards cards = jcards.getCards();
			cards.deleteCard(selection);
			if (cards.size() > 0)
			{
				if (selection == cards.size())
				{
					selection--;
				}

				cardsTable.getSelectionModel().setSelectionInterval(cardsTable.getSelectedRow(), cardsTable.getSelectedRow());
			}
		}
		
		log.exiting("JCardsFrame", "deleteCard");
	}
	
	/** Restore the deleted cards */
	public void undeleteCard()
	{
		jcards.getCards().undeleteCard();
	}
	
	/**
	 * Replace a card entry
	 * @param cardIndex the index of the card to replace
	 * @param card the replacement card
	 */
	public void replaceCard(int cardIndex, Card card)
	{
		log.entering("JCardsFrame", "replaceCard", cardIndex);
		
		int index = jcards.getCards().setCard(cardIndex, card);
		cardsTable.getSelectionModel().setAnchorSelectionIndex(index);
		
		log.exiting("JCardsFrame", "replaceCard", index);
	}
	
	/**
	 * Get the minimum size of the app
	 * @return the minimum size
	 */
	@Override public Dimension getMinimumSize()
	{
		return new Dimension(240, 282);
	}
	
	/**
	 * Get the maximum possible size
	 * @return the maximum size
	 */
	@Override public Dimension getMaximumSize()
	{
		Toolkit tk = Toolkit.getDefaultToolkit();
		return tk.getScreenSize();
	}
	
	/**
	 * Set the widths of the columns
	 * @param colWidths List of column widths
	 * @todo replace with setTableSettings to cover alignment as well
	 */
	public void setColWidths(List<Integer> colWidths)
	{
		log.entering("JCardsFrame", "setColWidths", colWidths);
		
		if (colWidths != null)
		{
			TableColumn tc;
			TableColumnModel tcm = cardsTable.getColumnModel();
			if (colWidths.size() <= tcm.getColumnCount())
			{
				int col = 0;
				for (int i : colWidths)
				{
					tc = tcm.getColumn(col++);
					tc.setPreferredWidth(i);
				}
				cardsTable.validate();
			}
			else
			{
				log.log(Level.WARNING, "Too many column widths: {0}", colWidths.size());
			}
		}
		
		log.exiting("JCardsFrame", "setColWidths");
	}
	
	/** Get the column widths */
	public Element getColWidths()
	{
		log.entering("JCardsFrame", "getColWidths");
		
		Element el = new Element(Elements.COL_WIDTHS);
		Element col;
		TableColumn tc;
		Enumeration<TableColumn> enm = cardsTable.getColumnModel().getColumns();
		while (enm.hasMoreElements())
		{
			tc = enm.nextElement();
			col = new Element(Elements.WIDTH);
			col.setText(Integer.toString(tc.getWidth()));
			el.addContent(col);
		}
		
		log.exiting("JCardsFrame", "getColumns");
		
		return el;
	}
	
	/**
	 * Get the current sort fields
	 * @return List of SortDirectives
	 */
	public SortDirective getSortedColumn()
	{
		log.entering("JCardsFrame", "getSortedColumn");
		
		SortDirective sortDrv = null;
		TableColumn sortedCol = cardsTable.getSortedColumn();
		if (sortedCol != null)
		{
			DBDef def = DBDef.getInstance();
			FieldDef fd = def.getFieldDef(sortedCol.getModelIndex());
			SortOrder so = cardsTable.getSortOrder(cardsTable.convertColumnIndexToView(sortedCol.getModelIndex()));
			sortDrv = new SortDirective(fd.getName(), so);
		}

		log.exiting("JCardsFrame", "getSortedColumn", sortDrv);
		
		return sortDrv;
	}
	
	/**
	 * Set the sorting columns
	 * @param sortColumns list of SortDirective
	 */
	public void setSortedColumn(SortDirective sortedColumn)
	{
		log.entering("JCardsFrame", "setSortColumns");

		// Check to see if the frame is visible. On startup temporarily store the sortColumns
		log.log(Level.FINER, "isVisible={0}", isVisible());

		if (sortedColumn != null)
		{
		    List<String> visFields = DBDef.getInstance().getVisibleFields();
		    int col = visFields.indexOf(sortedColumn.colName);
		    if (col > -1) {
			cardsTable.setSortOrder(col, sortedColumn.sortOrder);
		    }
		}

		log.exiting("JCardsFrame", "setSortColumns");
	}
	
	/** Print the table */
	public void printTable()
	{
		log.entering("JCardsFrame", "printTable");

		PrintRequestAttributeSet printAttrs = new HashPrintRequestAttributeSet();
		printAttrs.add(new RequestingUserName(System.getProperty("user.name"), null));
		printAttrs.add(OrientationRequested.LANDSCAPE);
		MessageFormat headerFormat = null;
		if (jcards.getCardFile() != null)
		{
			String fileName = jcards.getCardFile().getName();
			headerFormat = new MessageFormat(fileName);
			printAttrs.add(new JobName("JCards: "+fileName, null));
		}
		else
		{
			printAttrs.add(new JobName("JCards print", null));
		}
		MessageFormat footerFormat = new MessageFormat(JCards.getLabelString("page")+" {0}");
		try {
			cardsTable.print(JTable.PrintMode.NORMAL, headerFormat, footerFormat, true, printAttrs, true);
			setStatus(JCards.getMessage("printed.table"));
		}
		catch (PrinterException pex)
		{
			log.log(Level.WARNING, "Cannot print table", pex);
			showError(JCards.getMessage("error"), pex.getMessage(), "JCardsFrame.printTable", pex, Level.WARNING);
		}
		catch (NullPointerException npex)
		{
			log.warning("JDK printing bug encountered");
			showError(JCards.getMessage("warning"), JCards.getMessage("jdk.print.bug"), "JCardsFrame.printTable", npex, Level.WARNING);
		}
		
		log.exiting("JCardsFrame", "printTable");
	}
	
	/**
	 * Display a URL in a web browser
	 * @param url the URL to display
	 */
	public void displayURL(URL url)
	{
		log.entering("JCardsFrame", "displayURL", url);

		assert url != null : "URL is null!";
		
		if (JCards.getUserSettings().isUseSystemBrowser())
		{
			try
			{
				Desktop.getDesktop().browse(url.toURI());
			}
			catch (Exception ex)
			{
				log.log(Level.WARNING, "Cannot display URL "+url, ex);
				String msg = JCards.getFormattedMessage("cannot.display.url", url);
				showError(JCards.getMessage("error"), msg, "JCardsFrame.displayURL", ex, Level.WARNING);
			}
		}
		else // use user defined browser
		{
			String browser = JCards.getUserSettings().getBrowserLocation();

			// cannot find it
			if (browser == null)
			{
				JOptionPane.showMessageDialog(this, JCards.getMessage("no.browser"), JCards.getMessage("error"), JOptionPane.ERROR_MESSAGE);
			}
			else
			{
				try
				{
					new ProcessBuilder(browser, url.toString()).start();
				}
				catch (IOException ioex)
				{
					log.log(Level.WARNING, "Cannot start browser for url "+url, ioex);
					String msg = JCards.getFormattedMessage("cannot.display.url", url);
					showError(JCards.getMessage("error"), msg, "JCardsFrame.displayURL", ioex, Level.WARNING);
				}
			}
		}
		log.exiting("JCardsFrame", "displayURL");
	}
	
	/** The font in use has changed - update the GUI */
	public void fontChanged()
	{
		log.entering("JCardsFrame", "fontChanged");
		
		FontUIResource font = new FontUIResource(JCards.getUserSettings().getFont());
		UIManager.put("Button.font", font);
		UIManager.put("CheckBox.font", font);
		UIManager.put("CheckBoxMenuItem.font", font);
		UIManager.put("ColorChooser.font", font);
		UIManager.put("ComboBox.font", font);
		UIManager.put("EditorPane.font", font);
		UIManager.put("Label.font", font);
		UIManager.put("List.font", font);
		UIManager.put("MenuBar.font", font);
		UIManager.put("Menu.font", font);
		UIManager.put("MenuItem.font", font);
		UIManager.put("OptionPane.font", font);
		UIManager.put("Panel.font", font);
		UIManager.put("PasswordField.font", font);
		UIManager.put("PopupMenu.font", font);
		UIManager.put("ProgressBar.font", font);
		UIManager.put("RadioButton.font", font);
		UIManager.put("RadioButtonMenuItem.font", font);
		UIManager.put("ScrollPane.font", font);
		UIManager.put("TabbedPane.font", font);
		UIManager.put("Table.font", font);
		UIManager.put("TableHeader.font", font);
		UIManager.put("TextArea.font", font);
		UIManager.put("TextField.font", font);
		UIManager.put("TextPane.font", font);
		UIManager.put("TitledBorder.font", font);
		UIManager.put("ToggleButton.font", font);
		UIManager.put("ToolBar.font", font);
		UIManager.put("ToolTip.font", font);
		UIManager.put("Tree.font", font);
		UIManager.put("Viewport.font", font);
		
		SwingUtilities.updateComponentTreeUI(this);
		pack();
		
		log.exiting("JCardsFrame", "fontChanged");
	}
	
	/**
	 * Display a message in the status bar
	 * @param message the text to display
	 */
	public final void setStatus(String message)
	{
		log.entering("JCardsFrame", "setStatus");
		statusBar.setStatus(message);
	}
	
	/**
	 * Remember the size of the card dialog
	 * @param dim the size to remember
	 */
	public void setCardDialogSize(Dimension dim)
	{
		log.entering("JCardsFrame", "setCardDialogSize", dim);
		cardDialogSize = dim;
	}
	
	/**
	 * Remember the position of the card dialog
	 * @param pt the position to remember
	 */
	public void setCardDialogPosition(Point pt)
	{
		log.entering("JCardsFrame", "setCardDialogPosition", pt);
		cardDialogPosition = pt;
	}
	
	/** Show the comment dialog */
	public void showCommentDialog()
	{
		log.entering("JCardsFrame", "showCommentDialog");
		CommentDialog commentDialog = new CommentDialog(jcards.getFrame(), jcards.getCards());
		commentDialog.pack();
		commentDialog.setLocationRelativeTo(jcards.getFrame());
		commentDialog.setVisible(true);
		log.exiting("JCardsFrame", "showCommentDialog");
	}
	
	/** Display the password dialog */
	public void showPasswordDialog()
	{
		log.entering("JCardsFrame", "showPasswordDialog");
		PasswordDialog passDialog = new PasswordDialog(jcards, null, jcards.getPassword(), true);
		passDialog.pack();
		passDialog.setLocationRelativeTo(this);
		passDialog.setVisible(true);
		log.exiting("JCardsFrame", "showPasswordDialog");
	}
	
	/**
	 * Indicate to the status bar the encrypted status of the current file
	 * @param passwordSet
	 */
	public void passwordSet(boolean passwordSet)
	{
		statusBar.setEncrypted(passwordSet);
	}
	
	/**
	 * Display an error message and allow the user to report a bug
	 * @param title title of the error dialog
	 * @param message message to display
	 * @param category where the error occurred
	 * @param error exception
	 * @param errorLevel level of the error
	 */
	public void showError(String title, String message, String category, Throwable error, Level errorLevel)
	{
		JXErrorPane errorPane = new JXErrorPane();
		ErrorInfo errorInfo = new ErrorInfo(title, message, null, category, error, errorLevel, null);
		errorPane.setErrorInfo(errorInfo);
		ErrorReporter emailReporter = new ErrorReporter() {
			@Override
			public void reportError(ErrorInfo info) throws NullPointerException
			{
				try
				{
					String subject = "JCards%20v"+JCards.VERSION+"%20bug";
					Desktop.getDesktop().mail(new URI("mailto:nsydenham@yahoo.co.uk?subject="+subject));
				}
				catch (Exception mex)
				{
					log.log(Level.SEVERE, "Cannot send email", mex);
				}
			}
		};
		errorPane.setErrorReporter(emailReporter);
		JDialog errorDialog = JXErrorPane.createDialog(this, errorPane);
		errorDialog.setVisible(true);
	}

	/**
	 * Load an image from the current JCards DB
	 * @param imgName internal image ref
	 * @return InputStream or null
	 */
	public InputStream loadImageFromDB(String imgName)
	{
		log.entering("JCardsFrame", "loadImageFromDB", imgName);
		try {
			return jcards.loadImageFromDB(imgName);
		} catch (JCAException jcaex) {
			showError(JCards.getMessage("error"), JCards.getMessage("cannot.load.image"), "JCardsFrame.loadImageFromDB", jcaex, Level.WARNING);
		} finally {
			log.exiting("JCardsFrame", "loadImageFromDB");
		}
		
		return null;
	}
}
