/*
 * JCards
 * Copyright (C) 2002-2009 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package nrs.jcards.gui;

import java.awt.Component;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.*;

/**
 * Combo box for selecting the Look and Feel
 * @author Copyright &#169; 2002-2009 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class LnFComboBox extends JComboBox
{
	private Component comp;

	private class LnFComboBoxModel extends DefaultComboBoxModel
	{
		private List<String> lnfList = new ArrayList<String>(5);
		private Map<String, UIManager.LookAndFeelInfo> lnfMap = new HashMap<String, UIManager.LookAndFeelInfo>(5);
		private int selectedIndex = 0;

		public LnFComboBoxModel()
		{
			for (UIManager.LookAndFeelInfo lnf : UIManager.getInstalledLookAndFeels())
			{
				lnfList.add(lnf.getName());
				lnfMap.put(lnf.getName(), lnf);
			}

			LookAndFeel lnf = UIManager.getLookAndFeel();
			if (lnf != null) {
				selectedIndex = lnfList.indexOf(lnf.getName());
				if (selectedIndex == -1) {
					selectedIndex = 0;
				}
			}
		}

		@Override
		public Object getElementAt(int index)
		{
			return lnfList.get(index);
		}

		@Override
		public int getIndexOf(Object anObject)
		{
			return lnfList.indexOf(anObject);
		}

		@Override
		public int getSize()
		{
			return lnfList.size();
		}

		@Override
		public Object getSelectedItem()
		{
			return lnfList.get(selectedIndex);
		}

		public void setSelectedIndex(int selectedIndex)
		{
			this.selectedIndex = selectedIndex;
		}

		/**
		 * Gets the selected theme
		 * @return class name for the selected theme
		 */
		public String getSelectedTheme()
		{
			return lnfMap.get(lnfList.get(selectedIndex)).getClassName();
		}

		/**
		 * Set the selected theme
		 * @param className theme class name
		 */
		public void setSelectedTheme(String className)
		{
			for (Entry<String, UIManager.LookAndFeelInfo> e : lnfMap.entrySet())
			{
				if (e.getValue().getClassName().equals(className))
				{
					selectedIndex = lnfList.indexOf(e.getKey());
					break;
				}
			}
		}

		@Override
		public void setSelectedItem(Object anObject)
		{
			if (anObject != null && lnfList.indexOf(anObject) > -1)
			{
				selectedIndex = lnfList.indexOf(anObject);
				String lnfName = (String)anObject;

				try {
					UIManager.setLookAndFeel(lnfMap.get(lnfName).getClassName());
					if (comp != null)
					{
						SwingUtilities.updateComponentTreeUI(comp);
						comp.repaint();
					}
				}
				catch (Exception ex)
				{
					Logger log = Logger.getLogger(LnFComboBoxModel.class.getName());
					log.log(Level.WARNING, "Cannot set LnF", ex);
				}
			}
		}
	}

	/**
	 * Create a new LnFComboBox
	 * @param comp component to update when a LnF is selected
	 */
	public LnFComboBox(Component comp)
	{
		this.comp = comp;
		setModel(new LnFComboBoxModel());
	}

	/**
	 * Return the class name of the selected theme
	 * @return
	 */
	public String getThemeClassName()
	{
		LnFComboBoxModel cbModel = (LnFComboBoxModel)getModel();
		return cbModel.getSelectedTheme();
	}

	/**
	 * Set the theme class name
	 * @param className
	 */
	public void setThemeClassName(String className)
	{
		LnFComboBoxModel cbModel = (LnFComboBoxModel)getModel();
		cbModel.setSelectedTheme(className);
	}
}
