/*
 * JCards
 * Copyright (C) 2002-2009 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package nrs.jcards.gui;

import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.io.*;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.*;
import nrs.jcards.JCards;
import nrs.jcards.UserSettings;
import nrs.jcards.fields.ImageFieldDef;
import nrs.jcards.io.JCAReader;
import nrs.jcards.io.JCAReader.SplitFilename;

/**
 * Picture display
 *
 * @author Copyright &#169; 2002-2009 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class ScrollablePicture extends JLabel implements ActionListener, MouseListener
{
	private boolean missingPicture;
	private boolean maintainAspectRatio;
	private Dimension resizeDimension;
	private String imgName;
	private JPopupMenu popupMenu;
	private JMenuItem exportMI, editMI;
	private JCardsFrame jcardsFrame;
	private static final Logger log = Logger.getLogger(ScrollablePicture.class.getName());

	private static final String EXPORT = "export";
	private static final String EDIT = "edit.pic";
	
	public ScrollablePicture(boolean maintainAspectRatio, Dimension resizeDimension, JCardsFrame jcardsFrame)
	{
		this.maintainAspectRatio = maintainAspectRatio;
		this.resizeDimension = resizeDimension;
		this.jcardsFrame = jcardsFrame;
		
		//Let the user scroll by dragging to outside the window.
		setAutoscrolls(true); //enable synthetic drag events
		addMouseListener(this);

		popupMenu = new JPopupMenu();

		// for internal images
		exportMI = new JCardsMenuItem(EXPORT);
		exportMI.setIcon(JCardsFrame.getImage(UserSettings.ICON_EXPORT));
		exportMI.addActionListener(this);
		popupMenu.add(exportMI);

		editMI = new JCardsMenuItem(EDIT);
		editMI.setIcon(JCardsFrame.getImage(UserSettings.ICON_IMAGE));
		editMI.addActionListener(this);
		popupMenu.add(editMI);
	}
	
	public void displayError()
	{
		missingPicture = true;
		setText(JCards.getLabelString("image.broken"));
		setHorizontalAlignment(CENTER);
		setOpaque(true);
		setBackground(Color.WHITE);
		setForeground(Color.BLACK);
	}

	/**
	 * Set the image to display
	 * @param imageInputStream
	 * @param imgName file name or internal JCards DB ref
	 */
	public void setImage(InputStream imageInputStream, String imgName)
	{
		log.entering("ScrollablePicture", "setImage", imgName);

		if (imageInputStream == null)
		{
			displayError();
		}
		else
		{
			try
			{
				ByteArrayOutputStream baos = new ByteArrayOutputStream(imageInputStream.available());
				byte[] buf = new byte[10000];
				BufferedInputStream bis = new BufferedInputStream(imageInputStream);
				int read;
				while ((read = bis.read(buf)) != -1)
				{
					baos.write(buf, 0, read);
				}
				bis.close();
				buf = null;
				log.finest("Read image bytes");
				Image img = new ImageIcon(baos.toByteArray()).getImage();
				log.finest("Converted to image");
				baos = null;

				if (resizeDimension != null)
				{
					log.log(Level.FINEST, "Resizing to {0}", resizeDimension);
					// figure out whether width or height is larger
					// set as floats so that the division below works
					float rwidth = resizeDimension.width;
					float rheight = resizeDimension.height;
					if (maintainAspectRatio)
					{
						if (img.getWidth(null) > img.getHeight(null))
						{
							rheight = (rwidth/img.getWidth(null))*img.getHeight(null);
						}
						else if (img.getHeight(null) > img.getWidth(null))
						{
							rwidth = (rheight/img.getHeight(null))*img.getWidth(null);
						}
					}
					img = img.getScaledInstance((int)rwidth, (int)rheight, Image.SCALE_SMOOTH);
					log.finest("Resized");
				}

				ImageIcon imgI = new ImageIcon(img);
				if (imgI.getImageLoadStatus() == MediaTracker.ERRORED) {
					displayError();
				}
				else
				{
					setIcon(imgI);
					setToolTipText(JCards.getLabelString("pic.tt"));
				}

				this.imgName = imgName;

				if (!imgName.startsWith(ImageFieldDef.PIC)) {
					exportMI.setVisible(false);
				}
			}
			catch (IOException ioe)
			{
				log.log(Level.WARNING, "Cannot display image "+imgName, ioe);
				displayError();
			}
		}

		log.exiting("ScrollablePicture", "setImage");
	}
	
	@Override
	public Dimension getPreferredSize()
	{
		Dimension dim = (missingPicture ? new Dimension(200, 50) : super.getPreferredSize());
		return dim;
	}

	@Override
	public void mouseClicked(MouseEvent e)
	{}

	@Override
	public void mousePressed(MouseEvent e)
	{
		if (!missingPicture && e.isPopupTrigger()) {
			popupMenu.show(this, e.getX(), e.getY());
		}
	}

	@Override
	public void mouseReleased(MouseEvent e)
	{
		if (!missingPicture && e.isPopupTrigger()) {
			popupMenu.show(this, e.getX(), e.getY());
		}
	}

	@Override
	public void mouseEntered(MouseEvent e)
	{}

	@Override
	public void mouseExited(MouseEvent e)
	{}

	private void exportImage(File imgFile)
	{
		log.entering("ScrollablePicture", "exportImage", imgFile.getAbsolutePath());

		InputStream imageIS = jcardsFrame.loadImageFromDB(imgName);
		if (imageIS != null)
		{
			try
			{
				BufferedOutputStream bos = new BufferedOutputStream(new FileOutputStream(imgFile));
				BufferedInputStream bis = new BufferedInputStream(imageIS);
				byte[] buffer = new byte[65536];
				int read;
				while ((read = bis.read(buffer)) != -1) {
					bos.write(buffer, 0, read);
				}
				bos.close();
				bis.close();
				jcardsFrame.setStatus(JCards.getFormattedMessage("img.exported", imgFile.getName()));
			} catch (FileNotFoundException fnex) {
				log.log(Level.WARNING, "Cannot export image to "+imgFile.getAbsolutePath(), fnex);
			} catch (IOException ioex) {
				log.log(Level.WARNING, "Cannot write to file "+imgFile.getAbsolutePath(), ioex);
			}
		}
		log.exiting("ScrollablePicture", "exportImage");
	}

	@Override
	public void actionPerformed(ActionEvent e)
	{
		String action = e.getActionCommand();
		if (action.equals(EXPORT))
		{
			JFileChooser fc = new JFileChooser();
			String name = "image";
			SplitFilename sf = JCAReader.splitFilename(imgName.substring(ImageFieldDef.PIC.length()));
			name += "." + sf.extension;
			fc.setSelectedFile(new File(name));

			if (fc.showSaveDialog(this) == JFileChooser.APPROVE_OPTION)
			{
				exportImage(fc.getSelectedFile());
			}
		}
		else if (action.equals(EDIT))
		{
			String imageToEdit = imgName;

			// export to temporary file first if internal
			if (imgName.startsWith(ImageFieldDef.PIC))
			{
				log.finer("Exporting internal image to temporary file");
				try
				{
					SplitFilename sf = JCAReader.splitFilename(imgName);
					File tempFile = File.createTempFile("image", "."+sf.extension);
					tempFile.deleteOnExit();
					exportImage(tempFile);
					imageToEdit = tempFile.getAbsolutePath();
				} catch (IOException ioex) {
					log.log(Level.WARNING, "Cannot export to temporary image", ioex);
					jcardsFrame.showError(JCards.getMessage("error"), JCards.getFormattedMessage("cannot.open.file", imgName),
							"ScrollablePicture.actionPerformed", ioex, Level.WARNING);
					return;
				}
			}

			String imgEditor = JCards.getUserSettings().getImageEditor();
			if (imgEditor != null)
			{
				File f = new File(imgEditor);
				if (f.exists())
				{
					try {
						new ProcessBuilder(imgEditor, imageToEdit).start();
						return;
					}
					catch (IOException ioex)
					{
						log.log(Level.WARNING, "Cannot open image "+imgName+" in editor "+imgEditor, ioex);
						jcardsFrame.showError(JCards.getMessage("error"), JCards.getFormattedMessage("cannot.open.file", imageToEdit),
							"ScrollablePicture.actionPerformed", ioex, Level.WARNING);
					}
				}
			}

			// if image editor is not set or failure above
			try {
				if (Desktop.isDesktopSupported())
				{
					Desktop d = Desktop.getDesktop();
					if (d.isSupported(Desktop.Action.EDIT)) {
						d.edit(new File(imageToEdit));
					} else if (d.isSupported(Desktop.Action.OPEN)) {
						d.open(new File(imageToEdit));
					}
				}
			} catch (IOException ioex) {
				log.log(Level.WARNING, "Cannot edit image "+imgName, ioex);
				jcardsFrame.showError(JCards.getMessage("error"), JCards.getFormattedMessage("cannot.open.file", imageToEdit),
					"ScrollablePicture.actionPerformed", ioex, Level.WARNING);
			}
		}
	}
}
