/*
 * JCards
 * Copyright (C) 2002-2011 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package nrs.jcards.gui;

import java.awt.Color;
import java.awt.SystemColor;
import java.awt.event.ActionEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.util.ArrayDeque;
import java.util.Deque;
import java.util.Observable;
import java.util.Observer;
import java.util.logging.Logger;
import javax.swing.*;
import nrs.jcards.CardUpdate;
import nrs.jcards.JCards;
import nrs.jcards.UserSettings;
import org.jdesktop.swingx.JXStatusBar;

/**
 * Status bar for displaying informational messages
 * 
 * @author Copyright &#169; 2002-2011 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class StatusBar extends JXStatusBar implements Observer
{
	private static int MAX_OLD_MSGS = 10;
	private final static Logger log = Logger.getLogger(StatusBar.class.getName());
	/** Timeout for blanking the status message */
	private final static int TIMEOUT = 5000;

	private JCardsFrame jcardsFrame;
	private JLabel encrypted = new JLabel();
	private JLabel comment = new JLabel();
    private JLabel records = new JLabel();
    private JLabel statusText = new JLabel();
	private Timer blankTimer;
	private Deque<String> oldMsgs = new ArrayDeque<String>(MAX_OLD_MSGS);
	
	/** Used to reset the status bar message text */
	private Action blankAction = new AbstractAction() {
		@Override
		public void actionPerformed(ActionEvent e) {
			statusText.setText(null);
		}
	};
	
	/** Creates new form StatusBar */
	public StatusBar(JCardsFrame jcardsFrame)
	{
		this.jcardsFrame = jcardsFrame;
		
        records.setHorizontalTextPosition(SwingConstants.CENTER);
		records.setToolTipText(JCards.getLabelString("num.records"));
        records.addMouseListener(new MouseAdapter() {
            @Override public void mouseClicked(MouseEvent evt) {
                resetFind();
            }
        });

		comment.setIcon(JCardsFrame.getImage(UserSettings.ICON_COMMENT));
		comment.setToolTipText(JCards.getLabelString("comment"));
		comment.setEnabled(false);
		comment.addMouseListener(new MouseAdapter() {
			@Override public void mouseClicked(MouseEvent e) {
				setComment();
			}
		});
		
        encrypted.setIcon(JCardsFrame.getImage(UserSettings.ICON_PASSWORD));
        encrypted.setToolTipText(JCards.getLabelString("file.unencrypted"));
        encrypted.setEnabled(false);
        encrypted.addMouseListener(new MouseAdapter() {
            @Override public void mouseClicked(MouseEvent evt) {
                setPassword();
            }
        });

		blankTimer = new Timer(TIMEOUT, blankAction);
		blankTimer.setRepeats(false);
		
		add(statusText, new JXStatusBar.Constraint(JXStatusBar.Constraint.ResizeBehavior.FILL));
		add(records, new JXStatusBar.Constraint(JXStatusBar.Constraint.ResizeBehavior.FIXED));
		add(comment, new JXStatusBar.Constraint(JXStatusBar.Constraint.ResizeBehavior.FIXED));
		add(encrypted, new JXStatusBar.Constraint(JXStatusBar.Constraint.ResizeBehavior.FIXED));
	}
	
	/**
	 * Display a short message
	 * @param text the text to display
	 */
	public void setStatus(String text)
	{
		log.entering("StatusBar", "setStatus", text);
		
		if (blankTimer.isRunning()) {
			blankTimer.stop();
		}
		statusText.setText(text);
		blankTimer.start();
		
		// store old messages
		if (oldMsgs.size() >= MAX_OLD_MSGS)
		{
			oldMsgs.removeLast();
		}
		oldMsgs.addFirst(text);
		
		// create tooltip text
		StringBuilder sb = new StringBuilder("<html><pre>");
		for (String msg : oldMsgs)
		{
			sb.append(msg);
			sb.append("\n");
		}
		sb.append("</pre></html>");
		statusText.setToolTipText(sb.toString());
		
		log.exiting("StatusBar", "setStatus");
	}
	
	/**
	 * Display the padlock icon if the file is encrypted
	 * @param fileEncrypted true if the file is encrypted, false otherwise
	 */
	public void setEncrypted(boolean fileEncrypted)
	{
		log.entering("StatusBar", "setEncrypted", fileEncrypted);
		encrypted.setEnabled(fileEncrypted);
		if (fileEncrypted) {
			encrypted.setToolTipText(JCards.getLabelString("file.encrypted"));
		} else {
			encrypted.setToolTipText(JCards.getLabelString("file.unencrypted"));
		}
		log.exiting("StatusBar", "setEncrypted");
	}
	
	/**
	 * Observe changes in the Cards object
	 * @param o
	 * @param arg
	 */
	@Override
	public void update(Observable o, Object arg)
	{
		log.entering("StatusBar", "update");
		CardUpdate cu = (CardUpdate)arg;
		
		switch (cu.cardOperation)
		{
			case CARD_DELETE:
			case CARD_NEW:
			case FIND:
			case LOAD:
				String text, tt = null;
				Color color;
				Icon icon = null;
				if (cu.foundCards > 0)
				{
					text = cu.foundCards + "/" + cu.numCards;
					tt = JCards.getLabelString("find.reset");
					icon = JCardsFrame.getImage(UserSettings.ICON_RESET);
					records.setHorizontalTextPosition(SwingConstants.RIGHT);
					color = Color.RED;
				}
				else
				{
					text = Integer.toString(cu.numCards);
					tt = JCards.getLabelString("num.records");
					color = SystemColor.textText;
				}
				records.setText(text);
				records.setToolTipText(tt);
				records.setForeground(color);
				records.setIcon(icon);
				records.repaint();
				break;
				
			case SET_INFO:
				if (cu.cardIndex == 0) {
					comment.setEnabled(false);
				} else {
					comment.setEnabled(true);
				}
				break;
		}
		
		log.exiting("StatusBar", "update");
	}

	/**
	 * Reset the table to show all results upon clicking the records indicator
	 * @param evt
	 */
	private void resetFind() {                           
		jcardsFrame.resetFind();
	}                          

	private void setPassword() {                             
		jcardsFrame.showPasswordDialog();
	}
	
	private void setComment() {
		jcardsFrame.showCommentDialog();
	}
}
