/*
 * JCards
 * Copyright (C) 2002-2009 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package nrs.jcards.gui.action;

import java.lang.reflect.Constructor;
import java.util.HashMap;
import java.util.Map;
import java.util.Observable;
import java.util.Observer;
import java.util.logging.Level;
import java.util.logging.Logger;
import nrs.jcards.CardUpdate;
import nrs.jcards.Cards;
import nrs.jcards.DBDef;
import nrs.jcards.DBDefUpdate;
import nrs.jcards.JCards;

/**
 * Manages actions - one reference to actions instead of many
 *
 * @author Copyright &#169; 2002-2009 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class ActionManager implements Observer
{
	private Map<String, JCardsAction> actionMap = new HashMap<String, JCardsAction>();
	private final static Logger log = Logger.getLogger(ActionManager.class.getName());

	/**
     * Creates a new instance of ActionManager.
     * Use class names defined in ActionEnum and reflection to create Actions
     */
	@SuppressWarnings("unchecked")
	public ActionManager(JCards jcards)
	{
		Class actionClass;
		Constructor<JCardsAction> actionConstructor;
		JCardsAction actionObject;

		log.entering("ActionManager", "()");
		try
		{
			for (ActionEnum action : ActionEnum.values())
			{
				log.fine("Creating action " + action.getClassName());
				actionClass = Class.forName("nrs.jcards.gui.action." + action.getClassName());
				log.finer("Created class");
				actionConstructor = actionClass.getConstructor(JCards.class, ActionEnum.class);
				log.finer("Created constructor");
				actionObject = actionConstructor.newInstance(jcards, action);
				log.finer("Created new instance");
				actionMap.put(action.getResourceName(), actionObject);
			}
		}
		catch (Exception fex)
		{
			log.log(Level.SEVERE, "Cannot create action", fex);
		}
		
		// register as listener for DBDef and Cards events
		DBDef.getInstance().addObserver(this);
		jcards.getCards().addObserver(this);
		
		log.exiting("ActionManager", "()");
	}

	/**
     * Get a defined action
     * @param action the action to find
     * @return the JCardsAction
     */
	public JCardsAction getAction(ActionEnum action)
	{
		return actionMap.get(action.getResourceName());
	}

	/**
     * Get a defined action
     * @see #getAction(ActionEnum)
     */
	public JCardsAction getAction(String resourceName)
	{
		return actionMap.get(resourceName);
	}
	
	/**
	 * Toggle whether an action is enabled or not
	 * @param actionEn
	 * @param enable
	 */
	public void toggleActionEnabled(ActionEnum actionEn, boolean enable)
	{
		getAction(actionEn).setEnabled(enable);
	}
	
	/**
	 * Internal method for enabling/disabling a number of actions in one go
	 * that are only relevant when at least one field is defined
	 */
	private void toggleFieldActions(boolean enable)
	{
		getAction(ActionEnum.FIELD_STATS).setEnabled(enable);
		getAction(ActionEnum.NEW_CARD).setEnabled(enable);
		getAction(ActionEnum.PRINT).setEnabled(enable);
		getAction(ActionEnum.SAVE).setEnabled(enable);
		getAction(ActionEnum.SAVE_AS).setEnabled(enable);
	}

	/**
	 * Internal method for enabling/disabling a number of actions in one go
	 * that are only relevant when at least one card is created
	 */
	private void toggleCardActions(boolean enable)
	{
		getAction(ActionEnum.COPY_CARD).setEnabled(enable);
		getAction(ActionEnum.COPY_CLIPBOARD).setEnabled(enable);
		getAction(ActionEnum.CSV_EXPORT).setEnabled(enable);
		getAction(ActionEnum.DB_STATS).setEnabled(enable);
		getAction(ActionEnum.DELETE_CARD).setEnabled(enable);
		getAction(ActionEnum.EDIT_CARD).setEnabled(enable);
		getAction(ActionEnum.FIND).setEnabled(enable);
		getAction(ActionEnum.FIND_REPLACE).setEnabled(enable);
		getAction(ActionEnum.HTML_CARD).setEnabled(enable);
		getAction(ActionEnum.HTML_TABLE).setEnabled(enable);
		getAction(ActionEnum.XML_EXPORT).setEnabled(enable);
	}

	/**
	 * Enable/disable actions depending on whether fields and cards are defined
	 * @param o may be either Cards or DBDef
	 * @param arg may be either CardUpdate or DBDefUpdate
	 */
	@Override public void update(Observable o, Object arg)
	{
		Object[] params = { o, arg };
		log.entering("ActionManager", "update", params);
		
		if (o instanceof DBDef)
		{
			log.finer("DBDef update");
			
			DBDefUpdate event = (DBDefUpdate)arg;
			log.finest("DBDefUpdate.event = "+event.getEvent());
				
			switch (event.getEvent())
			{
				case FIELD_ADD:
					toggleFieldActions(true);
					break;
					
				case FIELD_DELETE:
				case RESET:
					if (DBDef.getInstance().getNumberOfFields() == 0) {
						toggleFieldActions(false);
					}
					break;
					
				case CATEGORY_UPDATED:
					toggleActionEnabled(ActionEnum.CATEGORY_EXPORT, true);
					break;
					
				case CATEGORY_REMOVED:
					if (!DBDef.getInstance().areCategoriesDefined()) {
						toggleActionEnabled(ActionEnum.CATEGORY_EXPORT, false);
					}
					break;
			}
		}
		else if (o instanceof Cards)
		{
			log.finer("Cards update");
			
			Cards cards = (Cards)o;
			CardUpdate cu = (CardUpdate)arg;
			log.finest("CardUpdate.cardOperation = "+cu.cardOperation);
			
			switch (cu.cardOperation)
			{
				case LOAD:
					toggleCardActions((cards.size() > 0));
					break;
					
				case CARD_NEW:
					if (cards.size() > 0) {
						toggleCardActions(true);
						toggleActionEnabled(ActionEnum.UNDELETE, false);
					}
					break;
					
				case CARD_DELETE:
					if (cards.size() == 0) {
						toggleCardActions(false);
					}
					toggleActionEnabled(ActionEnum.UNDELETE, true);
					break;
			}
		}
		
		log.exiting("ActionManager", "update");
	}
}
