/*
 * JCards
 * Copyright (C) 2002-2009 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package nrs.jcards.gui.action;

import java.awt.event.ActionEvent;
import java.io.File;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.logging.Level;
import nrs.jcards.JCards;

/**
 * Show help in a web browser. Will display local file if available or
 * alternatively on the web.
 *
 * @author Copyright &#169; 2002-2009 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class HelpAction extends JCardsAction
{
	public static final String HELP_URL = "http://www.nsydenham.net/java/JCards/v6/help.html";
	private static final String HELP_FILE = "help.html";
	
	private static URL helpFile;
	
	/** Creates a new instance of HelpAction */
	public HelpAction(JCards jcards, ActionEnum action)
	{
		super(jcards, action);
	}

	/**
	 * Given a path to start searching from see if the help file can be found
	 * @param jcardsPath the path to search from that includes JCards
	 * @param index index of JCards
	 * @param fileName name of file to find
	 */
	final static String findHelpFile(String jcardsPath, int index, String fileName)
	{
		String path = null;
		
		String jcardsDir = jcardsPath.substring(0, index + 6);
		File f = new File(jcardsDir);
		if (f.isDirectory())
		{
			String fs = System.getProperty("file.separator");
			String file = jcardsDir + fs + "docs" + fs + "help" + fs + fileName;
			f = new File(file);
			path = (f.exists() ? file : null);
		}
		
		return path;
	}

	final static List<String> getPaths()
	{
		List<String> paths = new ArrayList<String>(3);
		String cp = System.getProperty("java.class.path");
		String[] cpPaths = cp.split(System.getProperty("path.separator"));
		paths.addAll(Arrays.asList(cpPaths));
		paths.add(System.getProperty("user.dir")); // working dir
		paths.add(System.getProperty("user.home")); // home

		return paths;
	}
	
	/**
	 * Utility method to find either the local help file or return a reference to the help URL
	 * @return URL of the help file
	 * @see #getHelpFile(java.lang.String)
	 */
	public static URL getHelpFile()
	{
		return getHelpFile(null);
	}
	
	/**
	 * Utility method to find either the local help file or return a reference to the help URL
	 * @param fragment optional location within the help file, e.g. prefs will be appended as #prefs
	 * @return URL of the help file
	 */
	public static URL getHelpFile(String fragment)
	{
		String path = null;
		
		// Fragments only work with the http scheme and are not allowed on the end of file
		// Seems perverse but I didn't define the standard!
		if (fragment == null)
		{
			// look through classpath and user dir for jcards directory
			// we can then guess help.html is in /docs/help/
			int index;
			for (String cp : getPaths())
			{
				index = cp.toLowerCase().indexOf("jcards");
				if (index > -1)
				{
					path = findHelpFile(cp, index, HELP_FILE);
					if (path != null)
					{
						break; // found it
					}
				}
			}
		}

		try
		{
			if (path == null) {
				helpFile = new URL(HELP_URL); // default to web location
				if (fragment != null) {
					helpFile = new URL(helpFile, "#"+fragment);
				}
			} else {
				helpFile = new File(path).toURI().toURL();
			}
		} catch (MalformedURLException mfex) {
			log.log(Level.SEVERE, "Cannot determine help URL", mfex);
		}

		return helpFile;
	}
	
	/**
	 * Invoked when an action occurs.
	 */
	@Override public void actionPerformed(ActionEvent e)
	{
		getHelpFile();
		jcards.getFrame().setStatus(JCards.getMessage("show.help"));
		jcards.getFrame().displayURL(helpFile);
	}
}
