/*
 * JCards
 * Copyright (C) 2002-2009 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package nrs.jcards.gui.action;

import java.util.logging.Logger;
import javax.swing.AbstractAction;
import javax.swing.AbstractButton;
import javax.swing.KeyStroke;
import nrs.jcards.JCards;
import nrs.jcards.gui.JCardsFrame;

/**
 * Abstract root class of all JCards actions. Handles mnemonics and other useful
 * features.
 *
 * @author Copyright &#169; 2002-2009 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public abstract class JCardsAction extends AbstractAction
{
	protected JCards jcards;
	private char mnemonic;
	private int mnIndex;
	protected final static Logger log = Logger.getLogger(JCardsAction.class.getName());
	/** Character representing mnemonic _ */
	public static final char MNEMONIC_CHAR = '_';

	/** Creates a new instance of JCardsAction */
	public JCardsAction(JCards jcardsRef, ActionEnum action)
	{
		jcards = jcardsRef;
		
		// set the name
		String mName = JCards.getMenuString(action.getResourceName());
		mnIndex = getMnemonicIndex(mName);
		if (mnIndex > -1 && mnIndex < mName.length())
		{
			mnemonic = getMnemonic(mName);
			putValue(NAME, getTextWithoutMnemonic(mName));
		}
		else
		{
			putValue(NAME, mName);
		}
		
		// see if an acclerator key has been defined
		String acc = JCards.getMenuString(action.getResourceName()+".acc");
		if (acc != null)
		{
			putValue(ACCELERATOR_KEY, KeyStroke.getKeyStroke(acc));
		}
		
		// see if an icon has been defined
		if (action.getIconName() != null)
		{
			putValue(SMALL_ICON, JCardsFrame.getImage(action.getIconName()));
		}
		
		// check for a tooltip
		String tt = JCards.getMenuString(action.getResourceName()+".tt");
		if (tt != null)
		{
			putValue(SHORT_DESCRIPTION, tt);
		}
	}
	
	public char getMnemonic()
	{
		return mnemonic;
	}
	
	public int getMnemonicIndex()
	{
		return mnIndex;
	}
	
	/**
	 * Utility method to get mnemonic from text
	 * @param text item text including mnemonic indicator
	 * @return char representing mnemonic or '0' if not found
	 * @see #MNEMONIC_CHAR
	 */
	public static char getMnemonic(String text)
	{
		assert text != null : "JCardsAction text is null";
		char mnemonic = '0';
		
		int _mnIndex = text.indexOf(MNEMONIC_CHAR);
		if (_mnIndex > -1 && _mnIndex < text.length())
		{
			mnemonic = text.charAt(_mnIndex+1);
		}
		
		return mnemonic;
	}
	
	/**
	 * Utility method to get index of mnemonic
	 * @param text item text including mnemonic indicator
	 * @return index of mnemonic or -1 if not found
	 * @see #MNEMONIC_CHAR
	 */
	public static int getMnemonicIndex(String text)
	{
		assert text != null : "JCardsAction text is null";
		return text.indexOf(MNEMONIC_CHAR);
	}
	
	/**
	 * Utility method to get menu text without any mnemonic indicator
	 * @param text item text including mnemonic indicator
	 * @return text without mnemonic indicator
	 * @see #MNEMONIC_CHAR
	 */
	public static String getTextWithoutMnemonic(String text)
	{
		assert text != null : "JCardsAction text is null";
		
		String menuText = null;
		
		int _mnIndex = text.indexOf(MNEMONIC_CHAR);
		if (_mnIndex > -1 && _mnIndex < text.length())
		{
			menuText = text.substring(0, _mnIndex) + text.substring(_mnIndex+1);
		}
		else
		{
			menuText = text;
		}
		
		return menuText;
	}

	/**
	 * Set button properties based on the text looked up from a resource bundle.
	 * If appropriate, text and mnemonic will be set
	 * @param text text that may include an underscore to indicate a mnemonic
	 * @param button button, menu item, etc
	 */
	public static void setButtonProperties(String text, AbstractButton button)
	{
		button.setText(JCardsAction.getTextWithoutMnemonic(text));
		int mnIndex = getMnemonicIndex(text);
		if (mnIndex > -1)
		{
			button.setMnemonic(JCardsAction.getMnemonic(text));
			button.setDisplayedMnemonicIndex(mnIndex);
		}
	}
}
