/*
 * JCards
 * Copyright (C) 2002-2009 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package nrs.jcards.gui.action;

import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;
import javax.swing.JCheckBox;
import javax.swing.JPanel;
import nrs.jcards.JCards;
import nrs.jcards.io.FileUtils;
import nrs.jcards.io.JCAReader;
import nrs.jcards.io.JCAReader.SplitFilename;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;

/**
 * Export to XML
 *
 * @author Copyright &#169; 2002-2009 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class XMLExportAction extends JCardsAction
{
	private File saveFile;
	static final String OPEN_IN_BROWSER = "OpenInBrowser";
	static final String ONLY_VISIBLE_FIELDS = "OnlyVisFields";
	static final String INCLUDE_EMPTY_FIELDS = "IncludeEmptyFields";
	static final String INCLUDE_COMMENTS = "IncludeComments";

	private static final String EXT = ".xml";
	
	/** Creates a new instance of XMLExportAction */
	public XMLExportAction(JCards jcards, ActionEnum action)
	{
		super(jcards, action);
		
		setEnabled(false);
	}
	
	/**
	 * Invoked when an action occurs.
	 */
	@Override public void actionPerformed(ActionEvent e)
	{
		JPanel accPanel = new JPanel();
		Map<String, JCheckBox> accComps = getAccessoriesPanel(accPanel, true);
		
		if (saveFile == null && jcards.getCardFile() != null)
		{
			SplitFilename sf = JCAReader.splitFilename(jcards.getCardFile().getAbsolutePath());
			saveFile = new File(sf.name+EXT);
		}
		saveFile = FileUtils.getFile(getTextWithoutMnemonic(JCards.getMenuString("xml.export")),
				jcards.getFrame(), saveFile, accPanel);
		if (saveFile == null)
		{
			return;
		}
		
		try
		{
			BufferedOutputStream bos = new BufferedOutputStream(new FileOutputStream(saveFile));
			XMLOutputter xmlOut = new XMLOutputter(Format.getPrettyFormat());
			// output prolog
			bos.write(JCAReader.PROLOG.getBytes(JCAReader.UTF8));
			boolean includeEmpty = accComps.get(INCLUDE_EMPTY_FIELDS).isSelected();
			boolean onlyVisFields = accComps.get(ONLY_VISIBLE_FIELDS).isSelected();
			boolean comments = accComps.get(INCLUDE_COMMENTS).isSelected();
			jcards.getCards().writeCardsToOutputStream(xmlOut, bos, includeEmpty, onlyVisFields, comments);
			bos.close();
			
			jcards.getFrame().setStatus(JCards.getFormattedMessage("exported.to", saveFile.getAbsolutePath()));
			
			if (accComps.get(OPEN_IN_BROWSER).isSelected())
			{
				jcards.getFrame().displayURL(saveFile.toURI().toURL());
			}
		}
		catch (IOException ioex)
		{
			jcards.getFrame().showError(JCards.getMessage("error"),
				JCards.getMessage("cannot.export.xml"), "XMLExportAction", ioex, Level.WARNING);
		}
	}
	
	/**
	 * Utility method to get an accessories panel that is displayed in the export dialogs for X(HT)ML
	 * @param accPanel the JPanel to be added to
	 * @param includeEmpty if true include a checkbox for exporting empty fields as well
	 * @return map of named JCheckBox components
	 */
	static Map<String, JCheckBox> getAccessoriesPanel(JPanel accPanel, boolean includeEmpty)
	{
		Map<String, JCheckBox> accComps = new HashMap<String, JCheckBox>(3);
		
		accPanel.setLayout(new GridBagLayout());
		GridBagConstraints gbc = new GridBagConstraints();
		gbc.insets = new Insets(10, 5, 10, 5);
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		gbc.anchor = GridBagConstraints.WEST;
		
		JCheckBox onlyVisFieldsCB = new JCheckBox(JCards.getLabelString("visible.fields.only"));
		accPanel.add(onlyVisFieldsCB, gbc);
		accComps.put(ONLY_VISIBLE_FIELDS, onlyVisFieldsCB);
		
		if (includeEmpty)
		{
			JCheckBox includeEmptyCB = new JCheckBox(JCards.getLabelString("include.empty.fields"));
			accPanel.add(includeEmptyCB, gbc);
			accComps.put(INCLUDE_EMPTY_FIELDS, includeEmptyCB);
		}

		JCheckBox includeCommentsCB = new JCheckBox(JCards.getLabelString("include.comments"));
		accPanel.add(includeCommentsCB, gbc);
		accComps.put(INCLUDE_COMMENTS, includeCommentsCB);

		JCheckBox openInBrowserCB = new JCheckBox(JCards.getLabelString("open.in.browser"));
		openInBrowserCB.setToolTipText(JCards.getLabelString("open.in.browser.tt"));
		accPanel.add(openInBrowserCB, gbc);
		accComps.put(OPEN_IN_BROWSER, openInBrowserCB);
		
		return accComps;
	}
}
