/*
 * JCards
 * Copyright (C) 2002-2009 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package nrs.jcards.gui.dialog;

import java.awt.BorderLayout;
import java.awt.FlowLayout;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.WindowConstants;
import nrs.jcards.Cards;
import nrs.jcards.JCards;
import nrs.jcards.UserSettings;
import nrs.jcards.gui.JCardsFrame;
import nrs.jcards.gui.JCardsLabel;
import nrs.jcards.gui.action.ActionEnum;
import nrs.jcards.gui.action.ContextHelpAction;
import nrs.jcards.io.CSVException;
import nrs.jcards.io.JCAReader;
import nrs.jcards.io.filter.CSVFileFilter;
import nrs.jcards.io.CSVIO;
import nrs.jcards.io.FileUtils;
import nrs.jcards.io.JCAReader.SplitFilename;

/**
 * Dialog for importing and export CSV files
 *
 * @author Copyright &#169; 2002-2009 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class CSVDialog extends BaseDialog implements ActionListener
{
	private final static Logger log = Logger.getLogger(CSVDialog.class.getName());
	
	/** Used in the constructor to specify the dialog type */
	public enum TYPE { IMPORT, EXPORT };
	
	/** Field delimiters */
	final static String[] FIELD_DELIM = { ",", "^", "$", "!", "#", ";", ":", "?", "/", "*", "|", "<space>", "<tab>"};
	/** Text delimiters */
	private final static String[] TEXT_DELIM = { "\"", "^", "$", ":", "?", "/", "*", "|", "<none>" };
	
	/** Character sets */
	final static String[] CHAR_SETS = { "UTF8", "ASCII", "ISO8859_1",
	"ISO8859_2", "ISO8859_3", "ISO8859_4", "ISO8859_5", "ISO8859_6",
	"ISO8859_7", "ISO8859_8", "ISO8859_9", "ISO8859_15", "KOI8_R",
	"JIS0201", "JIS0208", "JIS0212", "SJIS", "ISO2022KR" };
	
	private JCardsFrame jcardsFrame;
	private Cards cards;
	private TYPE type = TYPE.IMPORT;
	private JCheckBox headerRow, mlValues, concatNL;
	private JComboBox fieldDelimiter, textDelimiter, charSet;
	private JTextField csvFile;
	private File saveFile;
	
	/**
	 * Creates new form ImportDialog
	 * @param parent frame
	 * @param type import or export
	 */
	public CSVDialog(JCardsFrame frame, Cards cards, File cardFile, TYPE type)
	{
		super(frame, true);
		
		this.jcardsFrame = frame;
		this.cards = cards;
		this.type = type;
		
		int i;
		
		if (type == TYPE.IMPORT) {
			setTitle(JCards.getLabelString("csv.import"));
		} else {
			setTitle(JCards.getLabelString("csv.export"));
		}
		
		setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
		
		JPanel selectionPanel = new JPanel(new GridBagLayout());
		GridBagConstraints gbc = new GridBagConstraints();
		gbc.gridwidth = 1;
		gbc.anchor = GridBagConstraints.WEST;
		gbc.insets = new Insets(5, 5, 5, 5);
		
		JLabel label = new JCardsLabel("field.delimiter");
		selectionPanel.add(label, gbc);
		
		fieldDelimiter = new JComboBox();
		for (String delim : FIELD_DELIM)
		{
			fieldDelimiter.addItem(delim);
		}
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		label.setLabelFor(fieldDelimiter);
		selectionPanel.add(fieldDelimiter, gbc);
		
		label = new JCardsLabel("text.delimiter");
		gbc.gridwidth = 1;
		selectionPanel.add(label, gbc);
		
		textDelimiter = new JComboBox();
		for (String delim : TEXT_DELIM)
		{
			textDelimiter.addItem(delim);
		}
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		label.setLabelFor(textDelimiter);
		selectionPanel.add(textDelimiter, gbc);
		
		label = new JCardsLabel("file");
		gbc.gridwidth = 1;
		selectionPanel.add(label, gbc);
		
		csvFile = new JTextField(20);
		gbc.fill = GridBagConstraints.HORIZONTAL;
		gbc.weightx = 1.0;
		label.setLabelFor(csvFile);
		selectionPanel.add(csvFile, gbc);
		
		JButton browseButton = new JButton(JCards.getButtonString("browse"));
		browseButton.setActionCommand("browse");
		browseButton.addActionListener(this);
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		gbc.weightx = 0.0;
		gbc.fill = GridBagConstraints.NONE;
		selectionPanel.add(browseButton, gbc);
		
		label = new JCardsLabel("charset");
		gbc.gridwidth = 1;
		selectionPanel.add(label, gbc);
		
		charSet = new JComboBox();
		for (String cs : CHAR_SETS)
		{
			charSet.addItem(cs);
		}
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		label.setLabelFor(charSet);
		selectionPanel.add(charSet, gbc);
		
		gbc.gridwidth = 1;
		label = new JCardsLabel("header.row");
		label.setToolTipText(JCards.getLabelString("header.row.tt"));
		selectionPanel.add(label, gbc);
		
		headerRow = new JCheckBox();
		headerRow.setToolTipText(JCards.getLabelString("header.row.tt"));
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		label.setLabelFor(headerRow);
		selectionPanel.add(headerRow, gbc);
		
		if (type == TYPE.IMPORT)
		{
			gbc.gridwidth = 1;
			label = new JCardsLabel("multi.line.values");
			label.setToolTipText(JCards.getLabelString("multi.line.values.tt"));
			selectionPanel.add(label, gbc);
			
			mlValues = new JCheckBox();
			mlValues.setToolTipText(JCards.getLabelString("multi.line.values.tt"));
			gbc.gridwidth = GridBagConstraints.REMAINDER;
			label.setLabelFor(mlValues);
			selectionPanel.add(mlValues, gbc);
		}
		else if (type == TYPE.EXPORT)
		{
			label = new JCardsLabel("remove.newlines");
			gbc.gridwidth = 1;
			selectionPanel.add(label, gbc);
			
			concatNL = new JCheckBox();
			gbc.gridwidth = GridBagConstraints.REMAINDER;
			label.setLabelFor(concatNL);
			selectionPanel.add(concatNL, gbc);
			
			if (cardFile != null)
			{
				String fileName = cardFile.getAbsolutePath();
				SplitFilename sf = JCAReader.splitFilename(fileName);
				csvFile.setText(sf.name + ".csv");
			}
		}
		getContentPane().add(selectionPanel, BorderLayout.CENTER);
		
		// buttons
		JPanel toolbar = new JPanel(new FlowLayout(FlowLayout.RIGHT));
		
		ContextHelpAction help = (ContextHelpAction)jcardsFrame.getAction(ActionEnum.CONTEXT_HELP);
		help.setFragment("import.export");
		JButton helpButton = new JButton(help);
		toolbar.add(helpButton);
		
		JButton cancelButton = new JButton(JCards.getButtonString("cancel"), JCardsFrame.getImage(UserSettings.ICON_CANCEL));
		cancelButton.setActionCommand("cancel");
		cancelButton.addActionListener(this);
		toolbar.add(cancelButton);
		
		JButton okButton = new JButton(JCards.getButtonString("ok"), JCardsFrame.getImage(UserSettings.ICON_OK));
		okButton.setActionCommand("ok");
		okButton.addActionListener(this);
		toolbar.add(okButton);
				
		getContentPane().add(toolbar, BorderLayout.SOUTH);
	}
	
	@Override
	public void pack()
	{
		super.pack();
		csvFile.setCaretPosition(csvFile.getText().length());
		csvFile.requestFocus();
	}
	
	/** Invoked when an action occurs */
	@Override public void actionPerformed(ActionEvent e)
	{
		String ac = e.getActionCommand();
		
		if (ac.equals("browse")) {
			selectFile();
		} else if (ac.equals("cancel")) {
			dispose();
		} else if (ac.equals("ok")) {
			processCSV();
		}
	}
	
	/** Select a file */
	private void selectFile()
	{
		JFileChooser fDialog = new JFileChooser();
		fDialog.setDialogTitle(JCards.getMessage("csv.select"));
		if (saveFile == null)
		{
			saveFile = new File(FileUtils.getHomeDir(), "jcards.csv");
		}
		fDialog.setSelectedFile(saveFile);
		fDialog.addChoosableFileFilter(new CSVFileFilter());
		String btnText = (type == TYPE.IMPORT ? JCards.getButtonString("open") : JCards.getButtonString("save"));
		if (fDialog.showDialog(jcardsFrame, btnText) != JFileChooser.APPROVE_OPTION)
		{
			return;
		}
		saveFile = fDialog.getSelectedFile();
		if (saveFile != null)
		{
			csvFile.setText(saveFile.getAbsolutePath());
			csvFile.setCaretPosition(csvFile.getText().length()-1);
		}
	}
	
	/** Validate the user's actions */
	private void processCSV()
	{
		String td = (String)textDelimiter.getSelectedItem();
		String fd = (String)fieldDelimiter.getSelectedItem();
		
		if (td.equals(fd))
		{
			JOptionPane.showMessageDialog(jcardsFrame, JCards.getMessage("same.field.text.delim"), JCards.getMessage("error"), JOptionPane.INFORMATION_MESSAGE);
			return;
		}
		
		log.fine("Field delimiter="+fd);
		log.fine("Character delimiter="+td);
		
		String file = csvFile.getText();
		if (file.length() > 0)
		{
			File csv = new File(file);
			
			boolean hasHeader = headerRow.isSelected();
			String cs = (String)charSet.getSelectedItem();
			if (fd.equals("<space>"))
			{
				fd = " ";
			}
			else if (fd.equals("<tab>"))
			{
				fd = "\t";
			}
			
			if (td.equals("<none>"))
			{
				td = "";
			}
			
			if (type == TYPE.EXPORT)
			{
				exportCSV(csv, fd, td, hasHeader, cs);
			}
			else
			{
				importCSV(csv, fd, td, hasHeader, cs);
			}
		}
		else
		{
			JOptionPane.showMessageDialog(jcardsFrame, JCards.getMessage("no.file"), JCards.getMessage("error"), JOptionPane.ERROR_MESSAGE);
		}
	}
	
	/**
	 * Import a CSV file
	 * @param file the file to import
	 * @param fDelim file delimiter
	 * @param tDelim text delimiter
	 * @param hasHeader CSV file has field names as first row
	 * @param charset character set
	 */
	private void importCSV(File file, String fDelim, String tDelim, boolean hasHeader, String charset)
	{
		log.fine("Importing CSV");
		
		// check file exists
		if (!file.exists())
		{
			JOptionPane.showMessageDialog(jcardsFrame, JCards.getMessage("no.file"), JCards.getMessage("error"), JOptionPane.ERROR_MESSAGE);
			return;
		}
		
		CSVIO io = new CSVIO(cards, this);
		try
		{
			io.importCSV(file, fDelim, tDelim, charset, hasHeader, mlValues.isSelected());
			log.fine("Finished importing CSV");
		}
		catch (CSVException csex)
		{
			log.log(Level.WARNING, csex.getMessage(), csex.getCause());
			jcardsFrame.showError(JCards.getMessage("error"), JCards.getMessage("cannot.import.csv"),
				"CSVDialog.importCSV", csex, Level.WARNING);
		}
		
		jcardsFrame.updateViewMenu();
		
		// close the dialog
		dispose();
	}
	
	/**
	 * Export data to a CSV file
	 * @param file the file to export to
	 * @param fDelim file delimiter
	 * @param tDelim text delimiter
	 * @param hasHeader CSV file has field names as first row
	 * @param charset character set
	 */
	private void exportCSV(File file, String fDelim, String tDelim, boolean hasHeader, String charset)
	{
		log.fine("Exporting CSV");
		
		if (file.exists())
		{
			int ret = JOptionPane.showConfirmDialog(this, JCards.getMessage("file.exists"), JCards.getMessage("confirm"), JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE);
			if (ret == JOptionPane.NO_OPTION) {
				return;
			}
		}
		
		boolean concat = concatNL.isSelected();
		
		CSVIO io = new CSVIO(cards, this);
		try
		{
			io.exportCSV(file, fDelim, tDelim, charset, concat, hasHeader);
			
			jcardsFrame.setStatus(JCards.getFormattedMessage("exported.to", file.getAbsolutePath()));
			
			log.fine("Finished exporting CSV");
		}
		catch (CSVException csex)
		{
			log.log(Level.WARNING, csex.getMessage(), csex.getCause());
			jcardsFrame.showError(JCards.getMessage("error"), JCards.getMessage("cannot.export.csv"),
				"CSVDialog.exportCSV", csex, Level.WARNING);
		}
		
		dispose();
	}
	
	/**
	 * Show an error caused by parsing a CSV file
	 * @param message the message to display
	 */
	public void showError(String message)
	{
		JOptionPane.showMessageDialog(this, message, JCards.getMessage("error"), JOptionPane.ERROR_MESSAGE);
	}
}
