/*
 * JCards
 * Copyright (C) 2002-2009 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package nrs.jcards.gui.dialog;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.awt.print.PageFormat;
import java.awt.print.Printable;
import java.awt.print.PrinterException;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.print.DocFlavor;
import javax.print.DocPrintJob;
import javax.print.PrintException;
import javax.print.PrintService;
import javax.print.PrintServiceLookup;
import javax.print.ServiceUI;
import javax.print.SimpleDoc;
import javax.print.attribute.HashPrintRequestAttributeSet;
import javax.print.attribute.PrintRequestAttributeSet;
import javax.print.attribute.standard.JobName;
import javax.print.attribute.standard.OrientationRequested;
import javax.print.attribute.standard.RequestingUserName;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.RepaintManager;
import javax.swing.UIManager;
import nrs.jcards.Card;
import nrs.jcards.DBDef;
import nrs.jcards.fields.FieldDef;
import nrs.jcards.JCards;
import nrs.jcards.fields.InputCheck;
import nrs.jcards.gui.DisplayCardPanel;
import nrs.jcards.UserSettings;
import nrs.jcards.gui.JCardsButton;
import nrs.jcards.gui.JCardsFrame;

/**
 * Card Dialog allows the entry of new cards and the editing of existing cards
 * 
 * @author Copyright &#169; 2002-2009 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class CardDialog extends BaseDialog implements ActionListener, Printable, WindowListener
{
	private final static Logger log = Logger.getLogger(CardDialog.class.getName());
	private JCardsFrame jcardsFrame;
	private int cardIndex = -1;
	private boolean editCard = false;
	private boolean copyCard = false;
	private DisplayCardPanel dcp = null;
	private Dimension preferredSize;

	// action commands
	private static final String OK = "ok";
	private static final String CANCEL = "cancel";
	private static final String EDIT = "edit";
	private static final String NEXT = "next";
	private static final String PREVIOUS = "previous";
	private static final String PRINT = "print";
	private static final String RESET = "reset";
	private static final String UPDATE = "update";

	/**
     * Creates a new instance of CardDialog
     * @param jcardsFrame the parent frame
     * @param card Card that is being edited, may be null for a new one
     * @param copy set to true if copying a card
     * @param editing set to true to edit card
     * @param cardSelection if editing this is set to the card currently
     *			    selected, otherwise -1 for a new card
     */
	public CardDialog(JCardsFrame jcardsFrame, Card card, boolean copy, boolean editing, int cardSelection)
	{
		super(jcardsFrame, true);

		setTitle(JCards.getLabelString("edit.card"));

		this.jcardsFrame = jcardsFrame;
		this.copyCard = copy;
		this.editCard = editing;
		cardIndex = cardSelection;

		getContentPane().setLayout(new BorderLayout());
		
		boolean newCard = (card == null ? true : false);
		dcp = new DisplayCardPanel(this, newCard, editCard);
		dcp.display(card, jcardsFrame, this, copyCard);
		JScrollPane scrollPane = new JScrollPane(dcp);
		getContentPane().add(scrollPane, BorderLayout.CENTER);
		log.fine("Added DisplayCardPanel");

		// add toolbar
		JPanel toolbar = new JPanel(new FlowLayout(FlowLayout.RIGHT));

		JButton print = new JCardsButton(PRINT, JCardsFrame.getImage(UserSettings.ICON_PRINT));
		print.setText(null);
		print.addActionListener(this);
		toolbar.add(print);

		JButton cancel = new JCardsButton(CANCEL, JCardsFrame.getImage(UserSettings.ICON_CANCEL));
		cancel.setText(null);
		cancel.addActionListener(this);
		cancel.setMnemonic(KeyEvent.VK_ESCAPE);

		if (!editing)
		{
			JButton previous = new JCardsButton(PREVIOUS, JCardsFrame.getImage(UserSettings.ICON_PREVIOUS));
			previous.setText(null);
			previous.addActionListener(this);
			previous.setMnemonic(KeyEvent.VK_LEFT);
			toolbar.add(previous);

			JButton next = new JCardsButton(NEXT, JCardsFrame.getImage(UserSettings.ICON_NEXT));
			next.setText(null);
			next.addActionListener(this);
			next.setMnemonic(KeyEvent.VK_RIGHT);
			toolbar.add(next);

			toolbar.add(cancel);

			JButton edit = new JCardsButton(EDIT, JCardsFrame.getImage(UserSettings.ICON_EDIT));
			edit.setText(null);
			edit.addActionListener(this);
			toolbar.add(edit);
			getRootPane().setDefaultButton(edit);
		}
		else
		{
			// only need an update button if calculated fields are defined
			if (DBDef.getInstance().areCalculationsDefined())
			{
				JButton update = new JCardsButton(UPDATE, JCardsFrame.getImage(UserSettings.ICON_CALCULATION));
				update.setText(null);
				update.addActionListener(this);
				toolbar.add(update);
			}

			if (!newCard && !copyCard)
			{
				JButton reset = new JCardsButton(RESET, JCardsFrame.getImage(UserSettings.ICON_RESET));
				reset.setText(null);
				reset.addActionListener(this);
				toolbar.add(reset);
			}
			
			toolbar.add(cancel);

			JButton ok = new JCardsButton(OK, JCardsFrame.getImage(UserSettings.ICON_OK));
			ok.setText(null);
			ok.setMnemonic(KeyEvent.VK_ENTER);
			ok.addActionListener(this);
			toolbar.add(ok);
			getRootPane().setDefaultButton(ok);
		}
		getContentPane().add(toolbar, BorderLayout.SOUTH);
		
		addWindowListener(this);
		
		log.finest("Leaving CardDialog constructor");
	}

	/** Invoked when one of the buttons is clicked */
	@Override public void actionPerformed(ActionEvent e)
	{
		String cmd = e.getActionCommand();

		// CardDialog command
		if (cmd.equals(OK))
		{
			if (DBDef.getInstance().areCalculationsDefined())
			{
				dcp.updateCalculations();
			}

			if (!updateValues())
			{
				return;
			}
			dispose();
		}
		else if (cmd.equals(EDIT))
		{
			dispose();
			jcardsFrame.editCard(false, true, false);
		}
		else if (cmd.equals(NEXT))
		{
			Card card = jcardsFrame.editOtherCard(1);
			dcp.display(card, jcardsFrame, this, false);
		}
		else if (cmd.equals(PREVIOUS))
		{
			Card card = jcardsFrame.editOtherCard(-1);
			dcp.display(card, jcardsFrame, this, false);
		}
		else if (cmd.equals(UPDATE))
		{
			dcp.updateCalculations();
		}
		else if (cmd.equals(CANCEL))
		{
			if (cardIndex == -1 || copyCard)
			{
				dcp.rollbackCounters();
			}
			dispose();
		}
		else if (cmd.equals(PRINT))
		{
			print();
		}
		else if (cmd.equals(RESET))
		{
			dcp.setValues(true);
		}
	}

	/**
     * Update the values
     * @return true if updated successfully
     */
	private boolean updateValues()
	{
		Map<String, JComponent> entries = dcp.getEntries();
		Map<String, Object> values = new HashMap<String, Object>(entries.size());
		JComponent comp;
		FieldDef fd;
		DBDef def = DBDef.getInstance();

		for (String fieldName : entries.keySet())
		{
			fd = def.getFieldDef(fieldName);
			comp = entries.get(fieldName);
			InputCheck ic = fd.checkInputValue(comp);
			if (ic.isValid())
			{
				values.put(fd.getName(), ic.getValue());
			}
			else
			{
				JOptionPane.showMessageDialog(this, ic.getMessage(), JCards.getMessage("incorrect.value"), JOptionPane.ERROR_MESSAGE);
				return false;
			}
		}

		if (copyCard || cardIndex < 0)
		{
			jcardsFrame.addCard(new Card(values));
		}
		else
		{
			jcardsFrame.replaceCard(cardIndex, new Card(values));
		}

		return true;
	}

	/** Print the card */
	private void print()
	{
		PrintService[] services = PrintServiceLookup.lookupPrintServices(null, null);
		PrintService svc = PrintServiceLookup.lookupDefaultPrintService();
		PrintRequestAttributeSet attrs = new HashPrintRequestAttributeSet();
		attrs.add(new RequestingUserName(System.getProperty("user.name"), null));
		attrs.add(new JobName("JCards Card Print", null));
		attrs.add(OrientationRequested.LANDSCAPE);
		try
		{
			PrintService selection = ServiceUI.printDialog(null, 100, 100, services, svc, null, attrs);
			if (selection != null)
			{
				try
				{
					DocPrintJob dpj = svc.createPrintJob();
					SimpleDoc doc = new SimpleDoc(this, DocFlavor.SERVICE_FORMATTED.PRINTABLE, null);
					dpj.print(doc, attrs);
					jcardsFrame.setStatus(JCards.getMessage("printed.card"));
				}
				catch (PrintException pex)
				{
					log.log(Level.WARNING, "Print exception", pex);
					jcardsFrame.showError(JCards.getMessage("error"), JCards.getMessage("cannot.print"),
						"CardDialog.print", pex, Level.WARNING);
				}
			}
		}
		catch (NullPointerException npex)
		{
			log.log(Level.WARNING, "JDK printing bug encountered", npex);
			JOptionPane.showMessageDialog(this, JCards.getMessage("jdk.print.bug"), JCards.getMessage("warning"), JOptionPane.WARNING_MESSAGE);
		}
	}

	/**
     * Get the preferred size of the CardDialog
     * @return Dimension
     */
	@Override public Dimension getPreferredSize()
	{
		return (preferredSize == null ? new Dimension(500, 400) : preferredSize);
	}
	
	/**
	 * Set the preferred size of the CardDialog
	 * @param dim
	 */
	@Override public void setPreferredSize(Dimension dim)
	{
		preferredSize = dim;
	}
	
	/**
	 * Set the background of the components to enable printing
	 * @param array of Components
	 */
	private void setPrintBackground(Component[] comps, Color col)
	{
		for (Component c : comps)
		{
			c.setBackground(col);
			if (c instanceof Container)
			{
				setPrintBackground(((Container)c).getComponents(), col);
			}
		}
	}
	
	/** Callback method for printing the card */
	@Override public int print(Graphics graphics, PageFormat pageFormat, int pageIndex) throws PrinterException
	{
		int printStatus;
		
		if (pageIndex == 0)
		{
			Graphics2D g2d = (Graphics2D)graphics;
			g2d.translate(pageFormat.getImageableX(), pageFormat.getImageableY());
			RepaintManager currentManager = RepaintManager.currentManager(dcp);
			setPrintBackground(dcp.getComponents(), Color.WHITE);
			currentManager.setDoubleBufferingEnabled(false);
			dcp.paint(g2d);
			currentManager.setDoubleBufferingEnabled(true);
			Color backCol = Color.PINK;
			backCol = UIManager.getColor("window");
			setPrintBackground(dcp.getComponents(), backCol);
			dcp.update(graphics);
			printStatus = Printable.PAGE_EXISTS;
		}
		else
		{
			printStatus = Printable.NO_SUCH_PAGE;
		}
		
		return printStatus;
	}
	
	@Override public void windowActivated(WindowEvent e) {}
	@Override public void windowClosed(WindowEvent e) {}

	@Override public void windowClosing(WindowEvent e)
	{
		jcardsFrame.setCardDialogSize(getSize());
		jcardsFrame.setCardDialogPosition(getLocation());
	}
	
	@Override public void windowDeactivated(WindowEvent e) {}
	@Override public void windowDeiconified(WindowEvent e) {}
	@Override public void windowIconified(WindowEvent e) {}
	@Override public void windowOpened(WindowEvent e) {}
}
