/*
 * JCards
 * Copyright (C) 2002-2009 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package nrs.jcards.gui.dialog;

import java.awt.BorderLayout;
import java.awt.FlowLayout;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.*;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.WindowConstants;
import nrs.jcards.DBDef;
import nrs.jcards.JCards;
import nrs.jcards.UserSettings;
import nrs.jcards.gui.JCardsFrame;
import nrs.jcards.gui.JCardsLabel;
import nrs.jcards.gui.action.ActionEnum;
import nrs.jcards.gui.action.ContextHelpAction;

/**
 * Dialog used for exporting categories to a file
 *
 * @author Copyright &#169; 2002-2009 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class CatExportDialog extends BaseDialog implements ActionListener
{
	private JCardsFrame jcardsFrame;
	private JTextField fileName;
	private JComboBox charSet, fieldDelimiter;

	/** Creates a new instance of CatExportDialog */
	public CatExportDialog(JCardsFrame jcardsFrame)
	{
		super(jcardsFrame, true);

		this.jcardsFrame = jcardsFrame;

		setTitle(JCards.getLabelString("cat.export"));

		setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);

		JPanel inputPanel = new JPanel(new GridBagLayout());
		GridBagConstraints gbc = new GridBagConstraints();
		gbc.anchor = GridBagConstraints.WEST;
		gbc.insets = new Insets(5, 5, 5, 5);

		JLabel label = new JCardsLabel("field.delimiter");
        inputPanel.add(label, gbc);

		fieldDelimiter = new JComboBox();
		for (String delim : CSVDialog.FIELD_DELIM) {
			fieldDelimiter.addItem(delim);
		}
		label.setLabelFor(fieldDelimiter);
        gbc.gridwidth = GridBagConstraints.REMAINDER;
        inputPanel.add(fieldDelimiter, gbc);

		label = new JCardsLabel("file");
		gbc.gridwidth = 1;
		inputPanel.add(label, gbc);

		fileName = new JTextField(20);
		label.setLabelFor(fileName);
		gbc.fill = GridBagConstraints.HORIZONTAL;
		gbc.weightx = 1.0;
		inputPanel.add(fileName, gbc);

		JButton select = new JButton(JCards.getButtonString("browse"));
		select.setActionCommand("browse");
		select.addActionListener(this);
		gbc.fill = GridBagConstraints.NONE;
		gbc.weightx = 0.0;
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		inputPanel.add(select, gbc);

		label = new JCardsLabel("charset");
		gbc.gridwidth = 1;
        inputPanel.add(label, gbc);

		charSet = new JComboBox();
		for (int i=0; i<CSVDialog.CHAR_SETS.length; i++) {
			charSet.addItem(CSVDialog.CHAR_SETS[i]);
		}
		label.setLabelFor(charSet);
        gbc.gridwidth = GridBagConstraints.REMAINDER;
        inputPanel.add(charSet, gbc);

		JPanel toolbar = new JPanel(new FlowLayout(FlowLayout.RIGHT, 5, 5));

		ContextHelpAction help = (ContextHelpAction)jcardsFrame.getAction(ActionEnum.CONTEXT_HELP);
		help.setFragment("import.export");
		JButton helpButton = new JButton(help);
		toolbar.add(helpButton);
		
		JButton cancelButton = new JButton(JCards.getButtonString("cancel"), JCardsFrame.getImage(UserSettings.ICON_CANCEL));
		cancelButton.setActionCommand("cancel");
		cancelButton.addActionListener(this);
		toolbar.add(cancelButton);

		JButton okButton = new JButton(JCards.getButtonString("ok"), JCardsFrame.getImage(UserSettings.ICON_OK));
		okButton.setActionCommand("ok");
		okButton.addActionListener(this);
		toolbar.add(okButton);

		getContentPane().setLayout(new BorderLayout(5, 10));
		getContentPane().add(inputPanel, BorderLayout.CENTER);
		getContentPane().add(toolbar, BorderLayout.SOUTH);
	}

	/** User clicked a button */
	@Override public void actionPerformed(ActionEvent e)
	{
		String ac = e.getActionCommand();
		if (ac.equals("cancel")) {
			dispose();
		} else if (ac.equals("browse")) {
			selectFile();
		} else if (ac.equals("ok")) {
			exportCategories();
		}
	}

	/** Export the categories to the selected file */
	private void exportCategories()
	{
		String file = fileName.getText();
		String cs = (String)charSet.getSelectedItem();
		String fd = (String)fieldDelimiter.getSelectedItem();
		if (file.length() > 0)
		{
			File f = new File(file);
			try
			{
				Map<String, List<String>> cats = DBDef.getInstance().getCategories();
				PrintWriter print = new PrintWriter(new BufferedWriter(
					new OutputStreamWriter(new FileOutputStream(f), cs)));
				StringBuilder sb;
				for (String catName : cats.keySet())
				{
					sb = new StringBuilder(catName);
					sb.append("=");
					for (String item : cats.get(catName))
					{
						sb.append(item);
						sb.append(fd);
					}
					sb.setLength(sb.length()-1);
					print.println(sb.toString());
				}
				print.close();

				JOptionPane.showMessageDialog(this, JCards.getMessage("exported"), JCards.getMessage("exported"), JOptionPane.INFORMATION_MESSAGE);
				dispose();
			}
			catch (IOException ioex)
			{
				jcardsFrame.showError(JCards.getMessage("error"), JCards.getMessage("cannot.export.cats"),
					"CatExportDialog.exportCategories", ioex, Level.WARNING);
			}
		}
		else
		{
			JOptionPane.showMessageDialog(this, JCards.getMessage("no.file"), JCards.getMessage("error"), JOptionPane.ERROR_MESSAGE);
		}
	}

	/** Select a file to save to */
	private void selectFile()
	{
		JFileChooser fDialog = new JFileChooser();
		fDialog.setDialogTitle(JCards.getMessage("file.select"));
		fDialog.setDialogType(JFileChooser.SAVE_DIALOG);
		fDialog.setVisible(true);
		if (fDialog.showSaveDialog(jcardsFrame) == JFileChooser.APPROVE_OPTION)
		{
			File file = fDialog.getSelectedFile();
			if (file != null)
			{
				fileName.setText(file.getAbsolutePath());
				fileName.setCaretPosition(fileName.getText().length()-1);
			}
		}
	}
}
