/*
 * JCards
 * Copyright (C) 2002-2009 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package nrs.jcards.gui.dialog;

import java.awt.BorderLayout;
import java.awt.FlowLayout;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.io.File;
import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JPasswordField;
import javax.swing.WindowConstants;
import nrs.jcards.JCards;
import nrs.jcards.gui.JCardsLabel;
import nrs.jcards.gui.action.ActionEnum;
import nrs.jcards.gui.action.CancelAction;
import nrs.jcards.gui.action.ContextHelpAction;
import nrs.jcards.gui.action.OKActionI;
import nrs.jcards.gui.action.OKAction;
import nrs.jcards.io.JCAException;
import nrs.jcards.io.JCAReader;

/**
 * Password dialog
 *
 * @author Copyright &#169; 2002-2009 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class PasswordDialog extends BaseDialog implements OKActionI
{
	private JCards jcards;
	private JPasswordField passTF1, passTF2;
	private boolean verify = true;
	private File jcaFile;
	
	/**
	 * Creates a new instance of PasswordDialog
	 * @param jcards app class
	 * @param openFile file being opened
	 * @param currentPassword the current password or null if not set
	 * @param verify if true use two lines for verifying the password, otherwise just enter it once
	 */
	public PasswordDialog(JCards jcards, File openFile, String currentPassword, boolean verify)
	{
		super(jcards.getFrame(), true);
		
		this.jcards = jcards;
		this.verify = verify;
		this.jcaFile = openFile;
		
		setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
		
		if (openFile != null)
		{
			setTitle(openFile.getName());
		}
		else
		{
			setTitle(JCards.getLabelString("enter.password"));
		}
		
		JPanel panel = new JPanel(new BorderLayout());
		panel.setBorder(BorderFactory.createEmptyBorder(2, 5, 2, 5));
		JPanel inputPanel = new JPanel();
		JPanel buttonPanel = new JPanel(new FlowLayout(FlowLayout.RIGHT, 10, 5));
		
		if (verify) {
			inputPanel.setLayout(new GridLayout(2, 2, 4, 10));
		} else {
			inputPanel.setLayout(new GridLayout(1, 2, 4, 10));
		}
		
		JLabel label = new JCardsLabel("enter.password", true);
		inputPanel.add(label);
		passTF1 = new JPasswordField(10);
		if (currentPassword != null && currentPassword.length() > 0)
		{
			passTF1.setText(currentPassword);
		}
		label.setLabelFor(passTF1);
		inputPanel.add(passTF1);
		
		ContextHelpAction help = (ContextHelpAction)jcards.getFrame().getAction(ActionEnum.CONTEXT_HELP);
		help.setFragment("encryption");
		JButton helpButton = new JButton(help);
		buttonPanel.add(helpButton);

		if (verify)
		{
			label = new JCardsLabel("reenter.password", true);
			inputPanel.add(label);
			passTF2 = new JPasswordField(10);
			if (currentPassword != null && currentPassword.length() > 0)
			{
				passTF2.setText(currentPassword);
			}
			label.setLabelFor(passTF2);
			inputPanel.add(passTF2);
			
			JButton clearButton = new JButton(JCards.getButtonString("clear"));
			clearButton.setMnemonic(KeyEvent.VK_C);
			clearButton.addActionListener(new ActionListener()
			{
				@Override public void actionPerformed(ActionEvent e)
				{
					passTF1.setText(null);
					if (PasswordDialog.this.verify)
					{
						passTF2.setText(null);
					}
					passTF1.requestFocus();
				}
			});
			buttonPanel.add(clearButton);
		}
		
		JButton cancelButton = new JButton(new CancelAction(this));
		buttonPanel.add(cancelButton);
		
		JButton okButton = new JButton(new OKAction(this));
		getRootPane().setDefaultButton(okButton);
		buttonPanel.add(okButton);
		
		panel.add(inputPanel, BorderLayout.CENTER);
		panel.add(buttonPanel, BorderLayout.SOUTH);
		
		getContentPane().add(panel);
	}
	
	/** Set focus to first password field */
	@Override
	public void setVisible(boolean show)
	{
		if (show)
		{
			super.setVisible(show);
			passTF1.requestFocus();
		}
	}
	
	/** OK Button was clicked */
	@Override public void okClicked()
	{
		String password = null;
		String pass1 = new String(passTF1.getPassword());
		
		if (verify)
		{
			String pass2 = new String(passTF2.getPassword());
			if (pass1.equals(pass2))
			{
				password = pass1;
			}
			else
			{
				JOptionPane.showMessageDialog(this, JCards.getMessage("pass.no.match"), JCards.getMessage("incorrect.password"), JOptionPane.ERROR_MESSAGE);
				return;
			}
		}
		else
		{
			password = pass1;
			try
			{
				if (!JCAReader.checkPassword(jcaFile, password))
				{
					JOptionPane.showMessageDialog(this, JCards.getMessage("incorrect.password"), JCards.getMessage("error"), JOptionPane.ERROR_MESSAGE);
					return;
				}
			} catch (JCAException je) {
				JOptionPane.showMessageDialog(this, je.getMessage(), JCards.getMessage("incorrect.password"), JOptionPane.ERROR_MESSAGE);
			}
		}
		
		if (password != null && password.length() == 0) {
			password = null;
		}
		
		jcards.setPassword(password);
		
		dispose();
	}
	
//    public Dimension getPreferredSize()
//    {
//	return new Dimension(300, 120);
//    }
}
