/*
 * JCards
 * Copyright (C) 2002-2011 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package nrs.jcards.gui.dialog;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.GridLayout;
import java.awt.Insets;
import java.awt.event.*;
import java.awt.event.ItemListener;
import java.io.File;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.Map;
import javax.swing.BorderFactory;
import javax.swing.DefaultComboBoxModel;
import javax.swing.DefaultListModel;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTabbedPane;
import javax.swing.JTextField;
import javax.swing.JToolBar;
import javax.swing.ListCellRenderer;
import javax.swing.SortOrder;
import javax.swing.SwingConstants;
import javax.swing.SwingUtilities;
import javax.swing.WindowConstants;
import nrs.jcards.DBDef;
import nrs.jcards.JCards;
import nrs.jcards.UserSettings;
import nrs.jcards.gui.FontChooser;
import nrs.jcards.gui.HelpLabel;
import nrs.jcards.gui.I18NIcon;
import nrs.jcards.gui.JCardsFrame;
import nrs.jcards.gui.JCardsLabel;
import nrs.jcards.gui.LnFComboBox;
import nrs.jcards.gui.action.CancelAction;
import nrs.jcards.gui.action.JCardsAction;
import nrs.jcards.gui.action.OKAction;
import nrs.jcards.gui.action.OKActionI;
import org.jdesktop.swingx.JXList;

/**
 * Preferences dialog window
 * 
 * @author Copyright &#169; 2002-2011 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class PrefDialog extends BaseDialog implements ActionListener, OKActionI
{
    private JCards jcards;
	private JCardsFrame jcardsFrame;
	private JButton fontButton;
	private Font font;
	private JCheckBox openLastCB,  autoEditCB,  wrapCB,  hideCB,  showColonsCB, tableEditCB, systemBrowserCB;
	private JTextField imageEditorTF;
	private LnFComboBox lnfCB;
	private DefaultComboBoxModel browserModel;
	private JXList allIconsList,  selectedIconsList;
	private JDialog fontDialog;
	private UserSettings settings;

	// action commands
	private static final String BROWSE = "browse";
	private static final String IBROWSE = "ibrowse";
	private static final String FONT = "font";
	private static final String ICON_ADD = "iconAdd";
	private static final String ICON_DOWN = "iconDown";
	private static final String ICON_UP = "iconUp";
	private static final String ICON_REMOVE = "iconRemove";
	private static final String ICON_RESET = "iconReset";
	
	class MyCellRenderer extends JLabel implements ListCellRenderer
	{
		@Override
		public Component getListCellRendererComponent(JList list, Object value, int index,
			boolean isSelected, boolean cellHasFocus)
		{
			I18NIcon icon = (I18NIcon)value;
			setText(icon.getName());
			setIcon(JCardsFrame.getImage(icon.getGraphic()));
			if (isSelected)
			{
				setBackground(list.getSelectionBackground());
				setForeground(list.getSelectionForeground());
			}
			else
			{
				setBackground(list.getBackground());
				setForeground(list.getForeground());
			}
			setEnabled(list.isEnabled());
			setFont(list.getFont());
			setOpaque(true);
			return this;
		}
	}
	
	/**
	 * Creates a new instance of PrefDialog
	 * @param jcardsFrame the parent frame
	 * @param settings UserSettings
	 */
	public PrefDialog(JCards jcards)
	{
		super(jcards.getFrame(), true);

		this.jcards = jcards;
		this.jcardsFrame = jcards.getFrame();
		this.settings = JCards.getUserSettings();

		setForeground(Color.black);
		setBackground(Color.white);

		setTitle(JCards.getLabelString("prefs"));

		setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);

		getContentPane().setLayout(new BorderLayout());

		// add toolbar
		JToolBar toolbar = new JToolBar();
		toolbar.setFloatable(false);
		toolbar.setLayout(new FlowLayout(FlowLayout.TRAILING));

		JButton cancel = new JButton(new CancelAction(this));
		toolbar.add(cancel);

		JButton ok = new JButton(new OKAction(this));
		toolbar.add(ok);
		getRootPane().setDefaultButton(ok);

		getContentPane().add(toolbar, BorderLayout.SOUTH);

		JTabbedPane tabMan = new JTabbedPane();
		DBDef def = DBDef.getInstance();

		// app settings
		String name = JCards.getLabelString("tab.global", true);
		tabMan.add(JCardsAction.getTextWithoutMnemonic(name), createGlobalPanel());
		tabMan.setMnemonicAt(0, JCardsAction.getMnemonic(name));
		tabMan.setDisplayedMnemonicIndexAt(0, JCardsAction.getMnemonicIndex(name));
		tabMan.setToolTipTextAt(0, JCards.getLabelString("tab.global.tt"));
		tabMan.setIconAt(0, JCardsFrame.getImage(UserSettings.ICON_OPEN_URL));

		// database settings
		name = JCards.getLabelString("tab.db", true);
		tabMan.add(JCardsAction.getTextWithoutMnemonic(name), createDBPanel(def));
		tabMan.setMnemonicAt(1, JCardsAction.getMnemonic(name));
		tabMan.setDisplayedMnemonicIndexAt(1, JCardsAction.getMnemonicIndex(name));
		tabMan.setToolTipTextAt(1, JCards.getLabelString("tab.db.tt"));
		tabMan.setIconAt(1, JCardsFrame.getImage(UserSettings.ICON_DB));

		// icons
		name = JCards.getLabelString("tab.icons", true);
		tabMan.add(JCardsAction.getTextWithoutMnemonic(name), createIconPanel());
		tabMan.setMnemonicAt(2, JCardsAction.getMnemonic(name));
		tabMan.setDisplayedMnemonicIndexAt(2, JCardsAction.getMnemonicIndex(name));
		tabMan.setToolTipTextAt(2, JCards.getLabelString("tab.icons.tt"));
		tabMan.setIconAt(2, JCardsFrame.getImage(UserSettings.ICON_ICONS));

		getContentPane().add(tabMan, BorderLayout.CENTER);
	}

	/**
	 * Create the application panel
	 * @todo option to display in system tray? - use java.awt.SystemTray
	 */
	private Component createGlobalPanel()
	{
		JLabel l;
		GridBagLayout gbl = new GridBagLayout();

		// label constraints
		GridBagConstraints gbcLabel = new GridBagConstraints();
		gbcLabel.anchor = GridBagConstraints.WEST;
		gbcLabel.weightx = 0.0;
		gbcLabel.weighty = 0.0;
		gbcLabel.gridwidth = 1;
		gbcLabel.insets = new Insets(2, 2, 2, 5);

		// component constraints
		GridBagConstraints gbcComp = new GridBagConstraints();
		gbcComp.anchor = GridBagConstraints.WEST;
		gbcComp.fill = GridBagConstraints.NONE;
		gbcComp.weightx = 1.0;
		gbcComp.weighty = 0.0;
		gbcComp.gridwidth = GridBagConstraints.REMAINDER;
		gbcComp.insets = new Insets(2, 2, 2, 2);

		/* Application settings */
		JPanel appPanel = new JPanel(gbl);
		appPanel.setBorder(BorderFactory.createEmptyBorder(5, 5, 5, 5));

		// load last file
		l = new JCardsLabel("load.last");
		appPanel.add(l, gbcLabel);

		openLastCB = new JCheckBox();
		openLastCB.setSelected(settings.isOpenLastFile());
		openLastCB.setToolTipText(JCards.getLabelString("load.last.tt"));
		l.setLabelFor(openLastCB);
		appPanel.add(openLastCB, gbcComp);

		// auto edit cards
		l = new JCardsLabel("auto.edit");
		appPanel.add(l, gbcLabel);

		autoEditCB = new JCheckBox();
		autoEditCB.setSelected(settings.isAutoEdit());
		autoEditCB.setToolTipText(JCards.getLabelString("auto.edit.tt"));
		l.setLabelFor(autoEditCB);
		appPanel.add(autoEditCB, gbcComp);

		// hide fields on view
		l = new JCardsLabel("word.wrap");
		appPanel.add(l, gbcLabel);

		wrapCB = new JCheckBox();
		wrapCB.setSelected(settings.isWordWrap());
		wrapCB.setToolTipText(JCards.getLabelString("word.wrap.tt"));
		l.setLabelFor(wrapCB);
		appPanel.add(wrapCB, gbcComp);
		
		// hide fields on view
		l = new JCardsLabel("table.edit");
		appPanel.add(l, gbcLabel);

		tableEditCB = new JCheckBox();
		tableEditCB.setSelected(settings.isTableEdit());
		tableEditCB.setToolTipText(JCards.getLabelString("table.edit.tt"));
		l.setLabelFor(tableEditCB);
		appPanel.add(tableEditCB, gbcComp);
		
		// system browser option
		l = new JCardsLabel("browser.system");
		appPanel.add(l, gbcLabel);
		
		systemBrowserCB = new JCheckBox();
		systemBrowserCB.setSelected(settings.isUseSystemBrowser());
		systemBrowserCB.setToolTipText(JCards.getLabelString("browser.system.tt"));
		l.setLabelFor(systemBrowserCB);
		appPanel.add(systemBrowserCB, gbcComp);
		
		// browser
		l = new JCardsLabel("browser");
		appPanel.add(l, gbcLabel);

		List<String> browsers = UserSettings.findWebBrowser();
		String currentBrowser = settings.getBrowserLocation();
		browserModel = new DefaultComboBoxModel(browsers.toArray());
		if (currentBrowser != null)
		{
			if (!browsers.contains(currentBrowser))
			{
				browserModel.insertElementAt(currentBrowser, 0);
			}
			browserModel.setSelectedItem(currentBrowser);
		}
		final JComboBox webBrowser = new JComboBox(browserModel);
		webBrowser.setToolTipText(JCards.getLabelString("browser.tt"));
		webBrowser.setEnabled(!settings.isUseSystemBrowser());
		gbcComp.gridwidth = 1;
		gbcComp.weightx = 1.0;
		l.setLabelFor(webBrowser);
		appPanel.add(webBrowser, gbcComp);

		final JButton fileButton = new JButton(JCards.getButtonString("browse"));
		fileButton.setActionCommand(BROWSE);
		fileButton.addActionListener(this);
		fileButton.setEnabled(!settings.isUseSystemBrowser());
		gbcComp.weightx = 0.0;
		gbcComp.gridwidth = GridBagConstraints.REMAINDER;
		appPanel.add(fileButton, gbcComp);

		// add listener to toggle browser options
		systemBrowserCB.addItemListener(new ItemListener() {
			@Override public void itemStateChanged(ItemEvent e)
			{
				JCheckBox cb = (JCheckBox)e.getItem();
				webBrowser.setEnabled(!cb.isSelected());
				fileButton.setEnabled(!cb.isSelected());
			}
		});

		// image editor
		l = new JCardsLabel("image.editor");
		gbcLabel.anchor = GridBagConstraints.NORTHWEST;
		gbcLabel.weighty = 0.0;
		appPanel.add(l, gbcLabel);

		imageEditorTF = new JTextField(settings.getImageEditor(), 20);
		l.setLabelFor(imageEditorTF);
		gbcComp.gridwidth = 1;
		gbcComp.weightx = 1.0;
		gbcComp.anchor = GridBagConstraints.NORTHWEST;
		appPanel.add(imageEditorTF, gbcComp);

		JButton ifileButton = new JButton(JCards.getButtonString("browse"));
		ifileButton.setActionCommand(IBROWSE);
		ifileButton.addActionListener(this);
		gbcComp.weightx = 0.0;
		gbcComp.gridwidth = GridBagConstraints.REMAINDER;
		appPanel.add(ifileButton, gbcComp);

		// font
		l = new JCardsLabel("font");
		gbcLabel.anchor = GridBagConstraints.NORTHWEST;
		gbcLabel.weighty = 0.0;
		appPanel.add(l, gbcLabel);

		String fontString = settings.getFont().getFontName() + " " + settings.getFont().getSize();
		fontButton = new JButton(fontString);
		fontButton.setActionCommand(FONT);
		fontButton.addActionListener(this);
		l.setLabelFor(fontButton);
		appPanel.add(fontButton, gbcComp);

		// theme
		l = new JCardsLabel("theme");
		gbcLabel.anchor = GridBagConstraints.NORTHWEST;
		gbcLabel.weighty = 1.0;
		appPanel.add(l, gbcLabel);

		lnfCB = new LnFComboBox(this);
		lnfCB.setThemeClassName(settings.getTheme());
		gbcComp.gridwidth = GridBagConstraints.REMAINDER;
		gbcComp.weightx = 1.0;
		gbcComp.weighty = 1.0;
		gbcComp.anchor = GridBagConstraints.NORTHWEST;
		appPanel.add(lnfCB, gbcComp);
		l.setLabelFor(lnfCB);

		// help button
		l = new HelpLabel(jcardsFrame, "prefs.global");
		gbcLabel.anchor = GridBagConstraints.SOUTHEAST;
		gbcLabel.gridwidth = GridBagConstraints.REMAINDER;
		appPanel.add(l, gbcLabel);
		
		JScrollPane scroll = new JScrollPane(appPanel);

		return scroll;
	}

	/** Create the DB panel */
	private Component createDBPanel(DBDef def)
	{
		GridBagLayout gbl = new GridBagLayout();
		JPanel dbPanel = new JPanel(gbl);
		dbPanel.setBorder(BorderFactory.createEmptyBorder(5, 5, 5, 5));

		GridBagConstraints gbcLabel = new GridBagConstraints();
		gbcLabel.gridwidth = 1;
		gbcLabel.weightx = 0.0;
		gbcLabel.insets = new Insets(10, 0, 0, 10);
		gbcLabel.anchor = GridBagConstraints.WEST;

		GridBagConstraints gbcComp = new GridBagConstraints();
		gbcComp.gridwidth = GridBagConstraints.REMAINDER;
		gbcComp.insets = new Insets(10, 0, 0, 10);
		gbcComp.anchor = GridBagConstraints.WEST;
		gbcComp.weightx = 1.0;
		gbcComp.fill = GridBagConstraints.NONE;

		// show field labels
		JLabel l = new JCardsLabel("show.labels");
		dbPanel.add(l, gbcLabel);

		hideCB = new JCheckBox();
		hideCB.setToolTipText(JCards.getLabelString("show.labels.tt"));
		hideCB.setSelected(def.displayFieldNames());
		l.setLabelFor(hideCB);
		dbPanel.add(hideCB, gbcComp);

		l = new JCardsLabel("show.colon");
		gbcLabel.weighty = 1.0;
		gbcLabel.anchor = GridBagConstraints.NORTHWEST;
		dbPanel.add(l, gbcLabel);

		showColonsCB = new JCheckBox();
		showColonsCB.setToolTipText(JCards.getLabelString("show.colon.tt"));
		showColonsCB.setSelected(def.displayColons());
		l.setLabelFor(showColonsCB);
		gbcComp.weighty = 1.0;
		gbcComp.anchor = GridBagConstraints.NORTHWEST;
		dbPanel.add(showColonsCB, gbcComp);

		// help button
		l = new HelpLabel(jcardsFrame, "prefs.db");
		gbcLabel.anchor = GridBagConstraints.SOUTHEAST;
		gbcLabel.gridwidth = GridBagConstraints.REMAINDER;
		dbPanel.add(l, gbcLabel);
		
		return dbPanel;
	}

	/** Create the icons panel */
	private JPanel createIconPanel()
	{
		GridBagLayout gbl = new GridBagLayout();
		GridBagConstraints gbc = new GridBagConstraints();

		JPanel iconPanel = new JPanel(gbl);
		iconPanel.setBorder(BorderFactory.createEmptyBorder(2, 2, 2, 2));

		JLabel allLabel = new JCardsLabel("all", false);
		Font f = settings.getFont();
		Font boldFont = f.deriveFont(Font.BOLD, f.getSize() + 2);
		allLabel.setFont(boldFont);
		allLabel.setHorizontalAlignment(SwingConstants.CENTER);
		gbc.gridwidth = 1;
		gbc.fill = GridBagConstraints.NONE;
		gbc.weightx = 0.5;
		gbc.insets = new Insets(2, 2, 2, 2);
		gbc.anchor = GridBagConstraints.CENTER;
		iconPanel.add(allLabel, gbc);

		// spacer
		JLabel l = new JLabel();
		gbc.weightx = 0;
		iconPanel.add(l, gbc);

		JLabel selLabel = new JCardsLabel("selected", false);
		selLabel.setFont(boldFont);
		selLabel.setHorizontalAlignment(SwingConstants.CENTER);
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		gbc.weightx = 0.5;
		iconPanel.add(selLabel, gbc);

		allIconsList = new JXList(new DefaultListModel());
		allIconsList.setAutoCreateRowSorter(true);
		allIconsList.setSortOrder(SortOrder.ASCENDING);
		allIconsList.setComparator(new Comparator()
		{
			@Override
			public int compare(Object o1, Object o2)
			{
				I18NIcon i1 = (I18NIcon)o1;
				I18NIcon i2 = (I18NIcon)o2;

				return i1.getName().compareTo(i2.getName());
			}
		});
		allIconsList.setCellRenderer(new MyCellRenderer());
		allLabel.setLabelFor(allIconsList);
		gbc.gridwidth = 1;
		gbc.gridheight = 5;
		gbc.fill = GridBagConstraints.BOTH;
		gbc.weightx = 0.5;
		gbc.weighty = 1.0;
		gbc.anchor = GridBagConstraints.CENTER;
		JScrollPane scroll = new JScrollPane(allIconsList);
		iconPanel.add(scroll, gbc);

		JPanel buttons = createIconButtonPanel();
		gbc.gridheight = 1;
		gbc.fill = GridBagConstraints.NONE;
		gbc.weightx = 0.0;
		gbc.weighty = 1.0;
		iconPanel.add(buttons, gbc);

		selectedIconsList = new JXList(new DefaultListModel());
		selectedIconsList.setCellRenderer(new MyCellRenderer());
		selLabel.setLabelFor(selectedIconsList);
		gbc.gridwidth = GridBagConstraints.REMAINDER;
		gbc.gridheight = 5;
		gbc.fill = GridBagConstraints.BOTH;
		gbc.weightx = 0.5;
		gbc.weighty = 1.0;
		scroll = new JScrollPane(selectedIconsList);
		iconPanel.add(scroll, gbc);

		// set the icons in each
		resetIcons(false);

		return iconPanel;
	}

	/** Create the icon buttons panel */
	private JPanel createIconButtonPanel()
	{
		JPanel toolbar = new JPanel(new GridLayout(5, 1, 10, 10));
		toolbar.setMinimumSize(new Dimension(100, 200));

		JButton add = new JButton(JCardsFrame.getImage(UserSettings.ICON_NEXT));
		add.setActionCommand(ICON_ADD);
		add.addActionListener(this);
		add.setToolTipText(JCards.getLabelString("select.tt"));
		toolbar.add(add);

		JButton remove = new JButton(JCardsFrame.getImage(UserSettings.ICON_PREVIOUS));
		remove.setActionCommand(ICON_REMOVE);
		remove.addActionListener(this);
		remove.setToolTipText(JCards.getLabelString("remove"));
		toolbar.add(remove);

		JButton reset = new JButton(JCardsFrame.getImage(UserSettings.ICON_RESET));
		reset.setActionCommand(ICON_RESET);
		reset.addActionListener(this);
		reset.setToolTipText(JCards.getLabelString("reset.tt"));
		toolbar.add(reset);

		JButton up = new JButton(JCardsFrame.getImage(UserSettings.ICON_UP));
		up.setActionCommand(ICON_UP);
		up.addActionListener(this);
		up.setToolTipText(JCards.getLabelString("up.tt"));
		toolbar.add(up);

		JButton down = new JButton(JCardsFrame.getImage(UserSettings.ICON_DOWN));
		down.setActionCommand(ICON_DOWN);
		down.addActionListener(this);
		down.setToolTipText(JCards.getLabelString("down.tt"));
		toolbar.add(down);

		return toolbar;
	}

	private String getFilePath()
	{
		String filePath = null;

		JFileChooser fc = new JFileChooser();
		fc.setApproveButtonText(JCards.getButtonString("select"));
		if (fc.showOpenDialog(this) == JFileChooser.APPROVE_OPTION)
		{
			File f = fc.getSelectedFile();
			if (!f.isDirectory()) {
				filePath = f.getAbsolutePath();
			}
		}

		return filePath;
	}

	/** Called when one of the buttons is clicked */
	@Override
	public void actionPerformed(ActionEvent actionEvent)
	{
		String cmd = actionEvent.getActionCommand();

		if (cmd.startsWith("icon"))
		{
			if (cmd.equals(ICON_ADD))
			{
				addIcon(selectedIconsList, allIconsList);
			}
			else if (cmd.equals(ICON_REMOVE))
			{
				removeIcon(selectedIconsList, allIconsList);
			}
			else if (cmd.equals(ICON_RESET))
			{
				resetIcons(true);
			}
			else if (cmd.equals(ICON_UP))
			{
				iconUp(selectedIconsList);
			}
			else if (cmd.equals(ICON_DOWN))
			{
				iconDown(selectedIconsList);
			}
		}
		else if (cmd.equals(BROWSE))
		{
			String filePath = getFilePath();
			if (filePath != null)
			{
				browserModel.removeElement(filePath);
				browserModel.insertElementAt(filePath, 0);
				browserModel.setSelectedItem(filePath);
			}
		}
		else if (cmd.equals(IBROWSE))
		{
			String filePath = getFilePath();
			if (filePath != null)
			{
				imageEditorTF.setText(filePath);
			}
		}
		else if (cmd.equals(FONT))
		{
			final FontChooser fontChooser = new FontChooser(settings.getFont());
			ActionListener okListener = new ActionListener() {
				@Override public void actionPerformed(ActionEvent e)
				{
					font = fontChooser.getUserFont();
					fontDialog.dispose();
					String fontString = font.getFontName() + " " + font.getSize();
					fontButton.setText(fontString);
					fontButton.setFont(font);
				}
			};
			
			fontDialog = FontChooser.createDialog(this, JCards.getLabelString("font.set"), true, fontChooser, okListener, null);
			fontDialog.pack();
			fontDialog.setLocationRelativeTo(this);
			fontDialog.setVisible(true);
		}
	}
	
	/** Move an icon/field up */
	private void iconUp(JList selectedList)
	{
		int index = selectedList.getSelectedIndex();
		if (index < 1)
		{
			return;
		}

		I18NIcon icon = (I18NIcon)selectedList.getSelectedValue();
		DefaultListModel model = (DefaultListModel) selectedList.getModel();
		model.remove(index);
		model.add(index - 1, icon);
		selectedList.setSelectedIndex(index - 1);
	}

	/** Move an icon/field down */
	private void iconDown(JList selectedList)
	{
		int index = selectedList.getSelectedIndex();
		if (index == -1 || index >= selectedList.getModel().getSize() - 1)
		{
			return;
		}

		I18NIcon icon = (I18NIcon)selectedList.getSelectedValue();
		DefaultListModel model = (DefaultListModel) selectedList.getModel();
		model.remove(index);
		model.add(index + 1, icon);
		selectedList.setSelectedIndex(index + 1);
	}

	/** Add an icon/field to display */
	private void addIcon(JList selectedList, JList allList)
	{
		int index = allList.getSelectedIndex();
		if (index == -1)
		{
			return;
		}

		I18NIcon icon = (I18NIcon)allList.getSelectedValue();
		DefaultListModel allModel = (DefaultListModel)allList.getModel();
		allModel.removeElement(icon);
		DefaultListModel selModel = (DefaultListModel)selectedList.getModel();
		selModel.addElement(icon);

		selectedList.setSelectedIndex(selModel.getSize()-1);

		if (allModel.getSize() > 0) {
			allList.setSelectedIndex((index < allModel.getSize() ? index : index-1));
		}
	}

	/** Remove an icon/field from display */
	private void removeIcon(JList selectedList, JList allList)
	{
		int index = selectedList.getSelectedIndex();
		if (index == -1)
		{
			return;
		}

		I18NIcon icon = (I18NIcon)selectedList.getSelectedValue();
		DefaultListModel selModel = (DefaultListModel) selectedList.getModel();
		selModel.removeElement(icon);
		DefaultListModel allModel = (DefaultListModel) allList.getModel();
		allModel.addElement(icon);

		if (selModel.getSize() > 0)	{
			selectedList.setSelectedIndex((index < selModel.getSize() ? index : index-1));
		}
	}

	/**
	 * Reset the icons in each list
	 * @param reset if false just display the icons; otherwise do a reset first
	 */
	private void resetIcons(boolean reset)
	{
		// reset user settings
		if (reset) {
			settings.resetIcons();
		}

		DefaultListModel selModel = (DefaultListModel)selectedIconsList.getModel();
		DefaultListModel allModel = (DefaultListModel)allIconsList.getModel();

		// clear lists
		selModel.clear();
		allModel.clear();

		Map<String, I18NIcon> allIcons = settings.getAllIcons(false);
		List<I18NIcon> userIcons = settings.getUserIcons();		

		// set selected list
		for (I18NIcon icon : userIcons)
		{
			selModel.addElement(icon);
		}
		
		// set remainder
		for (Map.Entry<String, I18NIcon> entry : allIcons.entrySet())
		{
			if (!userIcons.contains(entry.getValue())) {
				allModel.addElement(entry.getValue());
			}
		}
		selectedIconsList.repaint();
		allIconsList.repaint();

		allIconsList.setSelectedIndex(0);
		selectedIconsList.setSelectedIndex(0);
	}

	/** OK button clicked */
	@Override public void okClicked()
	{
		// set app prefs
		settings.setOpenLastFile(openLastCB.isSelected());
		settings.setAutoEdit(autoEditCB.isSelected());
		settings.setWordWrap(wrapCB.isSelected());
		settings.setUseSystemBrowser(systemBrowserCB.isSelected());
		settings.setBrowserLocation((String) browserModel.getSelectedItem());
		settings.setImageEditor(imageEditorTF.getText());
		settings.setTableEdit(tableEditCB.isSelected());

		jcardsFrame.setTableEditable(tableEditCB.isSelected());
		
		// set font
		if (font != null)
		{
			settings.setFont(font);
			jcardsFrame.fontChanged();
		}

		// set theme
		settings.setTheme(lnfCB.getThemeClassName());
		SwingUtilities.updateComponentTreeUI(jcardsFrame);
		jcardsFrame.repaint();

		// set user icons
		DefaultListModel selModel = (DefaultListModel)selectedIconsList.getModel();
		List<I18NIcon> icons = new ArrayList<I18NIcon>(selModel.size());
		for (int i=0; i<selModel.size(); i++)
		{
			icons.add((I18NIcon)selModel.elementAt(i));
		}
		settings.setUserIcons(icons);
		jcardsFrame.createToolbar();

		// set db prefs
		DBDef def = DBDef.getInstance();
		if (def.displayFieldNames() != hideCB.isSelected())
		{
			def.setDisplayFieldNames(hideCB.isSelected());
			jcards.setChanged();
		}

		if (def.displayColons() != showColonsCB.isSelected())
		{
			def.setDisplayColons(showColonsCB.isSelected());
			jcards.setChanged();
		}

		// set LnF
		SwingUtilities.updateComponentTreeUI(jcardsFrame);
		jcardsFrame.pack();
		
		dispose();
	}
}
