/*
 * JCards
 * Copyright (C) 2002-2009 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package nrs.jcards.gui.table;

import java.awt.Toolkit;
import java.util.Observable;
import java.util.Observer;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.table.AbstractTableModel;
import nrs.jcards.*;
import nrs.jcards.fields.CategoryFieldDef;
import nrs.jcards.fields.FieldDef;
import nrs.jcards.fields.FieldType;

/**
 * Observes the Cards object for changes and acts as the model for the main table.
 * Observes DBDef for changes in visible fields.
 *
 * @todo alignment of columns (e.g. numbers centre aligned by default). Right click for alignment options.
 *
 * @author Copyright &#169; 2002-2009 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class CardsTableModel extends AbstractTableModel implements Observer
{
	private JCards jcards;
	private Cards cards;
	private final static Logger log = Logger.getLogger(CardsTableModel.class.getName());

	/** Creates a new instance of CardsTableModel */
	public CardsTableModel(JCards jcards)
	{
		this.jcards = jcards;
		cards = jcards.getCards();
		cards.addObserver(this);
		DBDef.getInstance().addObserver(this);
	}

	@Override
	public Object getValueAt(int rowIndex, int columnIndex)
	{
		// log.finest("CardsTableModel.getValueAt (" + rowIndex + ", " + columnIndex + ")");
		return cards.getValueAt(rowIndex, columnIndex);
	}

	@Override
	public int getRowCount()
	{
		// log.exiting("CardsTableModel", "getRowCount", cards.getRowCount());
		return cards.getRowCount();
	}

	@Override
	public int getColumnCount()
	{
		// log.exiting("CardsTableModel", "getColumnCount", cards.getColumnCount());
		return cards.getColumnCount();
	}

	/** Listen for Cards or DBDef visible fields changes */
	@Override
	public void update(Observable o, Object arg)
	{
		log.entering("CardsTableModel", "update", arg);

		if (arg instanceof CardUpdate)
		{
			CardUpdate cu = (CardUpdate)arg;

			switch (cu.cardOperation)
			{
				case LOAD:
					fireTableStructureChanged();
					fireTableDataChanged();
					break;

				case CARD_NEW:
					fireTableRowsInserted(cu.cardIndex, cu.cardIndex);
					break;

				case CARD_DELETE:
					fireTableRowsDeleted(cu.cardIndex, cu.cardIndex);
					break;

				case CARD_UPDATE:
					fireTableRowsUpdated(cu.cardIndex, cu.cardIndex);
					break;

				case FIND:
					fireTableDataChanged();
					break;
			}
		}
		else if (arg instanceof DBDefUpdate)
		{
			DBDefUpdate evt = (DBDefUpdate) arg;

			if (evt.getEvent() == DBDefUpdate.Event.VISIBLE_FIELDS || evt.getEvent() == DBDefUpdate.Event.RESET)
			{
				fireTableStructureChanged();
			}
		}

		log.exiting("CardsTableModel", "update");
	}

	/** Get column name */
	@Override
	public String getColumnName(int column)
	{
		return DBDef.getInstance().getVisibleFields().get(column);
	}

	/** Some cells can be edited */
	@Override
	public boolean isCellEditable(int rowIndex, int columnIndex)
	{
		boolean editable = true;
		
		DBDef dbDef = DBDef.getInstance();
		String fieldName = dbDef.getVisibleFields().get(columnIndex);
		FieldDef fd = dbDef.getFieldDef(fieldName);

		// some types are definitely not editable
		switch (fd.getType())
		{
			case CALCULATED:
			case COUNTER:
			case MEMO:
				editable = false;
				break;
		}

		return editable;
	}

	/**
     * Returns class of column
     * @param columnIndex  the column being queried
     * @return the correct class
     */
	@Override
	public Class getColumnClass(int columnIndex)
	{
		log.entering("CardsTableModel", "getColumnClass", columnIndex);
		DBDef dbDef = DBDef.getInstance();
		String fieldName = dbDef.getVisibleFields().get(columnIndex);
		FieldDef fd = dbDef.getFieldDef(fieldName);
		assert fd != null : "getColumnClass: Cannot get FieldDef for " + columnIndex + ", " + fieldName;
		log.log(Level.FINEST, "Got FieldDef for {0}", fieldName);
		log.exiting("CardsTableModel", "getColumnClass", fd.getFieldClass());
		return fd.getFieldClass();
	}

	/**
     *  Set value in table cell
     *  @param  aValue   value to assign to cell
     *  @param  rowIndex   row of cell
     *  @param  columnIndex  column of cell
     */
	@Override
	public void setValueAt(Object aValue, int rowIndex, int columnIndex)
	{
		Object[] params = {aValue, rowIndex, columnIndex};
		log.entering("CardsTableModel", "setValueAt", params);

		DBDef dbDef = DBDef.getInstance();
		String fieldName = dbDef.getVisibleFields().get(columnIndex);
		FieldDef fd = dbDef.getFieldDef(fieldName);
		assert fd != null : "setValueAt: Cannot get FieldDef for " + columnIndex;
		
		Object parsedValue = null;
		if (aValue != null)
		{
			switch (fd.getType())
			{
				case DATE:
					// date picker ensures only dates get entered
					parsedValue = aValue;
					break;
					
				default:
					parsedValue = fd.parseValue(aValue.toString());
					break;
			}
		}

		// check for null in a mandatory field
		if (fd.isMandatory() && (parsedValue == null || parsedValue.toString().length() == 0))
		{
			Toolkit.getDefaultToolkit().beep();
			return;
		}

		// add new values to categories on the fly
		if (parsedValue != null && fd.getType().equals(FieldType.CATEGORY))
		{
			CategoryFieldDef cfd = (CategoryFieldDef)fd;
			if (!dbDef.getCategory(cfd.getCategory()).contains(parsedValue.toString())) {
				dbDef.addCategoryItem(cfd.getCategory(), parsedValue.toString());
			}
		}
		
		// see if value has changed
		Card card = cards.getCard(rowIndex);
		assert card != null : "Cannot get card at row "+rowIndex;
		Object currentVal = card.getValue(fd.getName());
		boolean noChange = false;
		if (currentVal == null)
		{
			if (parsedValue == null)
			{
				noChange = true;
			}
		}
		else
		{
			if (parsedValue != null)
			{
				if (currentVal.toString().equals(parsedValue.toString()))
				{
					noChange = true;
				}
			}
		}

		if (noChange)
		{
			log.log(Level.FINEST, "No change in value from {0}", currentVal);
		}
		else
		{
			log.log(Level.FINEST, "Value has changed from {0}", currentVal);
			card.setValue(fd.getName(), parsedValue);
			jcards.setChanged(JCards.getMessage("table.entry.updated"));

			// check to see if any calculations are involved
			dbDef.updateAllCalculations(fieldName, card);

			fireTableRowsUpdated(rowIndex, rowIndex);
		}

		log.exiting("CardsTableModel", "setValueAt", noChange);
	}
}
