/*
 * JCards
 * Copyright (C) 2002-2011 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package nrs.jcards.gui.table;

import java.awt.event.*;
import java.util.List;
import javax.swing.*;
import nrs.jcards.DBDef;
import nrs.jcards.JCards;
import nrs.jcards.UserSettings;
import nrs.jcards.fields.FieldType;
import nrs.jcards.gui.JCardsFrame;
import nrs.jcards.gui.JCardsMenu;
import nrs.jcards.gui.JCardsMenuItem;
import nrs.jcards.gui.JCardsRadioButtonMenuItem;

/**
 * Allows the user to select the visible columns. This button is displayed in the top
 * right corner of the scroll panel.
 *
 * @author Copyright &#169; 2002-2011 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class SelectVisibleColumnsButton extends JButton implements ActionListener, ItemListener, MouseListener
{
	private JTable cardsTable;
	private JMenu resizeMenu;
	private JRadioButtonMenuItem all, last, next, sub, off;
	private final static String ALL = "all";
	
	/** Creates a new instance of SelectVisibleColumnsAction */
	public SelectVisibleColumnsButton(JTable cardsTable)
	{
		super(JCardsFrame.getImage(UserSettings.ICON_COLUMNS));
		this.cardsTable = cardsTable;
		setToolTipText(JCards.getButtonString("columns.tt"));
		addMouseListener(this);
		setRequestFocusEnabled(false);
	}
	
	@Override
	public void mousePressed(MouseEvent e)
	{
		List<String> fieldNames = DBDef.getInstance().getFieldNames();
		JPopupMenu popup = new JPopupMenu();
		JCheckBoxMenuItem mi;
		
		if (!fieldNames.isEmpty())
		{
			List<String> visFields = DBDef.getInstance().getVisibleFields();

			DBDef dbDef = DBDef.getInstance();
			for (String fieldName : fieldNames)
			{
				// can't display images
				if (dbDef.getFieldDef(fieldName).getType() != FieldType.IMAGE)
				{
					mi = new JCheckBoxMenuItem(fieldName);
					mi.setSelected(visFields.contains(fieldName));
					mi.addItemListener(this);
					popup.add(mi);
				}
			}
			
			popup.addSeparator();
			
			JMenuItem miall = new JCardsMenuItem(ALL);
			miall.setActionCommand(ALL);
			miall.addActionListener(this);
			popup.add(miall);
		}
		else
		{
			mi = new JCheckBoxMenuItem(JCards.getMessage("no.fields.def"));
			popup.add(mi);
		}
		
		popup.addSeparator();
		createResizeMenu();
		popup.add(resizeMenu);
		
		JButton button = (JButton)e.getSource();
		popup.show(button, e.getX(), e.getY());
	}
	
	/** Menu used for setting table column resizing mode */
	private void createResizeMenu()
	{
		// cached for future use
		if (resizeMenu == null)
		{
			resizeMenu = new JCardsMenu("resize.mode");
			ButtonGroup btnGrp = new ButtonGroup();
			
			all = new JCardsRadioButtonMenuItem(ALL);
			all.setActionCommand(Integer.toString(JTable.AUTO_RESIZE_ALL_COLUMNS));
			all.addActionListener(this);
			resizeMenu.add(all);
			btnGrp.add(all);
			
			last = new JCardsRadioButtonMenuItem("last");
			last.setActionCommand(Integer.toString(JTable.AUTO_RESIZE_LAST_COLUMN));
			last.addActionListener(this);
			resizeMenu.add(last);
			btnGrp.add(last);
			
			next = new JCardsRadioButtonMenuItem("next");
			next.setActionCommand(Integer.toString(JTable.AUTO_RESIZE_NEXT_COLUMN));
			next.addActionListener(this);
			resizeMenu.add(next);
			btnGrp.add(next);
			
			sub = new JCardsRadioButtonMenuItem("subsequent");
			sub.setActionCommand(Integer.toString(JTable.AUTO_RESIZE_SUBSEQUENT_COLUMNS));
			sub.addActionListener(this);
			resizeMenu.add(sub);
			btnGrp.add(sub);
			
			off = new JCardsRadioButtonMenuItem("off");
			off.setActionCommand(Integer.toString(JTable.AUTO_RESIZE_OFF));
			off.addActionListener(this);
			resizeMenu.add(off);
			btnGrp.add(off);
		}
		
		// set the selected mode
		switch (cardsTable.getAutoResizeMode())
		{
			case JTable.AUTO_RESIZE_ALL_COLUMNS:
				all.setSelected(true);
				break;
				
			case JTable.AUTO_RESIZE_LAST_COLUMN:
				last.setSelected(true);
				break;
				
			case JTable.AUTO_RESIZE_NEXT_COLUMN:
				next.setSelected(true);
				break;
				
			case JTable.AUTO_RESIZE_OFF:
				off.setSelected(true);
				break;
				
			case JTable.AUTO_RESIZE_SUBSEQUENT_COLUMNS:
				sub.setSelected(true);
				break;
		}
	}
	
	@Override public void mouseReleased(MouseEvent e)
	{}
	@Override public void mouseExited(MouseEvent e)
	{}
	@Override public void mouseEntered(MouseEvent e)
	{}
	@Override public void mouseClicked(MouseEvent e)
	{}
	
	/** One of the menu items was selected */
	@Override
	public void itemStateChanged(ItemEvent e)
	{
		DBDef def = DBDef.getInstance();
		String col = null;
		
		List<String> visFields = def.getVisibleFields();
		JCheckBoxMenuItem cbMI = (JCheckBoxMenuItem)e.getSource();
		col = cbMI.getText();
		if (cbMI.isSelected())
		{
			visFields.add(col);
		}
		else
		{
			visFields.remove(col);
			if (visFields.isEmpty())
			{
				visFields.add(col);
			}
		}
		
		def.setVisibleFields(visFields);
	}
	
	@Override
	public void actionPerformed(ActionEvent e)
	{
		if (e.getActionCommand().equals(ALL))
		{
			DBDef def = DBDef.getInstance();
			List<String> visFields = def.getVisibleFields();
			for (String name : def.getFieldNames())
			{
				if (!visFields.contains(name) && (def.getFieldDef(name).getType() != FieldType.IMAGE))
				{
					visFields.add(name);
				}
			}
			def.setVisibleFields(visFields);
		} else {
			cardsTable.setAutoResizeMode(Integer.parseInt(e.getActionCommand()));
		}
	}
}
