/*
 * JCards
 * Copyright (C) 2002-2009 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package nrs.jcards.gui.table;

import java.awt.event.KeyEvent;
import static java.awt.event.KeyEvent.VK_ENTER;
import java.awt.event.KeyListener;
import java.text.MessageFormat;
import java.util.regex.Pattern;
import java.util.regex.PatternSyntaxException;
import nrs.jcards.JCards;
import nrs.jcards.gui.JCardsFrame;
import org.jdesktop.swingx.JXTable;

/**
 * Listens for key events to enable quick searches
 * 
 * @author Copyright &#169; 2002-2009 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class TableKeyListener implements KeyListener
{
	private JXTable table;
	private JCardsFrame jcardsFrame;
	private StringBuilder keys = new StringBuilder(3);
	private long lastKeyPressTime;
	private int currentRow;
	private static final int SEARCH_FLAGS = Pattern.CASE_INSENSITIVE | Pattern.UNICODE_CASE | Pattern.LITERAL;
	private static final Pattern VALID_PATTERN = Pattern.compile("\\p{Graph}|\\p{Zs}|\\p{L}|\\p{N}|\\p{P}|\\p{S}");

	private final static int TIMEOUT = 2000;

	public TableKeyListener(JXTable table, JCardsFrame jcardsFrame)
	{
		this.table = table;
		this.jcardsFrame = jcardsFrame;
	}

	private boolean isSearchKey(KeyEvent e)
	{
		if (table.isEditing() || (e.getModifiersEx() != 0) ||
			!VALID_PATTERN.matcher(Character.toString(e.getKeyChar())).matches())
		{
			return false;
		}

		return true;
	}

	private void search()
	{
		if (keys.length() == 0) {
			return;
		}
		
		try
		{
			Pattern p = Pattern.compile(keys.toString(), SEARCH_FLAGS);
			if ((currentRow = table.getSearchable().search(p, currentRow)) == -1) {
				if ((currentRow = table.getSearchable().search(p, 0)) > -1) {
					jcardsFrame.setStatus(MessageFormat.format(JCards.getMessage("search.found"), keys.toString()));
				} else {
					jcardsFrame.setStatus(MessageFormat.format(JCards.getMessage("search.not.found"), keys.toString()));
				}
			} else {
				jcardsFrame.setStatus(MessageFormat.format(JCards.getMessage("search.found"), keys.toString()));
			}
		} catch (PatternSyntaxException psex) {
			jcardsFrame.setStatus(MessageFormat.format(JCards.getMessage("invalid.search"), keys.toString()));
			keys.setLength(0);
		}
	}

	@Override
	public void keyPressed(KeyEvent e)
	{
		if (isSearchKey(e))
		{
			e.consume();
		}
	}

	@Override
	public void keyReleased(KeyEvent e)
	{
		if (e.getKeyCode() == VK_ENTER && !table.isEditing() && keys.length() > 0)
		{
			search();
			return;
		}

		if (isSearchKey(e))
		{
			e.consume();
		}
		else
		{
			keys.setLength(0);
			lastKeyPressTime = 0;
			if (table.isEditing()) {
				table.getCellEditor().cancelCellEditing();
			}
		}
	}
	
	@Override
	public void keyTyped(KeyEvent e)
	{
		if (!isSearchKey(e))
		{
			return;
		}

		if (lastKeyPressTime == 0) {
			lastKeyPressTime = e.getWhen();
		}

		if ((e.getWhen() - lastKeyPressTime) > TIMEOUT)
		{
			keys.setLength(0);
		}
		keys.append(e.getKeyChar());
		
		lastKeyPressTime = e.getWhen();

		search();
		
		e.consume();
	}
}
