/*
 * JCards
 * Copyright (C) 2002-2009 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package nrs.jcards.gui.table;

import java.awt.Point;
import java.awt.Toolkit;
import java.awt.datatransfer.Clipboard;
import java.awt.datatransfer.StringSelection;
import java.awt.event.*;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Pattern;
import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.JTable;
import javax.swing.SwingUtilities;
import nrs.jcards.*;
import nrs.jcards.fields.FieldDef;
import nrs.jcards.gui.JCardsFrame;
import nrs.jcards.gui.JCardsMenuItem;
import nrs.jcards.gui.action.ActionEnum;
import nrs.jcards.gui.action.ActionManager;
import nrs.jcards.gui.action.JCardsAction;

/**
 * Mouse listener for table
 *
 * @author Copyright &#169; 2002-2009 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class TableMouseListener extends MouseAdapter implements ActionListener
{
	private JCardsFrame jcardsFrame;
	private Cards cards;
	private JTable table;
	private ActionManager actions;
	private JPopupMenu popup;
	private JMenuItem urlMI,  clipMI;
	/** Position of a mouse click */
	private int x,  y;

	private final static String COPY_CLIP = "copy.clip";
	private final static String OPEN_URL = "open.url";

	// Regex used for checking if a value starts with a protocol that a web browser can open
	private static final Pattern VALID_PROTOCOL = Pattern.compile("^(https?|ftp)://\\p{Alpha}+", Pattern.CASE_INSENSITIVE);
	
	private final static Logger log = Logger.getLogger(TableMouseListener.class.getName());
	
	/** Creates a new instance of TableMouseListener */
	public TableMouseListener(JCardsFrame jcardsFrame, JTable table, Cards cards, ActionManager actions)
	{
		this.jcardsFrame = jcardsFrame;
		this.table = table;
		this.cards = cards;
		this.actions = actions;
	}

	/**
	 * Open the edit card dialog for the card at the specified point
	 * @param p Point where the user clicked on the table
	 */
	private void editCard(Point p)
	{
		log.entering("TableMouseListener", "editCard", p);
		
		int row = table.rowAtPoint(p);

		// edit card
		if (row > -1)
		{
			// set selected row as middle click won't do it
			table.getSelectionModel().setSelectionInterval(row, row);
			jcardsFrame.editCard(false, JCards.getUserSettings().isAutoEdit(), false);
		}
		
		log.exiting("TableMouseListener", "editCard");
	}
	
	/**
	 * Check for user clicking on table
	 */
	@Override
	public void mouseClicked(MouseEvent e)
	{
		log.entering("TableMouseListener", "mouseClicked", e);
		
		if (SwingUtilities.isMiddleMouseButton(e) ||
			((e.getModifiers() & InputEvent.CTRL_MASK) == InputEvent.CTRL_MASK))
		{
			editCard(e.getPoint());
		}
		else if (SwingUtilities.isRightMouseButton(e))
		{
			showPopupMenu(e);
		}
		else if (e.getClickCount() > 1 && !JCards.getUserSettings().isTableEdit())
		{
			editCard(e.getPoint());
		}
		
		log.exiting("TableMouseListener", "mouseClicked");
	}

	/** Show the popup menu */
	private void showPopupMenu(MouseEvent e)
	{
		log.entering("TableMouseListener", "showPopupMenu", e);
		
		// show popup menu
		if (popup == null)
		{
			popup = new JPopupMenu();

			JCardsAction[] actionList = {actions.getAction(ActionEnum.EDIT_CARD),
				actions.getAction(ActionEnum.DELETE_CARD),
				actions.getAction(ActionEnum.COPY_CARD)};

			JMenuItem mi;
			for (JCardsAction jca : actionList)
			{
				mi = new JCardsMenuItem(jca);
				popup.add(mi);
			}
		}

		// see if showing a URL
		if (urlMI != null)
		{
			popup.remove(urlMI);
			urlMI = null;
		}

		if (clipMI != null)
		{
			popup.remove(clipMI);
			clipMI = null;
		}

		Point pt = new Point(e.getX(), e.getY());
		int col = table.columnAtPoint(pt);
		int visrow = table.rowAtPoint(pt);
		int row = table.convertRowIndexToModel(visrow);
		if (row > -1 && col > -1)
		{
			table.setRowSelectionInterval(visrow, visrow);
			Object obj = table.getModel().getValueAt(row, col);
			if (obj != null)
			{
				String val = obj.toString();
				if (VALID_PROTOCOL.matcher(obj.toString()).find())
				{
					// check it's a well-formed URL
					try
					{
						URL u = new URL(val);
						urlMI = new JCardsMenuItem(OPEN_URL);
						urlMI.setIcon(JCardsFrame.getImage(UserSettings.ICON_OPEN_URL));
						urlMI.addActionListener(this);
						popup.add(urlMI);
					} catch (MalformedURLException mfex) {
						log.log(Level.FINER, "Invalid URL specified by user: {0}", val);
					}
				}

				// copy value to clipboard
				clipMI = new JCardsMenuItem(COPY_CLIP);
				clipMI.setIcon(JCardsFrame.getImage(UserSettings.ICON_CLIPBOARD));
				clipMI.addActionListener(this);
				popup.add(clipMI);

				x = e.getX();
				y = e.getY();
			}
			popup.show(table, e.getX(), e.getY());
		}
		
		log.exiting("TableMouseListener", "showPopupMenu");
	}

	/** User selected a URL */
	@Override
	public void actionPerformed(ActionEvent e)
	{
		log.entering("TableMouseListener", "actionPerformed", e);
		
		Point pt = new Point(x, y);
		int row = table.convertRowIndexToModel(table.rowAtPoint(pt));
		Card card = cards.getCard(row);
		assert card != null : "Cannot get card at row " + row;
		int col = table.columnAtPoint(pt);
		FieldDef fd = DBDef.getInstance().getVisibleFieldDef(col);
		assert fd != null : "Cannot get FieldDef at " + col;
		Object val = card.getValue(fd.getName());
		assert val != null : "Value is null";

		JCardsMenuItem mi = (JCardsMenuItem)e.getSource();
		if (mi.getActionCommand().equals(COPY_CLIP))
		{
			Clipboard clip = Toolkit.getDefaultToolkit().getSystemClipboard();
			StringSelection selection = new StringSelection(fd.formatValue(val));
			clip.setContents(selection, selection);
			jcardsFrame.setStatus(JCards.getMessage("copy.clip"));
		}
		else if (mi.getActionCommand().equals(OPEN_URL))
		{
			try {
				URL url = new URL(val.toString());
				jcardsFrame.displayURL(url);
			} catch (MalformedURLException mfex) {
				log.log(Level.WARNING, "Invalid user specified url "+val.toString(), mfex);
			}
		}
		
		log.exiting("TableMouseListener", "actionPerformed");
	}
}
