/*
 * JCards
 * Copyright (C) 2002-2009 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package nrs.jcards.io;

import java.io.*;
import java.util.HashMap;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.logging.Level;
import java.util.logging.Logger;
import nrs.jcards.Card;
import nrs.jcards.Cards;
import nrs.jcards.DBDef;
import nrs.jcards.JCards;
import nrs.jcards.fields.FieldDef;
import nrs.jcards.fields.FieldType;
import nrs.jcards.gui.dialog.CSVDialog;
import nrs.util.CSVListener;
import nrs.util.CSVParser;

/**
 * CSV input/output
 *
 * @author Copyright &#169; 2002-2009 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class CSVIO implements CSVListener
{
	private final static Logger log = Logger.getLogger(CSVIO.class.getName());
	private Cards cards;
	private CSVDialog csvDialog;
	private DBDef def = DBDef.getInstance();
	private CSVParser csvParser;
	private Map<String, Object> csvFields;
	private int fieldCount = 0;
	private int numOfFields = 0;
	private int oldLineNum = -1;
	private boolean multiLineValues;

	/** Creates a new instance of CSVIO */
	public CSVIO(Cards cards, CSVDialog csvDialog)
	{
		this.cards = cards;
		this.csvDialog = csvDialog;
	}

	/**
	 * Define the database's fields on the fly
	 * @param line line of fields
	 * @param fieldDelimiter character separating fields
	 */
	private void defineFields(String line, String fieldDelimiter)
	{
		log.entering("CSVIO", "defineFields", line);
		
		String fieldName;
		FieldDef newFD;

		StringTokenizer st = new StringTokenizer(line, fieldDelimiter);
		while (st.hasMoreTokens())
		{
			fieldName = st.nextToken();
			log.log(Level.FINE, "Defining new field {0}", fieldName);
			newFD = FieldDef.createFieldDef(fieldName, FieldType.STRING);
			def.addFieldDef(newFD);
		}

		numOfFields = def.getNumberOfFields();
		log.log(Level.FINE, "Defined {0} fields", numOfFields);
		
		log.exiting("CSVIO", "defineFields");
	}

	/**
	 * Import a CSV file
	 * @param file the file to import
	 * @param fDelim field delimiter
	 * @param tDelim text delimiter
	 * @param charSet character set
	 * @param header if true parse the first line as containing field names
	 * @para, multiLineValues values on multiple lines
	 * @exception CSVException error importing file
	 */
	public void importCSV(File file, String fDelim, String tDelim, String charSet,
		boolean header, boolean multiLineValues) throws CSVException
	{
		log.entering("CSVIO", "importCSV");

		assert fDelim != null : "Field delimiter is null";
		assert tDelim != null : "Text delimiter is null";

		BufferedReader reader = null;
		this.multiLineValues = multiLineValues;

		try
		{
			reader = new BufferedReader(
				new InputStreamReader(new FileInputStream(file), charSet));
			// read header row
			if (header)
			{
				log.fine("Reading header row");
				String line = reader.readLine();
				if (line == null)
				{
					throw new CSVException(JCards.getMessage("eof"));
				}

				// automatically define fields
				if (def.getNumberOfFields() == 0)
				{
					log.fine("Automatically defining fields");
					defineFields(line, fDelim);
				}
			}
			else if (def.getNumberOfFields() == 0)
			{
				throw new CSVException(JCards.getMessage("no.fields.def.msg"));
			}

			// read data
			log.finest("Reading data");
			char td = (tDelim.length() == 0 ? '\0' : tDelim.charAt(0));
			csvParser = new CSVParser(reader, fDelim.charAt(0), td, this);
			if (tDelim.length() == 0)
			{
				csvParser.setNoTextDelimiter(true);
			}
			csvParser.parse();

			if (!csvParser.hasReadAll())
			{
				throw new CSVException(JCards.getMessage("csv.read.failed"));
			}
		}
		catch (Exception ex)
		{
			throw new CSVException(JCards.getMessage("invalid.file"), ex);
		}
		finally
		{
			try
			{
				if (reader != null)
				{
					reader.close();
				}
			}
			catch (Exception ex)
			{
				log.log(Level.WARNING, "Cannot close reader", ex);
			}
			reader = null;
		}
		
		log.exiting("CSVIO", "importCSV");
	}

	/**
	 * Export data to a CSV file
	 * @param file the file to export to
	 * @param fDelim field delimiter
	 * @param tDelim text delimiter
	 * @param charSet character set
	 * @param concatNL concatenate new lines
	 * @param CSVException error exporting CSV
	 */
	public void exportCSV(File file, String fDelim, String tDelim, String charSet,
		boolean concatNL, boolean outputFieldNames) throws CSVException
	{
		log.entering("CSVIO", "exportCSV");
		
		try
		{
			PrintWriter writer = new PrintWriter(
				new BufferedWriter(
				new OutputStreamWriter(new FileOutputStream(file), charSet)));
			if (outputFieldNames)
			{
				writer.println(DBDef.getInstance().getCSV(fDelim));
				log.fine("Written db definition");
			}
			writer.println(cards.getCSV(fDelim, tDelim, concatNL));
			log.fine("Written data");
			writer.close();
		}
		catch (IOException ioex)
		{
			throw new CSVException(JCards.getMessage("error"), ioex);
		}
		finally
		{
			log.exiting("CSVIO", "exportCSV");
		}
	}

	/**
	 * A new CSV value has been found
	 * @param field the new field
	 * @param lineNum the line number the field was initially read from
	 * @param fieldNum the field number
	 */
	@Override public void newCSVValue(String value, int lineNum, int fieldNum)
	{
		Object[] params = {value, lineNum, fieldNum};
		log.entering("CSVIO", "newCSVValue", params);

		if (fieldCount == def.getNumberOfFields() ||
			(lineNum != oldLineNum && !multiLineValues && csvFields != null))
		{
			Card card = new Card(csvFields);
			cards.addCardSilently(card);
			csvFields = new HashMap<String, Object>(def.getNumberOfFields());
			fieldCount = 0;
		}

		if (csvFields == null)
		{
			csvFields = new HashMap<String, Object>(def.getNumberOfFields());
		}

		// check to see if there are too many fields
		if (oldLineNum == lineNum && fieldCount == 0)
		{
			csvParser.stop();
			log.warning("Too many fields per line");
			csvDialog.showError(JCards.getFormattedMessage("csv.too.many.fields", Integer.valueOf(lineNum + 1)));
			return;
		}

		FieldDef fd = def.getFieldDef(fieldCount++);
		Object val = fd.formatValue(value);
		if (val != null)
		{
			csvFields.put(fd.getName(), val);
		}

		oldLineNum = lineNum;

		log.exiting("CSVIO", "newCSVValue");
	}

	/** No more values will be read from the CSV file */
	@Override public void noMoreCSVValues()
	{
		log.entering("CSVIO", "noMoreCSVValues");
		
		// add the last line
		if (fieldCount == def.getNumberOfFields() && csvFields != null && !csvFields.isEmpty())
		{
			Card card = new Card(csvFields);
			cards.addCardSilently(card);
		}
		
		cards.finishedLoad();
		log.exiting("CSVIO", "noMoreCSVValues");
	}
}
