/*
 * JCards
 * Copyright (C) 2002-2009 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package nrs.jcards.io;

import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import nrs.jcards.DBDef;
import nrs.jcards.fields.DateFieldDef;
import nrs.jcards.fields.FieldDef;
import nrs.jcards.fields.FieldType;
import nrs.jcards.fields.NumberFieldDef;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

/**
 * Parse field definitions
 *
 * @author Copyright &#169; 2002-2009 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class DBDefFieldParser extends DefaultHandler
{
	private DBDef dbDef = DBDef.getInstance();
	private String defaultDateFormat;
	private boolean defaultPrepopDate = false;
	private String name;
	private Map<String, String> fieldDetails;
	private StringBuilder contents = new StringBuilder();
	private final static Logger log = Logger.getLogger(DBDefFieldParser.class.getName());

	/**
	 * Helper class to handle differences between JCards v5 files which use upper case attributes
	 * and v6 files which use lower case elements
	 * @param <K> key
	 * @param <V> value
	 */
	private class UpperKeyMap<K extends String, V extends String> extends HashMap<K, V>
	{
		@Override public V get(Object key)
		{
			String upperKey = (key == null ? null : key.toString().toUpperCase());
			return super.get(upperKey);
		}
		
		@SuppressWarnings("unchecked")
		@Override public V put(K key, V value)
		{
			String upperKey = (key == null ? null : key.toString().toUpperCase());
			return super.put((K)upperKey, value);
		}
		
		@Override
		public boolean containsKey(Object key)
		{
			String upperKey = (key == null ? null : key.toString().toUpperCase());
			return super.containsKey(upperKey);
		}
	}
	
	/**
     * Create a new field parser
     * @param defaultDateFormat default date format in old JCA format
     * @param defaultPrepopDate default date prepopulation in old JCA format
     */
	public DBDefFieldParser(String defaultDateFormat, boolean defaultPrepopDate)
	{
		this.defaultDateFormat = defaultDateFormat;
		this.defaultPrepopDate = defaultPrepopDate;
	}

	/** Start a new element */
	@Override
	public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException
	{
		log.entering("DBDefFieldParser", "startElement", qName);
		
		contents.setLength(0);
		
		if (qName.equalsIgnoreCase(Elements.FIELD))
		{
			fieldDetails = new UpperKeyMap<String, String>();
			
			// parse attributes
			String attrName;
			for (int i = 0; i < attributes.getLength(); i++)
			{
				attrName = attributes.getQName(i);
				log.log(Level.FINEST, "Attr {0}={1}", new Object[]{attrName, attributes.getValue(i)});
				if (attrName.equalsIgnoreCase(Elements.NAME))
				{
					name = attributes.getValue(i);
				}
				else
				{
					fieldDetails.put(attrName, attributes.getValue(i));
				}
			}
			
			if (name == null)
			{
				log.severe("No FieldDef name");
				throw new SAXException("No FieldDef name");
			}
		}

		log.exiting("DBDefFieldParser", "startElement");
	}

	@Override
	public void endElement(String uri, String localName, String qName) throws SAXException
	{
		log.entering("DBDefFieldParser", "endElement", qName);
		
		if (qName.equalsIgnoreCase(Elements.FIELD))
		{
			String type = fieldDetails.get(Elements.TYPE);
			if (type == null)
			{
				log.severe("No FieldDef type");
				throw new SAXException("No FieldDef type");
			}

			FieldType ft = FieldType.valueOf(type.toUpperCase());
			log.finer("Creating FieldDef");
			FieldDef fd = FieldDef.createFieldDef(name, ft);
			log.finer("Created FieldDef");

			// set any field specific values
			log.finer("Setting field XML");
			fd.setFieldDetails(fieldDetails);
			log.finer("Set field XML");
			
			// handle old files that didn't have field date formats
			if (ft.equals(FieldType.DATE) && defaultDateFormat != null)
			{
				DateFieldDef dfd = (DateFieldDef)fd;
				dfd.setDateFormat(defaultDateFormat);
				dfd.setDatePrepopulated(defaultPrepopDate);
			}

			if (type.equalsIgnoreCase("Integer"))
			{
				NumberFieldDef nfd = (NumberFieldDef)fd;
				nfd.setNumType(NumberFieldDef.NUMBER_TYPE.INTEGER);
			}
			else if (type.equalsIgnoreCase("Double"))
			{
				NumberFieldDef nfd = (NumberFieldDef)fd;
				nfd.setDecimalPlaces(2);
			}
			dbDef.addFieldDef(fd);

			fieldDetails = null;
		}
		else
		{
			if (contents != null && fieldDetails != null) {
				fieldDetails.put(qName, contents.toString());
			}
		}
		log.exiting("DBDefFieldParser", "endElement");
	}

	@Override
	public void characters(char[] ch, int start, int length) throws SAXException
	{
		if (contents != null) {
			contents.append(ch, start, length);
		}
	}
}
