/*
 * JCards
 * Copyright (C) 2002-2011 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package nrs.jcards.io;

import java.text.SimpleDateFormat;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;
import nrs.jcards.Counter;
import nrs.jcards.DBDef;
import nrs.jcards.fields.FieldDef;
import nrs.jcards.fields.FieldType;
import nrs.jcards.gui.SortDirective;
import org.jdom.Element;

/**
 * Helper class for DBDef to transform the definitions into XML
 *
 * @author Copyright &#169; 2002-2011 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class DBDefXML
{
	private final static Logger log = Logger.getLogger(DBDefXML.class.getName());
	
	/** Add counters to XML */
	private static void addCounters(DBDef def, Element dbdEl)
	{
		log.entering("DBDefXML", "addCounters");
		
		if (def.getCounters().size() > 0)
		{
			Element countersEl = new Element(Elements.COUNTERS);
			dbdEl.addContent(countersEl);
			Element countEl;
			Counter c;
			Map<String, Counter> counters = def.getCounters();
			Iterator<String> it = counters.keySet().iterator();
			while (it.hasNext())
			{
				countEl = new Element(Elements.COUNTER);
				c = counters.get(it.next());
				countEl.setAttribute(Elements.NAME, c.getName());
				countEl.setAttribute(Elements.VALUE, Double.toString(c.getValue()));
				countEl.setAttribute(Elements.INCREMENT, Double.toString(c.getIncrement()));
				countersEl.addContent(countEl);
			}
		}
		
		log.exiting("DBDefXML", "addCounters");
	}
	
	/** Add categories to XML */
	private static void addCategories(DBDef def, Element dbdEl)
	{
		log.entering("DBDefXML", "addCategories");
		
		if (def.getCategories().size() > 0)
		{
			Element catsEl = new Element(Elements.CATEGORIES);
			dbdEl.addContent(catsEl);
			String catVal;
			List<String> catItems;
			Element catEl, catItemEl;
			Iterator<String> cit;
			for (String catName : def.getCategories().keySet())
			{
				catEl = new Element(Elements.CATEGORY);
				catEl.setAttribute(Elements.NAME, catName);
				
				// add category items
				catItems = def.getCategory(catName);
				cit = catItems.iterator();
				while (cit.hasNext())
				{
					catVal = cit.next();
					catItemEl = new Element(Elements.CAT_ITEM);
					catItemEl.setText(catVal);
					catEl.addContent(catItemEl);
				}
				catsEl.addContent(catEl);
			}
		}
		
		log.exiting("DBDefXML", "addCategories");
	}
	
	/**
	 * Get the data as XML
	 * @param def DBDef DBDef reference
	 * @param sortColumns sort columns
	 * @param version5 save as version 5 file
	 * @return XMLElement
	 */
	public static Element getXML(DBDef def, SortDirective sortedColumn, boolean version5)
	{
		log.entering("DBDefXML", "getXML");
		
		Element dbdEl = new Element(Elements.DBDEF);
		
		// save settings
		Element setEl = new Element(Elements.SETTINGS);
		
		if (version5)
		{
			Element mlEl = new Element(Elements.MEMO_LINES);
			mlEl.setText("5");
			setEl.addContent(mlEl);
			
			Element dfEl = new Element(Elements.DATE_FORMAT);
			dfEl.setText((new SimpleDateFormat()).toPattern());
			setEl.addContent(dfEl);
			
			Element pdEl = new Element(Elements.PREPOPULATE_DATE);
			pdEl.setText(Boolean.TRUE.toString());
			setEl.addContent(pdEl);
			
			Element dfnEl = new Element(Elements.DISPLAY_FIELD_NAMES);
			dfnEl.setText(Boolean.toString(def.displayFieldNames()));
			setEl.addContent(dfnEl);
			
			Element sfEl = new Element(Elements.SORT_FIELD);
			sfEl.setAttribute(Elements.SORT_DIR, JCardsSortOrder.JUNSORTED.getValue());
			sfEl.setText(def.getFieldDef(0).getName());
		}
		else
		{
			Element dfnEl = new Element(Elements.DISPLAY_FIELD_NAMES);
			dfnEl.setText(Boolean.toString(def.displayFieldNames()));
			setEl.addContent(dfnEl);
			
			Element dcEl = new Element(Elements.DISPLAY_COLONS);
			dcEl.setText(Boolean.toString(def.displayColons()));
			setEl.addContent(dcEl);
		}
		
		// sort fields
		if (sortedColumn != null)
		{
			Element sortEl = new Element(Elements.SORT_FIELD);
			sortEl.setText(sortedColumn.colName);
			sortEl.setAttribute(Elements.SORT_DIR, JCardsSortOrder.getValue(sortedColumn.sortOrder));
			setEl.addContent(sortEl);
		}
		
		// visible fields
		Element visEl = new Element(Elements.VIS_FIELDS);
		dbdEl.addContent(setEl);
		Element vfEl;
		for (String fieldName : def.getVisibleFields())
		{
			vfEl = new Element(Elements.VIS_FIELD);
			vfEl.setText(fieldName);
			visEl.addContent(vfEl);
		}
		setEl.addContent(visEl);
		
		// save Categories
		addCategories(def, dbdEl);
		
		// save Counters
		addCounters(def, dbdEl);
		
		// save Fields
		Element fieldsEl = new Element(Elements.FIELDS);
		dbdEl.addContent(fieldsEl);
		Element fieldEl;
		FieldDef fd;
		Map<String, String> fieldDef;
		for (String fieldName : def.getFieldNames())
		{
			fd = def.getFieldDef(fieldName);
			fieldDef = fd.getFieldDetails();
			fieldEl = new Element(Elements.FIELD);
			
			if (version5) // v5 format using attributes
			{
				filterV5FieldDefinition(fd, fieldDef);
				for (String key: fieldDef.keySet())
				{
					fieldEl.setAttribute(key, fieldDef.get(key));
				}
			}
			else // modern format using elements
			{
				fieldEl.setAttribute(Elements.NAME, fieldDef.get(Elements.NAME));
				Element fdEl;
				for (String key: fieldDef.keySet())
				{
					if (!key.equals(Elements.NAME))
					{
						fdEl = new Element(key).addContent(fieldDef.get(key));
						fieldEl.addContent(fdEl);
					}
				}
			}
			fieldsEl.addContent(fieldEl);
		}
		
		log.exiting("DBDefXML", "getXML");
		
		return dbdEl;
	}
	
	/**
	 * Filter a field definition so that it is suitable for saving as a V5 file
	 * @param fd FieldDef
	 * @param fieldDefinition Map of name/value pairs
	 */
	private static void filterV5FieldDefinition(FieldDef fd, Map<String, String> fieldDefinition)
	{
		switch (fd.getType())
		{
			case CALCULATED:
			case COUNTER:
				fieldDefinition.put(Elements.TYPE, FieldType.STRING.getName());
				fieldDefinition.remove(Elements.DECIMAL_PLACES);
				fieldDefinition.remove(Elements.NUM_TYPE);
				break;
				
			case DATE:
				fieldDefinition.remove(Elements.DATE_FORMAT);
				fieldDefinition.remove(Elements.PREPOPULATE_DATE);
				break;
				
			case IMAGE:
				fieldDefinition.put(Elements.TYPE, FieldType.STRING.getName());
				fieldDefinition.remove(Elements.AUTO_FOLD);
				break;
				
			case MEMO:
				fieldDefinition.remove(Elements.DISPLAY_AS_HTML);
				fieldDefinition.remove(Elements.AUTO_FOLD);
				break;
				
			case NUMBER:
				fieldDefinition.put(Elements.TYPE, FieldType.STRING.getName());
				fieldDefinition.remove(Elements.DECIMAL_PLACES);
				fieldDefinition.remove(Elements.NUM_TYPE);
				break;
		}
	}
}
