/*
 * JCards
 * Copyright (C) 2002-2011 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package nrs.jcards.io;

import java.io.*;
import java.util.*;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.zip.*;
import nrs.encryption.Decryptor;
import nrs.encryption.Encryptor;
import nrs.jcards.Cards;
import nrs.jcards.DBDef;
import nrs.jcards.JCards;
import nrs.jcards.fields.ImageFieldDef;
import nrs.jcards.gui.SortDirective;
import org.bouncycastle.util.encoders.Base64Encoder;
import org.jdom.Element;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;
import org.xml.sax.InputSource;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.XMLReaderFactory;

/**
 * Class to read and write the JCards file format.
 * <p />
 * Encrypted files are:
 * <ol>
 * <li>zipped first to ensure a minimum file size</li>
 * <li>encrypted</li>
 * <li>written to the zip file</li>
 * </ol>
 *
 * Unencrypted files are written directly to the zip file as characters.
 * 
 * @author Copyright Copyright &#169; 2002-2011 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class JCAReader
{
	private JCards jcards;
	private SortDirective sortedColumn;
	private List<Integer> colWidths;
	private static int SALT_SIZE = 128;
	private final static Logger log = Logger.getLogger(JCAReader.class.getName());
	
	/** Zipfile entry constants */
	private final static String CARD_DATA = "Card Data";
	private final static String DIGEST = "Digest";
	private final static String HEADER = "Header";
	private static final String PICS_DIR = "pics/";
	
	/** Header constants */
	public final static String ENCRYPTED = "Encrypted";
	public final static String VERSION = "Version";
	public final static String CREATED = "Created";
	public final static String MODIFIED = "Modified";
	public final static String SALT = "Salt";
	
	/** Encodings */
	private final static String ASCII = "ASCII";
	public final static String UTF8 = "UTF-8";
	public final static String PROLOG = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n\n";
	
	/** XML Elements */
	public  final static String JCARDS_NS = "http://www.nsydenham.net/java/JCards/6";
	private final static String JCARDS_START = "<JCards xmlns=\""+JCARDS_NS+"\">";
	private final static String JCARDS_END = "</JCards>";

	public static class SplitFilename
	{
		public String name;
		public String extension;
	}

	/** Construct the reader with a reference to the main class for logging, etc */
	public JCAReader(JCards jcards)
	{
		this.jcards = jcards;
	}
	
	/**
	 * Read the bytes from a zip file entry
	 * @param zf ZipFIle
	 * @param zeName name of the zip entry to read
	 * @exception IOException can't read from file
	 */
	private static byte[] readZipBytes(ZipFile zf, String zeName) throws IOException, JCAException
	{
		Object[] params = { zf, zeName };
		log.entering("JCAReader", "readZipBytes", params);
		ZipEntry ze = zf.getEntry(zeName);
		if (ze == null)
		{
			throw new JCAException(JCards.getMessage("invalid.file"));
		}
		log.exiting("JCAReader", "readZipBytes");
		
		return readInputStream(zf.getInputStream(ze));
	}
	
	/**
	 * Read bytes from an InputStream
	 * @param is InputStream
	 * @return bytes read from stream
	 * @exception IOException couldn't read from stream
	 */
	private static byte[] readInputStream(InputStream is) throws IOException
	{
		log.entering("JCAReader", "readInputStream");
		BufferedInputStream bis = new BufferedInputStream(is);
		ByteArrayOutputStream baos = new ByteArrayOutputStream(bis.available());
		int bytesRead;
		byte[] bytes = new byte[4096];
		while ((bytesRead = bis.read(bytes)) > -1)
		{
			baos.write(bytes, 0, bytesRead);
		}
		baos.flush();
		baos.close();
		bis.close();
		
		log.exiting("JCAReader", "readInputStream");
		return baos.toByteArray();
	}
	
	/**
	 * Write bytes to a zip file
	 * @param zos ZipOutputStream
	 * @param zeName name of the zip entry
	 * @param bytes data to write
	 * @exception IOException can't write to file
	 */
	private void writeZipBytes(ZipOutputStream zos, String zeName, byte[] bytes) throws IOException
	{
		log.entering("JCAReader", "writeZipBytes");
		ZipEntry ze = new ZipEntry(zeName);
		zos.putNextEntry(ze);
		zos.write(bytes, 0, bytes.length);
		zos.closeEntry();
		log.exiting("JCAReader", "writeZipBytes");
	}

	/**
	 * Copy any images across to the file being saved to and update field values
	 * @param zos file being written to
	 * @param origFile original file containing any existing images
	 * @param imageFieldDefs set of image field defs
	 */
	private void saveImages(ZipOutputStream zos, File origFile, Set<ImageFieldDef> imageFieldDefs, Cards cards)
	{
		Map<String, String> jcaFileNames = new HashMap<String, String>();
		for (ImageFieldDef ifd : imageFieldDefs)
		{
			if (ifd.isStoreInDB())
			{
				Map<String, String> imageFileNames = cards.setDBImageFileNamesBeforeSave(ifd.getName());
				jcaFileNames.putAll(imageFileNames);
			}
		}

		// copy images from file system into DB
		String path, zipPath;
		byte[] buffer = new byte[65536];
		int readBytes;
		ZipEntry newZE;
		for (String jcaFileName : jcaFileNames.keySet())
		{
			path = jcaFileNames.get(jcaFileName);
			if (path != null)
			{
				zipPath = PICS_DIR+jcaFileName.substring(ImageFieldDef.PIC.length());
				try
				{
					newZE = new ZipEntry(zipPath);
					newZE.setComment(path);
					zos.putNextEntry(newZE);
					BufferedOutputStream bos = new BufferedOutputStream(zos);
					BufferedInputStream bis = new BufferedInputStream(new FileInputStream(path));
					while ((readBytes = bis.read(buffer)) != -1) {
						bos.write(buffer, 0, readBytes);
					}
					bis.close();
					bos.flush();
					zos.closeEntry();
				} catch (FileNotFoundException fnex) {
					log.log(Level.WARNING, "File appears to have moved "+path, fnex);
				} catch (IOException ioex) {
					log.log(Level.SEVERE, "Cannot copy image into DB "+path, ioex);
				}
			}
		}

		// move across existing images that haven't been deleted
		try
		{
			ZipFile zipFile = new ZipFile(origFile);
			ZipEntry ze;
			String name;
			for (String jcaFileName : jcaFileNames.keySet())
			{
				path = jcaFileNames.get(jcaFileName);
				if (path == null)
				{
					name = PICS_DIR+jcaFileName.substring(ImageFieldDef.PIC.length());
					ze = zipFile.getEntry(name);
					if (ze != null)
					{
						newZE = new ZipEntry(name);
						zos.putNextEntry(newZE);
						BufferedOutputStream bos = new BufferedOutputStream(zos);
						BufferedInputStream bis = new BufferedInputStream(zipFile.getInputStream(ze));
						while ((readBytes = bis.read(buffer)) != -1) {
							bos.write(buffer, 0, readBytes);
						}
						bis.close();
						bos.flush();
						zos.closeEntry();
					}
				}
			}
			zipFile.close();
		} catch (ZipException zex) {
			log.log(Level.SEVERE, "Error saving to file", zex);
		} catch (IOException ioex) {
			log.log(Level.SEVERE, "Error saving to file", ioex);
		}
	}

	private File getTempFile(File originalFile, String label)
	{
		SplitFilename sf = splitFilename(originalFile.getName());
		String ext = (sf.extension == null ? "tmp" : sf.extension);
		String name = sf.name + label + "." + ext;
		File tempFile = new File(originalFile.getParent(), name);
		tempFile.deleteOnExit();

		return tempFile;
	}

	/**
	 * Save the data to the specified file
	 * @param saveFile file to save to
	 * @param password password to encrypt with
	 * @param colWidths column widths
	 * @exception Exception save didn't work
	 */
	public void saveFile(File saveFile, String password, Element colWidths, boolean version5)
		throws FileNotFoundException, UnsupportedEncodingException, IOException, JCAException
	{
		log.entering("JCAReader", "saveFile", new Object[]{saveFile, password});

		// save to temp file first in case something goes wrong
		File tempFile = getTempFile(saveFile, "-temp");
		log.log(Level.FINEST, "Saving to temp file {0}", tempFile.getAbsolutePath());

		DBDef def = DBDef.getInstance();
		ZipOutputStream zos = new ZipOutputStream(new FileOutputStream(tempFile));
		ZipOutputStream encZos = null;
		BufferedWriter writer;
		Encryptor enc = null;
		ZipEntry ze;
		
		// write header
		StringBuilder sb = new StringBuilder("# JCards - http://www.nsydenham.net/java/JCards/\n");
		if (version5) {
			sb.append(VERSION+": "+JCards.VERSION5+"\n");
		} else {
			sb.append(VERSION+": "+JCards.VERSION+"\n");
		}
		String p = (password != null ? "true" : "false");
		sb.append(ENCRYPTED + ": ").append(p).append("\n");
		sb.append(CREATED + ": ").append(jcards.getCreationDate()).append("\n");
		sb.append(MODIFIED + ": ").append(jcards.getModifiedDate()).append("\n");
		byte[] salt = new byte[SALT_SIZE];
		if (password != null && password.length() > 0)
		{
			sb.append(DIGEST + ": ").append(Encryptor.DIGEST.SHA256).append("\n");
			Random rand = new Random();
			rand.nextBytes(salt);
			Base64Encoder b64e = new Base64Encoder();
			ByteArrayOutputStream baos = new ByteArrayOutputStream(salt.length);
			b64e.encode(salt, 0, salt.length, baos);
			sb.append(SALT + ": ").append(new String(baos.toByteArray(), UTF8)).append("\n");
		}
		
		byte[] headerBytes = sb.toString().getBytes(ASCII);
		writeZipBytes(zos, HEADER, headerBytes);
		headerBytes = null;
		log.finer("Written header");

		// write images, if set to store in DB, and update ImageFieldDef entries
		saveImages(zos, jcards.getCardFile(), def.getImages(), jcards.getCards());
		
		// do we need to encrypt it
		if (password != null && password.length() > 0)
		{
			log.finer("Encrypting");
			// zip before encrypting to ensure a smaller file size
			ByteArrayOutputStream encBytes = new ByteArrayOutputStream(SALT_SIZE);
			encBytes.write(salt);
			encBytes.write(password.getBytes(UTF8));
			enc = new Encryptor(encBytes.toByteArray(), true);
			enc.setOutputStream(zos);

			writeZipBytes(zos, DIGEST, Encryptor.getDigest(encBytes.toByteArray(), Encryptor.DIGEST.SHA256));
			log.finer("Written digest");
			
			ze = new ZipEntry(CARD_DATA);
			zos.putNextEntry(ze);
			
			encZos = new ZipOutputStream(enc);
			ZipEntry encZe = new ZipEntry("encrypted");
			encZos.putNextEntry(encZe);
			
			writer = new BufferedWriter(new OutputStreamWriter(encZos, UTF8));
		}
		else
		{
			ze = new ZipEntry(CARD_DATA);
			zos.putNextEntry(ze);
			writer = new BufferedWriter(new OutputStreamWriter(zos, UTF8));
		}
		
		writer.write(PROLOG);
		writer.write(JCARDS_START);
		
		// write DBDef
		XMLOutputter xmlOut = new XMLOutputter(Format.getRawFormat());
		xmlOut.output(DBDefXML.getXML(def, jcards.getFrame().getSortedColumn(), version5), writer);
		log.finer("Written DBDef");
		
		// write Cards
		jcards.getCards().writeCardsToWriter(xmlOut, writer, false, false);
		log.finer("Written Cards");
		
		// write ColWidths
		xmlOut.output(colWidths, writer);
		log.finer("Written ColWidths");
		
		writer.write(JCARDS_END);
		writer.close();
		log.finest("Finished saving");
		
		// move temp file back over original
		if (!tempFile.renameTo(saveFile))
		{
			// can't overwrite using a rename operation
			log.log(Level.WARNING, "Cannot overwrite original file {0} with {1}", new Object[]{saveFile.getAbsolutePath(), tempFile.getAbsolutePath()});

			File newTempFile = getTempFile(saveFile, "-bak");
			if (saveFile.renameTo(newTempFile)) // should succeed as new file path doesn't exist
			{
				tempFile.renameTo(saveFile);
				newTempFile.delete();
				log.finest("Backed up original and rename worked");
			}
			else // oh dear!
			{
				throw new JCAException("Cannot save to file "+saveFile.getAbsolutePath());
			}
		}

		log.exiting("JCAReader", "saveFile");
	}
	
	/**
	 * Check if the file is a JCA file
	 * @param file the file to check
	 * @return true if it's a JCA file, false otherwise
	 */
	public static boolean isJCAFile(File file)
	{
		log.entering("JCAReader", "isJCAFile", file.getPath());
		
		boolean jcaFile = false;
		if (file.exists())
		{
			String fileName = file.getName();
			if (fileName.endsWith(".jca") || fileName.endsWith(".jc5"))
			{
				jcaFile = true;
			}
			else
			{
				log.warning("File doesn't end with .jca or .jc5");
			}

			try
			{
				ZipFile zf = new ZipFile(file);
				ZipEntry ze = zf.getEntry(HEADER);
				if (ze == null) {
					jcaFile = false;
				}
				zf.close();
			}
			catch (IOException ioex)
			{
				log.log(Level.WARNING, "{0} isn''t a zip file", file.getAbsolutePath());
				jcaFile = false;
			}
		}
		
		log.exiting("JCAReader", "isJCAFile", jcaFile);
		return jcaFile;
	}

	/**
	 * Check whether the specified file is encrypted
	 * @param file the file to check
	 * @return true if encrypted; false otherwise
	 * @exception IOException cannot open the file
	 */
	public static boolean isEncrypted(File file) throws IOException
	{
		log.entering("JCAReader", "isEncrypted", file);
		
		if (!isJCAFile(file))
		{
			throw new IOException(JCards.getMessage("not.jcards.file"));
		}
		
		boolean encrypted = false;
		Map<String, String> headerFields = readHeader(file);
		if (headerFields.containsKey(ENCRYPTED))
		{
			encrypted = Boolean.parseBoolean(headerFields.get(ENCRYPTED));
		}

		log.exiting("JCAReader", "isEncrypted", encrypted);
		return encrypted;
	}
	
	/**
	 * Check if a password is valid
	 * @param file
	 * @param password
	 * @return true if a valid password, false otherwise
	 * @exception IOException cannot open the file
	 */
	public static boolean checkPassword(File file, String password) throws JCAException
	{
		assert password != null : "Cannot check a null password";

		log.entering("JCAReader", "checkPassword", file);

		if (!isJCAFile(file))
		{
			throw new JCAException(JCards.getMessage("not.jcards.file"));
		}

		Map<String, String> header = readHeader(file);
		boolean valid = false;
		try
		{
			ByteArrayOutputStream baos = new ByteArrayOutputStream(SALT_SIZE+12);
			if (header.containsKey(SALT))
			{
				Base64Encoder b64e = new Base64Encoder();
				byte[] saltBytes = header.get(SALT).getBytes(UTF8);
				b64e.decode(saltBytes, 0, saltBytes.length, baos);
			}
			baos.write(password.getBytes(UTF8));

			// figure out digest to use
			String digestName = Encryptor.DIGEST.MD5.toString();
			if (header.get(DIGEST) != null)	{
				digestName = header.get(DIGEST);
			}
			Encryptor.DIGEST digestType = Encryptor.DIGEST.valueOf(digestName);

			// get the digest for the entered password
			String pdigest = new String(Encryptor.getDigest(baos.toByteArray(), digestType), UTF8);

			// check with the digest stored in the JCA file
			ZipFile zf = new ZipFile(file);
			String digest = new String(readZipBytes(zf, DIGEST), UTF8);
			valid = pdigest.equals(digest);
		}
		catch (Exception ex)
		{
			log.log(Level.WARNING, "Cannot check password", ex);
			throw new JCAException(JCards.getFormattedMessage("cannot.open.file", file.getName()));
		}

		log.exiting("JCAReader", "checkPassword", valid);
		return valid;
	}

	/**
	 * Read JCA Header details
	 * @param file the file to read
	 * @return a map of properties
	 */
	public static Map<String, String> readHeader(File file)
	{
		log.entering("JCAReader", "readHeader", file);
		
		Map<String, String> props = new HashMap<String, String>();
		if (file != null && file.exists() && isJCAFile(file))
		{
			ZipFile zf = null;
			try
			{
				zf = new ZipFile(file);
				// read header
				ZipEntry headZE = zf.getEntry(HEADER);
				// header isn't mandatory but should be present
				if (headZE != null)
				{
					log.finer("Reading header");
					String line;
					int colonIndex;
					BufferedReader reader = new BufferedReader(new InputStreamReader(zf.getInputStream(headZE), ASCII));
					while ((line = reader.readLine()) != null)
					{
						log.finest(line);
						if (line.trim().length() > 0 && !line.startsWith("#"))
						{
							colonIndex = line.indexOf(':');
							if (colonIndex > -1)
							{
								props.put(line.substring(0, colonIndex), line.substring(colonIndex+2));
							}
						}
					}
					reader.close();
				}
			}
			catch (Exception zex)
			{
				log.log(Level.WARNING, "Cannot read JCA Header", zex);
			}
			finally
			{
				if (zf != null) {
					try {
						zf.close();
					} catch (Exception ex) {
						log.log(Level.WARNING, "Cannot close zip file", ex);
					}
				}
			}
		}		
		log.exiting("JCAReader", "readHeader", props);
		
		return props;
	}
	
	public SortDirective getSortedColumn()
	{
		return sortedColumn;
	}
	
	public List<Integer> getColumnWidths()
	{
		return colWidths;
	}
	
	/**
	 * Open the specified file
	 * @param file file to open
	 * @param password optional password
	 * @exception Exception couldn't open file
	 */
	public void openFile(File file, String password) throws JCAException
	{
		boolean encrypted = false;
		boolean oldFile = false;
		byte[] salt = new byte[0];
		
		log.entering("JCAReader", "openFile", file);
		
		assert file != null : "Cannot open null file";
		
		// check it's a JCA file
		if (!isJCAFile(file))
		{
			throw new JCAException(JCards.getMessage("not.jcards.file"));
		}
		
		try
		{
			Map<String, String> headerProps = readHeader(file);
			// header isn't mandatory but should be present
			if (!headerProps.isEmpty())
			{
				if (headerProps.containsKey(VERSION) && headerProps.get(VERSION).startsWith("3"))
				{
					oldFile = true;
				}
				
				if (headerProps.containsKey(ENCRYPTED))
				{
					encrypted = Boolean.parseBoolean(headerProps.get(ENCRYPTED));
				}
				
				if (headerProps.containsKey(CREATED))
				{
					try
					{
						jcards.setCreationDate(Long.parseLong(headerProps.get(CREATED)));
					}
					catch (Exception ex)
					{
						log.log(Level.WARNING, "Invalid creation date", ex);
					}
				}
				
				if (headerProps.containsKey(MODIFIED))
				{
					try
					{
						jcards.setModifiedDate(Long.parseLong(headerProps.get(MODIFIED)));
					}
					catch (Exception ex)
					{
						log.log(Level.WARNING, "Invalid modification date", ex);
					}
				}

				if (headerProps.containsKey(SALT))
				{
					try
					{
						salt = headerProps.get(SALT).getBytes(UTF8);
					}
					catch (Exception ex)
					{
						log.log(Level.SEVERE, "Cannot read salt", ex);
					}
				}
			}
			log.finer("Read header");
			
			// open the file
			ZipFile zf = new ZipFile(file);
			
			// check if password is required
			if (encrypted && (password == null || password.length() == 0))
			{
				throw new JCAException(JCards.getMessage("missing.password"));
			}
			
			log.finer("Reading db definition");
			ZipEntry ze = zf.getEntry(CARD_DATA);
			if (ze == null)
			{
				throw new JCAException(JCards.getMessage("invalid.file"));
			}
			
			InputStream inputStream = zf.getInputStream(ze);
			if (encrypted)
			{
				ByteArrayOutputStream baos = new ByteArrayOutputStream(SALT_SIZE+12);
				if (salt.length > 0)
				{
					Base64Encoder b64e = new Base64Encoder();
					b64e.decode(salt, 0, salt.length, baos);
				}
				baos.write(password.getBytes(UTF8));
				Decryptor dec = new Decryptor(inputStream, baos.toByteArray());
				ZipInputStream zis = new ZipInputStream(dec);
				zis.getNextEntry();
				inputStream = zis;
			}
			XMLReader reader = XMLReaderFactory.createXMLReader();
			JCASAXParser jcaParser = new JCASAXParser(jcards);
			reader.setContentHandler(jcaParser);
			InputSource inputSource = new InputSource(inputStream);
			if (!oldFile)
			{
				inputSource.setEncoding(UTF8);
			}
			reader.parse(inputSource);
			zf.close();
			
			sortedColumn = jcaParser.getSortedColumn();
			colWidths = jcaParser.getColWidths();
		}
		catch (ZipException zipex)
		{
			log.log(Level.SEVERE, "Zip exception opening file", zipex);
			throw new JCAException(zipex.getMessage());
		}
		catch (IOException ioex)
		{
			log.log(Level.SEVERE, "Error opening file", ioex);
			throw new JCAException(ioex.getMessage());
		}
		catch (Exception ex)
		{
			log.log(Level.SEVERE, "Error reading jca file", ex);
			throw new JCAException(ex.getMessage(), ex);
		}
		finally
		{
			log.exiting("JCAReader", "openFile");
		}
	}

	/**
	 * Load an image from a JCards DB
	 * @param cardFile JCards DB file
	 * @param imgName name of the image - internal JCards reference
	 * @return InputStream may be null if the picture can't be found
	 * @throws JCAException
	 */
	public static InputStream getImageInputStream(File cardFile, String imgName) throws JCAException
	{
		log.entering("JCAReader", "getImageInputStream", imgName);

		assert cardFile != null : "Null file specified";
		assert imgName != null : "Cannot get a null image";
		
		try
		{
			ZipFile zf = new ZipFile(cardFile);
			String picPath = PICS_DIR+imgName.substring(ImageFieldDef.PIC.length());
			ZipEntry ze = zf.getEntry(picPath);

			if (ze == null)
			{
				log.log(Level.SEVERE, "Cannot load zip image {0}", imgName);
				throw new JCAException("Missing zip image "+imgName);
			}
			else {
				return zf.getInputStream(ze);
			}
		}
		catch (ZipException zex)
		{
			log.log(Level.SEVERE, "Zip exception opening file "+cardFile.getAbsolutePath(), zex);
			throw new JCAException(zex.getMessage());
		}
		catch (IOException ioex)
		{
			log.log(Level.SEVERE, "Error opening file "+cardFile.getAbsolutePath(), ioex);
			throw new JCAException(ioex.getMessage());
		}
		finally
		{
			log.exiting("JCAReader", "getImageInputStream");
		}
	}

	/**
	 * Split a file name into a name and an extension
	 * @param filename
	 * @return SplitFilename
	 */
	public static SplitFilename splitFilename(String filename)
	{
		SplitFilename sf = new SplitFilename();
		int index = filename.lastIndexOf(".");
		if (index > 0)
		{
			sf.name = filename.substring(0, index);
			sf.extension = filename.substring(index+1);
		}
		else
		{
			sf.name = filename;
		}
		return sf;
	}
}
