/*
 * JCards
 * Copyright (C) 2002-2011 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package nrs.jcards.io;

import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import nrs.jcards.JCards;
import nrs.jcards.gui.SortDirective;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.helpers.DefaultHandler;

/**
 * SAX Factory for determining which parser to use for which parts of a document
 *
 * @author Copyright &#169; 2002-2011 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class JCASAXParser extends DefaultHandler
{
	private JCards jcards;
	private DefaultHandler subParser;
	private String defaultDateFormat;
	private boolean defaultPrepopDate = false;
	private SortDirective sortedColumn;
	private List<Integer> colWidths;
	private final static Logger log = Logger.getLogger(JCASAXParser.class.getName());
	
	/**
	 * Creates a new instance of JCASAXFactory
	 * @param parser SAX parser
	 */
	public JCASAXParser(JCards jcards)
	{
		this.jcards = jcards;
	}
	
	@Override
	public void characters(char[] ch, int start, int length) throws org.xml.sax.SAXException
	{
		if (subParser != null)
		{
			subParser.characters(ch, start, length);
		}
	}
	
	@Override
	public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException
	{
		log.entering("JCASAXParser", "startElement", qName);
		
		if (qName.equalsIgnoreCase(Elements.SETTINGS))
		{
			subParser = new DBDefSettingsParser();
		}
		else if (qName.equalsIgnoreCase(Elements.CATEGORIES))
		{
			subParser = new DBDefCategoryParser();
		}
		else if (qName.equalsIgnoreCase(Elements.COUNTERS))
		{
			subParser = new DBDefCountersParser();
		}
		else if (qName.equalsIgnoreCase(Elements.FIELDS))
		{
			subParser = new DBDefFieldParser(defaultDateFormat, defaultPrepopDate);
		}
		else if (qName.equalsIgnoreCase(Elements.CARDS))
		{
			subParser = new CardsParser(jcards);
		}
		else if (qName.equalsIgnoreCase(Elements.COL_WIDTHS))
		{
			subParser = new ColWidthsParser();
		}
		else if (subParser != null)
		{
			subParser.startElement(uri, localName, qName, attributes);
		}
		
		log.exiting("JCASAXParser", "startElement");
	}
	
	@Override
	public void endElement(String uri, String localName, String qName) throws SAXException
	{
		log.entering("JCASAXParser", "endElement", qName);
		
		if (qName.equalsIgnoreCase(Elements.SETTINGS))
		{
			DBDefSettingsParser settingsParser = (DBDefSettingsParser)subParser;
			defaultDateFormat = settingsParser.getDefaultDateFormat();
			defaultPrepopDate = settingsParser.isDefaultPrepopDate();
			sortedColumn = settingsParser.getSortedColumn();
		}
		else if (qName.equalsIgnoreCase(Elements.COL_WIDTHS))
		{
			ColWidthsParser cwParser = (ColWidthsParser)subParser;
			colWidths = cwParser.getColumnWidths();
		}
		else if (subParser != null)
		{
			subParser.endElement(uri, localName, qName);
		}
		
		log.exiting("JCASAXParser", "endElement");
	}
	
	/**
	 * Table sort columns
	 * @return List of SortDirectives
	 */
	public SortDirective getSortedColumn()
	{
		return sortedColumn;
	}
	
	/**
	 * Table column widths
	 * @return List of widths
	 */
	public List<Integer> getColWidths()
	{
		return colWidths;
	}
	
	@Override
	public void warning(SAXParseException e) throws SAXException
	{
		String msg = "Warning at "+e.getLineNumber()+","+e.getColumnNumber()+"-"+e.getMessage();
		log.log(Level.WARNING, msg, e);
	}
	
	@Override
	public void error(SAXParseException e) throws SAXException
	{
		String msg = "Warning at "+e.getLineNumber()+","+e.getColumnNumber()+"-"+e.getMessage();
		log.log(Level.SEVERE, msg, e);
	}
}
