/*
 * PostfixCalculator
 * Copyright (C) 2003-2008 Nick Sydenham <nsydenham@yahoo.co.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package nrs.util;

import java.lang.reflect.*;
import java.util.*;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * Postfix calculator.
 * <p>
 * Specify a calculation using postfix notation and then call calculate to get the value.
 * In addition to numbers and operators this calculator also understands functions from
 * the java.lang.Math class which take one or two <i>double</i> arguments. It also
 * understands the random function which takes no arguments.
 * </p>
 * <p>To see an example of how this class works try the following:</p>
 * <p><em>java nrs.util.PostfixCalculator "36 2 * toRadians cos"</em></p>
 * <p>which in normal terms can be translated as cos(toRadians(36 * 2)) with an answer of
 * 0.30901699437494745.</p>
 * <p>In programmatic use, you would create a new instance of PostfixCalculator and then
 * use the <em>add*</em> methods to add numbers and operands to the stack. The alternative
 * method is to use the <em>parseCalculation</em> method which will parse a space separated String.
 * The <em>getResult</em> method then returns the answer as a double.</p>
 *
 * <p>The supported mathematical operators are:
 * <table border="1">
 * <tr><th>Operator</th><th>Usage</th><th>Description</th><th>Example</th></tr>
 * <tr><th>+</th><td>op1 op2 +</td><td>Add</td><td>3 2 + = 5</td></tr>
 * <tr><th>-</th><td>op1 op2 -</td><td>Subtract</td><td>3 2 - = 1</td></tr>
 * <tr><th>*</th><td>op1 op2 *</td><td>Multiply</td><td>3 2 * = 6</td></tr>
 * <tr><th>/</th><td>op1 op2 /</td><td>Divide<td></td><td>3 2 / = 1.5</td></tr>
 * <tr><th>%</th><td>op1 op2 %</td><td>Remainder</td><td>3 2 % = 1</td></tr>
 * <tr><th>&</th><td>op1 op2 &</td><td>Bitwise AND</td><td>3 2 & = 2</td></tr>
 * <tr><th>|</th><td>op1 op2 |</td><td>Bitwise OR</td><td>3 2 | = 3</td></tr>
 * <tr><th>^</th><td>op1 op2 ^</td><td>Bitwise XOR</td><td>3 2 ^ = 1</td></tr>
 * <tr><th>p</th><td>op1 op2 p</td><td>Power</td><td>3 2 p = 9</td></tr>
 * <tr><th>~</th><td>op1 ~</td><td>Bitwise complement</td><td>3 ~ = -4</td></tr>
 * <tr><th>&lt;&lt;</th><td>op1 op2 &lt;&lt;</td><td>Shift bits of op1 left by distance op2; fills with zero bits on the right-hand side</td><td>3 2 &lt;&lt; = 12</td></tr>
 * <tr><th>&gt;&gt;</th><td>op1 op2 &gt;&gt;</td><td>Shift bits of op1 right by distance op2; fills with highest (sign) bit on the left-hand side</td><td>12 2 &gt;&gt; = 3</td></tr>
 * <tr><th>&gt;&gt;&gt;</td><td>op1 op2 &gt;&gt;&gt;</td><td>Shift bits of op1 right by distance op2; fills with zero bits on the left-hand side</td><td>12 2 &gt;&gt;&gt; = 3</td></tr>
 * </table>
 * </p>
 * 
 * <p>The supported mathematical constants are PI and E.</p>
 *
 * @author Copyright &#169; 2003-2008 Nick Sydenham &lt;nsydenham@yahoo.co.uk&gt;
 */
public class PostfixCalculator
{
	/** Allowed operands are + - / * % & | ^ p ~ << >> <<< */
	private final static String[] OPERANDS = {"+", "-", "/", "*", "%", "&", "|", "^", "p", "~", "<<", ">>", ">>>"};
	
	private final static String[] CONSTANTS = {"E", "PI"};
	
	private Stack<Object> stack = new Stack<Object>();
	
	private final static Logger log = Logger.getLogger(PostfixCalculator.class.getName());
	
	/** Cache how many arguments are required by each Math method */
	private static Map<String, Integer> args = new HashMap<String, Integer>(22);
	
	static
	{
		log.setLevel(Level.ALL);
		
		// sort Operands
		Arrays.sort(OPERANDS);
		
		Method[] methods = Math.class.getDeclaredMethods();
		Method method;
		Class[] methodArgs;
		
		for (int i=0; i<methods.length; i++)
		{
			method = methods[i];
			methodArgs = method.getParameterTypes();
			if (methodArgs.length == 1 && methodArgs[0].equals(double.class))
			{
				args.put(method.getName(), 1);
			}
			else if (methodArgs.length == 2 &&
					methodArgs[0].equals(double.class) && methodArgs[1].equals(double.class))
			{
				args.put(method.getName(), 2);
			}
			else if (methodArgs.length == 0 && method.getName().equals("random"))
			{
				args.put(method.getName(), 0);
			}
		}
	}
	
	/**
	 * Checks to see if the specified parameter is an operand
	 * @param param parameter to check
	 * @return true if it is an operand, false otherwise
	 */
	public static boolean isOperand(String param)
	{
		return (param != null && Arrays.binarySearch(OPERANDS, param) > -1);
	}
	
	/**
	 * Checks to see if the specified parameter is a valid constant. Valid constants
	 * are PI and E and they have to be uppercase.
	 * @param param parameter to check
	 * @return true if it is a constant, false otherwise
	 */
	public static boolean isConstant(String param)
	{
		return (param != null && Arrays.binarySearch(CONSTANTS, param) > -1);
	}
	
	/**
	 * Checks to see if the specified parameter is a java.lang.Math method
	 * @param param parameter to check
	 * @return true if it is a Math method, false otherwise
	 */
	public static boolean isMathFunction(String param)
	{
		return (param != null && args.get(param) != null);
	}
	
	/**
	 * Add edither a number or operator to the stack
	 * @param item either Number or String (operand or Math method)
	 * @exception IllegalArgumentException item wasn't a Number, operand or Math method
	 */
	private void add(Object item) throws IllegalArgumentException
	{
		log.entering("PostfixCalculator", "add", item);
		
		if (item == null)
		{
			throw new NullPointerException("Item is null");
		}
		
		if (item instanceof Number)
		{
			addNumber((Number)item);
			return;
		}
		
		if (item instanceof String)
		{
			try
			{
				addNumber(Double.valueOf(item.toString()));
				return;
			}
			catch (NumberFormatException nfex)
			{
				// log.log(Level.WARNING, "Cannot add number "+item, nfex);
			}
			
			String sitem = item.toString();
			if (isOperand(sitem))
			{
				addCheckedOperand(sitem);
				return;
			}
			else if (isConstant(sitem))
			{
				addConstant(sitem);
				return;
			}
			else if (isMathFunction(sitem))
			{
				addMathFunction(sitem);
				return;
			}
		}
		
		throw new IllegalArgumentException(item+" isn't a number, operand or function");
	}
	
	/** Add a long */
	private void add(long num)
	{
		addNumber(Long.valueOf(num));
	}
	
	/** Add a double */
	private void add(double num)
	{
		addNumber(Double.valueOf(num));
	}
	
	/**
	 * Add a number to the stack
	 * @param num Number to add
	 */
	private void addNumber(Number num)
	{
		log.entering("PostfixCalculator", "addNumber", num);
		stack.push(num);
	}
	
	private void addConstant(String constant)
	{
		log.entering("PostfixCalculator", "addConstant", constant);
		if (constant.equals("PI")) {
			stack.push(Math.PI);
		} else if (constant.equals("E")) {
			stack.push(Math.E);
		}
	}
	
	private void addCheckedOperand(String op)
	{
		log.entering("PostfixCalculator", "addCheckedOperand", op);
		
		int size = stack.size();
		
		if ((size == 0) || (op.equals("~") && size < 1) || (!op.equals("~") && size < 2))
		{
			throw new IllegalArgumentException("Not enough numbers on stack for operand "+op);
		}
		
		stack.push(op);
		calculate();
		log.exiting("PostfixCalculator", "addCheckedOperand");
	}
	
	/**
	 * Add an operand
	 * @param op must be one of the allowed operands
	 * @exception IllegalArgumentException either incorrect op or stack doesn't have any numbers on it yet
	 * @see #allowedOperands
	 */
	private void addOperand(String op) throws IllegalArgumentException
	{
		log.entering("PostfixCalculator", "addOperand", op);
		if (isOperand(op))
		{
			addCheckedOperand(op);
		}
		else
		{
			throw new IllegalArgumentException(op+" is not an allowed operand");
		}
		log.exiting("PostfixCalculator", "addOperand");
	}
	
	/**
	 * Get the number of arguments required for a math function
	 * @param function the math function
	 * @return number of arguments
	 */
	private int getMathArguments(String function)
	{
		log.entering("PostfixCalculator", "getMathArguments");
		
		// lookup number of args from the cache
		Integer argI = args.get(function);
		if (argI == null)
		{
			throw new IllegalArgumentException(function + " is not a legal function");
		}
		
		log.exiting("PostfixCalculator", "getMathArguments");
		return argI.intValue();
	}
	
	/**
	 * Add a math function to the stack
	 * @param function function as defined by java.lang.Math.
	 * The methods allowed are random and methods which take one or two double arguments.
	 * @see java.lang.Math
	 */
	private void addMathFunction(String function) throws IllegalArgumentException
	{
		Method meth;
		
		log.entering("PostfixCalculator", "addMathFunction", function);
		
		if (function == null)
		{
			throw new IllegalArgumentException("No function specified");
		}
		
		// handle random
		if (function.equals("random"))
		{
			addNumber(new Double(Math.random()));
			return;
		}
		
		// check argument count
		int argCount = getMathArguments(function);
		Class[] methArgs = null;
		if (argCount == 1)
		{
			methArgs = new Class[1];
			methArgs[0] = double.class;
		}
		else if (argCount == 2)
		{
			methArgs = new Class[2];
			methArgs[0] = double.class;
			methArgs[1] = double.class;
		}
		else if (argCount != 0)
		{
			throw new IllegalArgumentException(function+" is not a legal function");
		}
		
		try
		{
			meth = Math.class.getDeclaredMethod(function, methArgs);
			stack.add(meth);
			calculate();
			return;
		}
		catch (NoSuchMethodException nsmex)
		{
			log.log(Level.WARNING, "Math function not found "+function, nsmex);
		}
		
		throw new IllegalArgumentException(function+" is not a legal function");
	}
	
	/** Reset the calculator */
	public void reset()
	{
		stack.removeAllElements();
	}
	
	/**
	 * Returns the top value on the stack. This may be null if the stack
	 * is empty or the top of the stack isn't a number.
	 * @return the top of the stack converted to a Double
	 */
	public Double getResult()
	{
		log.entering("PostfixCalculator", "getResult");
		
		Double d = null;
		
		if (stack.size() > 0)
		{
			Object o = stack.peek();
			if (o instanceof Number)
			{
				Number num = (Number)o;
				d = new Double(num.doubleValue());
			}
		}
		
		log.exiting("PostfixCalculator", "getResult", d);
		return d;
	}
	
	/** Checks that the top of the stack is a number */
	private void checkTopOfStackIsNumber()
	{
		Object item = stack.peek();
		if (!(item instanceof Number))
		{
			throw new IllegalStateException("Top of stack isn't a number");
		}
	}
	
	/**
	 * Use a math function to calculate the result
	 * @param meth the method to invoke
	 */
	private void calculateMethod(Method meth)
	{
		log.entering("PostfixCalculator", "calculateMethod", meth);
		
		Object[] cargs;
		int i = getMathArguments(meth.getName());
		
		checkTopOfStackIsNumber();
		Number num1 = (Number)stack.pop();
		if (i == 1)
		{
			cargs = new Object[1];
			cargs[0] = Double.valueOf(num1.doubleValue());
		}
		else
		{
			checkTopOfStackIsNumber();
			cargs = new Object[2];
			Number num2 = (Number)stack.pop();
			cargs[0] = Double.valueOf(num2.doubleValue());
			cargs[1] = Double.valueOf(num1.doubleValue());
		}
		
		try
		{
			stack.push((Double)meth.invoke(null, args));
		}
		catch (Exception ex)
		{
			log.log(Level.WARNING, "Cannot invoke "+meth.getName(), ex);
			throw new IllegalArgumentException("Cannot invoke "+meth.getName());
		}
		log.exiting("PostfixCalculator", "calculateMethod");
	}
	
	private void calculateOperator(String op)
	{
		log.entering("PostfixCalculator", "calculateOperator", op);
		
		Number num1 = (Number)stack.pop();
		if (op.equals("~"))
		{
			int i = ~num1.intValue();
			stack.push(Integer.valueOf(i));
			return;
		}
		
		Number num2 = (Number)stack.pop();
		Number temp = num1;
		num1 = num2;
		num2 = temp;
		
		if (op.equals("+"))
		{
			stack.push(add(num1, num2));
		}
		else if (op.equals("-"))
		{
			stack.push(subtract(num1, num2));
		}
		else if (op.equals("*"))
		{
			stack.push(multiply(num1, num2));
		}
		else if (op.equals("/"))
		{
			stack.push(divide(num1, num2));
		}
		else if (op.equals("%"))
		{
			stack.push(remainder(num1, num2));
		}
		else if (op.equals("p"))
		{
			stack.push(power(num1, num2));
		}
		else if (op.equals("^"))
		{
			stack.push(xor(num1, num2));
		}
		else if (op.equals("|"))
		{
			stack.push(or(num1, num2));
		}
		else if (op.equals("&"))
		{
			stack.push(and(num1, num2));
		}
		else if (op.equals("<<"))
		{
			stack.push(leftShift(num1, num2));
		}
		else if (op.equals(">>"))
		{
			stack.push(rightShift(num1, num2));
		}
		else if (op.equals(">>>"))
		{
			stack.push(rightShiftZeroFill(num1, num2));
		}
		log.exiting("PostfixCalculator", "calculateOperator");
	}
	
	/**
	 * Apply the last operand/function to the numbers
	 * @exception IllegalArgumentException calculation is incorrect
	 * @exception IllegalStateException stack doesn't contain enough arguments to perform a calculation
	 */
	private void calculate() throws IllegalArgumentException, IllegalStateException
	{
		log.entering("PostfixCalculator", "calculate");
		
		Object obj = stack.peek();
		if (obj instanceof Number)
		{
			throw new IllegalArgumentException("Last item isn't an operand or function");
		}
		
		if (stack.size() < 2)
		{
			throw new IllegalStateException("Not enough arguments to perform calculation");
		}
		
		Object op = stack.peek();
		if (!(op instanceof String || op instanceof Method))
		{
			throw new IllegalStateException("Top of stack isn't an operand or function");
		}
		op = stack.pop();
		if (op instanceof String) // process operator
		{
			calculateOperator((String)op);
		}
		else if (op instanceof Method) // process function
		{
			calculateMethod((Method)op);
		}
		
		log.exiting("PostfixCalculator", "calculate");
	}
	
	private Double add(Number num1, Number num2)
	{
		return Double.valueOf(num1.doubleValue() + num2.doubleValue());
	}
	
	private Double subtract(Number num1, Number num2)
	{
		return Double.valueOf(num1.doubleValue() - num2.doubleValue());
	}
	
	private Double multiply(Number num1, Number num2)
	{
		return Double.valueOf(num1.doubleValue() * num2.doubleValue());
	}
	
	private Double divide(Number num1, Number num2)
	{
		return Double.valueOf(num1.doubleValue() / num2.doubleValue());
	}
	
	private Double power(Number num1, Number num2)
	{
		return Double.valueOf(Math.pow(num1.doubleValue(), num2.doubleValue()));
	}
	
	private Double remainder(Number num1, Number num2)
	{
		return Double.valueOf(num1.doubleValue() % num2.doubleValue());
	}
	
	private Integer and(Number num1, Number num2)
	{
		return Integer.valueOf(num1.intValue() & num2.intValue());
	}
	
	private Integer or(Number num1, Number num2)
	{
		return Integer.valueOf(num1.intValue() | num2.intValue());
	}
	
	private Integer xor(Number num1, Number num2)
	{
		return Integer.valueOf(num1.intValue() ^ num2.intValue());
	}
	
	private Integer leftShift(Number num1, Number num2)
	{
		return Integer.valueOf(num1.intValue() << num2.intValue());
	}
	
	private Integer rightShift(Number num1, Number num2)
	{
		return Integer.valueOf(num1.intValue() >> num2.intValue());
	}
	
	private Integer rightShiftZeroFill(Number num1, Number num2)
	{
		return Integer.valueOf(num1.intValue() >>> num2.intValue());
	}
	
	/**
	 * Instead of passing number and operands one at a time they can be passed in altogether
	 * @param calculation space separated string of numbers, operands and functions
	 */
	public void parseCalculation(String calculation)
	{
		log.entering("PostfixCalculator", "parseCalculation", calculation);
		
		String token;
		StringTokenizer st = new StringTokenizer(calculation.trim(), " ");
		while (st.hasMoreTokens())
		{
			token = st.nextToken();
			add(token);
		}
		log.exiting("PostfixCalculator", "parseCalculation");
	}
	
	/** Remove all previously added values, functions and operands */
	public void clear()
	{
		stack.clear();
	}
	
	/** Get a description of the internal Stack */
	@Override
	public String toString()
	{
		String item;
		StringBuffer sb = new StringBuffer(stack.size()*2);
		Enumeration enm = stack.elements();
		while (enm.hasMoreElements())
		{
			item = enm.nextElement().toString();
			sb.append(item);
			sb.append("\n");
		}
		
		return sb.toString();
	}
	
	/** Main method used for testing*/
	public static void main(String[] args)
	{
		if (args.length < 1)
		{
			System.out.println("Please specify the calculation on the command line separated by spaces");
			System.exit(-1);
		}
		
		PostfixCalculator pc = new PostfixCalculator();
		StringBuffer sb = new StringBuffer(args.length);
		for (int i=0; i<args.length; i++)
		{
			sb.append(args[i]);
			sb.append(" ");
		}
		
		try
		{
			pc.parseCalculation(sb.toString().trim());
			Double result = pc.getResult();
			System.out.println(sb.toString().trim()+" = "+result.doubleValue());
		}
		catch (Exception ex)
		{
			System.out.println(ex.getMessage());
		}
	}
}